/******************************************************************************/
/* Copyright (c) 2005-2011, SAS Institute Inc., Cary, NC, USA,                */
/*                       All Rights Reserved.                                 */
/*                                                                            */
/* Name:    migrate_datastore_mo53m3_mo54ga.sas                               */
/*                                                                            */
/* Purpose: Convert an MO 5.3 M3 datastore to an MO 5.4 GA datastore.         */
/*                                                                            */
/*          It is intended that this macro be invoked from the MO SDM (see    */
/*          Usage Notes below).                                               */
/*                                                                            */
/*          Alternatively, this macro can be run manually from an interactive */
/*          SAS session.                                                      */
/*                                                                            */
/*          NOTE 1: The determination of whether this program is being run    */
/*                  from a batch SAS session or an interactive SAS session is */
/*                  made by the program itself and typically requires no user */
/*                  intervention (see below).                                 */
/*                                                                            */
/* Usage:                                                                     */
/*          IF USING THIS FILE MANUALLY, PERFORM THE FOLLOWING:               */
/*          1. Start an interactive SAS session and perform the following     */
/*             steps within that session.                                     */
/*          2. In the Program Editor, open and execute the Marketing          */
/*             Optimiation autoexec (MarketingOptimization_autoexec.sas)      */
/*             if not previously run.                                         */
/*          3. In the Program Editor, open and this program and verify that   */
/*             the following values are correctly set:                        */
/*             - The location of the MO datastore directory.                  */
/*             - The location of the MO uncompiled macros (ucmacros)          */
/*               directory.                                                   */
/*             - The operating system type (WIN or UNIX).                     */
/*          4. In the Program Editor, execute this program.                   */
/*          5. Verify it completes successfully with no ERRORs or unexpected  */
/*             WARNINGs in the log.                                           */
/*                                                                            */
/*          IF THIS FILE IS BEING INVOKED FROM THE SAS SOFTWARE DEPLOYMENT    */
/*          WIZARD FOR MO, PERFORM THE FOLLOWING:                             */
/*          1. The program invocation must specify the fully qualified path   */
/*             and file name of this program in the -SYSIN option.            */
/*          2. The program invocation must specify the -SYSPARM option        */
/*             which specifies the fully qualified path and file name of a    */
/*             text file containing input parameters.                         */
/*             This file should contain a line for each input parameter       */
/*             with the format of each line being an equals-separated pair    */
/*             as follows:                                                    */
/*                  <parameter name>=<parameter value>                        */
/*             like a Java or ANT properties file.                            */
/*             Specifically, the following is a list of parameter names with  */
/*             the possible values listed for each name. levconfig is the     */
/*             required parameter. Missing parameters will be assumed to be   */
/*             as specified below.                                            */
/*                  datastoreDir=<fully qualified path name to directory where*/
/*                    the Marketing Optimization 5.3M3-format datastore tables*/
/*                    to be migrated to 5.4GA-format reside>                  */
/*                  ucmacrosDir=<fully qualified path name to directory where */
/*                    the Marketing Optimization 5.4 uncompiled macros (for   */
/*                    SASAUTOS option) reside>                                */
/*                  osType=<WIN || UNX>                                       */
/*                    (Missing parm or <blank> value --> WIN)                 */
/*          3. The program invocation should specify the -LOG option with     */
/*             the fully qualified path and file name to which the SAS log    */
/*             should be directed.                                            */
/*                                                                            */
/* Example SAS program invocation command:                                    */
/* (Note: The line breaks in the example should be eliminated from the        */
/*  command used in the script.)                                              */
/*                                                                            */
/* "C:\Program Files\SAS\SASFoundation\9.2\sas"                               */
/*  -sysin "C:\Program Files\SAS\9.1.3\mktopt\                                */
/*          sasmisc\migrate_datastore_mo53m3_mo54ga.sas"                      */
/*  -sysparm "C:\SAS\Config\Lev1\Applications\SASMarketingOptimization5.4\    */
/*            config\mo_properties_set_server_locale.properties               */
/*  -log "C:\SAS\Config\Lev1\SASApp\Applications\SASMarketingOptimization5.4\ */
/*        Logs                                                                */
/******************************************************************************/

%macro migrate_datastore_mo53m3_mo54ga;

   options notes;
   ODS _ALL_ CLOSE;

   %local macname;
   %let macname=&sysmacroname;
   %let macname=%lowcase(&macname);


   %*-----------------;
   %* Initialization. ;
   %*-----------------;
   %local inputParametersFile;
   %let inputParametersFile =;

   %local interactive_sas_session;
   %let interactive_sas_session = Y;

   %local datastoreDir;
   %let datastoreDir =;

   %local ucmacrosDir;
   %let ucmacrosDir =;

   %local osType;
   %let osType = WIN;

   %local sasautos_prev;
   %let sasautos_prev =;

   %local datastore_migratable;
   %let datastore_migratable = Y;

   %local datastore_notmigrated_reasons;
   %let datastore_notmigrated_reasons = ;


   %*-----------------------------------------------------------------------;
   %* Since the MO autoexec is not run when this program is run under a     ;
   %* batch SAS session, global macro variables must be defined manually.   ;
   %*-----------------------------------------------------------------------;
   %if (%symexist(mometa) eq 0) %then %do;
      %local mometa;
   %end;  
   %if (%symexist(operating_system) eq 0) %then %do;
      %local operating_system;
   %end;   
   %if (%symexist(separator) eq 0) %then %do;
      %local separator;
   %end;  
   %if (%symexist(momacro) eq 0) %then %do;
      %local momacro;
   %end;  
   %if (%symexist(moiter) eq 0) %then %do;
      %local moiter;
      %let moiter = 60;
   %end;  
   %if (%symexist(mosleep) eq 0) %then %do;
      %local mosleep;
      %let mosleep = 3;
   %end;  
   %if (%symexist(modebug) eq 0) %then %do;
      %global modebug;
      %let modebug = 0;
   %end;


   %*--------------------------------------------------------------------------;
   %* Determine whether this program is being run from a batch SAS session or  ;
   %* an interactive SAS session by checking whether or not the SYSPARM option ;
   %* has a value (batch session) or not (interactive session).                ;
   %*                                                                          ;
   %* If this is a batch session, retrieve the input parameters to the program ;
   %* by reading the file specified by the SYSPARM option.                     ;
   %*                                                                          ;
   %* If this is an interactive session, update the hardcoded input parameters.;
   %*--------------------------------------------------------------------------;
   data _null_;
      length inputParametersFile $ 200;
      inputParametersFile = sysparm();
      if (inputParametersFile ~= ' ') then do;
         call symput("inputParametersFile", left(trim(inputParametersFile)));
      end;
   run;

   %if (%nrbquote(&inputParametersFile) eq %str()) %then %do; /* Assume interactive */
      %let interactive_sas_session = Y;

      %*!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!;
      %* Verify the following hard-coded parameter settings.                   ;
      %* 1. The SAS install directory where the MO uncompiled macros reside.   ;
      %* 2. The directory where the MO datastore resides.                      ;
      %* 3. The OS type: either WIN or UNIX.                                   ;
      %*                                                                       ;
      %* NOTE: UNIX specifications are case-sensitive!                         ;
      %*!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!;
      %let ucmacrosDir = %nrbquote(C:\Program Files\SAS\SASFoundation\9.2\mktopt\ucmacros);                           /* UPDATE */

      %let datastoreDir = %nrbquote(C:\SAS\Config\Lev1\Applications\SASMarketingOptimization5.4\datastore);           /* UPDATE */

      %let osType = WIN;              /* UPDATE */  
   %end;
   %else %do;
      %let interactive_sas_session = N;

      %if (not %sysfunc(fileexist(%nrbquote(&inputParametersFile)))) %then %do;
         %put ERROR: Input Parameters File specified in the -SYSPARM option does not exist.;
         %goto ERROR_EXIT;
      %end;

      %let datastoreDir =;
      %let ucmacrosDir =;
      %let osType =;
      
      filename parms "%nrbquote(&inputParametersFile)";
      data _null_;
         length parmName $ 32;
         length parmValue $ 2048;

         infile parms;
         input;

         parmName = left(trim(scan(_INFILE_, 1, "=")));
         parmValue = left(trim(scan(_INFILE_, 2, "=")));

         if (upcase(parmName) = "DATASTOREDIR") then do;
            call symput("datastoreDir", parmValue);
         end;
         else if (upcase(parmName) = "UCMACROSDIR") then do;
            call symput("ucmacrosDir", parmValue);
         end;
         else if (upcase(parmName) = "OSTYPE") then do;
            call symput("osType", parmValue);
         end;
      run;

      %let datastoreDir = %trim(%left(%nrbquote(&datastoreDir)));
      %let ucmacrosDir = %trim(%left(%nrbquote(&ucmacrosDir)));
      %let osType = %upcase(%trim(%left(&osType)));
   %end;

   %*-----------------------------------------------------------------------;
   %* Process type of OS indicated.                                         ;
   %*-----------------------------------------------------------------------;
   %if (&osType eq WIN) %then %do;
      %let operating_system = WIN;
      %let separator = %str(\);
   %end;
   %else %do;
      %let operating_system = UNIX;
      %let separator = %str(/);
   %end;
 
   %put NOTE: DATASTOREDIR: %nrbquote(&datastoreDir);
   %put NOTE: USMACROSDIR: %nrbquote(&ucmacrosDir);
   %put NOTE: OSTYPE: &osType;
   %put NOTE: SEPARATOR: %nrbquote(&separator);
   %put NOTE: OPERATING_SYSTEM: %nrbquote(&operating_system);


   %*-----------------------------------------------------------------------;
   %* Since the MO autoexec is not run when this program is run under a     ;
   %* batch SAS session, the SASAUTOS option to include the path to the MO  ;
   %* uncompiled macros must be set up "manually", necessary global         ;
   %* macro variables must be defined, and the libref to the MO datastore   ;
   %* must be defined.  For consistency, use names that are used by the MO  ;
   %* autoexec.                                                             ;
   %*-----------------------------------------------------------------------;
   %let momacro = %nrbquote(&ucmacrosDir);
   %if (not %sysfunc(fileexist(%nrbquote(&momacro)))) %then %do;
      %put ERROR: Directory %upcase(%nrbquote(&momacro)) does not exist.;
      %put ERROR: Specify a valid fully qualified pathname to the file system location of the SAS Marketing Optimization 5.4 uncompiled macros directory.;
      %goto ERROR_EXIT;
   %end;
   %let sasautos_prev = %sysfunc(compress(%sysfunc(getoption(SASAUTOS)),()));
   options sasautos = ("%nrbquote(&momacro)" &sasautos_prev);

   %mo_constants


   %*-----------------------------------------------------------------------;
   %* Verify that a SAS Marketing Optimization 5.4 datastore directory      ;
   %* exists, is accessible via a connection to the Share Server, and is    ;
   %* not empty.                                                            ;
   %*-----------------------------------------------------------------------;
   %let mometa = %nrbquote(&datastoreDir);
   %if (not %sysfunc(fileexist(%nrbquote(&mometa)))) %then %do;
      %put; %put;
      %put WARNING: Directory %nrbquote(&mometa) either does not exist or is not accessible from the SAS tier machine.;
      %put WARNING: Make sure that directory %nrbquote(&mometa) is the intended location for your SAS Marketing Optimization 5.4 datastore.;

      %let datastore_migratable = N;
      %if (%nrbquote(&datastore_notmigrated_reasons) eq %str()) %then %do;
         %let datastore_notmigrated_reasons = Datastore Directory Not Accessible;
      %end;
      %else %do;
         %let datastore_notmigrated_reasons = %nrbquote(&datastore_notmigrated_reasons, )Datastore Directory Not Accessible;
      %end;
   %end;
   %else %do;
      %if (&interactive_sas_session eq N) %then %do;
         libname mometa "%nrbquote(&mometa)"; /* SDM: SHARE Server not needed */
      %end;
      %else %do;
         libname mometa "&mometa" server=&moshare authdomain="&modatausr_authdomain";
      %end;
      %if (&SYSLIBRC ne 0) %then %do;
         %put; %put;
         %put WARNING: Cannot access directory %nrbquote(&mometa) as a SAS library. The SAS libname statement failed to associate this directory with a libref value (syslibrc=&syslibrc).;
         %put WARNING: Make sure that file system permissions to directory %nrbquote(&mometa) allow read and write access for MO administrators and users, the SAS GeneralServers account (e.g., sassrv) and the MO Data Users account (e.g., modatausr).;
         %put WARNING: Also make sure the SAS/SHARE server is running.;

         %let datastore_migratable = N;
         %if (%nrbquote(&datastore_notmigrated_reasons) eq %str()) %then %do;
            %let datastore_notmigrated_reasons = Libname Failure;
         %end;
         %else %do;
            %let datastore_notmigrated_reasons = %nrbquote(&datastore_notmigrated_reasons, )Libname Failure;
         %end;
      %end;
      %else %do;
         proc contents data=mometa._all_ out=work.tableNames_tmp (keep=memname) noprint nodetails;
         run;
         %if (&syserr gt 4) %then %do;
            %let &_rc = 100;
            %put %mo_error Error encountered using proc contents to get the list of tables in MOMETA (syserr=&syserr).;
            %goto ERROR_EXIT;
         %end;

         %local memberCount;
         %let memberCount = 0;
         proc sql noprint;
            select count(*) into :memberCount
            from work.tableNames_tmp
            where (memname ~= "");
         quit;
         %if (&sqlrc gt 4) %then %do;
            %put ERROR: Error encountered while counting SAS members of directory %nrbquote(&mometa) (sqlrc=&sqlrc).;
            %goto ERROR_EXIT;
         %end;

         %if (&memberCount eq 0) %then %do;
            %put; %put;
            %put WARNING: Directory %nrbquote(&mometa) is empty.;
            %put WARNING: Make sure that directory %nrbquote(&mometa) is the intended location for your SAS Marketing Optimization 5.4 datastore.;

            %let datastore_migratable = N;
            %if (%nrbquote(&datastore_notmigrated_reasons) eq %str()) %then %do;
               %let datastore_notmigrated_reasons = Empty Datastore Directory;
            %end;
            %else %do;
               %let datastore_notmigrated_reasons = %nrbquote(&datastore_notmigrated_reasons, )Empty Datastore Directory;
            %end;
         %end;
      %end;
   %end;

   
   %*------------------------------------------------------------------------;
   %* If any problems precluding datastore migration were found above,       ;
   %* exit.                                                                  ;
   %*------------------------------------------------------------------------;
   %if (&datastore_migratable eq N) %then %do;
      %goto DS_NOT_MIGRATABLE_EXIT;
   %end;


   %*------------------------------------------------------;
   %* Backup the MO Datastore directory before proceeding. ;
   %*------------------------------------------------------;
   %mo_datastore_backup(fq_backup_path =
                       ,_rc = dbrc);
   %if (&dbrc ne 0) %then %do;
      %put ERROR: Error encountered backing up the MO datastore directory.;
      %goto ERROR_EXIT;
   %end;


   %*------------------;
   %* Initialization.  ;
   %*------------------;
   %local col_names;

   %mo_constants


   %*-------------------------------------------------------------------;      
   %* TABLE = INPUT_DATA (new columns in 5.4 GA since 5.3 M3)           ;     
   %* 1. Add prioritization-related columns (uses_prioritization,       ;
   %*    customer_rank_measure_name, priority_column_name).             ;
   %* 2. Add contact history-related columns (uses_contact_history,     ;
   %*    contact_history_table_name, communication_history_table_name). ;
   %*-------------------------------------------------------------------;
   %let col_names=;
   proc contents noprint data=mometa.input_data out=work.col_names (keep=name length);
   run;
   proc sql noprint;
      select name into :col_names separated by " " from work.col_names;
   quit;

   %if ((%index(%upcase(&col_names), USES_PRIORITIZATION) = 0) or
        (%index(%upcase(&col_names), CUSTOMER_RANK_MEASURE_NAME) = 0) or
        (%index(%upcase(&col_names), PRIORITY_COLUMN_NAME) = 0)) %then %do;
      %mo_create_tables(lib=work,table=input_data,_rc=cudrc)
      %if (&cudrc ne 0) %then %do;
         %put ERROR: Error encountered creating WORK.INPUT_DATA table.;
         %goto ERROR_EXIT;
      %end;

      /* Concatenate the existing metadata table to the working table, in */
      /* order to populate the working table with the existing rows.      */
      data work.input_data;
         set work.input_data 
             mometa.input_data;
         uses_prioritization = 0;
         customer_rank_measure_name = "";
         priority_column_name = "";
      run;

      /* Copy the working table, now populated with existing input_data   */
      /* rows, over to the metadata library.                              */
      data mometa.input_data;
         set work.input_data;
      run;
   %end;

   %if ((%index(%upcase(&col_names), USES_CONTACT_HISTORY) = 0) or
        (%index(%upcase(&col_names), CONTACT_HISTORY_TABLE_NAME) = 0) or
        (%index(%upcase(&col_names), COMMUNICATION_HISTORY_TABLE_NAME) = 0) or
        (%index(%upcase(&col_names), NUMBER_OF_HISTORICAL_CONTACTS) = 0)) %then %do;
      %mo_create_tables(lib=work,table=input_data,_rc=cudrc)
      %if (&cudrc ne 0) %then %do;
         %put ERROR: Error encountered creating WORK.INPUT_DATA table.;
         %goto ERROR_EXIT;
      %end;

      /* Concatenate the existing metadata table to the working table, in */
      /* order to populate the working table with the existing rows.      */
      data work.input_data;
         set work.input_data 
             mometa.input_data;
         uses_contact_history = 0;
         contact_history_table_name = "";
         communication_history_table_name = "";
         number_of_historical_contacts = 0;
      run;

      /* Copy the working table, now populated with existing input_data   */
      /* rows, over to the metadata library.                              */
      data mometa.input_data;
         set work.input_data;
      run;
   %end;

   %*-------------------------------------------------------------------;      
   %* TABLE = INPUT_DATA_TEMPLATE (new columns in 5.4 GA since 5.3 GA)  ;     
   %* 1. Add contact history-related columns                            ;
   %*    (contact_history_table_name, communication_history_table_name).;
   %*-------------------------------------------------------------------;
   %let col_names=;
   proc contents noprint data=mometa.input_data_template out=work.col_names (keep=name length);
   run;
   proc sql noprint;
      select name into :col_names separated by " " from work.col_names;
   quit;

   %if ((%index(%upcase(&col_names), CONTACT_HISTORY_TABLE_NAME) = 0) or
        (%index(%upcase(&col_names), COMMUNICATION_HISTORY_TABLE_NAME) = 0)) %then %do;
      %mo_create_tables(lib=work,table=input_data_template,_rc=cudrc)
      %if (&cudrc ne 0) %then %do;
         %put ERROR: Error encountered creating WORK.INPUT_DATA_TEMPLATE table.;
         %goto ERROR_EXIT;
      %end;

      /* Concatenate the existing metadata table to the working table, in */
      /* order to populate the working table with the existing rows.      */
      data work.input_data_template;
         set work.input_data_template 
             mometa.input_data_template;
         contact_history_table_name = "";
         communication_history_table_name = "";
      run;

      /* Copy the working table, now populated with existing              */
      /* rows, over to the metadata library.                              */
      data mometa.input_data_template;
         set work.input_data_template;
      run;
   %end;

   %*-------------------------------------------------------------------------;
   %* TABLE = INPUT_DATA_MOMA_PARAMETERS (new columns in 5.4 GA since 5.3 M3) ;
   %* 1. Add optimization horizon columns (uses_finite_optimization_horizon,  ;
   %*    optimization_horizon_in_days).                                       ;
   %* 2. Add contact history days column (contact_history_in_days)            ;
   %* 3. Add uses contact history column (uses_contact_history)               ;
   %*-------------------------------------------------------------------------;
   %let col_names=;
   proc contents noprint data=mometa.input_data_moma_parameters out=work.col_names (keep=name length);
   run;
   proc sql noprint;
      select name into :col_names separated by " " from work.col_names;
   quit;

   %if ((%index(%upcase(&col_names), CONTACT_HISTORY_IN_DAYS) = 0) or
        (%index(%upcase(&col_names), OPTIMIZATION_HORIZON_IN_DAYS) = 0) or
        (%index(%upcase(&col_names), USES_FINITE_OPTIMIZATION_HORIZON) = 0)) %then %do;
      %mo_create_tables(lib=work,table=input_data_moma_parameters,_rc=cudrc)
      %if (&cudrc ne 0) %then %do;
         %put ERROR: Error encountered creating WORK.INPUT_DATA_MOMA_PARAMETERS table.;
         %goto ERROR_EXIT;
      %end;

      /* Concatenate the existing metadata table to the working table, in */
      /* order to populate the working table with the existing rows.      */
      data work.input_data_moma_parameters;
         set work.input_data_moma_parameters 
             mometa.input_data_moma_parameters;
         contact_history_in_days = 90;
         uses_finite_optimization_horizon = 0; 
         optimization_horizon_in_days = 14;
      run;

      /* Copy the working table, now populated with existing input_data   */
      /* rows, over to the metadata library.                              */
      data mometa.input_data_moma_parameters;
         set work.input_data_moma_parameters;
      run;
   %end;

   %if (%index(%upcase(&col_names), USES_CONTACT_HISTORY) = 0) %then %do;
      %mo_create_tables(lib=work,table=input_data_moma_parameters,_rc=cudrc)
      %if (&cudrc ne 0) %then %do;
         %put ERROR: Error encountered creating WORK.INPUT_DATA_MOMA_PARAMETERS table.;
         %goto ERROR_EXIT;
      %end;

      /* Concatenate the existing metadata table to the working table, in */
      /* order to populate the working table with the existing rows.      */
      data work.input_data_moma_parameters;
         set work.input_data_moma_parameters 
             mometa.input_data_moma_parameters;
         uses_contact_history = 0;
      run;

      /* Copy the working table, now populated with existing input_data   */
      /* rows, over to the metadata library.                              */
      data mometa.input_data_moma_parameters;
         set work.input_data_moma_parameters;
      run;
   %end;

   %*---------------------------------------------------------------;      
   %* TABLE = COMMUNICATIONS (new columns in 5.4 GA since 5.3 M3)   ;     
   %* 1. Add 3 new flags for mandatory inclusions (mandatory,       ;
   %*    exclude_from_constraints, and exclude_from_contact_pols) ;
   %*---------------------------------------------------------------;
   %let col_names=;
   proc contents noprint data=mometa.communications out=work.col_names (keep=name length);
   run;
   proc sql noprint;
      select name into :col_names separated by " " from work.col_names;
   quit;

   %if ((%index(%upcase(&col_names), MANDATORY) = 0) or
        (%index(%upcase(&col_names), EXCLUDE_FROM_CONSTRAINTS) = 0) or
        (%index(%upcase(&col_names), EXCLUDE_FROM_CONTACT_POLS) = 0)) %then %do;
      %mo_create_tables(lib=work,table=communications,_rc=cudrc)
      %if (&cudrc ne 0) %then %do;
         %put ERROR: Error encountered creating WORK.COMMUNICATIONS table.;
         %goto ERROR_EXIT;
      %end;

      /* Concatenate the existing metadata table to the working table, in */
      /* order to populate the working table with the existing rows.      */
      data work.communications;
         set work.communications 
             mometa.communications;
         mandatory = 0;
         exclude_from_constraints = 0;
         exclude_from_contact_pols = 0;
      run;

      /* Copy the working table, now populated with existing communications */
      /* rows, over to the metadata library.                                */
      data mometa.communications;
         set work.communications;
      run;
   %end;

   %*---------------------------------------------------------------;      
   %* TABLE = REFRESH_MESSAGES (new columns in 5.4 GA since 5.3 M3) ;     
   %* 1. Add message_type field                                     ;
   %*---------------------------------------------------------------;
   %let col_names=;
   proc contents noprint data=mometa.refresh_messages out=work.col_names (keep=name length);
   run;
   proc sql noprint;
      select name into :col_names separated by " " from work.col_names;
   quit;

   %if (%index(%upcase(&col_names), MESSAGE_TYPE) = 0) %then %do;
      %mo_create_tables(lib=work,table=refresh_messages,_rc=cudrc)
      %if (&cudrc ne 0) %then %do;
         %put ERROR: Error encountered creating WORK.REFRESH_MESSAGES table.;
         %goto ERROR_EXIT;
      %end;

      /* Concatenate the existing metadata table to the working table, in */
      /* order to populate the working table with the existing rows.      */
      data work.refresh_messages;
         set work.refresh_messages 
             mometa.refresh_messages;
         message_type = &_SCENARIO_REFRESH_;
      run;

      /* Copy the working table, now populated with existing refresh_messages */
      /* rows, over to the metadata library.                                  */
      data mometa.refresh_messages;
         set work.refresh_messages;
      run;
   %end;

   %*------------------------------------------------------------------------;
   %* TABLE = SCENARIO_ADVANCED_OPTIONS (new columns in 5.4 GA since 5.3 M3) ;
   %* 1. Add contact history flag (active_contact_history).                  ;
   %*------------------------------------------------------------------------;
   %let col_names=;
   proc contents noprint data=mometa.scenario_advanced_options out=work.col_names (keep=name length);
   run;
   proc sql noprint;
      select name into :col_names separated by " " from work.col_names;
   quit;

   %if (%index(%upcase(&col_names), ACTIVE_CONTACT_HISTORY) = 0) %then %do;
      %mo_create_tables(lib=work,table=scenario_advanced_options,_rc=cudrc)
      %if (&cudrc ne 0) %then %do;
         %put ERROR: Error encountered creating WORK.SCENARIO_ADVANCED_OPTIONS table.;
         %goto ERROR_EXIT;
      %end;

      /* Concatenate the existing metadata table to the working table, in */
      /* order to populate the working table with the existing rows.      */
      data work.scenario_advanced_options;
         set work.scenario_advanced_options 
             mometa.scenario_advanced_options;
         active_contact_history = 0;
      run;

      /* Copy the working table, now populated with existing input_data   */
      /* rows, over to the metadata library.                              */
      data mometa.scenario_advanced_options;
         set work.scenario_advanced_options;
      run;
   %end;

   %*----------------------------------------------------------;
   %* TABLE = INPUT_DATA_USAGE_TRACKING (new tabln in 5.4 GA)  ;
   %*----------------------------------------------------------;
   %if (not %sysfunc(exist(mometa.input_data_usage_tracking))) %then %do;
      %mo_create_tables(lib=mometa,table=input_data_usage_tracking,_rc=cudrc)
      %if (&cudrc ne 0) %then %do;
         %put ERROR: Error encountered creating MOMETA.INPUT_DATA_USAGE_TRACKING table.;
         %goto ERROR_EXIT;
      %end;
   %end;


   %*--------------------------------;
   %* Exit processing                ;
   %*--------------------------------;
   %put;
   %mo_no_notes(on)
   %put NOTE: MO 5.3 M3 datastore successfully migrated to MO 5.4 GA.;
   %goto EXIT;


   %ERROR_EXIT:
   %put;
   %put ERROR: Error encountered while migrating MO 5.3 M3 datastore to MO 5.4 GA.;
   %goto EXIT;


   %DS_NOT_MIGRATABLE_EXIT:
   %put; %put;
   %put WARNING: Unable to migrate the SAS Marketing Optimization datastore for the following reason(s) - %nrbquote(&datastore_notmigrated_reasons).; 
   %put; %put;
   %put WARNING: After verifying and correcting the above items, perform the following steps.;
   %put WARNING: Perform manual migration of the 5.3 M3 datastore to the 5.4 GA datastore by running program migrate_datastore_mo53m3_mo54ga.sas from an interactive SAS session, following the instructions in the comments.;


   %EXIT:
   %if (&syscc le 4) %then %do;
      %let syscc = 0;
   %end;

%mend migrate_datastore_mo53m3_mo54ga;


%migrate_datastore_mo53m3_mo54ga;
