/******************************************************************************/
/* Copyright (c) 2005-2011, SAS Institute Inc., Cary, NC, USA,                */
/*                       All Rights Reserved.                                 */
/*                                                                            */
/* Name:    mo_compare_comm_num_attr_dflts                                    */
/*                                                                            */
/* Purpose: Compare a set of MO input data comm numeric attribute values with */
/*          a set of MO input data comm numeric attribute defaults and        */
/*          capture the differences.                                          */
/*                                                                            */
/* Input:   input_data_id_FROM   - "From" input data, i.e., the "values"      */
/*          scenario_id_FROM     - "From" scenario, i.e., the "values"        */
/*          input_data_id_TO     - "To" input data, i.e., the "defaults"      */
/*          _dflts_to_add        - Macro variable to hold IDs of defaults     */
/*                                 to add to the FROM data so that it         */
/*                                 would match the defaults in the TO data.   */
/*          _vals_to_delete      - Macro variable to hold IDs of values       */
/*                                 to delete from the FROM data so that it    */
/*                                 would match the defaults in the TO data.   */
/*          _attrs_to_swap_FROM  - Macro variable to hold IDs of values       */
/*                                 in the FROM data that match defaults in    */
/*                                 the TO data                                */
/*          _attrs_to_swap_TO    - Macro variable to hold IDs of defaults     */
/*                                 in the TO data that match values in        */
/*                                 the FROM data                              */
/*                                                                            */
/* Output:  &_rc = return code                                                */
/*                                                                            */
/* *** ASSUMPTION *** A CELL_CODE COLUMN MAY HAVE ALREADY BEEN JOINED TO THE  */
/*                    COMM_NUMERIC_ATTRIBUTE_values TABLE IN THE SPECIFIED    */
/*                    LIBRARIES IF IT IS NEEDED... OR NOT.                    */
/*                                                                            */
/* HISTORY:                                                                   */
/* 5.4                                                                        */
/* 09Apr10  Creation for scenario import (KEC)                                */
/*                                                                            */
/* F1MO54, F0MO541                                                            */
/* 09Jan12 S0823209 Replace use of SQL query of Dictionay Tables with use     */
/*         of SAS Proc Contents.                                              */
/*                                                                            */
/******************************************************************************/

%macro mo_compare_comm_num_attr_dflts(input_data_id_FROM  =
                                     ,scenario_id_FROM    =
                                     ,input_data_id_TO    =
                                     ,library_FROM        = mometa
                                     ,library_TO          = mometa
                                     ,id_FROM_mapping_tbl =
                                     ,id_FROM_mapping_col =
                                     ,id_TO_mapping_col   =
                                     ,_dflts_to_add       =
                                     ,_vals_to_delete     =
                                     ,_attrs_to_swap_FROM =
                                     ,_attrs_to_swap_TO   =
                                     ,_rc                 =
                                     );

   %local macname;
   %let macname=&sysmacroname;
   %let macname=%lowcase(&macname);

   %if %bquote(&_rc) eq %str() %then %do;
      %local rc;
      %let _rc = rc;
   %end;
   %else %global &_rc;
   %let &_rc = 0;

   %if %bquote(&_dflts_to_add) eq %str() %then %do;
      %local dflts_to_add;
      %let _dflts_to_add = dflts_to_add;
   %end;
   %else %global &_dflts_to_add;
   %let &_dflts_to_add =;

   %if %bquote(&_vals_to_delete) eq %str() %then %do;
      %local vals_to_delete;
      %let _vals_to_delete = vals_to_delete;
   %end;
   %else %global &_vals_to_delete;
   %let &_vals_to_delete =;

   %if %bquote(&_attrs_to_swap_FROM) eq %str() %then %do;
      %local attrs_to_swap_FROM;
      %let _attrs_to_swap_FROM = attrs_to_swap_FROM;
   %end;
   %else %global &_attrs_to_swap_FROM;
   %let &_attrs_to_swap_FROM =;

   %if %bquote(&_attrs_to_swap_TO) eq %str() %then %do;
      %local attrs_to_swap_TO;
      %let _attrs_to_swap_TO = attrs_to_swap_TO;
   %end;
   %else %global &_attrs_to_swap_TO;
   %let &_attrs_to_swap_TO =;


   %mo_constants


   %local local_debug;
   %let local_debug = 0;
   %if (&local_debug eq 1) %then %do;
      %mo_no_notes(on)
      %put;
      %put %mo_note input_data_id_FROM: %nrbquote(&input_data_id_FROM);
      %put %mo_note input_data_id_TO: %nrbquote(&input_data_id_TO);
      %put %mo_note library_FROM: %nrbquote(&library_FROM);
      %put %mo_note library_TO: %nrbquote(&library_TO);
      %mo_no_notes(off)
   %end;


   %mo_no_notes(on)
   %put;
   %put %mo_note Starting comparison of comm numeric attribute values.;
   %mo_no_notes(off)


   %*--------------------------------;
   %* Validate parms                 ;
   %*--------------------------------;
   %if (%nrbquote(&input_data_id_FROM) eq %str()) %then %do;
      %put %mo_error No "FROM" input data ID was specified.;
      %let &_rc = 102;
      %goto ERROR_EXIT;
   %end;
   %if (%nrbquote(&input_data_id_TO) eq %str()) %then %do;
      %put %mo_error No "TO" input data ID was specified.;
      %let &_rc = 102;
      %goto ERROR_EXIT;
   %end;
   %if (%nrbquote(&id_FROM_mapping_tbl) ne %str()) %then %do;
      %if (%nrbquote(&id_FROM_mapping_col) eq %str()) %then %do;
         %put %mo_error No "ID FROM" mapping column was specified.;
         %let &_rc = 102;
         %goto ERROR_EXIT;
      %end;
      %else %if (%nrbquote(&id_TO_mapping_col) eq %str()) %then %do;
         %put %mo_error No "ID TO" mapping column was specified.;
         %let &_rc = 102;
         %goto ERROR_EXIT;
      %end;
   %end;


   %put;
   %put %mo_note Starting comparison of comm numeric attribute values.;
   %put;


   %*-------------------------;
   %* Create base tables from ;
   %* FROM library tables     ;
   %*-------------------------;
   %local cell_cd_column_status_FROM uses_cell_code_FROM;
   %let cell_cd_column_status_FROM =;
   %let uses_cell_code_FROM = 0;
   proc sql noprint;
      select cell_cd_column_status into :cell_cd_column_status_FROM
      from &library_FROM..input_data
      where (input_data_id = &input_data_id_FROM);
   quit;
   %if ((&sqlrc gt 4) or (&sqlobs eq 0)) %then %do;
      %let &_rc = 100;
      %put %mo_error Error occurred while reading &library_FROM..input_data (sqlrc=&sqlrc, sqlob =&sqlobs).;
      %goto ERROR_EXIT;
   %end;
   %let cell_cd_column_status_FROM = %trim(%left(&cell_cd_column_status_FROM));

   %if ((&cell_cd_column_status_FROM eq &_CELL_CD_UNIQUE_) or
        (&cell_cd_column_status_FROM eq &_CELL_CD_UNIQUE_UNNEEDED_)) %then %do;
      %let uses_cell_code_FROM = 1;
   %end;

   %local has_cell_code_FROM;
   %let has_cell_code_FROM = 0;
   proc contents data=&library_FROM..comm_numeric_attribute_defaults out=work.hasCellCodeFrom(keep=name) noprint;
   run;
   %if (&syserr gt 4) %then %do;
      %put %mo_note Error encountered retrieving contents of the %upcase(&library_FROM..comm_numeric_attribute_defaults) table (syserr=&syserr).;
      %let &_rc = 100;
      %goto ERROR_EXIT;
   %end;

   proc sql noprint;
      select count(*) into :has_cell_code_FROM
      from work.hasCellCodeFrom(keep=name)
      where (upcase(name) = "CELL_CODE");
   quit;
   %if (&sqlrc gt 4) %then %do;
      %let &_rc = 107;
      %put %mo_error SQL error encountered obtaining information about %upcase(&library_FROM..comm_numeric_attribute_defaults) (sqlrc=&sqlrc).;
      %goto ERROR_EXIT;
   %end;

   %if ((&uses_cell_code_FROM eq 1) and (&has_cell_code_FROM eq 0)) %then %do;
      proc sql noprint;
         create table work.defaults_base_temp_prelim as
         select *
         from &library_FROM..comm_numeric_attribute_defaults
         where (input_data_id = &input_data_id_FROM);

         create table work.defaults_base_temp as
         select a.*, d.comm_attribute_value as cell_code
         from work.defaults_base_temp_prelim as a,
              &library_FROM..comm_character_attribute_values as b,
              &library_FROM..comm_character_attributes as c,
              &library_FROM..comm_character_distinct_values as d
         where (a.campaign_id = b.campaign_id) and
               (a.communication_id = b.communication_id) and
               (b.comm_char_attribute_id = c.comm_char_attribute_id) and
               (lowcase(c.comm_attribute_name) = "cell_cd") and
               (b.comm_attribute_distinct_value_id = d.comm_attribute_distinct_value_id);

         create table work.values_base_temp_prelim as
         select *
         from &library_FROM..comm_numeric_attribute_values
         where (scenario_id = &scenario_id_FROM);

         create table work.values_base_temp as
         select a.*, d.comm_attribute_value as cell_code
         from work.values_base_temp_prelim as a,
              &library_FROM..comm_character_attribute_values as b,
              &library_FROM..comm_character_attributes as c,
              &library_FROM..comm_character_distinct_values as d
         where (a.campaign_id = b.campaign_id) and
               (a.communication_id = b.communication_id) and
               (b.comm_char_attribute_id = c.comm_char_attribute_id) and
               (lowcase(c.comm_attribute_name) = "cell_cd") and
               (b.comm_attribute_distinct_value_id = d.comm_attribute_distinct_value_id);
      quit;
      %if (&sqlrc gt 4) %then %do;
         %let &_rc = 107;
         %put %mo_error Error occurred while creating work.values_base_temp (sqlrc=&sqlrc, sqlob =&sqlobs).;
         %goto ERROR_EXIT;
      %end;

      %let has_cell_code_FROM = 1;
   %end;
   %else %do;
      data work.defaults_base_temp;
         set &library_FROM..comm_numeric_attribute_defaults;
         if (input_data_id = &input_data_id_FROM);
      run;
      %if (&syserr gt 4) %then %do;
         %let &_rc = 100;
         %put %mo_error Error occurred while creating work.defaults_base_temp (sqlrc=&sqlrc, sqlob =&sqlobs).;
         %goto ERROR_EXIT;
      %end;

      data work.values_base_temp;
         set &library_FROM..comm_numeric_attribute_values;
         if (scenario_id = &scenario_id_FROM);
      run;
      %if (&syserr gt 4) %then %do;
         %let &_rc = 100;
         %put %mo_error Error occurred while creating work.values_base_temp (sqlrc=&sqlrc, sqlob =&sqlobs).;
         %goto ERROR_EXIT;
      %end;
   %end;

   proc sql noprint;
      create table work.defaults_base as
      select a.comm_attribute_value_id, 
             upcase(b.campaign_code) as campaign_code, 
             upcase(c.communication_code) as communication_code, 
             %if (&has_cell_code_FROM eq 1) %then %do;
                upcase(a.cell_code) as cell_code,
             %end;
             upcase(d.comm_attribute_name) as comm_attribute_name,
             a.comm_attribute_value
      from work.defaults_base_temp as a, 
           &library_FROM..campaigns as b,
           &library_FROM..communications as c, 
           &library_FROM..comm_numeric_attributes as d
      where (a.input_data_id = &input_data_id_FROM) and 
            (a.campaign_id = b.campaign_id) and 
            (a.communication_id = c.communication_id) and
            (a.comm_attribute_id = d.comm_attribute_id);

      create table work.values_base as
      select a.comm_attribute_value_id, 
             upcase(b.campaign_code) as campaign_code, 
             upcase(c.communication_code) as communication_code, 
             %if (&has_cell_code_FROM eq 1) %then %do;
                upcase(a.cell_code) as cell_code,
             %end;
             upcase(d.comm_attribute_name) as comm_attribute_name,
             a.comm_attribute_value
      from work.values_base_temp as a, 
           &library_FROM..campaigns as b,
           &library_FROM..communications as c, 
           &library_FROM..comm_numeric_attributes as d
      where (a.input_data_id = &input_data_id_FROM) and 
            (a.campaign_id = b.campaign_id) and 
            (a.communication_id = c.communication_id) and
            (a.comm_attribute_id = d.comm_attribute_id);
   quit;
   %if (&sqlrc gt 4) %then %do;
      %let &_rc = 107;
      %put %mo_error Error occurred creating work.values_base (sqlrc=&sqlrc).;
   %end;

   proc sort data=work.values_base;
      by campaign_code communication_code 
         %if (&has_cell_code_FROM eq 1) %then %do;
            cell_code
         %end;
         comm_attribute_name comm_attribute_value;
   run;
   %if (&syserr gt 4) %then %do;
      %let &_rc = 100;
      %put %mo_error Error occurred sorting work.values_base (syserr=&syserr).;
   %end;


   %*-------------------------;
   %* Create comparison table ;
   %* from TO library table   ;
   %*-------------------------;
   %local cell_cd_column_status_TO uses_cell_code_TO;
   %let cell_cd_column_status_TO =;
   %let uses_cell_code_TO = 0;
   proc sql noprint;
      select cell_cd_column_status into :cell_cd_column_status_TO
      from &library_TO..input_data
      where (input_data_id = &input_data_id_TO);
   quit;
   %if ((&sqlrc gt 4) or (&sqlobs eq 0)) %then %do;
      %let &_rc = 100;
      %put %mo_error Error occurred while reading &library_TO..input_data (sqlrc=&sqlrc, sqlob =&sqlobs).;
      %goto ERROR_EXIT;
   %end;
   %let cell_cd_column_status_TO = %trim(%left(&cell_cd_column_status_TO));

   %if ((&cell_cd_column_status_TO eq &_CELL_CD_UNIQUE_) or
        (&cell_cd_column_status_TO eq &_CELL_CD_UNIQUE_UNNEEDED_)) %then %do;
      %let uses_cell_code_TO = 1;
   %end;

   %let has_cell_code_TO = 0;
   proc contents data=&library_TO..comm_numeric_attribute_defaults out=work.hasCellCodeTo(keep=name) noprint;
   run;
   %if (&syserr gt 4) %then %do;
      %put %mo_note Error encountered retrieving contents of the %upcase(&library_TO..comm_numeric_attribute_defaults) table (syserr=&syserr).;
      %let &_rc = 100;
      %goto ERROR_EXIT;
   %end;

   proc sql noprint;
      select count(*) into :has_cell_code_TO
      from work.hasCellCodeTo(keep=name)
      where (upcase(name) = "CELL_CODE");
   quit;
   %if (&sqlrc gt 4) %then %do;
      %let &_rc = 107;
      %put %mo_error SQL error encountered obtaining information about %upcase(&library_TO..comm_numeric_attribute_defaults) (sqlrc=&sqlrc).;
      %goto ERROR_EXIT;
   %end;

   %if ((&uses_cell_code_TO eq 1) and (&has_cell_code_TO eq 0)) %then %do;
      proc sql noprint;
         create table work.defaults_comp_temp_prelim as
         select *
         from &library_TO..comm_numeric_attribute_defaults
         where (input_data_id = &input_data_id_TO);

         create table work.defaults_comp_temp as
         select a.*, d.comm_attribute_value as cell_code
         from work.defaults_comp_temp_prelim as a,
              &library_TO..comm_character_attribute_values as b,
              &library_TO..comm_character_attributes as c,
              &library_TO..comm_character_distinct_values as d
         where (a.campaign_id = b.campaign_id) and
               (a.communication_id = b.communication_id) and
               (b.comm_char_attribute_id = c.comm_char_attribute_id) and
               (lowcase(c.comm_attribute_name) = "cell_cd") and
               (b.comm_attribute_distinct_value_id = d.comm_attribute_distinct_value_id);
      quit;
      %if (&sqlrc gt 4) %then %do;
         %let &_rc = 107;
         %put %mo_error Error occurred while creating work.defaults_comp_temp (sqlrc=&sqlrc, sqlob =&sqlobs).;
         %goto ERROR_EXIT;
      %end;

      %let has_cell_code_TO = 1;
   %end;
   %else %do;
      data work.defaults_comp_temp;
         set &library_TO..comm_numeric_attribute_defaults;
         if (input_data_id = &input_data_id_TO);
      run;
      %if (&syserr gt 4) %then %do;
         %let &_rc = 100;
         %put %mo_error Error occurred while creating work.defaults_comp_temp (sqlrc=&sqlrc, sqlob =&sqlobs).;
         %goto ERROR_EXIT;
      %end;
   %end;

   proc sql noprint;
      create table work.defaults_comp as
      select a.comm_attribute_value_id, 
             upcase(b.campaign_code) as campaign_code, 
             upcase(c.communication_code) as communication_code, 
             %if (&has_cell_code_TO eq 1) %then %do;
                upcase(a.cell_code) as cell_code,
             %end;
             upcase(d.comm_attribute_name) as comm_attribute_name,
             a.comm_attribute_value
      from work.defaults_comp_temp as a, 
           &library_TO..campaigns as b,
           &library_TO..communications as c, 
           &library_TO..comm_numeric_attributes as d
      where (a.input_data_id = &input_data_id_TO) and 
            (a.campaign_id = b.campaign_id) and 
            (a.communication_id = c.communication_id) and
            (a.comm_attribute_id = d.comm_attribute_id);
   quit;
   %if (&sqlrc gt 4) %then %do;
      %let &_rc = 107;
      %put %mo_error Error occurred creating work.defaults_comp (sqlrc=&sqlrc).;
   %end;

   proc sort data=work.defaults_comp;
      by campaign_code communication_code 
         %if (&has_cell_code_TO eq 1) %then %do;
            cell_code
         %end;
         comm_attribute_name comm_attribute_value;
   run;
   %if (&syserr gt 4) %then %do;
      %let &_rc = 100;
      %put %mo_error Error occurred sorting work.defaults_comp (syserr=&syserr).;
   %end;


   %*-------------------------------------------;
   %* Compare the two tables and capture        ;
   %* the differences into the macro variables. ;
   %*-------------------------------------------;
   data work.values_merged;
      length status $8.;
      merge work.values_base (in=a rename=(comm_attribute_value_id=id_FROM)) 
            work.defaults_comp (in=b rename=(comm_attribute_value_id=id_TO));
      by campaign_code communication_code 
         %if ((&has_cell_code_FROM eq 1) and (&has_cell_code_TO eq 1)) %then %do;
            cell_code
         %end;
         comm_attribute_name comm_attribute_value;

      if (a and b) then status='MATCH';
      else if (b and not a) then status='ADD';
      else if (a and not b) then status='DELETE';
   run;
   %if (&syserr gt 4) %then %do;
      %let &_rc = 100;
      %put %mo_error Error occurred merging work.values_base to work.defaults_comp (syserr=&syserr).;
   %end;
   
   proc sql noprint;
      select id_FROM, id_TO 
      into :&_attrs_to_swap_FROM separated by ' ',
           :&_attrs_to_swap_TO separated by ' '
      from work.values_merged
      where (status eq 'MATCH')
      order by id_FROM;
   quit;
   %if (&sqlrc gt 4) %then %do;
      %let &_rc = 100;
      %put %mo_error Error occurred identifying _attrs_to_swap (sqlrc=&sqlrc).;
   %end;

   proc sql noprint;
      select id_TO into :&_dflts_to_add separated by ' '
      from work.values_merged
      where (status eq 'ADD')
      order by id_TO;
   quit;
   %if (&sqlrc gt 4) %then %do;
      %let &_rc = 100;
      %put %mo_error Error occurred identifying _dflts_to_add (sqlrc=&sqlrc).;
   %end;

   proc sql noprint;
      select id_FROM into :&_vals_to_delete separated by ' '
      from work.values_merged
      where (status eq 'DELETE')
      order by id_FROM;
   quit;
   %if (&sqlrc gt 4) %then %do;
      %let &_rc = 100;
      %put %mo_error Error occurred identifying _vals_to_delete (sqlrc=&sqlrc).;
   %end;


   %* Post-process the list of comm numeric attribute values to be deleted. ;
   %if (&&&_vals_to_delete ne %str()) %then %do; 

      %* If a comm numeric attribute value has been overridden from its ;
      %* corresponding comm numeric attribute default, do not delete    ;
      %* its comm_numeric_attribute_value row and do not add the        ;
      %* corresponding TO comm_numeric_attribute_default row.           ;   
      %local do_not_delete_ids_FROM do_not_add_ids_TO;
      %let do_not_delete_ids_FROM =;
      %let do_not_add_ids_TO =;

      proc sql noprint;
         create table work.do_not_delete_ids as
         select a.id_FROM, a.campaign_code, a.communication_code, 
         %if ((&has_cell_code_FROM eq 1) and (&has_cell_code_TO eq 1)) %then %do;
            a.cell_code,
         %end;
         a.comm_attribute_name
         from work.values_merged as a, work.defaults_base as b
         where (a.status = "DELETE") and 
               (a.campaign_code = b.campaign_code) and
               (a.communication_code = b.communication_code) and 
               %if ((&has_cell_code_FROM eq 1) and (&has_cell_code_TO eq 1)) %then %do;
                  (a.cell_code = b.cell_code) and
               %end;
               (a.comm_attribute_name = b.comm_attribute_name) and
               (a.comm_attribute_value ~= b.comm_attribute_value);
      quit;
      %if (&sqlrc gt 4) %then %do;
         %let &_rc = 100;
         %put %mo_error Error occurred determining overridden comm numeric attribute values (1) (sqlrc=&sqlrc).;
      %end;

      %if (&sqlobs ne 0) %then %do;
         proc sql noprint;
            select id_FROM into :do_not_delete_ids_FROM separated by ' '
            from work.do_not_delete_ids;

            create table work.do_not_add_ids as
            select a.id_TO
            from work.values_merged as a, work.do_not_delete_ids as b
            where (a.status = "ADD") and 
                  (a.campaign_code = b.campaign_code) and
                  (a.communication_code = b.communication_code) and 
                  %if ((&has_cell_code_FROM eq 1) and (&has_cell_code_TO eq 1)) %then %do;
                     (a.cell_code = b.cell_code)
                  %end;
                  (a.comm_attribute_name = b.comm_attribute_name);
         quit;
         %if (&sqlrc gt 4) %then %do;
            %let &_rc = 100;
            %put %mo_error Error occurred determining overridden comm numeric attribute values (2) (sqlrc=&sqlrc).;
         %end;
         %if (&sqlobs ne 0) %then %do;
            proc sql noprint;
               select id_TO into :do_not_add_ids_TO separated by ' '
               from work.do_not_add_ids;
            quit;
            %if (&sqlrc gt 4) %then %do;
               %let &_rc = 100;
               %put %mo_error Error occurred determining overridden comm numeric attribute values (3) (sqlrc=&sqlrc).;
            %end;
         %end;
      %end;

      %if (&do_not_delete_ids_FROM ne %str()) %then %do;
         /* Rebuild list of IDs, excluding those found above. */
         %local temp_vals_to_delete;
         %let temp_vals_to_delete = &&&_vals_to_delete; 
         %let &_vals_to_delete = ; 

         %let numNoDel = %mo_word_count(&do_not_delete_ids_FROM);
         %do noDelNm = 1 %to &numNoDel;
            %let noDelID = %scan(&do_not_delete_ids_FROM, &noDelNm);

            %let numDel = %mo_word_count(&temp_vals_to_delete);
            %do delNm = 1 %to &numDel;
               %let delID = %scan(&temp_vals_to_delete, &delNm);
               %if (&delID ne &noDelID) %then %do;
                  %let &_vals_to_delete = &&&_vals_to_delete &delID; 
               %end;
            %end;
         %end;
      %end;

      %if (&do_not_add_ids_TO ne %str()) %then %do;
         /* Rebuild list of IDs, excluding those found above. */
         %local temp_dflts_to_add;
         %let temp_dflts_to_add = &&&_dflts_to_add; 
         %let &_dflts_to_add = ; 

         %let numNoAdd = %mo_word_count(&do_not_add_ids_TO);
         %do noAddNm = 1 %to &numNoAdd;
            %let noAddID = %scan(&do_not_add_ids_TO, &noAddNm);

            %let numAdd = %mo_word_count(&temp_dflts_to_add);
            %do addNm = 1 %to &numAdd;
               %let addID = %scan(&temp_dflts_to_add, &addNm);
               %if (&addID ne &noAddID) %then %do;
                  %let &_dflts_to_add = &&&_dflts_to_add &addID; 
               %end;
            %end;
         %end;
      %end;

      %* Re-map ID if mapping table provided. ;
      %if (%nrbquote(&id_FROM_mapping_tbl) ne %str()) %then %do;
         %local temp_vals_to_delete;
         %let temp_vals_to_delete = &&&_vals_to_delete; 
         %let &_vals_to_delete = ; 

         proc sql noprint;
            select &id_TO_mapping_col into :&_vals_to_delete separated by ' '
            from &id_FROM_mapping_tbl
            where (&id_FROM_mapping_col in (&temp_vals_to_delete)); 
         quit;
         %if (&sqlrc gt 4) %then %do;
            %let &_rc = 100;
            %put %mo_error Error occurred identifying _vals_to_delete (sqlrc=&sqlrc).;
         %end;
      %end;
   %end;

   %if (&local_debug eq 1) %then %do;
      %mo_no_notes(on)
      %put;
      %put %mo_note _dflts_to_add: %nrbquote(&&&_dflts_to_add);
      %put %mo_note _vals_to_delete: %nrbquote(&&&_vals_to_delete);
      %put %mo_note _attrs_to_swap_FROM: %nrbquote(&&&_attrs_to_swap_FROM);
      %put %mo_note _attrs_to_swap_TO: %nrbquote(&&&_attrs_to_swap_TO);
      %mo_no_notes(off)
   %end;


   %*--------------------------------;
   %* Exit processing                ;
   %*--------------------------------;
   %* Successful completion;
   %mo_no_notes(on)
   %put;
   %put %mo_note Successfully compared comm numeric attribute values from library &library_FROM (input_data_id_FROM = &input_data_id_FROM) to library &library_TO (input_data_id_TO = &input_data_id_TO).;
   %goto EXIT;


   %ERROR_EXIT:
   %put;
   %put %mo_error Error encountered during comparison of comm numeric attribute values from library &library_FROM (input_data_id_FROM = &input_data_id_FROM) to library &library_TO (input_data_id_TO = &input_data_id_TO).;


   %EXIT:
   %mo_no_notes(off)

%mend mo_compare_comm_num_attr_dflts;
