/******************************************************************************/
/* Copyright (c) 2005-2011, SAS Institute Inc., Cary, NC, USA,                */
/*                       All Rights Reserved.                                 */
/*                                                                            */
/* Name:    mo_copy_project                                                   */
/*                                                                            */
/* Purpose: Copy a MO project definition and all related scenarios            */
/*                                                                            */
/* Input:   original_project_id                                               */
/*          new_project_name                                                  */
/*          new_project_description (optional)                                */
/*          new_input_data_id (optional)                                      */
/*          user_id                                                           */
/*          datetime_created                                                  */
/*          save_as = 0 (no - default), 1 (yes)                               */
/*          refresh = 0 (no - default), 1 (yes)                               */
/*          lock = LOCK/NOLOCK (set NOLOCK if calling from another macro      */
/*                 that has already obtained a lock)                          */
/*          new_directory = 1 (create new directory), = (do not)              */
/*                                                                            */
/* Output:  &_rc = return code                                                */
/*          &_id = unique identifier                                          */
/*                                                                            */
/* Usage:   set lock=NOLOCK if calling from another macro that has already    */
/*          obtained a lock                                                   */
/*                                                                            */
/*                                                                            */
/* HISTORY:                                                                   */
/* 03Mar04 S0237007 Creation (dwh)                                            */
/* 14Jul04 S0259026 Remove check for project and scenario description (dwh)   */
/* 14Jul04 S0258942 Return &_id with new project_id (dwh)                     */
/* 05Aug04 S0263527 Do not copy description if none provided (dwh)            */
/* 17Aug04 S0256229 Implement SAVE AS functionality (dwh)                     */
/*                                                                            */
/* 4.3:                                                                       */
/* 08Mar05 S0273454 Refresh project_input definitions (dwh)                   */
/* 16May05 S0296213 Allow numeric project_name and scenario_name (dwh)        */
/* 23Jun05 S0304001 Set datetime_modified for scheduled refresh (dwh)         */
/* 25Jun05 S0307230 Preserve status of imported scenario (dwh)                */
/*                                                                            */
/* 4.4:                                                                       */
/* 13Dec08 S0308904 Do not update imported scenario for refresh (dwh)         */
/* 19May06 S0360168 Cease copy across project_input if compare fails (dwh)    */
/*                                                                            */
/* 5.3:                                                                       */
/* 29Jul08 S0483854 MOMA support - Capture old and new scenario IDs into      */
/*         output parameters in support of restoring original IDs after a     */
/*         refresh. (KEC)                                                     */
/* 02Dec08 S0483854 Eliminate home-grown moutil catalog functions CREATE_     */
/*         DIRECTORY and RENAME_DIRECTORY.  Replace with corresponding SAS    */
/*         built-in functions DCREATE and RENAME.                             */
/* 09Jan08 S0555659 Pass datetime_refresh_started as a parameter to the       */
/*         mo_copy_project->mo_copy_scenario->mo_refresh_update_scen_tables   */
/*         call chain so that these macros do not (obscurely) refer back to   */
/*         a macro variable much higher up in the call chain (KEC)            */
/* 04Feb09 S0560001 Add "duplicate" field to input_data, projects, scenarios  */
/*         tables. The field is set to 1 in the new rows that are created by  */
/*         a refresh for the duration of the refresh (while the original      */
/*         input_data, projects, and scenarios tables rows still exist). Once */
/*         the refresh is complete and the original rows have been deleted,   */
/*         the "duplicate" field is reset to 0 in the new rows. (KEC)         */
/******************************************************************************/

%macro mo_copy_project(original_project_id       =
                      ,new_project_name          =
                      ,new_project_description   =
                      ,new_input_data_id         =
                      ,user_id                   =
                      ,datetime_created          =
                      ,save_as                   = 0
                      ,refresh                   = 0
                      ,datetime_refresh_started  =
                      ,lock                      = LOCK
                      ,new_directory             = 1
                      ,_rc                       =
                      ,_id                       =
                      ,_orig_scen_id_list        = 
                      ,_new_scen_id_list         =
                      );


   %local macname;
   %let macname=&sysmacroname;
   %let macname=%lowcase(&macname);

   %if %bquote(&_rc) eq %str() %then %do;
      %local rc;
      %let _rc = rc;
   %end;
   %else %global &_rc;
   %let &_rc = 0;

   %if %bquote(&_id) eq %str() %then %do;
      %local id;
      %let _id = id;
   %end;
   %else %global &_id;
   %let &_id =;

   %if %bquote(&_orig_scen_id_list) eq %str() %then %do;
      %local orig_scen_id_list;
      %let _orig_scen_id_list = orig_scen_id_list;
   %end;
   %else %global &_orig_scen_id_list;
   %let &_orig_scen_id_list =;

   %if %bquote(&_new_scen_id_list) eq %str() %then %do;
      %local new_scen_id_list;
      %let _new_scen_id_list = new_scen_id_list;
   %end;
   %else %global &_new_scen_id_list;
   %let &_new_scen_id_list =;

   %if (%bquote(&original_project_id) eq %str()) %then %do;
      %put %mo_error No original_project_id was specified.;
      %let &_rc = 102;
      %goto ERROR_EXIT;
   %end;
   %if (%bquote(X&new_project_name) eq %str(X)) %then %do;
      %put %mo_error No new_project_name was specified.;
      %let &_rc = 102;
      %goto ERROR_EXIT;
   %end;
   %if (%bquote(&datetime_created) eq %str()) %then %do;
      %put %mo_error No datetime_created was specified.;
      %let &_rc = 102;
      %goto ERROR_EXIT;
   %end;

   %mo_constants

   %if %bquote(&refresh) eq %str() %then %let refresh = 0;

   %if (&refresh ne 0) %then %do;
      %if (%bquote(&datetime_refresh_started) eq %str()) %then %do;
         %put %mo_error Refresh flag set, but no datetime_refresh_started was specified.;
         %let &_rc = 102;
         %goto ERROR_EXIT;
      %end;
   %end;

   %*--------------------------------;
   %* Obtain lock for metadata update;
   %*--------------------------------;
   %if (%upcase(&lock) eq LOCK) %then %do;
      %mo_lock_dataset(data=mometa.uid,_rc=lckrc)
      %if (&lckrc ne 0) %then %do;
         %put %mo_error Error obtaining lock of MOMETA.UID.;
         %let &_rc = 106;
         %goto ERROR_EXIT;
      %end;
   %end;


   %*--------------------------------;
   %* Retrieve metadata for original ;
   %*--------------------------------;
   %mo_no_notes(off)
   %local original_project_name original_user_id original_username;
   data _null_;
      set mometa.projects;
      if (project_id eq &original_project_id) then do;
         call symput('original_project_name',strip(project_name));
         call symput('original_user_id',user_id);
      end;
   run;
   %if (%bquote(X&original_project_name) eq %str(X)) %then %do;
      %let &_rc = 101;
      %put %mo_error Invalid metadata detected.;
      %put %mo_error Project_name not identified for project_id = &original_project_id..;
      %goto ERROR_EXIT;
   %end;
   
   data _null_;
      set mometa.users;
      if (user_id eq &original_user_id) then do;
         call symput('original_username',strip(username));
      end;
   run;
   %if (%bquote(&original_username) eq %str()) %then %do;
      %let &_rc = 101;
      %put %mo_error Invalid metadata detected.;
      %put %mo_error Username not identified for user_id = &original_user_id..;
      %goto ERROR_EXIT;
   %end;
   %if (&syserr gt 4) %then %do;
      %let &_rc = 100;
      %put %mo_error Error occurred while reading metadata (syserr=&syserr).;
      %goto ERROR_EXIT;
   %end;
   %mo_no_notes(on)

   %*--------------------------------;
   %* Retrieve metadata for new      ;
   %*--------------------------------;
   %mo_no_notes(off)
   %local username;
   data _null_;
      set mometa.users;
      if (user_id eq &user_id) then do;
         call symput('username',strip(username));
      end;
   run;
   %if (%bquote(&username) eq %str()) %then %do;
      %let &_rc = 101;
      %put %mo_error Invalid metadata detected.;
      %put %mo_error Username not identified for user_id = &user_id..;
      %goto ERROR_EXIT;
   %end;
   %if (&syserr gt 4) %then %do;
      %let &_rc = 100;
      %put %mo_error Error occurred while reading metadata (syserr=&syserr).;
      %goto ERROR_EXIT;
   %end;
   %mo_no_notes(on)


   %*--------------------------------;
   %* Generate new project ID        ;
   %*--------------------------------;
   %mo_gen_uid(r_macro=new_project_id, data=mometa.projects, id_var=project_id, _rc=uidrc)
   %if (&uidrc ne 0) %then %do;
      %let &_rc = 108;
      %put %mo_error Could not generate a unique project_id.;
      %goto ERROR_EXIT;
   %end;
   %let &_id = &new_project_id;


   %*--------------------------------;
   %* Create new project directory   ;
   %*--------------------------------;
   %if (&new_directory) %then %do;
      %let parent_dir=%mo_pathname(&moproj.&separator.&username.&separator);
      %let new_dir=%mo_pathname(&new_project_name);

      %let dir = %mo_pathname(&parent_dir.&new_dir);
      %if (%sysfunc(fileexist(&dir))) %then %do;
         %mo_delete_file_item(item=&dir,_rc=dirrc)
         %if (&dirrc ne 0) %then %do;
            %let &_rc = 115;
            %put %mo_error I/O error encountered deleting &dir for user = %upcase(&username) (dirrc=&dirrc).;
            %goto ERROR_EXIT;
         %end;
      %end;

      %local cdrc;
      %let cdrc=;
      data _null_;
         newdir = dcreate("&new_dir", "&parent_dir");
         call symput("cdrc", trim(left(newdir)));
      run;
      %if (%bquote(&cdrc) eq %str()) %then %do;
         %let &_rc = 105;
         %put %mo_error I/O error encountered creating &dir for user = %upcase(&username).;
         %goto ERROR_EXIT;
      %end;
   %end;


   %*--------------------------------;
   %* Update projects table          ;
   %*--------------------------------;
   %mo_no_notes(off)
   data work.projects;
      set mometa.projects;
      where (project_id eq &original_project_id);

      user_id = &user_id;
      project_id  = &new_project_id;
      project_name = "&new_project_name";
      project_description = "&new_project_description";
      %if (%bquote(&new_input_data_id) ne %str()) %then %do;
         input_data_id = &new_input_data_id;
      %end;
      datetime_created = &datetime_created;
      %if (&refresh) %then %do;
         duplicate = 1;
      %end;
      %else %do;
         duplicate = 0;
      %end;
   run;
   %if (&syserr gt 4) %then %do;
      %let &_rc = 100;
      %put %mo_error I/O error encountered while updating WORK.PROJECTS table (syserr=&syserr).;
      %goto ERROR_EXIT;
   %end;
   %mo_no_notes(on)

   %mo_record_insert(metadata_table = mometa.projects
                    ,update_table   = work.projects
                    ,by             = project_id
                    ,_rc            = updrc);
   %if (&updrc ne 0) %then %do;
      %let &_rc = 103;
      %goto ERROR_EXIT;
   %end;


   %*--------------------------------;
   %* Copy scenarios                 ;
   %*--------------------------------;
   %mo_no_notes(off)
   %local scenario_ids;
   proc sql noprint;
      select scenario_id into :scenario_ids
      separated by ' '
      from mometa.scenarios
      where project_id eq &original_project_id;
   quit;
   %if (&sqlrc gt 4) %then %do;
      %let &_rc = 200;
      %put %mo_error SQL error encountered while retrieving scenario_ids (sqlrc=&sqlrc).;
      %goto ERROR_EXIT;
   %end;
   %let &_orig_scen_id_list = &scenario_ids;

   %let have_done_the_compare = 0;
   %do g=1 %to %mo_word_count(&scenario_ids);
      %let scenario_id=%scan(&scenario_ids,&g);
      %local scenario_name scenario_description;
      data _null_;
         set mometa.scenarios;
         if (scenario_id eq &scenario_id) then do;
            call symput('scenario_name',strip(scenario_name));
            call symput('scenario_description',scenario_description);
         end;
      run;
      %if (%bquote(X&scenario_name) eq %str(X)) %then %do;
         %let &_rc = 200;
         %put %mo_error Invalid metadata detected.;
         %put %mo_error Scenario_name not identified for scenario_id = &scenario_id..;
         %goto ERROR_EXIT;
      %end;
      %if (&syserr gt 4) %then %do;
         %let &_rc = 200;
         %put %mo_error Error occurred while reading metadata (syserr=&syserr).;
         %goto ERROR_EXIT;
      %end;

      /* If doing a refresh, get the original datetime_created for the scenario, since
         this is a dummy copy_scenario operation. */
      %if (&refresh) %then %do;

         proc sql noprint;
            select put(datetime_created,best12.)
            into  :datetime_created
            from mometa.scenarios
            where (scenario_id eq &scenario_id);
         quit;
         %if (&sqlrc gt 4) %then %do;
            %let &_rc = 200;
            %put %mo_error SQL error encountered while retrieving scenario datetime created (sqlrc=&sqlrc).;
            %goto ERROR_EXIT;
         %end;
         %let datetime_created=%trim(%left(&datetime_created));
      %end;
      
      %let first_non_imported_scenario = 0;
      %if &have_done_the_compare = 0 %then %do;
         proc sql noprint;
            select count(*)
               into :first_non_imported_scenario
               from mometa.scenarios
               where ((scenario_id eq &scenario_id) and (status ne &_IMPORTED_));
         quit;
         %if &first_non_imported_scenario = 1 %then %let have_done_the_compare = 1;
      %end;

      %mo_copy_scenario(original_scenario_id          = &scenario_id
                       ,new_scenario_name             = &scenario_name
                       ,new_scenario_description      = &scenario_description
                       ,project_id                    = &new_project_id
                       ,datetime_created              = &datetime_created
                       ,refresh                       = &refresh
                       ,datetime_refresh_started      = &datetime_refresh_started
                       ,lock                          = NOLOCK
                       ,new_directory                 = &new_directory
                       %if (&first_non_imported_scenario = 1) %then %do;
                          ,compare                       = COMPARE
                       %end;
                       %else %do;
                          ,compare                       = NOCOMPARE
                       %end;
                       ,_rc                           = csrc
                       ,_id                           = scid
                       );
      %let &_new_scen_id_list = &&&_new_scen_id_list &scid;
      %if (&csrc ne 0) %then %do;
         %put %mo_error Error encountered copying scenario_id = &scenario_id from project_id = &new_project_id for user = %upcase(&username).;
         %let &_rc = 207;
         %if (&refresh) %then %goto ERROR_EXIT;
      %end;
   %end;

   %if (&refresh) %then %do;
      /* Update the datetime_modified for each of the new scenarios */
      %mo_no_notes(off)
      %local scenario_ids datetime_modified;
      proc sql noprint;
         select put(datetime_refresh_started,best12.)
            into :datetime_modified
            from mometa.input_data
            where input_data_id eq &new_input_data_id;
         select scenario_id 
            into :scenario_ids separated by ' '
            from mometa.scenarios
            where project_id eq &new_project_id;
      quit;
      %if (&sqlrc gt 4) %then %do;
         %let &_rc = 200;
         %put %mo_error SQL error encountered while retrieving datetime_modified and scenario_ids (sqlrc=&sqlrc).;
      %end;
      %let datetime_modified=%trim(%left(&datetime_modified));

      %do g=1 %to %mo_word_count(&scenario_ids);
         %let scenario_id=%scan(&scenario_ids,&g);

         %local stat;
         proc sql noprint;
            select status into :stat
            from mometa.scenarios
            where scenario_id eq &scenario_id;
         quit;
         %let stat=%trim(%left(&stat));

         %mo_edit_objective_parameters(scenario_id   = &scenario_id
                                       %if (&stat ne &_IMPORTED_) %then %do;
                                          ,datetime_modified = &datetime_modified
                                       %end;
                                       %if (&stat eq &_IMPORTED_) %then %do;
                                          ,status = &_IMPORTED_
                                       %end;
                                          ,_rc               = eoprc);
         %if (&eoprc ne 0) %then %do;
            %let &_rc = 200;
            %put %mo_error Error occurred updating datetime_modified for scenario_id = &scenario_id..;
            %goto ERROR_EXIT;
         %end;
     %end;
   %end;


   %*--------------------------------;
   %* Copy mo_opt_config.sas         ;
   %*--------------------------------;
   %if (&new_directory) %then %do;
      %let optcfg1=%mo_pathname(&moproj.&separator.&original_username.&separator.&original_project_name.&separator.mo_opt_config.sas);
      %let optcfg2=%mo_pathname(&moproj.&separator.&username.&separator.&new_project_name.&separator.mo_opt_config.sas);

      %if (%sysfunc(fileexist(&optcfg1)) eq 1) %then %do;
         data _null_;
            file "&optcfg2" lrecl=256;
            infile "&optcfg1";
            input;
            put _INFILE_;
         run;
         %if (&syserr gt 4) %then %do;
            %let &_rc = 200;
            %put %mo_error Error encountered while copying mo_opt_config.sas (syserr=&syserr).;
         %end;
      %end;
   %end;


   %*--------------------------------;
   %* Exit processing                ;
   %*--------------------------------;
   %* Successful completion;
   %put;
   %if (&&&_rc eq 0) %then
      %put %mo_note Successfully copied project = &original_project_name to project = &new_project_name with project_id = &new_project_id for user = %upcase(&username).;
   %else
      %put %mo_warning Abnormal condition encountered while copying project = &original_project_name to project = &new_project_name with project_id = &new_project_id for user = %upcase(&username).;
   %put;
   %goto EXIT;


   %ERROR_EXIT:
   %put;
   %put %mo_error Error encountered while copying project = &original_project_id (rc=&&&_rc).;


   %EXIT:
   %if (%upcase(&lock) eq LOCK) %then %do;
      %mo_unlock_dataset(data=mometa.uid)
   %end;
   %symdel new_project_id;

%mend mo_copy_project;
