/************************************************************************/
/* Copyright (c) 2005-2011, SAS Institute Inc., Cary, NC, USA,          */
/*                       All Rights Reserved.                           */
/*                                                                      */
/* Name:          mo_generate_report_data                               */
/*                                                                      */
/* Purpose:       Generate report datasets                              */
/*                                                                      */
/* Input:         scenario_id                                           */
/*                                                                      */
/* Output:        &_rc = return code                                    */
/*                                                                      */
/* Return Codes:                                                        */
/*                                                                      */
/*          100   generic error                                         */
/*          103   I/O error updating a metadata table                   */
/*          107   SQL error                                             */
/*          140   summary report job failed                             */
/*          141   histogram report job failed                           */
/*          142   sensitivity report job failed                         */
/*                                                                      */
/* HISTORY:                                                             */
/* 18May04 S0237007 Creation (dwh)                                      */
/* 19Jul04 S0258952 Add reports_finished information (dwh)              */
/* 04Aug04 S0261187 Add sysparm to indicate GUI submission (dwh)        */
/* 05Aug04 S0261277 Subtract GMT offset in hours (dwh)                  */
/* 31Aug04 S0266034 Correct typo in message (dwh)                       */
/* 25Jun05 S0306697 Subtract GMT offset instead of add (dwh)            */
/*                                                                      */
/* 4.4:                                                                 */
/* 31Mar06 S0341572 Automate setting of GMT offset (dwh)                */
/************************************************************************/

%macro mo_generate_report_data(scenario_id      =
                              ,_rc              =
                              );

   %local macname;
   %let macname=&sysmacroname;
   %let macname=%lowcase(&macname);

   %if %bquote(&_rc) eq %str() %then %do;
      %local rc;
      %let _rc = rc;
   %end;
   %else %global &_rc;
   %let &_rc = 0;

   %global sysparm;

   %if (&moerror ne 0) %then %goto OPT_EXIT;

   %*---------------------------------------;
   %* Note start of report job              ;
   %*---------------------------------------;
   %mo_create_tables(lib=work, table=optimization_jobs, _rc=cudrc)
   %if (&cudrc ne 0) %then %do;
      %let &_rc = 100;
      %put %mo_error Error occurred creating WORK.OPTIMIZATION_JOBS table.;
      %goto ERROR_EXIT;
   %end;
   %mo_no_notes(off)
   proc sql;
      insert into work.optimization_jobs
      select * from mometa.optimization_jobs
      where scenario_id eq &scenario_id;

      update work.optimization_jobs
      set
         reports_finished = 0
      ;
   quit;
   %if (&sqlrc gt 4) %then %do;
      %let &_rc = 107;
      %put %mo_error SQL error encountered while updating WORK.SCENARIOS table (sqlrc=&sqlrc).;
      %goto ERROR_EXIT;
   %end;
   %mo_no_notes(on)

   %mo_record_update(metadata_table = mometa.optimization_jobs
                    ,update_table   = work.optimization_jobs
                    ,by             = scenario_id
                    ,_rc            = updrc);
   %if (&updrc ne 0) %then %do;
      %let &_rc = 103;
      %goto ERROR_EXIT;
   %end;


   %*---------------------------------------;
   %* Summary reports                       ;
   %*---------------------------------------;
   %mo_rpt_offer_summary(_rc=offrc);
   %if (&offrc ne 0) %then %do;
      %let &_rc = 140;
      %put %mo_error Summary report data creation failed (offrc=&offrc).;
   %end;

   %if (&offrc eq 0) %then %do;
      %mo_create_tables(lib=work, table=scenarios, _rc=cudrc)
      %if (&cudrc ne 0) %then %do;
         %let &_rc = 100;
         %put %mo_error Error occurred creating WORK.SCENARIOS table.;
         %goto ERROR_EXIT;
      %end;
      %mo_no_notes(off)
      proc sql;
         insert into work.scenarios
         select * from mometa.scenarios
         where scenario_id eq &scenario_id;

         update work.scenarios
         set summary_report_status=1;
      quit;
      %if (&sqlrc gt 4) %then %do;
         %let &_rc = 107;
         %put %mo_error SQL error encountered while updating WORK.SCENARIOS table (sqlrc=&sqlrc).;
         %goto ERROR_EXIT;
      %end;
      %mo_no_notes(on)

      %mo_record_update(metadata_table = mometa.scenarios
                       ,update_table   = work.scenarios
                       ,by             = scenario_id
                       ,_rc            = updrc);
      %if (&updrc ne 0) %then %do;
         %let &_rc = 103;
         %goto ERROR_EXIT;
      %end;
   %end;


   %*---------------------------------------;
   %* Note completion of report job         ;
   %*---------------------------------------;
   %let elapsed_stop = %sysfunc(datetime());
   %let elapsed = %sysevalf(&elapsed_stop-&elapsed_start);

   %mo_create_tables(lib=work, table=optimization_jobs, _rc=cudrc)
   %if (&cudrc ne 0) %then %do;
      %let &_rc = 100;
      %put %mo_error Error occurred creating WORK.OPTIMIZATION_JOBS table.;
      %goto ERROR_EXIT;
   %end;
   %mo_no_notes(off)
   proc sql;
      insert into work.optimization_jobs
      select * from mometa.optimization_jobs
      where scenario_id eq &scenario_id;

      update work.optimization_jobs
      set
      %if (%bquote(&sysparm) eq GUI) %then %do;
         datetime_reports_finished = datetime_started+&elapsed
        ,reports_finished = 1
      %end;
      %else %do;
         datetime_reports_finished = &elapsed_start - (&mo_GMT_offset) + &elapsed
        ,reports_finished = 1
      %end;
      ;
   quit;
   %if (&sqlrc gt 4) %then %do;
      %let &_rc = 107;
      %put %mo_error SQL error encountered while updating WORK.SCENARIOS table (sqlrc=&sqlrc).;
      %goto ERROR_EXIT;
   %end;
   %mo_no_notes(on)

   %mo_record_update(metadata_table = mometa.optimization_jobs
                    ,update_table   = work.optimization_jobs
                    ,by             = scenario_id
                    ,_rc            = updrc);
   %if (&updrc ne 0) %then %do;
      %let &_rc = 103;
      %goto ERROR_EXIT;
   %end;


   %*--------------------------------;
   %* Exit processing                ;
   %*--------------------------------;
   %* Successful completion;
   %put;
   %if (&&&_rc eq 0) %then
      %put %mo_note Successfully generated report data for scenario_id = &scenario_id.;
   %else
      %put %mo_warning Report data generation completed with abnormal conditions for scenario_id = &scenario_id (rc=&&&_rc).;
   %put;
   %goto EXIT;


   %ERROR_EXIT:
   %put;
   %put %mo_error Error generating report data for scenario_id = &scenario_id (rc=&&&_rc).;
   %put;
   %goto EXIT;


   %OPT_EXIT:
   %put;
   %put %mo_error Report data for scenario_id = &scenario_id not generated due to optimization problems (moerror=&moerror).;

   %EXIT:

%mend mo_generate_report_data;
