/****************************************************************************/
/* Copyright (c) 2005-2011, SAS Institute Inc., Cary, NC, USA,              */
/*                       All Rights Reserved.                               */
/*                                                                          */
/* Name:    mo_import_scenario                                              */
/*                                                                          */
/* Purpose: Import an MO scenario definition                                */
/*                                                                          */
/* Input:   project_id             (required)                               */
/*          scenario_name          (required)                               */
/*          scenario_description   (optional)                               */
/*          fq_import_file_name    (required)                               */
/*          datetime_created       (required)                               */
/*          action                 (required)                               */
/*                                                                          */
/* Output:  &_id = scenarios table ID of imported scenario                  */
/*          &_rc = return code                                              */
/*                                                                          */
/* HISTORY:                                                                 */
/* 5.4:                                                                     */
/* 10Mar10 S0639443 Creation for scenario import (KEC)                      */
/* 20Apr10 Enhancements related to support for scenario import with         */
/*         validation (KEC)                                                 */
/* 28Jun10 S0661616 Don't copy UID table from MOMETA; create new one in     */
/*         validation directory instead (KEC).                              */
/* 12Jul10 S0667368 Fix cell_code column handling for MOMA input data (KEC).*/
/* 19Jul10 S0684342 Fix generated value where clause so that recursion is   */
/*         not prevented when filters-related tables occur multiple times   */
/*         in the ID mapping chain (KEC)                                    */
/* 23Aug10 Add check to disallow importing an analytics-based scenario into */
/*         prioritization-based input data.                                 */
/*                                                                          */
/* F1MO54, F0MO541                                                          */
/* 09Jan12 S0823209 Replace use of SQL query of Dictionay Tables with use   */
/*         of SAS Proc Contents.                                            */
/*                                                                          */
/****************************************************************************/

%macro mo_import_scenario(project_id           =
                         ,scenario_name        =
                         ,scenario_description =
                         ,fq_import_file_name  =
                         ,datetime_created     =
                         ,action               =
                         ,_has_import_msg      = 
                         ,_id                  =
                         ,_rc                  = 
                         );

   %local macname;
   %let macname=&sysmacroname;
   %let macname=%lowcase(&macname);

   %if %nrbquote(&_rc) eq %str() %then %do;
      %local rc;
      %let _rc = rc;
   %end;
   %else %global &_rc;
   %let &_rc = 0;

   %if %nrbquote(&_id) eq %str() %then %do;
      %local id;
      %let _id = id;
   %end;
   %else %global &_id;
   %let &_id = 0;

   %if %nrbquote(&_has_import_msg) eq %str() %then %do;
      %local has_import_msg;
      %let _has_import_msg = has_import_msg;
   %end;
   %else %global &_has_import_msg;
   %let &_has_import_msg = 0;

   %mo_constants

   %mo_no_notes(off)

   %local local_debug;
   %let local_debug = 0;
   %if (&local_debug eq 1) %then %do;
      %mo_no_notes(on)
      %put;
      %put %mo_note project_id: %nrbquote(&project_id);
      %put %mo_note scenario_name: %nrbquote(&scenario_name);
      %put %mo_note scenario_description: %nrbquote(&scenario_description);
      %put %mo_note fq_import_file_name: %nrbquote(&fq_import_file_name);
      %put %mo_note datetime_created: %nrbquote(&datetime_created);
      %put %mo_note action: %nrbquote(&action);
      %put;
      %mo_no_notes(off)
   %end;


   %*--------------------------------;
   %* Initialization.                ;
   %*--------------------------------;
   %local dsLocked;
   %let dsLocked = 0;

   %local importParentDir importFileName;
   %let importParentDir =;
   %let importFileName =;

   %local workingDatastoreLib;
   %let workingDatastoreLib =;

   %local numRefreshMessages;
   %let numRefreshMessages = 0;

   %local tbl refTbl;
   %let tbl = 0;
   %let refTbl = 0;


   %*--------------------------------;
   %* Validate parms                 ;
   %*--------------------------------;
   %if (%nrbquote(&project_id) eq %str()) %then %do;
      %put %mo_error No project_id was specified.;
      %let &_rc = 102;
      %goto ERROR_EXIT;
   %end;

   %if (%nrbquote(&scenario_name) eq %str()) %then %do;
      %put %mo_error No scenario_name was specified.;
      %let &_rc = 102;
      %goto ERROR_EXIT;
   %end;

   %if (%nrbquote(&datetime_created) eq %str()) %then %do;
      %put %mo_error No datetime_created was specified.;
      %let &_rc = 102;
      %goto ERROR_EXIT;
   %end;

   %if ((%nrbquote(&action) ne VALIDATE) and
        (%nrbquote(&action) ne COMMIT)) %then %do;
      %put %mo_error Action of &action is incorrect. Action must be either VALIDATE or COMMIT.;
      %let &_rc = 102;
      %goto ERROR_EXIT;
   %end;

   %if (%nrbquote(&fq_import_file_name) eq %str()) %then %do;
      %put %mo_error No fully qualified import file name was specified.;
      %let &_rc = 102;
      %goto ERROR_EXIT;
   %end;
   %else %do;
      %let fq_import_file_name= %nrbquote(%mo_pathname(%nrbquote(&fq_import_file_name)));
      %if (%lowcase(%nrbquote(%qscan(%nrbquote(&fq_import_file_name), -1, %str(.)))) ne %str(spk)) %then %do;
         %let &_rc = 102;
         %put %mo_error Fully qualified import file name must have an extension of .spk as required by the call to retrieve_package.;
         %goto ERROR_EXIT;
      %end;

      %mo_check_fq_file_item(fq_file_item = %nrbquote(&fq_import_file_name)
                            ,is_directory = 0
                            ,check_exists = 1
                            ,check_parent_exists = 0 
                            ,_rc = cfirc);
      %if (&cfirc ne 0) %then %do;
         %let &_rc = 102;
         %goto ERROR_EXIT;
      %end;       

      /* Make sure that mo_validate_import_file has been called by checking   */
      /* for existence of dpo and datasets subdirectories and existence of    */
      /* files within those subdirectories.                                   */
      %local dirRef drc dirID numFiles;
      %let importFileName = %nrbquote(%qscan(%nrbquote(&fq_import_file_name), -1, %str(&separator)));
      %let importParentDir = %nrbquote(%qsubstr(%nrbquote(&fq_import_file_name), 1, %eval(%index(%nrbquote(&fq_import_file_name), %nrbquote(&separator.&importFileName)))));

      %local importDSDir;
      %let importDSDir = %nrbquote(&importParentDir.datasets);
      %if (not %sysfunc(fileexist(%nrbquote(&importDSDir)))) %then %do;
         %let &_rc = 102;
         %put %mo_error Subdirectory datasets below directory %nrbquote(&importParentDir) does not exist.;
        %put %mo_error Macro mo_validate_import_file.sas must be executed so that the datasets subdirectory is properly created and populated.;
         %goto ERROR_EXIT;
      %end;
      %else %do;
         %let dirRef = dsDir;
         %let drc = %sysfunc(filename(dirRef, %nrbquote(&importDSDir)));
         %if (&drc ne 0) %then %do;
            %let &_rc = 105;
            %put %mo_warning Error encountered using FILENAME to assign a file reference to directory %nrbquote(&importDSDir) (rc=&drc).;
            %goto ERROR_EXIT;
         %end;

         %let dirID = %sysfunc(dopen(&dirRef));
         %if (&dirID eq 0) %then %do;
            %let &_rc = 105;
            %put %mo_warning Error encountered using DOPEN to open directory %nrbquote(&importDSDir).;
            %goto ERROR_EXIT;
         %end;
         %let numFiles = %sysfunc(dnum(&dirID));
         %let drc = %sysfunc(dclose(&dirID));

         %if (&numFiles eq 0) %then %do;
            %let &_rc = 102;
            %put %mo_error Subdirectory datasets below directory %nrbquote(&importParentDir) has 0 files.;
            %put %mo_error Macro mo_validate_import_file.sas must be executed so that the datasets subdirectory is properly created and populated.;
            %goto ERROR_EXIT;
         %end;
      %end;

      %local importDPODir;
      %let importDPODir = %nrbquote(&importParentDir.dpofiles); 
      %if (not %sysfunc(fileexist(%nrbquote(&importDPODir)))) %then %do;
         %let &_rc = 102;
         %put %mo_error Subdirectory dpofiles below directory %nrbquote(&importParentDir) does not exist.;
         %put %mo_error Macro mo_validate_import_file.sas must be executed so that the dpofiles subdirectory is properly created and populated.;
         %goto ERROR_EXIT;
      %end;
      %else %do;
         %let dirRef = dpoDir;
         %let drc = %sysfunc(filename(dirRef, %nrbquote(&importDPODir)));
         %if (&drc ne 0) %then %do;
            %let &_rc = 105;
            %put %mo_warning Error encountered using FILENAME to assign a file reference to directory %nrbquote(&importDPODir) (rc=&drc).;
            %goto ERROR_EXIT;
         %end;

         %let dirID = %sysfunc(dopen(&dirRef));
         %if (&dirID eq 0) %then %do;
            %let &_rc = 105;
            %put %mo_warning Error encountered using DOPEN to open directory %nrbquote(&importDPODir).;
            %goto ERROR_EXIT;
         %end;
         %let numFiles = %sysfunc(dnum(&dirID));
         %let drc = %sysfunc(dclose(&dirID));

         %if (&numFiles eq 0) %then %do;
            %let &_rc = 102;
            %put %mo_error Subdirectory datasets below directory %nrbquote(&importParentDir) has 0 files.;
            %put %mo_error Macro mo_validate_import_file.sas must be executed so that the dpofiles subdirectory is properly created and populated.;
            %goto ERROR_EXIT;
         %end;
      %end;
   %end;


   %*-------------------------------------------;
   %* Set up librefs to working subdirectories. ;
   %*-------------------------------------------;
   libname _dsimprt "&importDSDir";
   libname _dpomprt "&importDPODir";


   %*----------------------------------------------------------;
   %* Extract information from the package descriptor dataset. ;
   %* mo_validate_import_file already verified that it exists. ;
   %*----------------------------------------------------------;
   %local xprtPackageName xprtPackageDescription xprtScenarioName;
   %local xprtProjectName xprtInputDataName  xprtTimestamp xprtVersion;
   %let xprtPackageName =;
   %let xprtPackageDescription =;
   %let xprtScenarioName =;
   %let xprtProjectName =;
   %let xprtInputDataName =;
   %let xprtTimestamp =;
   %let xprtVersion =;
   data _null_;
      set _dsimprt.packageDescriptor;
      call symput("xprtPackageName", packageName);
      call symput("xprtPackageDescription", packageDescription);
      call symput("xprtScenarioName", scenarioName);
      call symput("xprtProjectName", projectName);
      call symput("xprtInputDataName", inputDataName);
      call symput("xprtTimestamp", put(timestamp, datetime.));
      call symput("xprtVersion", version);
   run;
   %if (&syserr gt 4) %then %do;
      %let &_rc = 100;
      %put %mo_error Error occurred reading _dsimprt.packageDescriptor (syserr=&syserr).;
      %goto ERROR_EXIT;
   %end;

   %let xprtPackageName = %trim(%left(&xprtPackageName));
   %let xprtPackageDescription = %trim(%left(&xprtPackageDescription));
   %let xprtScenarioName = %trim(%left(&xprtScenarioName));
   %let xprtProjectName = %trim(%left(&xprtProjectName));
   %let xprtInputDataName = %trim(%left(&xprtInputDataName));
   %let xprtTimestamp = %trim(%left(&xprtTimestamp));
   %let xprtVersion = %trim(%left(&xprtVersion));

   %if (&local_debug eq 1) %then %do;
      %mo_no_notes(on)
      %put;
      %put %mo_note xprtPackageName: %nrbquote(&xprtPackageName);
      %put %mo_note xprtPackageDescription: %nrbquote(&xprtPackageDescription);
      %put %mo_note xprtScenarioName: %nrbquote(&xprtScenarioName);
      %put %mo_note xprtProjectName: %nrbquote(&xprtProjectName);
      %put %mo_note xprtInputDataName: %nrbquote(&xprtInputDataName);
      %put %mo_note xprtTimestamp: %nrbquote(&xprtTimestamp);
      %put %mo_note xprtVersion: %nrbquote(&xprtVersion);
      %mo_no_notes(off)
   %end;


   %***** LATER: USE XPRTVERSION TO DETERMINE IF MIGRATION IS NEEDED *****;


   %*--------------------------------;
   %* Retrieve metadata              ;
   %*--------------------------------;
   %mo_no_notes(off)
   %local project_name input_data_id username actualname user_id; 
   %let project_name =;
   %let input_data_id =;
   %let username =; 
   %let user_id =; 
   proc sql noprint;
      select a.project_name, a.input_data_id, a.user_id, b.username, b.name
      into :project_name, :input_data_id, :user_id, :username, :actualname
      from mometa.projects as a, mometa.users as b
      where (a.project_id = &project_id) &
            (a.user_id = b.user_id);
   quit;
   %if (&sqlrc gt 4) %then %do;
      %let &_rc = 107;
      %put %mo_error Error occurred while reading metadata (sqlrc=&sqlrc).;
      %goto ERROR_EXIT;
   %end;

   %let project_name = %trim(%left(&project_name));
   %let input_data_id = %trim(%left(&input_data_id));
   %let username = %trim(%left(&username));
   %let actualname = %trim(%left(&actualname));
   %let user_id = %trim(%left(&user_id));
   %if (%nrbquote(&project_name) eq %str()) %then %do;
      %let &_rc = 101;
      %put %mo_error Invalid metadata detected.;
      %put %mo_error Project name not identified for project_id = &project_id..;
      %goto ERROR_EXIT;
   %end;
   %if (%nrbquote(&input_data_id) eq %str()) %then %do;
      %let &_rc = 101;
      %put %mo_error Invalid metadata detected.;
      %put %mo_error Input data ID not identified for project_id = &project_id..;
      %goto ERROR_EXIT;
   %end;
   %else %do;
      %local inputDataExists;
      %let inputDataExists = 0;
      proc sql noprint;
         select count(*) into :inputDataExists
         from mometa.input_data
         where (input_data_id = &input_data_id);
      quit;
      %if (&sqlrc gt 4) %then %do;
         %let &_rc = 107;
         %put %mo_error Error occurred while reading metadata (sqlrc=&sqlrc).;
         %goto ERROR_EXIT;
      %end;

      %if (&inputDataExists eq 0) %then %do;
         %let &_rc = 101;
         %put %mo_error Invalid metadata detected.;
         %put %mo_error Input data ID &input_data_id does not exist in the mometa.input_data table.;
         %goto ERROR_EXIT;
      %end;
   %end;
   %if (%nrbquote(&username) eq %str()) %then %do;
      %let &_rc = 101;
      %put %mo_error Invalid metadata detected.;
      %put %mo_error User name not identified for project_id = &project_id..; 
      %goto ERROR_EXIT;
   %end;
   %if (%nrbquote(&actualname) eq %str()) %then %do;
      %let &_rc = 101;
      %put %mo_error Invalid metadata detected.;
      %put %mo_error User actual name not identified for project_id = &project_id..; 
      %goto ERROR_EXIT;
   %end;
   %if (%nrbquote(&user_id) eq %str()) %then %do;
      %let &_rc = 101;
      %put %mo_error Invalid metadata detected.;
      %put %mo_error User ID not identified for project_id = &project_id..; 
      %goto ERROR_EXIT;
   %end;

   %local import_input_data_id import_project_id import_scenario_id; 
   %let import_input_data_id =; 
   %let import_project_id =; 
   %let import_scenario_id =; 
   proc sql noprint;
      select input_data_id into :import_input_data_id
      from _dsimprt.input_data;

      select project_id into :import_project_id
      from _dsimprt.projects;

      select scenario_id into :import_scenario_id
      from _dsimprt.scenarios;
   quit;
   %if (&sqlrc gt 4) %then %do;
      %let &_rc = 107;
      %put %mo_error Error occurred while obtaining input_data_id, project_id, and scenario_id from import scenario definition tables (sqlrc=&sqlrc).;
      %goto ERROR_EXIT;
   %end;
   %if ((&import_input_data_id eq %str()) or (&import_project_id eq %str()) or (&import_scenario_id eq %str())) %then %do;
      %let &_rc = 100;
      %put %mo_error One of the IDs from import scenario definition tables is null: input_data_id = &import_input_data_id, project_id = &import_project_id, scenario_id = &import_scenario_id (sqlrc=&sqlrc).;
      %goto ERROR_EXIT;
   %end; 


   %*------------------------------------------;
   %* Create datastore table mappings datasets ;
   %* to be used for generic processing below. ; 
   %*------------------------------------------;
   %mo_create_table_id_mappings(tableMappingDSName = work.tableIDMappings
                               ,attributeTypeMappingDSName = work.attrTypeMappings
                               ,custAttrTypeMappingDSName = work.custAttrTypeMappings
                               ,commAttrTypeMappingDSName = work.commAttrTypeMappings
                               ,scenarioItemTypeMappingDSName = work.scenItemMappings
                               ,_rc = ctimrc);
   %if (&ctimrc ne 0) %then %do;
      %let &_rc = 100;
      %goto ERROR_EXIT;
   %end;


   %*--------------------------------------------------------------------;
   %* If this is a validation request, create a directory into which to  ;
   %* carry out the validation and set up a libref to this directory.    ;
   %* Create an unofficial datastore in this directory containing tables ;
   %* and table entries that reflect the input data definition for the   ;
   %* input data ID that the scenario is to be validated against.        ;
   %*--------------------------------------------------------------------;
   %local scenarioValidationDir;
   %let scenarioValidationDir = %nrbquote(&importParentDir.validation);
   %if (&action eq VALIDATE) %then %do;

      %*--------------------------------------------------;
      %* Check whether this is an allowable import, i.e., ;
      %* an analytics-based scenario cannot be imported   ;
      %* into prioritization-based input data. The three  ;
      %* other permutations are allowed.                  ;
      %*--------------------------------------------------;
      %global prioritizationBasedInputData;
      %let prioritizationBasedInputData = 0;
      %mo_check_if_priority_based(item_type             = INPUT_DATA
                                 ,item_id               = &input_data_id
                                 ,_prioritization_based = prioritizationBasedInputData 
                                 ,_rc                   = cipbrc
                                 );
      %if (&cipbrc ne 0) %then %do;
         %let &_rc = &cipbrc;
         %goto ERROR_EXIT;
      %end;

      /* Cannot use mo_check_if_priority_based since not looking in mometa */
      %local prioritizationBasedScenario analyticsBasedScenario;
      %let prioritizationBasedScenario = 0;
      %let analyticsBasedScenario = 0;
      proc sql noprint;
         select uses_prioritization into :prioritizationBasedScenario 
         from _dsimprt.input_data
         where (input_data_id = &import_input_data_id);
      quit;
      %if (&sqlrc gt 4) %then %do;
         %let &_rc = 107;
         %put %mo_error Error occurred while reading _dsimprt.input_data to get uses_prioritization value for scenario to be imported (sqlrc=&sqlrc).;
         %goto ERROR_EXIT;
      %end;
      %let analyticsBasedScenario = %eval(not &prioritizationBasedScenario);

      %if ((&prioritizationBasedInputData eq 1) and 
           (&analyticsBasedScenario eq 1)) %then %do;
         %let &_rc = 170;
         %put %mo_error Importing an analytics-based scenario into prioritization-based input data is not allowed.;
         %goto ERROR_EXIT;
      %end;


      %*------------------------------;
      %* Create validation directory. ;
      %*------------------------------;
      %if (not %sysfunc(fileexist(%nrbquote(&scenarioValidationDir)))) %then %do;
         %mo_create_directory(parent_dir = %nrbquote(&importParentDir)
                             ,child_dir = validation
                             ,lock = NOLOCK
                             ,error_if_exists = 0
                             ,_rc = cdirrc);
         %if (&cdirrc ne 0) %then %do;
            %let &_rc = 105;
            %put %mo_error I/O error encountered creating subdirectory validation under parent directory %nrbquote(&importParentDir).;
            %goto ERROR_EXIT;
         %end;
      %end;

      %*-------------------------------------------;
      %* Set up libref to validation subdirectory  ;
      %* where the dataset results from the        ;
      %* validation action will be placed.         ;
      %*-------------------------------------------;
      libname _validat "&scenarioValidationDir";

      %*---------------------------------------;
      %* For VALIDATE, the "working" datastore ;
      %* library is the validation library.    ;
      %*---------------------------------------;
      %let workingDatastoreLib = _validat;

      %*------------------------------;
      %* Copy over input data tables  ;
      %* containing rows for only the ;
      %* input data against which the ;
      %* scenario is to be imported.  ;
      %*------------------------------;
      %local numTables;
      %let numTables = 0;
      %mo_get_input_data_tables(inputDataTablesDS = work.inputDataTables
                               ,_rc = gidtrc);
      %if (&gidtrc gt 4) %then %do;
         %let &_rc = 100;
         %put %mo_error Error occurred while obtaining datastore table names containing the input data definition (gidtrc=&gidtrc).;
         %goto ERROR_EXIT;
      %end;
      %mo_no_notes(off)

      proc sql noprint;
         select count(*) into :numTables
         from work.inputDataTables;
      quit;
      %if (&sqlrc gt 4) %then %do;
         %let &_rc = 107;
         %put %mo_error Error occurred while obtaining count of table names in work.inputDataTables (sqlrc=&sqlrc).;
         %goto ERROR_EXIT;
      %end;

      %local inputDataTable;
      %do tbl = 1 %to &numTables;
         data _null_;
            obsnum = &tbl;
            set work.inputDataTables point=obsnum;
            call symput("inputDataTable", strip(tableName));
            stop;
         run;
         %if (&syserr gt 4) %then %do;
            %let &_rc = 100;
            %put %mo_error Error occurred reading work.inputDataTables (syserr=&syserr).;
            %goto ERROR_EXIT;
         %end;
         %let inputDataTable = %trim(%left(&inputDataTable));

         data _validat.&inputDataTable;
            set mometa.&inputDataTable;
            if (input_data_id = &input_data_id);
         run;
         %if (&syserr gt 4) %then %do;
            %let &_rc = 100;
            %put %mo_error Error occurred creating _validat.&inputDataTable (syserr=&syserr).;
            %goto ERROR_EXIT;
         %end;
      %end;

      data _validat.users;
         set mometa.users;
      run;
      %if (&syserr gt 4) %then %do;
         %let &_rc = 100;
         %put %mo_error Error occurred creating _validat.users (syserr=&syserr).;
         %goto ERROR_EXIT;
      %end;

      %*----------------------------------------------------;
      %* No need to copy the UID table. If a refresh is in  ;
      %* progress, the attempt to copy it will fail. Since  ;
      %* the UID table in the validation directory is used  ;
      %* for generating throw-away table IDs, we can just   ;
      %* create a new one in the validation directory and   ;
      %* not risk contention for the one in the official    ;
      %* datastore.                                         ;
      %*----------------------------------------------------;
      %mo_create_tables(lib=_validat, table=uid, _rc = cudrc);
      %if (&cudrc ne 0) %then %do;
         %let &_rc = 100;
         %put %mo_error Error occurred creating _validat.uid (syserr=&syserr).;
         %goto ERROR_EXIT;
      %end;

      %*----------------------------------------;
      %* Determine if the input data uses time  ;
      %* periods, agents and/or contact history.;          
      %*----------------------------------------;
      %local input_data_uses_agents input_data_uses_history;
      %let input_data_uses_agents = 0;
      %let input_data_uses_history = 0;
      data _null_;
         set mometa.input_data;
         if (input_data_id eq &input_data_id) then do;
            if (agent_table_name ne '') and (agent_column_name ne '') then do;
               call symput('input_data_uses_agents', put(1, 1.));
            end;
            call symput('input_data_uses_history', put(uses_contact_history,1.));
         end;
      run;
      %if (&syserr gt 4) %then %do;
         %let &_rc = 100;
         %put %mo_error Error occurred while setting input_data_uses_xxx flags (for agents and contact history) (syserr=&syserr).;
         %goto ERROR_EXIT;
      %end;

      %local input_data_uses_timepds;
      %let input_data_uses_timepds = 0;
      proc sql noprint;
         select count(*) into :input_data_uses_timepds
         from _validat.time_periods
         where (input_data_id = &input_data_id);
      quit;
      %if (&sqlrc gt 4) %then %do;
         %let &_rc = 107;
         %put %mo_error Error occurred while obtaining count of time_periods in _validat.time_periods (sqlrc=&sqlrc).;
         %goto ERROR_EXIT;
      %end;
      %let input_data_uses_timepds = %trim(%left(&input_data_uses_timepds));
      %if (&input_data_uses_timepds gt 0) %then %do;
         %let input_data_uses_timepds = 1;
      %end;

      %*-----------------------------;
      %* Determine if the imported   ;
      %* scenario uses time periods. ;          
      %*-----------------------------;
      %local imported_data_uses_timepds;
      %let imported_data_uses_timepds = 0;
      proc sql noprint;
         select count(*) into :imported_data_uses_timepds
         from _dsimprt.time_periods
         where (input_data_id = &input_data_id);
      quit;
      %if (&sqlrc gt 4) %then %do;
         %let &_rc = 107;
         %put %mo_error Error occurred while obtaining count of time_periods in _validat.time_periods (sqlrc=&sqlrc).;
         %goto ERROR_EXIT;
      %end;
      %let imported_data_uses_timepds = %trim(%left(&imported_data_uses_timepds));
      %if (&imported_data_uses_timepds gt 0) %then %do;
         %let imported_data_uses_timepds = 1;
      %end;

      %*-------------------------------;
      %* Create empty scenario tables. ;
      %*-------------------------------;
      %let numTables = 0;
      %mo_get_scenario_tables(scenarioTablesDS = work.scenarioTables
                               ,_rc = gstrc);
      %if (&gstrc gt 4) %then %do;
         %let &_rc = 100;
         %put %mo_error Error occurred while obtaining datastore table names containing the scenario definition (gstrc=&gstrc).;
         %goto ERROR_EXIT;
      %end;
      %mo_no_notes(off)

      proc sql noprint;
         select count(*) into :numTables
         from work.scenarioTables;
      quit;
      %if (&sqlrc gt 4) %then %do;
         %let &_rc = 107;
         %put %mo_error Error occurred while obtaining count of table names in work.scenarioTables (sqlrc=&sqlrc).;
         %goto ERROR_EXIT;
      %end;

      %local scenarioTable;
      %do tbl = 1 %to &numTables;
         data _null_;
            obsnum = &tbl;
            set work.scenarioTables point=obsnum;
            call symput("scenarioTable", strip(tableName));
            stop;
         run;
         %if (&syserr gt 4) %then %do;
            %let &_rc = 100;
            %put %mo_error Error occurred reading work.scenarioTables (syserr=&syserr).;
            %goto ERROR_EXIT;
         %end;
         %let scenarioTable = %trim(%left(&scenarioTable));

         proc sql noprint;
            create table _validat.&scenarioTable 
            like mometa.&scenarioTable;
         run;
         %if (&sqlrc gt 4) %then %do;
            %let &_rc = 107;
            %put %mo_error Error occurred creating empty _validat.&scenarioTable (sqlrc=&sqlrc).;
            %goto ERROR_EXIT;
         %end;
      %end;

      %*----------------------------------------;
      %* Create empty refresh_messages table    ;
      %* in _validat to record the results of   ;
      %* the validation of the scenario against ;
      %* the input data definition.             ;
      %*----------------------------------------;
      proc sql noprint;
         create table _validat.refresh_messages 
         like mometa.refresh_messages;
      run;
      %if (&sqlrc gt 4) %then %do;
         %let &_rc = 107;
         %put %mo_error Error occurred creating empty _validat.refresh_messages (sqlrc=&sqlrc).;
         %goto ERROR_EXIT;
      %end;


      %*-----------------------------------------;
      %* Set up for special case processing for  ;
      %* certain input data tables.              ;
      %* - time_periods table                    ;
      %*-----------------------------------------;
      proc sql noprint;
         insert into _validat.time_periods
         set input_data_id = &input_data_id,
             time_period_id = .,
             time_period = "missing",
             order = 0;

         %if (&input_data_uses_timepds eq 1) %then %do; 
            insert into _validat.time_periods
            set input_data_id = &input_data_id,
                time_period_id = &_CONTACT_POLICY_TP_TABLE_RANGE_,
                time_period = "first_last",
                order = 0;
         %end;
      quit;
      %if (&sqlrc gt 4) %then %do;
         %let &_rc = 107;
         %put %mo_error Error occurred adding rows to _validat.time_periods (sqlrc=&sqlrc).;
         %goto ERROR_EXIT;
      %end;

      proc sql noprint;
         insert into _dsimprt.time_periods
         set input_data_id = &import_input_data_id,
             time_period_id = .,
             time_period = "missing",
             order = 0;

         %if (&input_data_uses_timepds eq 1) %then %do; 
            insert into _dsimprt.time_periods
            set input_data_id = &import_input_data_id,
                time_period_id = &_CONTACT_POLICY_TP_TABLE_RANGE_,
                time_period = "first_last",
                order = 0;
         %end;
      quit;
      %if (&sqlrc gt 4) %then %do;
         %let &_rc = 107;
         %put %mo_error Error occurred adding rows to _validat.time_periods (sqlrc=&sqlrc).;
         %goto ERROR_EXIT;
      %end;

      %*-----------------------------------------;
      %* Set up for special case processing for  ;
      %* certain input data tables.              ;
      %* - prepopulated constraints table        ;
      %* - prepopulated contact policies table   ;
      %*-----------------------------------------;
      data _validat.prepopulated_constraints;
         set _validat.prepopulated_constraints end=last;
         if (last = 1) then do;
            output;

            input_data_id = &input_data_id;
            prepopulated_constraint_id = .;
            constraint_name = "missing";
            constraint_type = .;
            constraint_scope = .;
            active = .;
            operator = .;
            rhs = .;
            numerator_attribute_name = "";
            denominator_attribute_name = "";
            computation_type = .;
            communication_filter_txt = "";
            customer_filter_txt = "";
            include_in_sensitivity_analysis = .;
            output;
         end;
         else do;
            output;
         end;
      run;
      %if (&syserr gt 4) %then %do;
         %let &_rc = 100;
         %put %mo_error Error occurred reading rewriting _validat.prepopulated_constraints (syserr=&syserr).;
         %goto ERROR_EXIT;
      %end;

      data _dsimprt.prepopulated_constraints;
         set _dsimprt.prepopulated_constraints end=last;

         idid = input_data_id;
         if (last = 1) then do;
            output;

            input_data_id = &input_data_id;
            prepopulated_constraint_id = .;
            constraint_name = "missing";
            constraint_type = .;
            constraint_scope = .;
            active = .;
            operator = .;
            rhs = .;
            numerator_attribute_name = "";
            denominator_attribute_name = "";
            computation_type = .;
            communication_filter_txt = "";
            customer_filter_txt = "";
            include_in_sensitivity_analysis = .;
            output;
         end;
         else do;
            output;
         end;
      run;
      %if (&syserr gt 4) %then %do;
         %let &_rc = 100;
         %put %mo_error Error occurred reading rewriting _dsimprt.prepopulated_constraints (syserr=&syserr).;
         %goto ERROR_EXIT;
      %end;


      data _validat.prepopulated_contact_policies;
         set _validat.prepopulated_contact_policies end=last;
         if (last = 1) then do;
            output;

            input_data_id = &input_data_id;
            prepopulated_contact_policy_id = .;
            contact_policy_name = "missing";
            contact_policy_type = .;
            target_level = .;
            active = .;
            operator = .;
            rhs = .;
            time_based = .;
            start_time_period_id = .;
            end_time_period_id = .;
            rolling_period_cnt = .;
            communication_filter_txt = "";
            customer_filter_txt = "";
            blocking_comm_filter_a_txt = "";
            blocking_comm_filter_b_txt = "";
            blocking_period_cnt = .;
            output;
         end;
         else do;
            output;
         end;
      run;
      %if (&syserr gt 4) %then %do;
         %let &_rc = 100;
         %put %mo_error Error occurred reading rewriting _validat.prepopulated_contact_policies (syserr=&syserr).;
         %goto ERROR_EXIT;
      %end;

      data _dsimprt.prepopulated_contact_policies;
         set _dsimprt.prepopulated_contact_policies end=last;

         idid = input_data_id;
         if (last = 1) then do;
            output;

            input_data_id = &input_data_id;
            prepopulated_contact_policy_id = .;
            contact_policy_name = "missing";
            contact_policy_type = .;
            target_level = .;
            active = .;
            operator = .;
            rhs = .;
            time_based = .;
            start_time_period_id = .;
            end_time_period_id = .;
            rolling_period_cnt = .;
            communication_filter_txt = "";
            customer_filter_txt = "";
            blocking_comm_filter_a_txt = "";
            blocking_comm_filter_b_txt = "";
            blocking_period_cnt = .;
            output;
         end;
         else do;
            output;
         end;
      run;
      %if (&syserr gt 4) %then %do;
         %let &_rc = 100;
         %put %mo_error Error occurred reading rewriting _dsimprt.prepopulated_contact_policies (syserr=&syserr).;
         %goto ERROR_EXIT;
      %end;


      %*-----------------------------------------;
      %* Set up for special case processing for  ;
      %* certain input data tables.              ;
      %* - communications table                  ;
      %* - comm_numeric_attribute_defaults table ;
      %*-----------------------------------------;
      %local _validat_cell_cd_column_status;
      %local _dsimprt_cell_cd_column_status;

      proc sql noprint;
         select cell_cd_column_status into :_validat_cell_cd_column_status
         from _validat.input_data
         where (input_data_id = &input_data_id);
      quit;
      %if ((&sqlrc gt 4) or (&sqlobs eq 0)) %then %do;
         %let &_rc = 107;
         %put %mo_error Error occurred while reading _validat.input_data (sqlrc=&sqlrc, sqlob =&sqlobs).;
         %goto ERROR_EXIT;
      %end;
      %let _validat_cell_cd_column_status = %trim(%left(&_validat_cell_cd_column_status));

      proc sql noprint;
         select cell_cd_column_status into :_dsimprt_cell_cd_column_status
         from _dsimprt.input_data;
      quit;
      %if ((&sqlrc gt 4) or (&sqlobs eq 0)) %then %do;
         %let &_rc = 107;
         %put %mo_error Error occurred while reading _dsimprt.input_data (sqlrc=&sqlrc, sqlob =&sqlobs).;
         %goto ERROR_EXIT;
      %end;
      %let _dsimprt_cell_cd_column_status = %trim(%left(&_dsimprt_cell_cd_column_status));

      %if ((&_validat_cell_cd_column_status eq &_CELL_CD_UNIQUE_) or
           (&_dsimprt_cell_cd_column_status eq &_CELL_CD_UNIQUE_)) %then %do;

         %if ((&_validat_cell_cd_column_status eq &_CELL_CD_UNIQUE_) or
              (&_validat_cell_cd_column_status eq &_CELL_CD_UNIQUE_UNNEEDED_)) %then %do;
            proc sql noprint;
               create table work.temp_comm as
               select *
               from _validat.communications
               where (input_data_id = &input_data_id);
   
               create table _validat.communications as
               select a.*, d.comm_attribute_value as cell_code
               from work.temp_comm as a,
                    _validat.comm_character_attribute_values as b,
                    _validat.comm_character_attributes as c,
                    _validat.comm_character_distinct_values as d
               where (a.campaign_id = b.campaign_id) and
                     (a.communication_id = b.communication_id) and
                     (b.comm_char_attribute_id = c.comm_char_attribute_id) and
                     (lowcase(c.comm_attribute_name) = "cell_cd") and
                     (b.comm_attribute_distinct_value_id = d.comm_attribute_distinct_value_id);
            quit;
            %if (&sqlrc gt 4) %then %do;
               %let &_rc = 107;
               %put %mo_error Error occurred while updating _validat.communications (sqlrc=&sqlrc, sqlob =&sqlobs).;
               %goto ERROR_EXIT;
            %end;
   
            proc sql noprint;
               create table work.comm_num_attr_def as
               select *
               from _validat.comm_numeric_attribute_defaults
               where (input_data_id = &input_data_id);
   
               create table _validat.comm_numeric_attribute_defaults as
               select a.*, d.comm_attribute_value as cell_code
               from work.comm_num_attr_def as a,
                    _validat.comm_character_attribute_values as b,
                    _validat.comm_character_attributes as c,
                    _validat.comm_character_distinct_values as d
               where (a.campaign_id = b.campaign_id) and
                     (a.communication_id = b.communication_id) and
                     (b.comm_char_attribute_id = c.comm_char_attribute_id) and
                     (lowcase(c.comm_attribute_name) = "cell_cd") and
                     (b.comm_attribute_distinct_value_id = d.comm_attribute_distinct_value_id);
            quit;
            %if (&sqlrc gt 4) %then %do;
               %let &_rc = 107;
               %put %mo_error Error occurred while updating _validat.comm_numeric_attribute_defaults (sqlrc=&sqlrc, sqlob =&sqlobs).;
               %goto ERROR_EXIT;
            %end;
         %end;
         %else %do;
            data _validat.communications;
               length cell_code $256;
               set _validat.communications;
               cell_code = "";
            run;
            %if (&sysrc gt 4) %then %do;
               %let &_rc = 100;
               %put %mo_error Error occurred while updating _validat.communications with blank cell_code (sysrc=&sysrc).;
               %goto ERROR_EXIT;
            %end;

            data _validat.comm_numeric_attribute_defaults;
               length cell_code $256;
               set _validat.comm_numeric_attribute_defaults;
               cell_code = "";
            run;
            %if (&sysrc gt 4) %then %do;
               %let &_rc = 100;
               %put %mo_error Error occurred while updating _validat.comm_numeric_attribute_defaults with blank cell_code (sysrc=&sysrc).;
               %goto ERROR_EXIT;
            %end;
         %end;
   
         %if ((&_dsimprt_cell_cd_column_status eq &_CELL_CD_UNIQUE_) or
              (&_dsimprt_cell_cd_column_status eq &_CELL_CD_UNIQUE_UNNEEDED_)) %then %do;
            proc sql noprint;
               create table work.temp_comm as
               select *
               from _dsimprt.communications
               where (input_data_id = &import_input_data_id);
   
               create table _dsimprt.communications as
               select a.*, d.comm_attribute_value as cell_code
               from work.temp_comm as a,
                    _dsimprt.comm_character_attribute_values as b,
                    _dsimprt.comm_character_attributes as c,
                    _dsimprt.comm_character_distinct_values as d
               where (a.campaign_id = b.campaign_id) and
                     (a.communication_id = b.communication_id) and
                     (b.comm_char_attribute_id = c.comm_char_attribute_id) and
                     (lowcase(c.comm_attribute_name) = "cell_cd") and
                     (b.comm_attribute_distinct_value_id = d.comm_attribute_distinct_value_id);
            quit;
            %if (&sqlrc gt 4) %then %do;
               %let &_rc = 107;
               %put %mo_error Error occurred while updating _dsimprt.communications (sqlrc=&sqlrc, sqlob =&sqlobs).;
               %goto ERROR_EXIT;
            %end;
   
            proc sql noprint;
               create table work.comm_num_attr_def as
               select *
               from _dsimprt.comm_numeric_attribute_defaults
               where (input_data_id = &import_input_data_id);
   
               create table _dsimprt.comm_numeric_attribute_defaults as
               select a.*, d.comm_attribute_value as cell_code
               from work.comm_num_attr_def as a,
                    _dsimprt.comm_character_attribute_values as b,
                    _dsimprt.comm_character_attributes as c,
                    _dsimprt.comm_character_distinct_values as d
               where (a.campaign_id = b.campaign_id) and
                     (a.communication_id = b.communication_id) and
                     (b.comm_char_attribute_id = c.comm_char_attribute_id) and
                     (lowcase(c.comm_attribute_name) = "cell_cd") and
                     (b.comm_attribute_distinct_value_id = d.comm_attribute_distinct_value_id);

               create table work.comm_num_attr_val as
               select *
               from _dsimprt.comm_numeric_attribute_values
               where (scenario_id = &import_scenario_id);
   
               create table _dsimprt.comm_numeric_attribute_values_cc as
               select a.*, d.comm_attribute_value as cell_code
               from work.comm_num_attr_val as a,
                    _dsimprt.comm_character_attribute_values as b,
                    _dsimprt.comm_character_attributes as c,
                    _dsimprt.comm_character_distinct_values as d
               where (a.campaign_id = b.campaign_id) and
                     (a.communication_id = b.communication_id) and
                     (b.comm_char_attribute_id = c.comm_char_attribute_id) and
                     (lowcase(c.comm_attribute_name) = "cell_cd") and
                     (b.comm_attribute_distinct_value_id = d.comm_attribute_distinct_value_id);
            quit;
            %if (&sqlrc gt 4) %then %do;
               %let &_rc = 107;
               %put %mo_error Error occurred while updating _dsimprt.comm_numeric_attribute_defaults and _dsimprt.comm_numeric_attribute_values_cc (sqlrc=&sqlrc, sqlob =&sqlobs).;
               %goto ERROR_EXIT;
            %end;
         %end;
         %else %do;
            data _dsimprt.communications;
               length cell_code $256;
               set _dsimprt.communications;
               cell_code = "";
            run;
            %if (&sysrc gt 4) %then %do;
               %let &_rc = 100;
               %put %mo_error Error occurred while updating _dsimprt.communications with blank cell_code (sysrc=&sysrc).;
               %goto ERROR_EXIT;
            %end;

            data _dsimprt.comm_numeric_attribute_defaults;
               length cell_code $256;
               set _dsimprt.comm_numeric_attribute_defaults;
               cell_code = "";
            run;
            %if (&sysrc gt 4) %then %do;
               %let &_rc = 100;
               %put %mo_error Error occurred while updating _dsimprt.comm_numeric_attribute_defaults with blank cell_code (sysrc=&sysrc).;
               %goto ERROR_EXIT;
            %end;

            data _dsimprt.comm_numeric_attribute_values_cc;
               length cell_code $256;
               set _dsimprt.comm_numeric_attribute_values;
               cell_code = "";
            run;
            %if (&sysrc gt 4) %then %do;
               %let &_rc = 100;
               %put %mo_error Error occurred while updating _dsimprt.comm_numeric_attribute_values_cc with blank cell_code (sysrc=&sysrc).;
               %goto ERROR_EXIT;
            %end;
         %end;
      %end;
   %end;

   %else %do;
      %local dirRef drc dirID numFiles;
      %if (not %sysfunc(fileexist(%nrbquote(&scenarioValidationDir)))) %then %do;
         %let &_rc = 102;
         %put %mo_error Subdirectory validation below directory %nrbquote(&importParentDir) does not exist.;
        %put %mo_error Macro mo_import_scenario.sas must first be executed with the VALIDATE option so that the validation subdirectory is properly created and populated.;
         %goto ERROR_EXIT;
      %end;
      %else %do;
         %let dirRef = dsDir;
         %let drc = %sysfunc(filename(dirRef, %nrbquote(&scenarioValidationDir)));
         %if (&drc ne 0) %then %do;
            %let &_rc = 105;
            %put %mo_warning Error encountered using FILENAME to assign a file reference to directory %nrbquote(&scenarioValidationDir) (rc=&drc).;
            %goto ERROR_EXIT;
         %end;

         %let dirID = %sysfunc(dopen(&dirRef));
         %if (&dirID eq 0) %then %do;
            %let &_rc = 105;
            %put %mo_warning Error encountered using DOPEN to open directory %nrbquote(&scenarioValidationDir).;
            %goto ERROR_EXIT;
         %end;
         %let numFiles = %sysfunc(dnum(&dirID));
         %let drc = %sysfunc(dclose(&dirID));

         %if (&numFiles eq 0) %then %do;
            %let &_rc = 102;
            %put %mo_error Subdirectory validation below directory %nrbquote(&importParentDir) has 0 files.;
            %put %mo_error Macro mo_import_scenario.sas must be first be executed with the VALIDATE option so that the validation subdirectory is properly created and populated.;
            %goto ERROR_EXIT;
         %end;
      %end;


      %*-------------------------------------------;
      %* Set up libref to validation subdirectory  ;
      %* where the dataset results reside from the ;
      %* prior validation action.                  ;
      %*-------------------------------------------;
      libname _validat "&scenarioValidationDir";

      %*---------------------------------------------;
      %* For COMMIT, the "working" datastore library ;
      %*  is the official MO datastore library.      ;
      %*---------------------------------------------;
      %let workingDatastoreLib = mometa;
   %end;


   %if (&local_debug eq 1) %then %do;
      %mo_no_notes(on)
      %put;
      %put %mo_note workingDatastoreLib: %nrbquote(&workingDatastoreLib);
      %mo_no_notes(off)
   %end;


   %*----------------------------------;
   %* Helper macro - Build where       ;
   %* clause for primary ID columns    ;
   %*----------------------------------;
   %macro buildIDWhereClause(valTbl
                            ,hasInpID
                            ,inpID 
                            ,hasUsrID
                            ,usrID
                            ,hasProjID
                            ,projID
                            ,hasScenID
                            ,scenID
                            ,officialLib
                            ,officialTblLtr
                            ,_whereClause
                            ,_rc
                            );

      %local macname;
      %let macname=&sysmacroname;
      %let macname=%lowcase(&macname);

      %if %nrbquote(&_rc) eq %str() %then %do;
         %local rc;
         %let _rc = rc;
      %end;
      %else %global &_rc;
      %let &_rc = 0;

      %if (&hasInpID eq 1) %then %do;
         %if (%nrbquote(&&&_whereClause) eq %str()) %then %do;
            %let &_whereClause = (&officialTblLtr..input_data_id = &inpID);
         %end;
         %else %do;
            %let &_whereClause = &&&_whereClause and (&officialTblLtr..input_data_id = &inpID);
         %end;
      %end;
      %if (&hasProjID eq 1) %then %do;
         %if (%nrbquote(&&&_whereClause) eq %str()) %then %do;
            %let &_whereClause = (&officialTblLtr..project_id = &projID);
         %end;
         %else %do;
            %let &_whereClause = &&&_whereClause and (&officialTblLtr..project_id = &projID);
         %end;
      %end;
      %if (&hasUsrID eq 1) %then %do;
         %if (%nrbquote(&&&_whereClause) eq %str()) %then %do;
            %let &_whereClause = (&officialTblLtr..user_id = &usrID);
         %end;
         %else %do;
            %let &_whereClause = &&&_whereClause and (&officialTblLtr..user_id = &usrID);
         %end;
      %end;
      %if (&hasScenID eq 1) %then %do;
         %if (%nrbquote(&&&_whereClause) eq %str()) %then %do;
            %let &_whereClause = (&officialTblLtr..scenario_id = &scenID);
         %end;
         %else %do;
            %let &_whereClause = &&&_whereClause and (&officialTblLtr..scenario_id = &scenID);
         %end;
      %end;

      %if (&local_debug eq 1) %then %do;
         %mo_no_notes(on)
         %put;
         %put %mo_note &_whereClause: %nrbquote(&&&_whereClause);
         %mo_no_notes(off)
      %end;

      %goto ID_WHERE_EXIT;


      %*--------------------------------;
      %* Exit processing                ;
      %*--------------------------------;
      %ID_WHERE_ERROR_EXIT:
      %put %mo_error Failed to construct table ID where clause for updating reference ID to table &valTbl for table &tableName.;

      %ID_WHERE_EXIT:

   %mend buildIDWhereClause;


   %*----------------------------------;
   %* Helper macro - Build where       ;
   %* clause for table value columns   ;
   %*----------------------------------;
   %macro buildValueWhereClause(valTblName
                               ,valColName
                               ,valSeparator
                               ,officialLib
                               ,importLib
                               ,officialTblLtr
                               ,importTblLtr
                               ,_fromTblLtr
                               ,_fromTables
                               ,_whereClause
                               ,_tblNameList
                               ,_rc
                               );

      %local macname;
      %let macname=&sysmacroname;
      %let macname=%lowcase(&macname);

      %if %nrbquote(&_rc) eq %str() %then %do;
         %local rc;
         %let _rc = rc;
      %end;
      %else %global &_rc;
      %let &_rc = 0;

      %if (&&&_fromTblLtr eq %str()) %then %do;
         %let &_fromTblLtr = 0;
      %end;

      %*----------------------------------------------------------------------;
      %* Handle the fact that the specified value column name may be a list   ;
      %* of one or more specifications of: 1) actual value column names,      ;
      %* 2) optional actual value column names, and 3) indirect references    ;
      %* to other tables where the actual value column name resides.          ;
      %*----------------------------------------------------------------------;

      %*----------------------------------------------------------------------;
      %* Parse the specified value column name into 3 separate lists of these ;
      %* types (column names, optional column names, indirect references).    ;
      %*----------------------------------------------------------------------;
      %local numColSpecs specNum colSpec;
      %let numColSpecs = %mo_word_count(&valColName, %str(&valSeparator));

      %local colNameList optColNameList indirectTblList;
      %let colNameList =;
      %let optColNameList =;
      %let indirectTblList =;

      %do specNum = 1 %to &numColSpecs;
         %let colSpec = %qscan(&valColName, &specNum, %str(&valSeparator));

         %if (%index(%nrbquote(&colSpec), %str(>)) eq 0) %then %do;
            %let colNameList = &colNameList &colSpec;
         %end;
         %else %do;
            %if (%index(%nrbquote(&colSpec), %str(optional)) ne 0) %then %do;
               %let optColNameList = &optColNameList %qscan(&colSpec, 2, %str(>));
            %end;
            %else %if (%index(%nrbquote(&colSpec), %str(indirect)) ne 0) %then %do;
               %let indirectTblList = &indirectTblList %qscan(&colSpec, 2, %str(>));
            %end;
            %else %do;
               %let &_rc = 102;
               %put %mo_error Unsupported value column name specification %nrbquote(&colSpec);
               %goto VAL_WHERE_ERROR_EXIT;
            %end;
         %end;
      %end;

      %*-----------------------------------------;
      %* Process the list of value column names. ;
      %*-----------------------------------------;
      %if (&colNameList ne %str()) %then %do;
         %local numColNames colNum colName;
         %let numColNames = %mo_word_count(&colNameList, %str( ));

         %do colNum = 1 %to &numColNames;
            %let colName = %qscan(&colNameList, &colNum, %str( ));

            proc contents data=&officialLib.._all_ out=work.valueColNames_tmp (keep=memname name type) noprint nodetails;
            run;
            %if (&syserr gt 4) %then %do;
               %let &_rc = 100;
               %put %mo_error Error encountered using proc contents to get the list of tables in %upcase(&officialLib.) (syserr=&syserr).;
               %goto VAL_WHERE_ERROR_EXIT;
            %end;

            %local colType;
            proc sql noprint;
               select type into :colType 
               from work.valueColNames_tmp
               where (upcase(memname) = upcase("&valTblName")) and
                     (upcase(name) = upcase("&colName"));
            quit;
            %if (&sqlrc gt 4) %then %do;
               %let &_rc = 107;
               %put %mo_error Error occurred while determining type of &colName in &officialLib..&valTblName (tableName=&tableName) (sqlrc=&sqlrc).;
               %goto VAL_WHERE_ERROR_EXIT;
            %end;
            %let colType = %trim(%left(&colType));

            %if (&colType eq char) %then %do;
               %if (%nrbquote(&&&_whereClause) eq %str()) %then %do;
                  %let &_whereClause = (upcase(&officialTblLtr..&colName) = upcase(&importTblLtr..&colName));
               %end;
               %else %do;
                  %let &_whereClause = &&&_whereClause and (upcase(&officialTblLtr..&colName) = upcase(&importTblLtr..&colName));
               %end;
            %end;
            %else %do;
               %if (%nrbquote(&&&_whereClause) eq %str()) %then %do;
                  %let &_whereClause = (&officialTblLtr..&colName = &importTblLtr..&colName);
               %end;
               %else %do;
                  %let &_whereClause = &&&_whereClause and (&officialTblLtr..&colName = &importTblLtr..&colName);
               %end;
            %end;
         %end;
      %end;


      %*--------------------------------------------------;
      %* Process the list of optional value column names. ;
      %*--------------------------------------------------;
      %if (&optColNameList ne %str()) %then %do;
         %local numOptColNames optColNum optColName;
         %let numOptColNames = %mo_word_count(&optColNameList, %str( ));

         %do optColNum = 1 %to &numOptColNames;
            %let optColName = %qscan(&optColNameList, &optColNum, %str( ));

            proc contents data=&officialLib.._all_ out=work.offValueColNames_tmp (keep=memname name type) noprint nodetails;
            run;
            %if (&syserr gt 4) %then %do;
               %let &_rc = 100;
               %put %mo_error Error encountered using proc contents to get the list of tables in %upcase(&officialLib.) (syserr=&syserr).;
               %goto VAL_WHERE_ERROR_EXIT;
            %end;

            proc contents data=&importLib.._all_ out=work.impValueColNames_tmp (keep=memname name type) noprint nodetails;
            run;
            %if (&syserr gt 4) %then %do;
               %let &_rc = 100;
               %put %mo_error Error encountered using proc contents to get the list of tables in %upcase(&importLib.) (syserr=&syserr).;
               %goto ERROR_EXIT;
            %end;

            %local officialTblHasCol importTblHasCol officialType importType;
            %let officialTblHasCol =;
            %let importTblHasCol =;

            proc sql noprint;
               select count(*), type into :officialTblHasCol, :officialType 
               from work.offValueColNames_tmp
               where (upcase(memname) = upcase("&valTblName")) and
                     (upcase(name) = upcase("&optColName"));

               select count(*), type into :importTblHasCol, :importType 
               from work.impValueColNames_tmp
               where (upcase(memname) = upcase("&valTblName")) and
                     (upcase(name) = upcase("&optColName"));
            quit;
            %if (&sqlrc gt 4) %then %do;
               %let &_rc = 107;
               %put %mo_error Error occurred while determining existence of &optColName in &officiaLib..&valTblName and &importLib..&valTblName (tableName=&tableName) (sqlrc=&sqlrc).;
               %goto VAL_WHERE_ERROR_EXIT;
            %end;
            %let officialType = %trim(%left(&officialType));
            %let importType = %trim(%left(&importType));

            %if ((&officialTblHasCol eq 1) and (&importTblHasCol eq 1)) %then %do;
               %if (&officialType eq char) %then %do;
                  %if (%nrbquote(&&&_whereClause) eq %str()) %then %do;
                     %let &_whereClause = (upcase(&officialTblLtr..&optColName) = upcase(&importTblLtr..&optColName));
                  %end;
                  %else %do;
                     %let &_whereClause = &&&_whereClause and (upcase(&officialTblLtr..&optColName) = upcase(&importTblLtr..&optColName));
                  %end;
               %end;
               %else %do;
                  %if (%nrbquote(&&&_whereClause) eq %str()) %then %do;
                     %let &_whereClause = (&officialTblLtr..&optColName = &importTblLtr..&optColName);
                  %end;
                  %else %do;
                     %let &_whereClause = &&&_whereClause and (&officialTblLtr..&optColName = &importTblLtr..&optColName);
                  %end;
               %end;
            %end;
         %end;
      %end;


      %*----------------------------------------------------------------------;
      %* Process the list of indirect table references to value column names. ;
      %*----------------------------------------------------------------------;
      %if (&indirectTblList ne %str()) %then %do;
         %local numIndrctTblNames indrctTblNum indrctTblName;
         %let numIndrctTblNames = %mo_word_count(&indirectTblList, %str( ));

         %do indrctTblNum = 1 %to &numIndrctTblNames;
            %let indrctTblName = %qscan(&indirectTblList, &indrctTblNum, %str( ));
            %let indrctTblName = %trim(%left(&indrctTblName));

            %if (%index(&indrctTblName, %str(-)) ne 0) %then %do; 

               %local diffIDColName indrctIDColName indrctValColName;
               %local offLtr impLtr;
               %let diffIDColName = %qscan(&indrctTblName, 2, %str(-)); 
               %let diffIDColName = %trim(%left(&diffIDColName));

               %let indrctTblName = %qscan(&indrctTblName, 1, %str(-)); 
               %let indrctTblName = %trim(%left(&indrctTblName));

               %mo_get_mapped_table_info(tblNm = %lowcase(&indrctTblName)
                                        ,_idColNm = indrctIDColName
                                        ,_valColNm = indrctValColName
                                        ,_rc = gmtirc);
               %if (&gmtirc ne 0) %then %do;
                  %let &_rc = &gmtirc;
                  %goto VALUE_WHERE_ERROR_EXIT;
               %end;

               %let offLtr = t&&&_fromTblLtr;
               %let &_fromTables = &&&_fromTables, &officialLib..&indrctTblName as &offLtr;

               %let impLtr = t%eval(&&&_fromTblLtr + 1);
               %let &_fromTables = &&&_fromTables, &importLib..&indrctTblName as &impLtr;

               %let &_fromTblLtr = %eval(&&&_fromTblLtr + 2);

               %if (%nrbquote(&&&_whereClause) eq %str()) %then %do;
                  %if (&diffIDColName ne %str()) %then %do;
                     %let &_whereClause = (&officialTblLtr..&diffIDColName = &offLtr..&indrctIDColName) and (&importTblLtr..&diffIDColName = &impLtr..&indrctIDColName);
                  %end;
                  %else %do;
                     %let &_whereClause = (&officialTblLtr..&indrctIDColName = &offLtr..&indrctIDColName) and (&importTblLtr..&indrctIDColName = &impLtr..&indrctIDColName);
                  %end;
               %end;
               %else %do;
                  %if (&diffIDColName ne %str()) %then %do;
                     %let &_whereClause = &&&_whereClause and (&officialTblLtr..&diffIDColName = &offLtr..&indrctIDColName) and (&importTblLtr..&diffIDColName = &impLtr..&indrctIDColName);
                  %end;
                  %else %do;
                     %let &_whereClause = &&&_whereClause and (&officialTblLtr..&indrctIDColName = &offLtr..&indrctIDColName) and (&importTblLtr..&indrctIDColName = &impLtr..&indrctIDColName);
                  %end;
               %end;


               %************************************************************;
               %* Do not incur recursion overhead if the indirect table was ;
               %* referenced by another table higher up in the chain. It    ;
               %* would be redundant and not worth the performance hit.     ;
               %* Exclude certain tables for which multiple references are  ;
               %* expected e.g., time periods table, values tables.         ;
               %************************************************************;
               %if (%index(&&&_tblNameList, &indrctTblName) eq 0) or
                   (%lowcase(&indrctTblName) eq time_periods) or
                   (%lowcase(&indrctTblName) eq comm_character_attributes) or
                   (%lowcase(&indrctTblName) eq comm_character_attribute_values) or
                   (%lowcase(&indrctTblName) eq comm_character_distinct_values) or
                   (%lowcase(&indrctTblName) eq comm_numeric_attributes) or
                   (%lowcase(&indrctTblName) eq comm_numeric_attribute_values) or
                   (%lowcase(&indrctTblName) eq comm_numeric_attribute_defaults) or
                   (%lowcase(&indrctTblName) eq customer_attributes) or
                   (%lowcase(&indrctTblName) eq customer_attribute_values) %then %do;

                  %let &_tblNameList = &&&_tblNameList &indrctTblName;
                  %buildValueWhereClause(&indrctTblName
                                        ,&indrctValColName
                                        ,&valSeparator
                                        ,&officialLib
                                        ,&importLib
                                        ,&offLtr
                                        ,&impLtr
                                        ,&_fromTblLtr
                                        ,&_fromTables
                                        ,&_whereClause
                                        ,&_tblNameList
                                        ,&_rc);
               %end;
            %end;
            %else %do; 
               %local whereSubclause;
               %let whereSubclause =;

               %local specialMappingTable specialTypeColName;
               %local indrctIDColName indrctTypeColName;
               %local offLtr impLtr;
               %if (&indrctTblName eq %str(useAttributeType)) %then %do;
                  %let specialMappingTable = work.attrTypeMappings;
                  %let specialTypeColName = attributeType;
                  %mo_get_mapped_table_info(tblNm = %lowcase(&valTblName)
                                           ,_attribTpColNm = indrctTypeColName
                                           ,_attribIDColNm = indrctIDColName
                                           ,_rc = gmtirc);
                  %if (&gmtirc ne 0) %then %do;
                     %let &_rc = &gmtirc;
                     %goto VAL_WHERE_ERROR_EXIT;
                  %end;
               %end;
               %else %if (&indrctTblName eq %str(useScenarioItem)) %then %do;
                  %let specialMappingTable = work.scenItemMappings;
                  %let specialTypeColName = itemType;
                  %mo_get_mapped_table_info(tblNm = %lowcase(&valTblName)
                                           ,_scenItemTpColNm = indrctTypeColName
                                           ,_scenItemIDColNm = indrctIDColName
                                           ,_rc = gmtirc);
                  %if (&gmtirc ne 0) %then %do;
                     %let &_rc = &gmtirc;
                     %goto ERROR_EXIT;
                  %end;
               %end;
               %else %if (&indrctTblName eq %str(useCommAttributeType)) %then %do;
                  %let specialMappingTable = work.commAttrTypeMappings;
                  %let specialTypeColName = commAttrType;
                  %mo_get_mapped_table_info(tblNm = %lowcase(&valTblName)
                                           ,_commAttribTpColNm = indrctTypeColName
                                           ,_commAttribIDColNm = indrctIDColName
                                           ,_rc = gmtirc);
                  %if (&gmtirc ne 0) %then %do;
                     %let &_rc = &gmtirc;
                     %goto ERROR_EXIT;
                  %end;
               %end;
               %else %if (&indrctTblName eq %str(useCustAttributeType)) %then %do;
                  %let specialMappingTable = work.custAttrTypeMappings;
                  %let specialTypeColName = custAttrType;
                  %mo_get_mapped_table_info(tblNm = %lowcase(&valTblName)
                                           ,_custAttribTpColNm = indrctTypeColName
                                           ,_custAttribIDColNm = indrctIDColName
                                           ,_rc = gmtirc);
                  %if (&gmtirc ne 0) %then %do;
                     %let &_rc = &gmtirc;
                     %goto ERROR_EXIT;
                  %end;
               %end;
               %else %do;
                  %let &_rc = 102;
                  %put %mo_error Unsupported indirect table reference value column specification %nrbquote(&indrctTblName);
                  %goto VAL_WHERE_ERROR_EXIT;
               %end;

               %***********************************************;
               %* Create temp tables in which to collect all   ;
               %* combinations of type, id, and name.          ;
               %***********************************************;
               proc sql noprint;
                  create table work.official_&indrctTblName
                     (type      num
                     ,id        num
                     ,charValue char(256)
                     ,numValue  num
                     );                  

                  create table work.import_&indrctTblName
                  like work.official_&indrctTblName;
               quit;
               %if (&sqlrc gt 4) %then %do;
                  %let &_rc = 107;
                  %put %mo_error SQL error occurred creating work.official_&indrctTblName and/or work.import_&indrctTblName (sqlrc=&sqlrc).;
                  %goto VAL_WHERE_ERROR_EXIT;
               %end;
              
               %local specialTypes specialTblNames;
               %local specialIDColNames specialValColNames;
               proc sql noprint;
                  select &specialTypeColName, tableName, idColName, valueColName
                  into :specialTypes separated by " ", 
                       :specialTblNames separated by " ", 
                       :specialIDColNames separated by " ", 
                       :specialValColNames separated by " "
                  from &specialMappingTable;
               quit;
               %if (&sqlrc gt 4) %then %do;
                  %let &_rc = 107;
                  %put %mo_error SQL error occurred obtaining mapping information from table &specialMappingTable (sqlrc=&sqlrc).;
                  %goto VAL_WHERE_ERROR_EXIT;
               %end;

               %local numSpecialTblNames specialTblNum;
               %local specialType specialTblName specialIDColName specialValColName;
               %let numSpecialTblNames = %mo_word_count(&specialTblNames, %str( ));

               %do specialTblNum = 1 %to &numSpecialTblNames;
                  %let specialType = %qscan(&specialTypes, &specialTblNum, %str( ));
                  %let specialTblName = %qscan(&specialTblNames, &specialTblNum, %str( ));
                  %let specialIDColName = %qscan(&specialIDColNames, &specialTblNum, %str( ));
                  %let specialValColName = %qscan(&specialValColNames, &specialTblNum, %str( ));

                  %let &_tblNameList = &&&_tblNameList &specialTblName;

                  proc contents data=&officialLib.._all_ out=work.specValueColNames_tmp (keep=memname name type) noprint nodetails;
                  run;
                  %if (&syserr gt 4) %then %do;
                     %let &_rc = 100;
                     %put %mo_error Error encountered using proc contents to get the list of tables in %upcase(&officialLib.) (syserr=&syserr).;
                     %goto VAL_WHERE_ERROR_EXIT;
                  %end;

                  %local colType;
                  proc sql noprint;
                     select type into :colType 
                     from work.specValueColNames_tmp
                     where (upcase(memname) = upcase("&specialTblName")) and
                           (upcase(name) = upcase("&specialValColName"));
                  quit;
                  %if (&sqlrc gt 4) %then %do;
                     %let &_rc = 107;
                     %put %mo_error Error occurred while determining type of &specialValColName in &officiaLib..&specialTblName (tableName=&tableName) (sqlrc=&sqlrc).;
                     %goto VAL_WHERE_ERROR_EXIT;
                  %end;
                  %let colType = %trim(%left(&colType));

                  proc sql noprint;
                     create table work.official_temp as
                     select &specialType as type,
                            &specialIDColName as id,
                            %if (&colType eq char) %then %do;
                               upcase(&specialValColName) as charValue length=256,
                               . as numValue
                            %end;
                            %else %do;
                               ' ' as charValue length=256,
                               &specialValColName as numValue
                            %end;
                     from &officialLib..&specialTblName;

                     create table work.import_temp as
                     select &specialType as type,
                            &specialIDColName as id,
                            %if (&colType eq char) %then %do;
                               upcase(&specialValColName) as charValue length=256, 
                               . as numValue
                            %end;
                            %else %do;
                               ' ' as charValue length=256, 
                               &specialValColName as numValue
                            %end;
                     from &importLib..&specialTblName;
                  quit;          
                  %if (&sqlrc gt 4) %then %do;
                     %let &_rc = 107;
                     %put %mo_error SQL error occurred creating work.official_temp and work.import_temp for &specialTypeColName &specialType (sqlrc=&sqlrc).;
                     %goto VAL_WHERE_ERROR_EXIT;
                  %end;

                  proc append base=work.official_&indrctTblName data=work.official_temp;
                  run;
                  %if (&syserr gt 4) %then %do;
                     %let &_rc = 100;
                     %put %mo_error Error occurred appending table work.official_temp to table work.official_&indrctTblName for &specialTypeColName &specialType (syserr=&syserr).;
                     %goto ERROR_EXIT;
                  %end;

                  proc append base=work.import_&indrctTblName data=work.import_temp;
                  run;
                  %if (&syserr gt 4) %then %do;
                     %let &_rc = 100;
                     %put %mo_error Error occurred appending table work.import_temp to table work.import_&indrctTblName for &specialTypeColName &specialType (syserr=&syserr).;
                     %goto ERROR_EXIT;
                  %end;
               %end;

               %let impLtr = t&&&_fromTblLtr;
               %let &_fromTables = &&&_fromTables, work.import_&indrctTblName as &impLtr;

               %let offLtr = t%eval(&&&_fromTblLtr + 1);
               %let &_fromTables = &&&_fromTables, work.official_&indrctTblName as &offLtr;

               %let &_fromTblLtr = %eval(&&&_fromTblLtr + 2);

               %let whereSubclause = (&importTblLtr..&indrctTypeColName = &impLtr..type) and (&importTblLtr..&indrctIDColName = &impLtr..id);
               %let whereSubclause = &whereSubclause and (&officialTblLtr..&indrctTypeColName = &offLtr..type) and (&officialTblLtr..&indrctIDColName = &offLtr..id); 
               %let whereSubclause = &whereSubclause and ((&impLtr..type = &offLtr..type) and (&impLtr..numValue = &offLtr..numValue) and (&impLtr..charValue = &offLtr..charValue));

               %if (%nrbquote(&whereSubclause) ne %str()) %then %do;
                  %if (%nrbquote(&&&_whereClause) eq %str()) %then %do;
                     %let &_whereClause = &whereSubclause;
                  %end;
                  %else %do;
                     %let &_whereClause = &&&_whereClause and &whereSubclause;
                  %end;
               %end;
            %end;
         %end;
      %end;

      %if (&local_debug eq 1) %then %do;
         %mo_no_notes(on)
         %put;
         %put %mo_note &_fromTables: %nrbquote(&&&_fromTables);
         %put %mo_note &_whereClause: %nrbquote(&&&_whereClause);
         %mo_no_notes(off)
      %end;

      %goto VAL_WHERE_EXIT;


      %*--------------------------------;
      %* Exit processing                ;
      %*--------------------------------;
      %VAL_WHERE_ERROR_EXIT:
      %put %mo_error Failed to construct table value where clause for updating reference ID to table &valTbl for table &tableName.;

      %VAL_WHERE_EXIT:

   %mend buildValueWhereClause;


   %*----------------------------------;
   %* Helper macro - Update ID to      ;
   %* referenced table                 ;
   %*----------------------------------;
   %macro updateReferenceID(rfTblNm      =
                           ,rfIDColNm    =
                           ,diffIDColNm  =
                           ,rfValColNm   =
                           ,officialLib  =
                           ,importLib    =
                           ,workLib      =
                           ,workTbl      =
                           ,workIDColNm  =
                           ,_rfTblNmList =
                           ,_rc          = 
                           );

      %local macname;
      %let macname=&sysmacroname;
      %let macname=%lowcase(&macname);

      %if %nrbquote(&_rc) eq %str() %then %do;
         %local rc;
         %let _rc = rc;
      %end;
      %else %global &_rc;
      %let &_rc = 0;

      %if (&local_debug eq 1) %then %do;
         data &workLib..preResolveRefs_&tbl._&refTbl;
            set &workLib..&tableName;
         run;
      %end;

      %put;
      %mo_no_notes(on)
      %put %mo_note Begin referenced table %upcase(&rfTblNm) %sysfunc(putn(%sysfunc(time()), time.));
      %mo_no_notes(off)

      %local rfHasInputDataID rfHasProjectID;
      %local rfHasScenarioID rfHasUserID;
      %mo_get_mapped_table_info(tblNm = %lowcase(&rfTblNm)
                               ,_hasInpID = rfHasInputDataID
                               ,_hasProjID = rfHasProjectID
                               ,_hasScenID = rfHasScenarioID
                               ,_hasUsrID = rfHasUserID
                               ,_rc = gmtirc);
      %if (&gmtirc ne 0) %then %do;
         %let &_rc = &gmtirc;
         %goto UPDATE_ERROR_EXIT;
      %end;

      %local usesIDColNm;
      %if (&diffIDColNm ne %str()) %then %do;
         %let usesIDColNm = &diffIDColNm;
      %end;
      %else %do;
         %let usesIDColNm = &rfIDColNm;
      %end;

      %***************************************;
      %* If the referenced table is an input  ;
      %* data table, resolve the ID by value. ;
      %***************************************;
      %if (&rfHasScenarioID eq 0) %then %do; 
         %put;
         %mo_no_notes(on)
         %put %mo_note Using "by value" to resolve reference to input data table %upcase(&rfTblNm).;
         %mo_no_notes(off)

         %local rfIDWhereClause;
         %let rfIDWhereClause =;
         %buildIDWhereClause(&rfTblNm
                            ,&rfHasInputDataID
                            ,&input_data_id
                            ,&rfHasUserID
                            ,&user_id
                            ,&rfHasProjectID
                            ,&project_id
                            ,&rfHasScenarioID
                            ,&scenario_id
                            ,&officialLib
                            ,a
                            ,rfIDWhereClause
                            ,bidwcrc);
         %if (&bidwcrc ne 0) %then %do;
            %let &_rc = &bidwcrc;
            %goto UPDATE_ERROR_EXIT;
         %end;
   
         %local rfValFromTblLtr rfValFromTables rfValWhereClause;
         %let rfValFromTblLtr = 0;
         %let rfValFromTables =;
         %let rfValWhereClause =;
         %let &_rfTblNmList = &&&_rfTblNmList &rfTblNm; 
         %buildValueWhereClause(&rfTblNm
                               ,&rfValColNm
                               ,%str(*)
                               ,&officialLib
                               ,&importLib
                               ,a
                               ,b
                               ,rfValFromTblLtr
                               ,rfValFromTables
                               ,rfValWhereClause
                               ,&_rfTblNmList
                               ,bvwcrc);
         %if (&bvwcrc ne 0) %then %do;
            %let &_rc = &bvwcrc;
            %goto UPDATE_ERROR_EXIT;
         %end;
   
         %if (&local_debug lt 2) %then %do;
            %mo_no_notes(on)
            %put;
            %put %mo_note Entire from table clause: &officialLib..&rfTblNm as a, &importLib..&rfTblNm as b, &workLib..&workTbl as c &rfValFromTables;  
            %put %mo_note Entire where clause:  &rfIDWhereClause and (c.&usesIDColNm = b.&rfIDColNm) and &rfValWhereClause;
            %mo_no_notes(off)
         %end;

         proc sql noprint;
            create table &workLib..temp_&tbl._&refTbl as
            select distinct c.*, a.&rfIDColNm as newRfTblID
            from &officialLib..&rfTblNm as a, &importLib..&rfTblNm as b,
                 &workLib..&workTbl as c  
                 %if (&rfValFromTables ne %str()) %then %do;
                    &rfValFromTables
                 %end;
            where %if (%nrbquote(&rfIDWhereClause) ne %str()) %then %do;
                     &rfIDWhereClause and
                  %end;
                  (c.&usesIDColNm = b.&rfIDColNm) and
                  %if (%nrbquote(&rfValWhereClause) ne %str()) %then %do;
                     &rfValWhereClause
                  %end;
                  ;
         quit;
         %if (&sqlrc gt 4) %then %do;
            %let &_rc = 107;
            %put %mo_error Error occurred while determining referenced table &rfTblNm IDs to store into table &workLib..temp_&tbl._&refTbl (tableName=&tableName) (sqlrc=&sqlrc).;
            %goto UPDATE_ERROR_EXIT;
         %end;
      %end;

      %**********************************************;
      %* If the referenced table is a scenario table ;
      %* resolve the ID by old-to-new ID mappings.   ;
      %**********************************************;
      %else %do; 
         %put;
         %mo_no_notes(on)
         %put %mo_note Using "ID mapping" to resolve reference to scenario table %upcase(&rfTblNm).;
         %mo_no_notes(off)

         proc sql noprint;
            create table &workLib..temp_&tbl._&refTbl as
            select c.*, b.&rfIDColNm as newRfTblID
            from &workLib..&workTbl as c, &workLib..&rfTblNm as b
            where (c.&usesIDColNm = b.origID);
         quit;
         %if (&sqlrc gt 4) %then %do;
            %let &_rc = 107;
            %put %mo_error Error occurred while determining referenced table &rfTblNm IDs to store into table &workLib..temp_&tbl._&refTbl (tableName=&tableName) (sqlrc=&sqlrc).;
            %goto UPDATE_ERROR_EXIT;
         %end;
      %end;
   
      %************************;
      %* Check for duplicates. ;
      %************************;
      %local numOrigRows numRefRows;
      %let numOrigRows = 0;
      %let numRefRows= 0;
      proc sql noprint;
         select count(*) into :numOrigRows
         from &workLib..&workTbl;
         select count(*) into :numRefRows
         from &workLib..temp_&tbl._&refTbl;
      quit;
      %if (&sqlrc gt 4) %then %do;
         %let &_rc = 107;
         %put %mo_error Error occurred while obtaining count of rows in &workLib..&workTbl and/or &workLib..temp_&tbl._&refTbl (sqlrc=&sqlrc).;
         %goto UPDATE_ERROR_EXIT;
      %end;
   
      %if (&numRefRows gt &numOrigRows) %then %do;
         %let &_rc = 100;
         %put %mo_error Duplicates found while determining referenced table &rfTblNm IDs to store into table &workLib..temp_&tbl._&refTbl (tableName=&tableName) (sqlrc=&sqlrc).;
         %goto UPDATE_ERROR_EXIT;
      %end;

      proc sql noprint;
         create table &workLib..tempLeftJoin_&tbl._&refTbl as
         select distinct a.*, b.newRfTblID 
         from &workLib..&workTbl as a left join &workLib..temp_&tbl._&refTbl as b
         on (a.&workIDColNm = b.&workIDColNm);
      quit;
      %if (&sqlrc gt 4) %then %do;
         %let &_rc = 107;
         %put %mo_error Error occurred while determining referenced table &rfTblNm IDs to store into table &workTbl (tableName=&tableName) (sqlrc=&sqlrc).;
         %goto UPDATE_ERROR_EXIT;
      %end;

      data &workLib..&workTbl (drop=newRfTblID);
         set &workLib..tempLeftJoin_&tbl._&refTbl;
         &usesIDColNm = newRfTblID;
      run;
      %if (&syserr gt 4) %then %do;
         %let &_rc = 100;
         %put %mo_error Error occurred updating reference table IDs of &rfTblNm in table &workTbl (tableName=&tableName) (syserr=&syserr).;
         %goto UPDATE_ERROR_EXIT;
      %end;

      %goto UPDATE_EXIT;


      %*--------------------------------;
      %* Exit processing                ;
      %*--------------------------------;
      %UPDATE_ERROR_EXIT:
      %put %mo_error Failed to update reference ID to &rfTblNm for table &tableName.;

      %UPDATE_EXIT:
      %put;
      %mo_no_notes(on)
      %put %mo_note End referenced table %upcase(&rfTblNm) %sysfunc(putn(%sysfunc(time()), time.));
      %mo_no_notes(off)

   %mend updateReferenceID;


   %*----------------------------------;
   %* Helper macro - Update ID based   ;
   %* on attribute type                ;
   %*----------------------------------;
   %macro updateIDByAttributeType(tableName        = 
                                 ,officialLib      =
                                 ,importLib        =
                                 ,workLib          = work
                                 ,refTableNameList =
                                 ,_rc              =
                                 );

      %local macname;
      %let macname=&sysmacroname;
      %let macname=%lowcase(&macname);

      %if %nrbquote(&_rc) eq %str() %then %do;
         %local rc;
         %let _rc = rc;
      %end;
      %else %global &_rc;
      %let &_rc = 0;

      %put;
      %mo_no_notes(on)
      %put %mo_note Begin update ID to referenced table %upcase(&tableName) by Attribute Type.;
      %mo_no_notes(off)

      %if (&local_debug eq 1) %then %do;
         data &workLib..preResolveAttrType_&tbl._&refTbl;
            set &workLib..&tableName;
         run;
      %end;

      %local attributeTypeColName attributeIDColName;
      %mo_get_mapped_table_info(tblNm = %lowcase(&tableName)
                               ,_attribTpColNm = attributeTypeColName
                               ,_attribIDColNm = attributeIDColName
                               ,_rc = gmtirc);
      %if (&gmtirc ne 0) %then %do;
         %let &_rc = &gmtirc;
         %goto UPDATE_BY_AT_ERROR_EXIT;
      %end;

      %local attributeTypeList;
      %let attributeTypeList =;
      proc sql noprint;
         select distinct &attributeTypeColName into :attributeTypeList separated by ' '
         from &workLib..&tableName;
      quit;
      %if (&sqlrc gt 4) %then %do;
         %let &_rc = 107;
         %put %mo_error SQL error occurred obtaining attribute types from column &attributeTypeColName in table &workLib..&tableName (sqlrc=&sqlrc).;
         %goto UPDATE_BY_AT_ERROR_EXIT;
      %end;

      %local numAttributeTypes attrTypeNum attributeType;
      %let numAttributeTypes = %mo_word_count(&attributeTypeList, %str( ));

      %do attrTypeNum = 1 %to &numAttributeTypes; 
         %let attributeType = %qscan(&attributeTypeList, &attrTypeNum, %str( ));
         %let attributeType = %trim(%left(&attributeType));

         proc sql noprint;
            create table &workLib..attrType_&attributeType as
            select *
            from &workLib..&tableName
            where (&attributeTypeColName = &attributeType);
         quit;
         %if (&sqlrc gt 4) %then %do;
            %let &_rc = 107;
            %put %mo_error SQL error occurred creating table &workLib..attrType_&attributeType from table &workLib..&tableName (sqlrc=&sqlrc).;
            %goto UPDATE_BY_AT_ERROR_EXIT;
         %end;

         %local attrTypeTableName attrTypeIDColName attrTypeValueColName;
         %mo_get_mapped_table_info(atrType = &attributeType
                                  ,_atrTblNm = attrTypeTableName 
                                  ,_atrIDColNm = attrTypeIDColName 
                                  ,_atrValColNm = attrTypeValueColName
                                  ,_rc = gmtirc);
         %if (&gmtirc ne 0) %then %do;
            %let &_rc = &gmtirc;
            %goto UPDATE_BY_AT_ERROR_EXIT;
         %end;
   
         %let refTbl = %eval(&refTbl + 1);
         %updateReferenceID(rfTblNm = &attrTypeTableName
                           ,rfIDColNm = &attrTypeIDColName
                           ,diffIDColNm = &attributeIDColName
                           ,rfValColNm = &attrTypeValueColName
                           ,officialLib = &officialLib
                           ,importLib = &importLib
                           ,workLib = work
                           ,workTbl = attrType_&attributeType
                           ,workIDColNm = &idColName
                           ,_rfTblNmList = refTableNameList
                           ,_rc = uridrc);
         %if (&uridrc ne 0) %then %do;
            %let &_rc = &uridrc;
            %goto UPDATE_BY_AT_ERROR_EXIT;
         %end;
      %end; /* For each attribute type ... */

      %**************************************************************;
      %* Re-create working table from the per-attribute type tables. ; 
      %**************************************************************;
      proc sql noprint;
         create table &workLib..&tableName
         like &workLib..&tableName;
      quit;
      %if (&sqlrc gt 4) %then %do;
         %let &_rc = 107;
         %put %mo_error SQL error occurred creating table &workLib..&tableName like &workLib..&tableName (sqlrc=&sqlrc).;
         %goto UPDATE_BY_AT_ERROR_EXIT;
      %end;

      %do attrTypeNum = 1 %to &numAttributeTypes; 
         %let attributeType = %qscan(&attributeTypeList, &attrTypeNum, %str( ));
         %let attributeType = %trim(%left(&attributeType));

         proc append base=&workLib..&tableName data=&workLib..attrType_&attributeType;
         run;
         %if (&syserr gt 4) %then %do;
            %let &_rc = 100;
            %put %mo_error Error occurred appending table &workLib..attrType_&attributeType to table &workLib..&tableName (syserr=&syserr).;
            %goto UPDATE_BY_AT_ERROR_EXIT;
         %end;
      %end; /* For each attribute type ... */

      %goto UPDATE_BY_AT_EXIT;

      %*--------------------------------;
      %* Exit processing                ;
      %*--------------------------------;
      %UPDATE_BY_AT_ERROR_EXIT:
      %put %mo_error Failed to update ID based on attribute type for table &tableName.;

      %UPDATE_BY_AT_EXIT:
      %put;
      %mo_no_notes(on)
      %put %mo_note End update ID to referenced table %upcase(&tableName) by Attribute Type.;
      %mo_no_notes(off)


   %mend updateIDByAttributeType;


   %*----------------------------------;
   %* Helper macro - Update ID based   ;
   %* on customer attribute type       ;
   %*----------------------------------;
   %macro updateIDByCustAttributeType(tableName        =
                                     ,officialLib      =
                                     ,importLib        =
                                     ,workLib          = work
                                     ,refTableNameList =
                                     ,_rc              =
                                     );

      %local macname;
      %let macname=&sysmacroname;
      %let macname=%lowcase(&macname);

      %if %nrbquote(&_rc) eq %str() %then %do;
         %local rc;
         %let _rc = rc;
      %end;
      %else %global &_rc;
      %let &_rc = 0;

      %put;
      %mo_no_notes(on)
      %put %mo_note Begin update ID to referenced table %upcase(&tableName) by Customer Attribute Type.;
      %mo_no_notes(off)

      %if (&local_debug eq 1) %then %do;
         data &workLib..preResolveCustAttrType_&tbl._&refTbl;
            set &workLib..&tableName;
         run;
      %end;

      %local custAttributeTypeColName custAttributeIDColName;
      %mo_get_mapped_table_info(tblNm = %lowcase(&tableName)
                               ,_custAttribTpColNm = custAttributeTypeColName
                               ,_custAttribIDColNm = custAttributeIDColName
                               ,_rc = gmtirc);
      %if (&gmtirc ne 0) %then %do;
         %let &_rc = &gmtirc;
         %goto UPDATE_BY_CSTAT_ERROR_EXIT;
      %end;

      %local custAttributeTypeList;
      %let custAttributeTypeList =;
      proc sql noprint;
         select distinct &custAttributeTypeColName into :custAttributeTypeList separated by ' '
         from &workLib..&tableName;
      quit;
      %if (&sqlrc gt 4) %then %do;
         %let &_rc = 107;
         %put %mo_error SQL error occurred obtaining customer attribute types from column &custAttributeTypeColName in table &workLib..&tableName (sqlrc=&sqlrc).;
         %goto UPDATE_BY_CSTAT_ERROR_EXIT;
      %end;

      %local numCustAttributeTypes custAttrTypeNum custAttributeType;
      %let numCustAttributeTypes = %mo_word_count(&custAttributeTypeList, %str( ));

      %do custAttrTypeNum = 1 %to &numCustAttributeTypes; 
         %let custAttributeType = %qscan(&custAttributeTypeList, &custAttrTypeNum, %str( ));
         %let custAttributeType = %trim(%left(&custAttributeType));

         proc sql noprint;
            create table &workLib..custAttrType_&custAttributeType as
            select *
            from &workLib..&tableName
            where (&custAttributeTypeColName = &custAttributeType);
         quit;
         %if (&sqlrc gt 4) %then %do;
            %let &_rc = 107;
            %put %mo_error SQL error occurred creating table &workLib..custAttrType_&custAttributeType from table &workLib..&tableName (sqlrc=&sqlrc).;
            %goto UPDATE_BY_CSTAT_ERROR_EXIT;
         %end;

         %local custAttrTypeTableName custAttrTypeIDColName custAttrTypeValueColName;
         %mo_get_mapped_table_info(cstAtrType = &custAttributeType
                                  ,_cstAtrTblNm = custAttrTypeTableName 
                                  ,_cstAtrIDColNm = custAttrTypeIDColName 
                                  ,_cstAtrValColNm = custAttrTypeValueColName
                                  ,_rc = gmtirc);
         %if (&gmtirc ne 0) %then %do;
            %let &_rc = &gmtirc;
            %goto UPDATE_BY_CSTAT_ERROR_EXIT;
         %end;
   
         %let refTbl = %eval(&refTbl + 1);
         %updateReferenceID(rfTblNm = &custAttrTypeTableName
                           ,rfIDColNm = &custAttrTypeIDColName
                           ,diffIDColNm = &custAttributeIDColName
                           ,rfValColNm = &custAttrTypeValueColName
                           ,officialLib = &officialLib
                           ,importLib = &importLib
                           ,workLib = work
                           ,workTbl = custAttrType_&custAttributeType
                           ,workIDColNm = &idColName
                           ,_rfTblNmList = refTableNameList
                           ,_rc = uridrc);
         %if (&uridrc ne 0) %then %do;
            %let &_rc = &uridrc;
            %goto UPDATE_BY_CSTAT_ERROR_EXIT;
         %end;
      %end; /* For each customer attribute type ... */

      %**************************************************************;
      %* Re-create working table from the per-attribute type tables. ; 
      %**************************************************************;
      proc sql noprint;
         create table &workLib..&tableName
         like &workLib..&tableName;
      quit;
      %if (&sqlrc gt 4) %then %do;
         %let &_rc = 107;
         %put %mo_error SQL error occurred creating table &workLib..&tableName like &workLib..&tableName (sqlrc=&sqlrc).;
         %goto UPDATE_BY_CSTAT_ERROR_EXIT;
      %end;

      %do custAttrTypeNum = 1 %to &numCustAttributeTypes; 
         %let custAttributeType = %qscan(&custAttributeTypeList, &custAttrTypeNum, %str( ));
         %let custAttributeType = %trim(%left(&custAttributeType));

         proc append base=&workLib..&tableName data=&workLib..custAttrType_&custAttributeType;
         run;
         %if (&syserr gt 4) %then %do;
            %let &_rc = 100;
            %put %mo_error Error occurred appending table &workLib..custAttrType_&custAttributeType to table &workLib..&tableName (syserr=&syserr).;
            %goto UPDATE_BY_CSTAT_ERROR_EXIT;
         %end;
      %end; /* For each customer attribute type ... */

      %goto UPDATE_BY_CSTAT_EXIT;

      %*--------------------------------;
      %* Exit processing                ;
      %*--------------------------------;
      %UPDATE_BY_CSTAT_ERROR_EXIT:
      %put %mo_error Failed to update ID based on attribute type for table &tableName.;


      %UPDATE_BY_CSTAT_EXIT:
      %put;
      %mo_no_notes(on)
      %put %mo_note End update ID to referenced table %upcase(&tableName) by Customer Attribute Type.;
      %mo_no_notes(off)

   %mend updateIDByCustAttributeType;


   %*----------------------------------;
   %* Helper macro - Update ID based   ;
   %* on communication attribute type  ;
   %*----------------------------------;
   %macro updateIDByCommAttributeType(tableName        =
                                     ,officialLib      =
                                     ,importLib        =
                                     ,workLib          = work
                                     ,refTableNameList =
                                     ,_rc              =
                                     );

      %local macname;
      %let macname=&sysmacroname;
      %let macname=%lowcase(&macname);

      %if %nrbquote(&_rc) eq %str() %then %do;
         %local rc;
         %let _rc = rc;
      %end;
      %else %global &_rc;
      %let &_rc = 0;

      %put;
      %mo_no_notes(on)
      %put %mo_note Begin update ID to referenced table %upcase(&tableName) by Communication Attribute Type.;
      %mo_no_notes(off)

      %if (&local_debug eq 1) %then %do;
         data &workLib..preResolveCommAttrType_&tbl._&refTbl;
            set &workLib..&tableName;
         run;
      %end;

      %local commAttributeTypeColName commAttributeIDColName;
      %mo_get_mapped_table_info(tblNm = %lowcase(&tableName)
                               ,_commAttribTpColNm = commAttributeTypeColName
                               ,_commAttribIDColNm = commAttributeIDColName
                               ,_rc = gmtirc);
      %if (&gmtirc ne 0) %then %do;
         %let &_rc = &gmtirc;
         %goto UPDATE_BY_COMAT_ERROR_EXIT;
      %end;

      %local commAttributeTypeList;
      %let commAttributeTypeList =;
      proc sql noprint;
         select distinct &commAttributeTypeColName into :commAttributeTypeList separated by ' '
         from &workLib..&tableName;
      quit;
      %if (&sqlrc gt 4) %then %do;
         %let &_rc = 107;
         %put %mo_error SQL error occurred obtaining communication attribute types from column &commAttributeTypeColName in table &workLib..&tableName (sqlrc=&sqlrc).;
         %goto UPDATE_BY_COMAT_ERROR_EXIT;
      %end;

      %local numCommAttributeTypes commAttrTypeNum commAttributeType;
      %let numCommAttributeTypes = %mo_word_count(&commAttributeTypeList, %str( ));

      %do commAttrTypeNum = 1 %to &numCommAttributeTypes; 
         %let commAttributeType = %qscan(&commAttributeTypeList, &commAttrTypeNum, %str( ));
         %let commAttributeType = %trim(%left(&commAttributeType));

         proc sql noprint;
            create table &workLib..commAttrType_&commAttributeType as
            select *
            from &workLib..&tableName
            where (&commAttributeTypeColName = &commAttributeType);
         quit;
         %if (&sqlrc gt 4) %then %do;
            %let &_rc = 107;
            %put %mo_error SQL error occurred creating table &workLib..commAttrType_&commAttributeType from table &workLib..&tableName (sqlrc=&sqlrc).;
            %goto UPDATE_BY_COMAT_ERROR_EXIT;
         %end;
   
         %local commAttrTypeTableName commAttrTypeIDColName commAttrTypeValueColName;
         %mo_get_mapped_table_info(comAtrType = &commAttributeType
                                  ,_comAtrTblNm = commAttrTypeTableName 
                                  ,_comAtrIDColNm = commAttrTypeIDColName 
                                  ,_comAtrValColNm = commAttrTypeValueColName
                                  ,_rc = gmtirc);
         %if (&gmtirc ne 0) %then %do;
            %let &_rc = &gmtirc;
            %goto UPDATE_BY_COMAT_ERROR_EXIT;
         %end;
   
         %let refTbl = %eval(&refTbl + 1);
         %updateReferenceID(rfTblNm = &commAttrTypeTableName
                           ,rfIDColNm = &commAttrTypeIDColName
                           ,diffIDColNm = &commAttributeIDColName
                           ,rfValColNm = &commAttrTypeValueColName
                           ,officialLib = &officialLib
                           ,importLib = &importLib
                           ,workLib = work
                           ,workTbl = commAttrType_&commAttributeType
                           ,workIDColNm = &idColName
                           ,_rfTblNmList = refTableNameList
                           ,_rc = uridrc);
         %if (&uridrc ne 0) %then %do;
            %let &_rc = &uridrc;
            %goto UPDATE_BY_COMAT_ERROR_EXIT;
         %end;
      %end; /* For each communication attribute type ... */

      %**************************************************************;
      %* Re-create working table from the per-attribute type tables. ; 
      %**************************************************************;
      proc sql noprint;
         create table &workLib..&tableName
         like &workLib..&tableName;
      quit;
      %if (&sqlrc gt 4) %then %do;
         %let &_rc = 107;
         %put %mo_error SQL error occurred creating table &workLib..&tableName like &workLib..&tableName (sqlrc=&sqlrc).;
         %goto UPDATE_BY_COMAT_ERROR_EXIT;
      %end;

      %do commAttrTypeNum = 1 %to &numCommAttributeTypes; 
         %let commAttributeType = %qscan(&commAttributeTypeList, &commAttrTypeNum, %str( ));
         %let commAttributeType = %trim(%left(&commAttributeType));

         proc append base=&workLib..&tableName data=&workLib..commAttrType_&commAttributeType;
         run;
         %if (&syserr gt 4) %then %do;
            %let &_rc = 100;
            %put %mo_error Error occurred appending table &workLib..commAttrType_&commAttributeType to table &workLib..&tableName (syserr=&syserr).;
            %goto UPDATE_BY_COMAT_ERROR_EXIT;
         %end;
      %end; /* For each communication attribute type ... */

      %goto UPDATE_BY_COMAT_EXIT;

      %*--------------------------------;
      %* Exit processing                ;
      %*--------------------------------;
      %UPDATE_BY_COMAT_ERROR_EXIT:
      %put %mo_error Failed to update ID based on attribute type for table &tableName.;

      %UPDATE_BY_COMAT_EXIT:
      %put;
      %mo_no_notes(on)
      %put %mo_note End update ID to referenced table %upcase(&tableName) by Communication Attribute Type.;
      %mo_no_notes(off)


   %mend updateIDByCommAttributeType;


   %*----------------------------------;
   %* Helper macro - Update ID based   ;
   %* on scenario item type            ;
   %*----------------------------------;
   %macro updateIDByScenarioItemType(tableName        =
                                    ,officialLib      =
                                    ,importLib        =
                                    ,workLib          = work
                                    ,refTableNameList =
                                    ,_rc              =
                                    );

      %local macname;
      %let macname=&sysmacroname;
      %let macname=%lowcase(&macname);

      %if %nrbquote(&_rc) eq %str() %then %do;
         %local rc;
         %let _rc = rc;
      %end;
      %else %global &_rc;
      %let &_rc = 0;

      %put;
      %mo_no_notes(on)
      %put %mo_note Begin update ID to referenced table %upcase(&tableName) by Scenario Item Type.;
      %mo_no_notes(off)

      %if (&local_debug eq 1) %then %do;
         data &workLib..preResolveScenItemType_&tbl._&refTbl;
            set &workLib..&tableName;
         run;
      %end;
   
      %local scenarioItemTypeColName scenarioItemIDColName;
      %mo_get_mapped_table_info(tblNm = %lowcase(&tableName)
                               ,_scenItemTpColNm = scenarioItemTypeColName
                               ,_scenItemIDColNm = scenarioItemIDColName
                               ,_rc = gmtirc);
      %if (&gmtirc ne 0) %then %do;
         %let &_rc = &gmtirc;
         %goto UPDATE_BY_SCNIT_ERROR_EXIT;
      %end;

      %local scenarioItemTypeList;
      %let scenarioItemTypeList =;
      proc sql noprint;
         select distinct &scenarioItemTypeColName into :scenarioItemTypeList separated by ' '
         from &workLib..&tableName;
      quit;
      %if (&sqlrc gt 4) %then %do;
         %let &_rc = 107;
         %put %mo_error SQL error occurred obtaining scenario item types from column &scenarioItemTypeColName in table &workLib..&tableName (sqlrc=&sqlrc).;
         %goto UPDATE_BY_SCNIT_ERROR_EXIT;
      %end;

      %local numScenarioItemTypes scenItemTypeNum scenarioItemType;
      %let numScenarioItemTypes = %mo_word_count(&scenarioItemTypeList, %str( ));

      %do scenItemTypeNum = 1 %to &numScenarioItemTypes; 
         %let scenarioItemType = %qscan(&scenarioItemTypeList, &scenItemTypeNum, %str( ));
         %let scenarioItemType = %trim(%left(&scenarioItemType));

         proc sql noprint;
            create table &workLib..scenItemType_&scenarioItemType as
            select *
            from &workLib..&tableName
            where (&scenarioItemTypeColName = &scenarioItemType);
         quit;
         %if (&sqlrc gt 4) %then %do;
            %let &_rc = 107;
            %put %mo_error SQL error occurred creating table &workLib..scenarioItem_&scenarioItemType from table &workLib..&tableName (sqlrc=&sqlrc).;
            %goto UPDATE_BY_SCNIT_ERROR_EXIT;
         %end;
   
         %local scenItemTableName scenItemIDColName scenItemValueColName;
         %mo_get_mapped_table_info(itmType = &scenarioItemType
                                  ,_itmTblNm = scenItemTableName 
                                  ,_itmIDColNm = scenItemIDColName 
                                  ,_itmValColNm = scenItemValueColName
                                  ,_rc = gmtirc);
         %if (&gmtirc ne 0) %then %do;
            %let &_rc = &gmtirc;
            %goto UPDATE_BY_SCNIT_ERROR_EXIT;
         %end;
   
         %let refTbl = %eval(&refTbl + 1);
         %updateReferenceID(rfTblNm = &scenItemTableName
                           ,rfIDColNm = &scenItemIDColName
                           ,diffIDColNm = &scenarioItemIDColName 
                           ,rfValColNm = &scenItemValueColName
                           ,officialLib = &officialLib
                           ,importLib = &importLib
                           ,workLib = work
                           ,workTbl = scenItemType_&scenarioItemType
                           ,workIDColNm = &idColName
                           ,_rfTblNmList = refTableNameList
                           ,_rc = uridrc);
         %if (&uridrc ne 0) %then %do;
            %let &_rc = &uridrc;
            %goto UPDATE_BY_SCNIT_ERROR_EXIT;
         %end;
      %end; /* For each scenario item type ... */

      %*********************************************************;
      %* Re-create working table from the per-item type tables. ; 
      %*********************************************************;
      proc sql noprint;
         create table &workLib..&tableName
         like &workLib..&tableName;
      quit;
      %if (&sqlrc gt 4) %then %do;
         %let &_rc = 107;
         %put %mo_error SQL error occurred creating table &workLib..&tableName like &workLib..&tableName (sqlrc=&sqlrc).;
         %goto UPDATE_BY_SCNIT_ERROR_EXIT;
      %end;

      %do scenItemTypeNum = 1 %to &numScenarioItemTypes; 
         %let scenarioItemType = %qscan(&scenarioItemTypeList, &scenItemTypeNum, %str( ));
         %let scenarioItemType = %trim(%left(&scenarioItemType));

         proc append base=&workLib..&tableName data=&workLib..scenItemType_&scenarioItemType;
         run;
         %if (&syserr gt 4) %then %do;
            %let &_rc = 100;
            %put %mo_error Error occurred appending table &workLib..scenItemType_&scenarioitemType to table &workLib..&tableName (syserr=&syserr).;
            %goto UPDATE_BY_SCNIT_ERROR_EXIT;
         %end;
      %end; /* For each scenario item type ... */

      %goto UPDATE_BY_SCNIT_EXIT;

      %*--------------------------------;
      %* Exit processing                ;
      %*--------------------------------;
      %UPDATE_BY_SCNIT_ERROR_EXIT:
      %put %mo_error Failed to update ID based on attribute type for table &tableName.;

      %UPDATE_BY_SCNIT_EXIT:
      %put;
      %mo_no_notes(on)
      %put %mo_note End update ID to referenced table %upcase(&tableName) by Scenario Item Type.;
      %mo_no_notes(off)


   %mend updateIDByScenarioItemType;


   %*----------------------------------------------------------------------;
   %* Segregate the scenario definition dataset names from allDef dataset. ;
   %* mo_validate_import_file already verified that it exists.             ;
   %*----------------------------------------------------------------------;
   data _dsimprt.scenarioDef (drop=tableType);
      set _dsimprt.allDef;
      if (tableType = "SCENARIO");
   run;
   %if (&syserr gt 4) %then %do;
      %let &_rc = 100;
      %put %mo_error Error occurred creating _dsimprt.scenarioDef (syserr=&syserr).;
      %goto ERROR_EXIT;
   %end;

   proc sort data=_dsimprt.scenarioDef;
      by relativeOrder tableName;
   run;
   %if (&syserr gt 4) %then %do;
      %let &_rc = 100;
      %put %mo_error Error occurred while sorting _dsimprt.scenarioDef (syserr=&yserr).;
      %goto ERROR_EXIT;
   %end;

   %local numTables;
   %let numTables = 0;
   proc sql noprint;
      select count(*) into :numTables
      from _dsimprt.scenarioDef;
   quit;
   %if (&sqlrc gt 4) %then %do;
      %let &_rc = 107;
      %put %mo_error Error occurred while obtaining count of table names in _dsimprt.scenarioDef (sqlrc=&sqlrc).;
      %goto ERROR_EXIT;
   %end;


   %***************************************************************************;
   %****************************** VALIDATION *********************************;
   %***************************************************************************;
   %if (&action eq VALIDATE) %then %do;
      %put;
      %mo_no_notes(on)
      %put %mo_note START VALIDATION AT %sysfunc(putn(%sysfunc(time()), time.));
      %mo_no_notes(off)

      %if (&local_debug eq 1) %then %do;
         %mo_no_notes(on)
         %put;
         %put %mo_note *** PERFORMING SCENARIO IMPORT VALIDATION ***;
         %mo_no_notes(off)
      %end;


      %*--------------------------------;
      %* Generate a unique scenario ID  ;
      %* from the designated library    ;
      %*--------------------------------;
      %mo_gen_uid(library=&workingDatastoreLib, r_macro=scenario_id, data=&workingDatastoreLib..scenarios, id_var=scenario_id, _rc=uidrc)
      %if (&uidrc ne 0) %then %do;
         %let &_rc = 108;
         %put %mo_error Could not generate a unique scenario_id.;
         %goto ERROR_EXIT;
      %end;


      %*-----------------------------------------------------------------------;
      %* For each scenario definition table, update the input data ID, project ;
      %* ID, and user ID, as appropriate, to those for which it is being       ;
      %* imported. In addition, if the scenario definition table is one that   ;
      %* has ID(s) for other tables, update those ID(s) by looking them up in  ;
      %* their corresponding datastore table(s). Obtain IDs for the scenario   ;
      %* definition table rows in the corresponding datastore table.           ;
      %*-----------------------------------------------------------------------;
   

      %***************************;
      %* For each table in the   *;
      %* scenario definition ... *;
      %***************************;
      %local tbl tableName idColName valueColName otherReferencedTables; 
      %local hasInputDataID hasProjectID hasScenarioID hasUserID;
      %local useAttributeType useCustAttributeType;
      %local useCommAttributetype useScenarioItem;
      %do tbl = 1 %to &numTables;
         data _null_;
            obsnum = &tbl;
            set _dsimprt.scenarioDef point=obsnum;
            call symput("tableName", strip(tableName));
            stop;
         run;
         %if (&syserr gt 4) %then %do;
            %let &_rc = 100;
            %put %mo_error Error occurred reading _dsimprt.scenarioDef (syserr=&syserr).;
            %goto ERROR_EXIT;
         %end;
         %let tableName = %trim(%left(&tableName));

         %put;
         %mo_no_notes(on)
         %put %mo_note Start table &tableName %sysfunc(putn(%sysfunc(time()), time.));
         %mo_no_notes(off)
   
         %mo_get_mapped_table_info(tblNm = %lowcase(&tableName)
                                  ,_idColNm = idColName
                                  ,_valColNm = valueColName
                                  ,_hasInpID = hasInputDataID
                                  ,_hasProjID = hasProjectID
                                  ,_hasScenID = hasScenarioID
                                  ,_hasUsrID = hasUserID
                                  ,_othrRefTbls = otherReferencedTables 
                                  ,_useAttribType = useAttributeType
                                  ,_useCustAttribType = useCustAttributeType
                                  ,_useCommAttribType = useCommAttributeType
                                  ,_useScenItem = useScenarioItem
                                  ,_rc = gmtirc);
         %if (&gmtirc ne 0) %then %do;
            %let &_rc = &gmtirc;
            %goto ERROR_EXIT;
         %end;
   
         %if (&local_debug eq 1) %then %do;
            %mo_no_notes(on)
            %put;
            %put %mo_note TABLE NAME: &tableName;
            %put %mo_note ID COL NAME: &idColName;
            %mo_no_notes(off)
         %end;

         %local refTableNameList; %* For this table, track all other   ;
                                  %* referenced tables, no matter what ;
                                  %* the reference mechanism (direct,  ;
                                  %* indirect/recursion, special).     ;
         %let refTableNameList = &tableName - ;
   

         %*******************************************;
         %* Does table have corresponding datastore *;
         %* datastore table with same column names? *;
         %*******************************************;
         %if (not %sysfunc(exist(%nrbquote(&workingDatastoreLib..&tableName)))) %then %do;
            %let &_rc = 100;
            %put %mo_error Dataset %nrbquote(&tableName) is not a recognized SAS Marketing Optimization datastore table name.;
            %goto ERROR_EXIT;
         %end;
   
         %local datastoreColNames importColNames;
         %let datastoreColNames=;
         proc contents noprint data=&workingDatastoreLib..&tableName out=work.datastoreColNames (keep=name);
         run;
         %if (&syserr gt 4) %then %do;
            %let &_rc = 100;
            %put %mo_error Error occurred using proc contents to obtain column names for &workingDatastoreLib..&tableName (syserr=&syserr).;
            %goto ERROR_EXIT;
         %end;
         proc sql noprint;
            select name into :datastoreColNames separated by " " 
            from work.datastoreColNames;
         quit; 
         %if (&sqlrc gt 4) %then %do;
            %let &_rc = 107;
            %put %mo_error Error occurred reading dataset work.datastoreColNames (sqlrc=&sqlrc).;
            %goto ERROR_EXIT;
         %end;
         %let datastoreColNames = %lowcase(&datastoreColNames);
   
         %let importColNames=;
         proc contents noprint data=_dsimprt.&tableName out=work.importColNames (keep=name);
         run;
         %if (&syserr gt 4) %then %do;
            %let &_rc = 100;
            %put %mo_error Error occurred using proc contents to obtain column names for &workingDatastoreLib..&tableName (syserr=&syserr).;
            %goto ERROR_EXIT;
         %end;
         proc sql noprint;
            select name into :importColNames separated by " " 
            from work.importColNames;
         quit; 
         %if (&sqlrc gt 4) %then %do;
            %let &_rc = 107;
            %put %mo_error Error occurred reading dataset work.importColNames (sqlrc=&sqlrc).;
            %goto ERROR_EXIT;
         %end;
         %let importColNames = %lowcase(&importColNames);
   
         %if (&importColNames ne &datastoreColNames) %then %do;
            %let &_rc = 100;
            %put %mo_error Dataset %nrbquote(&tableName) has different column names in the SAS Marketing Optimization datastore (&datastoreColNames) than in the scenario definition to be imported (&importColNames).;
            %goto ERROR_EXIT;
         %end;
   

         %************************************************;
         %* Set up working table from the _dsimprt table. ;
         %************************************************;
         %local numRows row;
         %let numRows = 0;
         proc sql noprint;
            select count(*) into :numRows
            from _dsimprt.&tableName;
         quit;
         %if (&sqlrc gt 4) %then %do;
            %let &_rc = 107;
            %put %mo_error Error occurred while obtaining count of table names in _dsimprt.&tableName (sqlrc=&sqlrc).;
            %goto ERROR_EXIT;
         %end;
         
         proc sql noprint;
            create table work.&tableName as
            select *
            from _dsimprt.&tableName;
         quit;
         %if (&sqlrc gt 4) %then %do;
            %let &_rc = 107;
            %put %mo_error SQL error occurred creating table work.&tableName from _dsimprt.&tableName (sqlrc=&sqlrc).;
            %goto ERROR_EXIT;
         %end;
         

         %***************************************;
         %* Generate new IDs for the table rows. ;
         %***************************************;
         %global newRowID;
         %let newRowID = -1;
         %if ((%lowcase(&tableName) ne scenarios) and
              (%lowcase(&tableName) ne scenario_advanced_options) and
              (%lowcase(&tableName) ne optimization_jobs)) %then %do;
            %mo_gen_uid(library=&workingDatastoreLib, r_macro=newRowID, data=&workingDatastoreLib..&tableName, id_var=&idColName, _rc=uidrc);
            %if ((&uidrc ne 0) or (&newRowID eq -1)) %then %do;
               %let &_rc = 108;
               %put %mo_error Could not generate a unique &idColName for &workingDatastoreLib..&tableName.;
               %goto ERROR_EXIT;
            %end;
         %end;
         %else %do;
            %let newRowID = &scenario_id;
         %end;
 
         data work.&tableName;
            set work.&tableName;
            origID = &idColName;
            &idColName = (_n_ - 1) + &newRowID;
         run;
         %if (&syserr gt 4) %then %do;
            %let &_rc = 100;
            %put %mo_error Error occurred updating row IDs for table work.&tableName (syserr=&syserr).;
            %goto ERROR_EXIT;
         %end;


         %********************************;
         %* Update referenced table IDs. *;
         %********************************;
         %local numRefTables refPair; 
         %local refTableName refIDColName usesDiffIDColName refValueColName;

         %let numRefTables = %mo_word_count(&otherReferencedTables, %str( ));
         %do refTbl = 1 %to &numRefTables;
            %let refPair = %qscan(&otherReferencedTables, &refTbl, %str( ));
            %let refTableName = %qscan(&refPair, 1, %str(-)); 
            %let usesDiffIDColName = %qscan(&refPair, 2, %str(-)); 

            %let refTableName = %trim(%left(&refTableName));
            %let usesDiffIDColName = %trim(%left(&usesDiffIDColName));
   
            %if (&local_debug eq 1) %then %do;
               %mo_no_notes(on)
               %put;
               %put %mo_note numRefTables: &numRefTables;
               %put %mo_note refPair: &refPair;
               %put %mo_note refTableName: &refTableName;
               %put %mo_note usesDiffIDColName: &usesDiffIDColName;
               %mo_no_notes(off)
            %end;
      
            %*****************************************;
            %* Certain referenced tables are handled *;
            %* separately. Do the following if the   *;
            %* referenced table is not one of these. *;
            %*****************************************;
            %if ((%lowcase(&refTableName) ne input_data) and
                 (%lowcase(&refTableName) ne projects) and
                 (%lowcase(&refTableName) ne scenarios) and
                 (%lowcase(&refTableName) ne users) and
                 (%lowcase(&refTableName) ne prepopulated_constraints) and
                 (%lowcase(&refTableName) ne prepopulated_contact_policies)) %then %do;

               %put;
               %mo_no_notes(on)
               %put %mo_note Begin update ID to referenced table %upcase(&refTableName) by normal mapping.;
               %mo_no_notes(off)
   
               %mo_get_mapped_table_info(tblNm = %lowcase(&refTableName)
                                        ,_idColNm = refIDColName
                                        ,_valColNm = refValueColName
                                        ,_rc = gmtirc);
               %if (&gmtirc ne 0) %then %do;
                  %let &_rc = &gmtirc;
                  %goto ERROR_EXIT;
               %end;

               %updateReferenceID(rfTblNm = &refTableName
                                 ,rfIDColNm = &refIDColName
                                 ,diffIDColNm = &usesDiffIDColName
                                 ,rfValColNm = &refValueColName
                                 ,officialLib = &workingDatastoreLib
                                 ,importLib = _dsimprt
                                 ,workLib = work
                                 ,workTbl = &tableName
                                 ,workIDColNm = &idColName
                                 ,_rfTblNmList = refTableNameList
                                 ,_rc = uridrc);
               %if (&uridrc ne 0) %then %do;
                  %let &_rc = &uridrc;
                  %goto ERROR_EXIT;
               %end;

               %put;
               %mo_no_notes(on)
               %put %mo_note End update ID to referenced table %upcase(&refTableName) by normal mapping.;
               %mo_no_notes(off)
            %end; /* If not input_data, projects, scenarios, users, prepopulated_constraints, prepopulated_contact_policies ... */
         %end; /* For each referenced table ... */
   

         %*****************************************;
         %* Update tables that use attribute type *;
         %* to identify the referenced table ID   *;
         %*****************************************;
         %if (&useAttributeType eq 1) %then %do;
            %updateIDByAttributeType(tableName = &tableName
                                    ,officialLib = &workingDatastoreLib
                                    ,importLib = _dsimprt
                                    ,workLib = work
                                    ,refTableNameList = refTableNameList
                                    ,_rc = ubatrc);
            %if (&ubatrc ne 0) %then %do;
               %let &_rc = &ubatrc;
               %goto ERROR_EXIT;
            %end;
         %end; /* If useAttributeType ... */
   

         %*********************************************;
         %* Update tables that use customer attribute *;
         %* type to identify the referenced table ID  *;
         %*********************************************;
         %if (&useCustAttributeType eq 1) %then %do;
            %updateIDByCustAttributeType(tableName = &tableName
                                        ,officialLib = &workingDatastoreLib
                                        ,importLib = _dsimprt
                                        ,workLib = work
                                        ,refTableNameList = refTableNameList
                                        ,_rc = ubcstatrc);
            %if (&ubcstatrc ne 0) %then %do;
               %let &_rc = &ubcstatrc;
               %goto ERROR_EXIT;
            %end;
         %end; /* If useCustAttributeType ... */

   
         %**************************************************;
         %* Update tables that use communication attribute *;
         %* type to identify the referenced table ID       *;
         %**************************************************;
         %if (&useCommAttributeType eq 1) %then %do;
            %updateIDByCommAttributeType(tableName = &tableName
                                        ,officialLib = &workingDatastoreLib
                                        ,importLib = _dsimprt
                                        ,workLib = work
                                        ,refTableNameList = refTableNameList
                                        ,_rc = ubcomatrc);
            %if (&ubcomatrc ne 0) %then %do;
               %let &_rc = &ubcomatrc;
               %goto ERROR_EXIT;
            %end;
         %end; /* If useCommAttributeType ... */
      

         %*********************************************;
         %* Update tables that use scenario item type *;
         %* to identify the referenced table ID       *;
         %*********************************************;
         %if (&useScenarioItem eq 1) %then %do;
            %updateIDByScenarioItemType(tableName = &tableName
                                       ,officialLib = &workingDatastoreLib
                                       ,importLib = _dsimprt
                                       ,workLib = work
                                       ,refTableNameList = refTableNameList
                                       ,_rc = ubscnitrc);
            %if (&ubscnitrc ne 0) %then %do;
               %let &_rc = &ubscnitrc;
               %goto ERROR_EXIT;
            %end;
         %end; /* If useScenarioItem ... */
   
   
         %***********************;
         %* Update primary IDs. *;
         %***********************;
         data work.&tableName;
            set work.&tableName;
            %if (&hasInputDataID eq 1) %then %do;
               input_data_id = &input_data_id;
            %end;
            %if (&hasProjectID eq 1) %then %do;
               project_id = &project_id;
            %end;
            %if (&hasUserID eq 1) %then %do;
               user_id = &user_id;
            %end;
            %if (&hasScenarioID eq 1) %then %do;
               scenario_id = &scenario_id;
            %end;
         run;
         %if (&syserr gt 4) %then %do;
            %let &_rc = 100;
            %put %mo_error Error occurred updating primary IDs in work.&tableName (syserr=&syserr).;
            %goto ERROR_EXIT;
         %end;
   

         %*********************************;
         %* Reset scenario status-related *;
         %* fields in relevant tables.    *;
         %*********************************;
         %if (%lowcase(&tableName) eq scenarios) %then %do;
            data work.&tableName;
               set work.&tableName;
               scenario_name = "&scenario_name";
               scenario_description = "&scenario_description";
               datetime_created = &datetime_created;
               datetime_modified = &datetime_created;
               status = &_UNOPTIMIZED_;
               objective_value = .;
               constraint_summary_report_status = 0;
               summary_report_status = 0;
               sensitivity_report_status = 0;
               job_flow_id = "";
               datetime_last_scheduled = .;
               recur_indefinitely = 0;
               promote = 0;
               promote_solution_dataset_name="";
               promote_solution_dataset_desc="";
               promote_num_communications = 0;
               duplicate = 0;
            run;
         %end;
         %else %if (%lowcase(&tableName) eq scenario_advanced_options) %then %do;
            data work.&tableName;
               set work.&tableName;
               secondary_obj_value = .;
               %if (&input_data_uses_agents = 0) %then %do;
                  active_agent_constraints = 0;
               %end;
               %if (&input_data_uses_history = 0) %then %do;
                  active_contact_history = 0;
               %end;
            run;
         %end;
         %else %if (%lowcase(&tableName) eq constraints) %then %do;
            data work.&tableName;
               set work.&tableName;
               binding = 0;
               infeasible = 0;
            run;
         %end;
         %else %if (%lowcase(&tableName) eq optimization_jobs) %then %do;
            data work.&tableName;
               set work.&tableName;
               datetime_started = .;
               datetime_finished = .;
               error_code = .;
               pid = .;
               datetime_reports_finished = .;
               reports_finished = .;
            run;
         %end;
         %else %if (%lowcase(&tableName) eq opt_group_assign) %then %do;
            data work.&tableName;
               set work.&tableName;
               optimization_group_id = "";
               description = "";
               datetime = .;
               active = 0;
               scenario_name_deleted = "";
               scenario_desc_deleted = "";
               project_name_deleted = "";
               project_desc_deleted = "";
               user_name_deleted = "";
               user_desc_deleted = "";
            run;
         %end;
         %if (&syserr gt 4) %then %do;
            %let &_rc = 100;
            %put %mo_error Error occurred resetting status fields in work.&tableName (syserr=&syserr).;
            %goto ERROR_EXIT;
         %end;
   
         %**********************************;
         %* Update corresponding datastore *;
         %* table with new rows.           *;
         %**********************************;
         %if (&numRows gt 0) %then %do;
            data work.noOrigID_&tbl;
               set work.&tableName (drop=origID);
            run;
            %if (&syserr gt 4) %then %do;
               %let &_rc = 100;
               %put %mo_error Error occurred dropping origID column from work.&tableName (syserr=&syserr).;
               %goto ERROR_EXIT;
            %end;

            %local deleteIDList;
            %let deleteIDList =;
            proc sql noprint;
               select &idColName into :deleteIDList separated by ", "
               from work.noOrigID_&tbl;
            quit;
            %if (&sqlrc gt 4) %then %do;
               %let &_rc = 107;
               %put %mo_error Error occurred obtaining IDs to delete from work.noOrigID_&tbl (sqlrc=&sqlrc).;
               %goto ERROR_EXIT;
            %end;
 
            %mo_record_delete(metadata_table = &workingDatastoreLib..&tableName
                             ,where          = &idColName in (&deleteIDList)
                             ,_rc            = delrc);
            %if (&delrc ne 0) %then %do;
               %let &_rc = 103;
               %goto ERROR_EXIT;
            %end;
            %mo_no_notes(off)
   
            %local byList;
            %let byList = ;
            %if (&hasInputDataID eq 1) %then %do;
               %let byList = &byList input_data_id;
            %end;
            %if (&hasProjectID eq 1) %then %do;
               %let byList = &byList project_id;
            %end;
            %if ((%lowcase(&tableName) ne scenarios) and
                 (%lowcase(&tableName) ne scenario_advanced_options) and
                 (%lowcase(&tableName) ne optimization_jobs) and
                 (&hasScenarioID eq 1)) %then %do;
               %let byList = &byList scenario_id;
            %end;
            %let byList = &byList &idColName;
   
            %if (&local_debug eq 1) %then %do;
               %mo_no_notes(on)
               %put;
               %put %mo_note byList: &byList;
               %mo_no_notes(off)
            %end;

            /* Create temp table without the origID column */
   
            %if (&numRows gt 1) %then %do;
               %mo_record_insert_multiple(metadata_table = &workingDatastoreLib..&tableName
                                         ,update_table   = work.noOrigID_&tbl
                                         ,by             = &byList
                                         ,_rc            = updrc);
               %if (&updrc ne 0) %then %do;
                  %let &_rc = 103;
                  %goto ERROR_EXIT;
               %end;
               %mo_no_notes(off)
            %end;
            %else %do;
               %mo_record_insert(metadata_table = &workingDatastoreLib..&tableName
                                ,update_table   = work.noOrigID_&tbl
                                ,by             = &byList
                                ,_rc            = updrc);
               %if (&updrc ne 0) %then %do;
                  %let &_rc = 103;
                  %goto ERROR_EXIT;
               %end;
               %mo_no_notes(off)
            %end;
         %end;

         %put;
         %mo_no_notes(on)
         %put %mo_note End table &tableName %sysfunc(putn(%sysfunc(time()), time.));
         %mo_no_notes(off)
      %end; /* For each table ... */



      %*---------------------------------------------------------------;
      %* If this is a validation request, determine which scenario     ;
      %* elements of the imported scenario definition do not reconcile ;
      %* to the input data against which it is to be imported.         ; 
      %* reconcile / update scenario elements to reflect the new input ;
      %* data against which it is to be imported.                      ;
      %*---------------------------------------------------------------;

      %**************************************************************;
      %* Scenario elements in general (re-use mo_copy_scenario code) ;
      %**************************************************************;
      %mo_refresh_precalc_measures(data_id_FROM = &import_input_data_id
                                  ,data_id_TO   = &input_data_id
                                  ,library_FROM = _dsimprt
                                  ,library_TO   = &workingDatastoreLib
                                  ,_rc          = rpmrc);
      %if (&rpmrc ne 0) %then %do;
         %let &_rc = 200;
         %put %mo_error Error occurred during precalculated measures check.;
      %end;

      %mo_refresh_check_scenario(scen_id_FROM = &import_scenario_id
                                ,scen_id_TO   = &scenario_id
                                ,library_FROM = _dsimprt
                                ,library_TO   = &workingDatastoreLib
                                ,_rc          = rcsrc);
      %if (&rcsrc ne 0) %then %do;
         %let &_rc = 200;
         %put %mo_error Error occurred during check for import scenario.;
      %end;

      %mo_refresh_update_scen_tables(scen_id_FROM = &import_scenario_id
                                    ,scen_id_TO   = &scenario_id
                                    ,library_FROM = _dsimprt
                                    ,library_TO   = &workingDatastoreLib
                                    ,datetime     = &datetime_created
                                    ,action       = &_SCENARIO_IMPORT_ 
                                    ,_rc          = rutrc);
      %if (&rutrc ne 0) %then %do;
         %let &_rc = 200;
         %put %mo_error Error occurred during update for import scenario.;
      %end;


      %***********************************************************;
      %* Scenario-level constraints and contact policies from     ;
      %* their corresponding prepopulated input data-level tables ;
      %***********************************************************;
      %*---------------------------------------;
      %* Take the artifically added missing    ;
      %* entries out of the:                   ;
      %* - prepopulated constraints table      ;
      %* - prepopulated contact policies table ;
      %*---------------------------------------;
      data &workingDatastoreLib..prepopulated_constraints;
         set &workingDatastoreLib..prepopulated_constraints;
         if ((prepopulated_constraint_id = .) and (constraint_name = "missing")) then do;
            delete;
         end;
      run;
      %if (&syserr gt 4) %then %do;
         %let &_rc = 100;
         %put %mo_error Error occurred reading rewriting _validat.prepopulated_constraints (syserr=&syserr).;
         %goto ERROR_EXIT;
      %end;

      data _dsimprt.prepopulated_constraints;
         set _dsimprt.prepopulated_constraints;
         if ((prepopulated_constraint_id = .) and (constraint_name = "missing")) then do;
            delete;
         end;
      run;
      %if (&syserr gt 4) %then %do;
         %let &_rc = 100;
         %put %mo_error Error occurred reading rewriting _dsimprt.prepopulated_constraints (syserr=&syserr).;
         %goto ERROR_EXIT;
      %end;

      data &workingDatastoreLib..prepopulated_contact_policies;
         set &workingDatastoreLib..prepopulated_contact_policies;
         if ((prepopulated_contact_policy_id = .) and (contact_policy_name = "missing")) then do;
            delete;
         end;
      run;
      %if (&syserr gt 4) %then %do;
         %let &_rc = 100;
         %put %mo_error Error occurred reading rewriting _validat.prepopulated_contact_policies (syserr=&syserr).;
         %goto ERROR_EXIT;
      %end;

      data _dsimprt.prepopulated_contact_policies;
         set _dsimprt.prepopulated_contact_policies;
         if ((prepopulated_contact_policy_id = .) and (contact_policy_name = "missing")) then do;
            delete;
         end;
      run;
      %if (&syserr gt 4) %then %do;
         %let &_rc = 100;
         %put %mo_error Error occurred reading rewriting _dsimprt.prepopulated_contact_policies (syserr=&syserr).;
         %goto ERROR_EXIT;
      %end;

      %global cons_to_add cons_to_del cons_to_swap_FROM cons_to_swap_TO;
      %global pols_to_add pols_to_del pols_to_swap_FROM pols_to_swap_TO;
      %let cons_to_add =;
      %let cons_to_del =;
      %let cons_to_swap_FROM =;
      %let cons_to_swap_TO =;
      %let pols_to_add =;
      %let pols_to_del =;
      %let pols_to_swap_FROM =;
      %let pols_to_swap_TO =;
      %mo_compare_prepop_constraints(original_input_data_id    = &import_input_data_id
                                    ,input_data_id             = &input_data_id
                                    ,library_FROM              = _dsimprt
                                    ,library_TO                = &workingDatastoreLib
                                    ,_constraints_to_add       = cons_to_add
                                    ,_constraints_to_delete    = cons_to_del
                                    ,_constraints_to_swap_FROM = cons_to_swap_FM
                                    ,_constraints_to_swap_TO   = cons_to_swap_TO
                                    ,_policies_to_add          = pols_to_add
                                    ,_policies_to_delete       = pols_to_del
                                    ,_policies_to_swap_FROM    = pols_to_swap_FM
                                    ,_policies_to_swap_TO      = pols_to_swap_TO
                                    ,_rc                       = cpcrc);
      %if (&cpcrc ne 0) %then %do;
         %let &_rc = 100;
         %put %mo_error Error occurred comparing the prepopulated constraints and contact policies (cpcrc=&cpcrc).;
      %end;

      %************************************************************************;
      %* Reconciliation of constraints and contact policies to the new input   ;
      %* data prepopulated constraints and prepopulated contact policies       ;
      %* was bypassed above. The addition of items only in the new input data, ;
      %* deletion of items only in the old input data, and swapping of the IDs ;
      %* of items in both the old and new input data (i.e., updating of all    ;
      %* aspects of the prepopulated field in those tables) is done here.      ;
      %************************************************************************;
      %mo_reconcile_prepop_constraints(constraints_to_add       = &cons_to_add
                                      ,constraints_to_delete    = &cons_to_del
                                      ,constraints_to_swap_FROM = &cons_to_swap_FM
                                      ,constraints_to_swap_TO   = &cons_to_swap_TO
                                      ,policies_to_add          = &pols_to_add
                                      ,policies_to_delete       = &pols_to_del
                                      ,policies_to_swap_FROM    = &pols_to_swap_FM
                                      ,policies_to_swap_TO      = &pols_to_swap_TO
                                      ,library_FROM             = _dsimprt
                                      ,library_TO               = &workingDatastoreLib
                                      ,input_data_id_FROM       = &import_input_data_id
                                      ,input_data_id_TO         = &input_data_id
                                      ,project_id               = &project_id
                                      ,scenario_id              = &scenario_id
                                      ,user_id                  = &user_id
                                      ,datetime_created         = &datetime_created
                                      ,_rc                      = rpcrc);
      %if (&rpcrc ne 0) %then %do;
         %let &_rc = 100;
         %put %mo_error Error occurred reconciling the prepopulated constraints and contact policies (rpcrc=&rpcrc).;
      %end;


      %*************************************************;
      %* Scenario-level precalculated attributes from   ;
      %* the input data-level input data measures table ;
      %*************************************************;
      %global meas_to_add meas_to_del meas_to_swap_FROM meas_to_swap_TO;
      %let meas_to_add =;
      %let meas_to_del =;
      %let meas_to_swap_FROM =;
      %let meas_to_swap_TO =;
      %mo_compare_input_data_measures(input_data_id_FROM     = &import_input_data_id
                                     ,input_data_id_TO       = &input_data_id
                                     ,library_FROM           = _dsimprt
                                     ,library_TO             = &workingDatastoreLib
                                     ,_measures_to_add       = meas_to_add
                                     ,_measures_to_delete    = meas_to_del
                                     ,_measures_to_swap_FROM = meas_to_swap_FM
                                     ,_measures_to_swap_TO   = meas_to_swap_TO
                                     ,_rc                    = cidmrc);
      %if (&cidmrc ne 0) %then %do;
         %let &_rc = 100;
         %put %mo_error Error occurred comparing the input data measures (cidmrc=&cidmrc).;
      %end;

      %************************************************************************;
      %* Reconciliation of precalculated attributes to the new input data      ;
      %* measures was NOT bypassed above with respect to deletion of items     ;
      %* only in the old input data measures, and swapping of the IDs of items ;
      %* in both the old and new input data measures, i.e., those actions have ;
      %* already been performed by the time we get here. All that is left is   ;
      %* to add items that are only in the new input data measures.            ;
      %************************************************************************;
      %mo_reconcile_input_data_measures(measures_to_add       = &meas_to_add
                                       ,measures_to_delete    = 
                                       ,measures_to_swap_FROM = 
                                       ,measures_to_swap_TO   =
                                       ,library_FROM          = _dsimprt
                                       ,library_TO            = &workingDatastoreLib
                                       ,input_data_id_FROM    = &import_input_data_id
                                       ,input_data_id_TO      = &input_data_id
                                       ,project_id            = &project_id
                                       ,scenario_id           = &scenario_id
                                       ,user_id               = &user_id
                                       ,_rc                   = ridmrc);
      %if (&ridmrc ne 0) %then %do;
         %let &_rc = 100;
         %put %mo_error Error occurred reconciling the input data measures (ridmrc=&ridmrc).;
      %end;


      %*********************************************************;
      %* Scenario-level comm numeric attribute values from the  ;
      %* input data-level comm numeric attribute defaults table ;
      %*********************************************************;
      
      %* If _dsimprt.comm_numeric_attribute_values_cc exists, then this ;
      %* "cell_codes" version needs to replace the "normal" version of  ;
      %*  _dsimport.comm_numeric_attribute_values temporarily for the   ;
      %* comparison below.                                              ;
      %if (%sysfunc(exist(_dsimprt.comm_numeric_attribute_values_cc))) %then %do;
         proc datasets library=_dsimprt nolist;
            change 
               comm_numeric_attribute_values=comm_numeric_attribute_vals_temp
               comm_numeric_attribute_values_cc=comm_numeric_attribute_values;
         run;
         quit;
         %if (&syserr gt 4) %then %do;
            %let &_rc = 100;
            %put %mo_error Error occurred saving/renaming _DSIMPRT.COMM_NUMERIC_ATTRIBUTE_VALUES (syserr=&syserr).;
            %goto ERROR_EXIT;
         %end;
      %end;

      %global attr_to_add attr_to_del attr_to_swap_FROM attr_to_swap_TO;
      %let attr_to_add =;
      %let attr_to_del =;
      %let attr_to_swap_FROM =;
      %let attr_to_swap_TO =;
      %mo_compare_comm_num_attr_dflts(input_data_id_FROM  = &import_input_data_id
                                     ,scenario_id_FROM    = &import_scenario_id
                                     ,input_data_id_TO    = &input_data_id
                                     ,library_FROM        = _dsimprt
                                     ,library_TO          = &workingDatastoreLib
                                     ,id_FROM_mapping_tbl = work.comm_numeric_attribute_values
                                     ,id_FROM_mapping_col = origID
                                     ,id_TO_mapping_col   = comm_attribute_value_id
                                     ,_dflts_to_add       = attr_to_add
                                     ,_vals_to_delete     = attr_to_del
                                     ,_attrs_to_swap_FROM = attr_to_swap_FM
                                     ,_attrs_to_swap_TO   = attr_to_swap_TO
                                     ,_rc                 = ccnadrc);
      %if (&cidmrc ne 0) %then %do;
         %let &_rc = 100;
         %put %mo_error Error occurred comparing the comm numeric attribute defaults (ccnadrc=&ccnadrc).;
      %end;

      %if (%sysfunc(exist(_dsimprt.comm_numeric_attribute_vals_temp))) %then %do;
         proc datasets library=_dsimprt nolist;
            change 
               comm_numeric_attribute_values=comm_numeric_attribute_values_cc
               comm_numeric_attribute_vals_temp=comm_numeric_attribute_values;
         run;
         quit;
         %if (&syserr gt 4) %then %do;
            %let &_rc = 100;
            %put %mo_error Error occurred restoring/renaming _DSIMPRT.COMM_NUMERIC_ATTRIBUTE_VALUES (syserr=&syserr).;
            %goto ERROR_EXIT;
         %end;
      %end;

      %************************************************************************;
      %* Reconcile of comm numeric attribute values to the new comm numeric    ;
      %* attribute defaults.                                                   ;
      %* Note: Swapping of campaign, communication and communication attribute ;
      %* has already been done by the main body of code so no need to repeat.  ;
      %************************************************************************;
      %mo_reconcile_comm_num_attr_dflts(dflts_to_add       = &attr_to_add
                                       ,vals_to_delete     = &attr_to_del
                                       ,attrs_to_swap_FROM = 
                                       ,attrs_to_swap_TO   = 
                                       ,library_FROM       = _dsimprt
                                       ,library_TO         = &workingDatastoreLib
                                       ,input_data_id_FROM = &import_input_data_id
                                       ,input_data_id_TO   = &input_data_id
                                       ,project_id_TO      = &project_id
                                       ,scenario_id_TO     = &scenario_id
                                       ,user_id_TO         = &user_id
                                       ,_rc                = rcnadmrc
                                       );
      %if (&rcnadmrc ne 0) %then %do;
         %let &_rc = 100;
         %put %mo_error Error occurred reconciling the comm numeric attribute defaults (rcnadrc=&rcnadrc).;
      %end;


      %**********************************************;
      %* Store import messages into refresh_messages ;
      %* table for surfacing at the UI.              ;
      %**********************************************;
      %if ((%nrbquote(&cons_to_del) ne %str()) or
           (%nrbquote(&pols_to_del) ne %str()) or
           (%nrbquote(&meas_to_del) ne %str()) or
           (%nrbquote(&attr_to_del) ne %str())) %then %do;

         %mo_create_tables(lib=work, table=refresh_messages, _rc = cudrc);
         %if (&cudrc ne 0) %then %do;
            %let &_rc = 100;
            %put %mo_error Error occurred creating WORK.REFRESH_MESSAGES table (cudrc=&cudrc).;
            %goto ERROR_EXIT;
         %end;

         proc sql noprint;
            insert into work.refresh_messages
            %if (%nrbquote(&cons_to_del) ne %str()) %then %do;
               set message = "refresh.msg.oneormore.constraints.deleted.fmt.txt,0",
                   reason = "refresh.msg.oneormore.constraints.deleted.reason.fmt.txt,0"
            %end; 

            %if (%nrbquote(&pols_to_del) ne %str()) %then %do;
               set message = "refresh.msg.oneormore.contactpols.deleted.fmt.txt,0",
                   reason = "refresh.msg.oneormore.contactpols.deleted.reason.fmt.txt,0"
            %end; 

            %if (%nrbquote(&meas_to_del) ne %str()) %then %do;
               set message = "refresh.msg.oneormore.precalcattrs.deleted.fmt.txt,0",
                   reason = "refresh.msg.oneormore.precalcattrs.deleted.reason.fmt.txt,0"
            %end; 
   
            %if (%nrbquote(&attr_to_del) ne %str()) %then %do;
               set message = "refresh.msg.oneormore.commnumattrvals.deleted.fmt.txt,0",
                   reason = "refresh.msg.oneormore.commnumattrvals.deleted.reason.fmt.txt,0"
            %end; 
            ;
         quit;
         %if (&sqlrc gt 4) %then %do;
            %let &_rc = 107;
            %put %mo_error Error occurred inserting message entries into WORK.REFRESH_MESSAGES table (sqlrc=&sqlrc).;
            %goto ERROR_EXIT;
         %end;

         data work.refresh_messages;
            set work.refresh_messages;

            datetime_refreshed = &datetime_created;
            message_type = &_SCENARIO_IMPORT_;     
            scenario_name = "&scenario_name";
            project_name = "&project_name";
            name = "&actualname";
            user_id = &user_id;
            project_id = &project_id;
            scenario_id = &scenario_id;
         run;
         %if (&syserr gt 4) %then %do;
            %let &_rc = 100;
            %put %mo_error Error occurred updating WORK.REFRESH_MESSAGES (syserr=&syserr).;
            %goto ERROR_EXIT;
         %end;

         data &workingDatastoreLib..refresh_messages;
            set work.refresh_messages &workingDatastoreLib..refresh_messages;
            sort_order = _n_;
         run;
         %if (&syserr gt 4) %then %do;
            %let &_rc = 200;
            %put %mo_warning Warning messages were not created properly (rarc=&rarc).;
         %end;

         %let numRefreshMessages = 0;
         proc sql noprint;
            select count(*) into :numRefreshMessages
            from &workingDatastoreLib..refresh_messages;
         quit;
         %if (&sqlrc gt 4) %then %do;
            %let &_rc = 107;
            %put %mo_error Error occurred counting the number of message entries in &workingDatastoreLib.refresh_messages table (sqlrc=&sqlrc).;
            %goto ERROR_EXIT;
         %end;
      %end; 

      %put;
      %mo_no_notes(on)
      %put %mo_note END VALIDATION AT %sysfunc(putn(%sysfunc(time()), time.));
      %mo_no_notes(off)
   %end;


   %***************************************************************************;
   %********************************** COMMIT *********************************;
   %***************************************************************************;
   %if (&action eq COMMIT) %then %do;
      %put;
      %mo_no_notes(on)
      %put %mo_note START COMMITMENT AT %sysfunc(putn(%sysfunc(time()), time.));
      %mo_no_notes(off)

      %*------------------------------------------------------------;
      %* Obtain lock to maintain datastore integrity during commit. ;
      %*------------------------------------------------------------;
      %local dsLocked;
      %let dsLocked = 0;
      %mo_lock_dataset(data=mometa.uid,_rc=lckrc)
      %if (&lckrc ne 0) %then %do;
         %put %mo_error Error obtaining lock of MOMETA.UID.;
         %let &_rc = 106;
         %goto ERROR_EXIT;
      %end;
      %let dsLocked = 1;


      %*--------------------------------;
      %* Generate a unique scenario ID  ;
      %* from the designated library    ;
      %*--------------------------------;
      %mo_gen_uid(library=&workingDatastoreLib, r_macro=scenario_id, data=&workingDatastoreLib..scenarios, id_var=scenario_id, _rc=uidrc)
      %if (&uidrc ne 0) %then %do;
         %let &_rc = 108;
         %put %mo_error Could not generate a unique scenario_id.;
         %goto ERROR_EXIT;
      %end;


      %*-----------------------------------------------------------------------;
      %* Copy each scenario table created in the validation directory by the   ;
      %* prior validation request to the specified datastore. At this point,   ;
      %* all of the input data level-IDs have been updated and scenario-level  ;
      %* elements have been cleaned up.                                        ;
      %* Obtain a new ID for each new row/entry in the scenario validation     ;
      %* tables since the ones created during that process were not based on   ;
      %* the official datastore tables.                                        ;
      %*-----------------------------------------------------------------------;
   
      %***************************;
      %* For each table in the   *;
      %* scenario definition ... *;
      %***************************;
      %local tableName;
      %local idColName hasInputDataID hasProjectID hasScenarioID hasUserID; 
      %local otherReferencedTables useAttributeType useCustAttributeType;
      %local useCommAttributetype useScenarioItem;
      %do tbl = 1 %to &numTables;
         data _null_;
            obsnum = &tbl;
            set _dsimprt.scenarioDef point=obsnum;
            call symput("tableName", strip(tableName));
            stop;
         run;
         %if (&syserr gt 4) %then %do;
            %let &_rc = 100;
            %put %mo_error Error occurred reading _dsimprt.scenarioDef (syserr=&syserr).;
            %goto ERROR_EXIT;
         %end;
         %let tableName = %trim(%left(&tableName));
   
         %mo_get_mapped_table_info(tblNm = %lowcase(&tableName)
                                  ,_idColNm = idColName
                                  ,_hasInpID = hasInputDataID
                                  ,_hasProjID = hasProjectID
                                  ,_hasScenID = hasScenarioID
                                  ,_hasUsrID = hasUserID
                                  ,_othrRefTbls = otherReferencedTables 
                                  ,_useAttribType = useAttributeType
                                  ,_useCustAttribType = useCustAttributeType
                                  ,_useCommAttribType = useCommAttributeType
                                  ,_useScenItem = useScenarioItem
                                  ,_rc = gmtirc);
         %if (&gmtirc ne 0) %then %do;
            %let &_rc = &gmtirc;
            %goto ERROR_EXIT;
         %end;
   
         %if (&local_debug eq 1) %then %do;
            %mo_no_notes(on)
            %put;
            %put %mo_note TABLE NAME: &tableName;
            %put %mo_note ID COLUMN NAME: &idColName;
            %mo_no_notes(off)
         %end;
   

         %************************************************;
         %* Set up working table from the _validat table. ;
         %************************************************;
         %local numRows row;
         %let numRows = 0;
         proc sql noprint;
            select count(*) into :numRows
            from _validat.&tableName;
         quit;
         %if (&sqlrc gt 4) %then %do;
            %let &_rc = 107;
            %put %mo_error Error occurred while obtaining count of table names in _validat.&tableName (sqlrc=&sqlrc).;
            %goto ERROR_EXIT;
         %end;
         
         proc sql noprint;
            create table work.&tableName as
            select *
            from _validat.&tableName;
         quit;
         %if (&sqlrc gt 4) %then %do;
            %let &_rc = 107;
            %put %mo_error SQL error occurred creating table work.&tableName from _validat.&tableName (sqlrc=&sqlrc).;
            %goto ERROR_EXIT;
         %end;


         %***************************************;
         %* Generate new IDs for the table rows. ;
         %***************************************;
         %global newRowID;
         %let newRowID = -1;
         %if ((%lowcase(&tableName) ne scenarios) and
              (%lowcase(&tableName) ne scenario_advanced_options) and
              (%lowcase(&tableName) ne optimization_jobs)) %then %do;
            %mo_gen_uid(library=&workingDatastoreLib, r_macro=newRowID, data=&workingDatastoreLib..&tableName, id_var=&idColName, _rc=uidrc);
            %if ((&uidrc ne 0) or (&newRowID eq -1)) %then %do;
               %let &_rc = 108;
               %put %mo_error Could not generate a unique &idColName for &workingDatastoreLib..&tableName.;
               %goto ERROR_EXIT;
            %end;
         %end;
         %else %do;
            %let newRowID = &scenario_id;
         %end;
 
         data work.&tableName;
            set work.&tableName;
            origID = &idColName;
            &idColName = (_n_ - 1) + &newRowID;
         run;
         %if (&syserr gt 4) %then %do;
            %let &_rc = 100;
            %put %mo_error Error occurred updating row IDs for table work.&tableName (syserr=&syserr).;
            %goto ERROR_EXIT;
         %end;
      

         %*****************************************************************;
         %* References to scenario element table IDs updated during the    ;
         %* validation phase must be revised to use the permanent scenario ;
         %* element IDs assigned during the commit phase. (References to   ;
         %* input data table IDs updated during the validation phase are   ;
         %* OK because the permanent input data IDs were used.)            ;
         %*****************************************************************;
   

         %********************************;
         %* Update referenced table IDs. *;
         %********************************;
         %local numRefTables refPair; 
         %local refTableName refIDColName refHasScenarioID;
         %local usesDiffIDColName refValueColName;

         %let numRefTables = %mo_word_count(&otherReferencedTables, %str( ));
         %do refTbl = 1 %to &numRefTables;
            %let refPair = %qscan(&otherReferencedTables, &refTbl, %str( ));
            %let refTableName = %qscan(&refPair, 1, %str(-)); 
            %let usesDiffIDColName = %qscan(&refPair, 2, %str(-)); 

            %let refTableName = %trim(%left(&refTableName));
            %let usesDiffIDColName = %trim(%left(&usesDiffIDColName));
   
            %if (&local_debug eq 1) %then %do;
               %mo_no_notes(on)
               %put;
               %put %mo_note numRefTables: &numRefTables;
               %put %mo_note refPair: &refPair;
               %put %mo_note refTableName: &refTableName;
               %put %mo_note usesDiffIDColName: &usesDiffIDColName;
               %mo_no_notes(off)
            %end;
   
            %mo_get_mapped_table_info(tblNm = %lowcase(&refTableName)
                                     ,_hasScenID = refHasScenarioID
                                     ,_rc = gmtirc);
            %if (&gmtirc ne 0) %then %do;
               %let &_rc = &gmtirc;
               %goto ERROR_EXIT;
            %end;
      
            %*******************************************;
            %* If referenced table is a scenario table *;
            %* (but not THE scenarios table) ...       *;
            %*******************************************;
            %if ((&refHasScenarioID eq 1) and
                 (%lowcase(&refTableName) ne scenarios)) %then %do;

               %put;
               %mo_no_notes(on)
               %put %mo_note Begin update ID to referenced table &refTableName by normal mapping.;
               %mo_no_notes(off)

               %mo_get_mapped_table_info(tblNm = &refTableName
                                        ,_idColNm = refIDColName
                                        ,_valColNm = refValueColName
                                        ,_rc = gmtirc);
               %if (&gmtirc ne 0) %then %do;
                  %let &_rc = &gmtirc;
                  %goto ERROR_EXIT;
               %end;

               %let refTbl = %eval(&refTbl + 1);
               %updateReferenceID(rfTblNm = &refTableName
                                 ,rfIDColNm = &refIDColName
                                 ,rfValColNm = &refValueColName
                                 ,officialLib = &workingDatastoreLib
                                 ,importLib = _validat
                                 ,workLib = work
                                 ,workTbl = &tableName
                                 ,workIDColNm = &idColName
                                 ,_rfTblNmList = refTableNameList
                                 ,_rc = uridrc);
               %if (&uridrc ne 0) %then %do;
                  %let &_rc = &uridrc;
                  %goto ERROR_EXIT;
               %end;
   
               %put;
               %mo_no_notes(on)
               %put %mo_note End update ID to referenced table &tableName by normal mapping.;
               %mo_no_notes(off)
            %end;
         %end;
   

         %*****************************************;
         %* Update tables that use attribute type *;
         %* to identify the referenced table ID   *;
         %*****************************************;
         %if (&useAttributeType eq 1) %then %do;
            %updateIDByAttributeType(tableName = &tableName
                                    ,officialLib = &workingDatastoreLib
                                    ,importLib = _validat
                                    ,workLib = work
                                    ,refTableNameList = refTableNameList
                                    ,_rc = ubatrc);
            %if (&ubatrc ne 0) %then %do;
               %let &_rc = &ubatrc;
               %goto ERROR_EXIT;
            %end;
         %end; /* If useAttributeType ... */
   

         %*********************************************;
         %* Update tables that use customer attribute *;
         %* type to identify the referenced table ID  *;
         %*********************************************;
         %if (&useCustAttributeType eq 1) %then %do;
            %updateIDByCustAttributeType(tableName = &tableName
                                        ,officialLib = &workingDatastoreLib
                                        ,importLib = _validat
                                        ,workLib = work
                                        ,refTableNameList = refTableNameList
                                        ,_rc = ubcstatrc);
            %if (&ubcstatrc ne 0) %then %do;
               %let &_rc = &ubcstatrc;
               %goto ERROR_EXIT;
            %end;
         %end; /* If useCustAttributeType ... */

   
         %**************************************************;
         %* Update tables that use communication attribute *;
         %* type to identify the referenced table ID       *;
         %**************************************************;
         %if (&useCommAttributeType eq 1) %then %do;
            %updateIDByCommAttributeType(tableName = &tableName
                                        ,officialLib = &workingDatastoreLib
                                        ,importLib = _validat
                                        ,workLib = work
                                        ,refTableNameList = refTableNameList
                                        ,_rc = ubcomatrc);
            %if (&ubcomatrc ne 0) %then %do;
               %let &_rc = &ubcomatrc;
               %goto ERROR_EXIT;
            %end;
         %end; /* If useCommAttributeType ... */


         %*********************************************;
         %* Update tables that use scenario item type *;
         %* to identify the referenced table ID       *;
         %*********************************************;
         %if (&useScenarioItem eq 1) %then %do;
            %local refTableNameList;
            %let refTableNameList = ;

            %updateIDByScenarioItemType(tableName = &tableName
                                       ,officialLib = &workingDatastoreLib
                                       ,importLib = _validat
                                       ,workLib = work
                                       ,refTableNameList = refTableNameList
                                       ,_rc = ubscnitrc);
            %if (&ubscnitrc ne 0) %then %do;
               %let &_rc = &ubscnitrc;
               %goto ERROR_EXIT;
            %end;
         %end; /* If useScenarioItem ... */


         %***********************;
         %* Update primary IDs. *;
         %***********************;
         data work.&tableName;
            set work.&tableName;
            %if (&hasInputDataID eq 1) %then %do;
               input_data_id = &input_data_id;
            %end;
            %if (&hasProjectID eq 1) %then %do;
               project_id = &project_id;
            %end;
            %if (&hasUserID eq 1) %then %do;
               user_id = &user_id;
            %end;
            %if (&hasScenarioID eq 1) %then %do;
               scenario_id = &scenario_id;
            %end;
         run;
         %if (&syserr gt 4) %then %do;
            %let &_rc = 100;
            %put %mo_error Error occurred updating primary IDs in work.&tableName (syserr=&syserr).;
            %goto ERROR_EXIT;
         %end;

   
         %**********************************;
         %* Update corresponding datastore *;
         %* table with new rows.           *;
         %**********************************;
         %if (&numRows gt 0) %then %do;
            data work.noOrigID_&tbl;
               set work.&tableName (drop=origID);
            run;
            %if (&syserr gt 4) %then %do;
               %let &_rc = 100;
               %put %mo_error Error occurred dropping origID column from work.noOrigID_&tbl (syserr=&syserr).;
               %goto ERROR_EXIT;
            %end;

            %local deleteIDList;
            %let deleteIDList =;
            proc sql noprint;
               select &idColName into :deleteIDList separated by ", "
               from work.noOrigID_&tbl;
            quit;
            %if (&sqlrc gt 4) %then %do;
               %let &_rc = 107;
               %put %mo_error Error occurred obtaining IDs to delete from work.noOrigID_&tbl (sqlrc=&sqlrc).;
               %goto ERROR_EXIT;
            %end;
 
            %mo_record_delete(metadata_table = &workingDatastoreLib..&tableName
                             ,where          = &idColName in (&deleteIDList)
                             ,_rc            = delrc);
            %if (&delrc ne 0) %then %do;
               %let &_rc = 103;
               %goto ERROR_EXIT;
            %end;
            %mo_no_notes(off)
   
            %local byList;
            %let byList = ;
            %if (&hasInputDataID eq 1) %then %do;
               %let byList = &byList input_data_id;
            %end;
            %if (&hasProjectID eq 1) %then %do;
               %let byList = &byList project_id;
            %end;
            %if ((%lowcase(&tableName) ne scenarios) and
                 (%lowcase(&tableName) ne scenario_advanced_options) and
                 (%lowcase(&tableName) ne optimization_jobs) and
                 (&hasScenarioID eq 1)) %then %do;
               %let byList = &byList scenario_id;
            %end;
            %let byList = &byList &idColName;
   
            %if (&local_debug eq 1) %then %do;
               %mo_no_notes(on)
               %put;
               %put %mo_note byList: &byList;
               %mo_no_notes(off)
            %end;
   
            %if (&numRows gt 1) %then %do;
               %mo_record_insert_multiple(metadata_table = &workingDatastoreLib..&tableName
                                         ,update_table   = work.noOrigID_&tbl
                                         ,by             = &byList
                                         ,_rc            = updrc);
               %if (&updrc ne 0) %then %do;
                  %let &_rc = 103;
                  %goto ERROR_EXIT;
               %end;
               %mo_no_notes(off)
            %end;
            %else %do;
               %mo_record_insert(metadata_table = &workingDatastoreLib..&tableName
                                ,update_table   = work.noOrigID_&tbl
                                ,by             = &byList
                                ,_rc            = updrc);
               %if (&updrc ne 0) %then %do;
                  %let &_rc = 103;
                  %goto ERROR_EXIT;
               %end;
               %mo_no_notes(off)
            %end;
         %end;
      %end; /* For each table ... */


      %*--------------------------------------;
      %* Store refresh messages in datastore. ;
      %*--------------------------------------;
      data work.refresh_messages;
         set _validat.refresh_messages;
         scenario_id = &scenario_id;
      run;
      %if (&syserr gt 4) %then %do;
         %let &_rc = 100;
         %put %mo_error Error occurred creating table work.refresh_messages from _validat.refresh_messages (syserr=&syserr).;
         %goto ERROR_EXIT;
      %end;

      %mo_record_append(metadata_table = &workingDatastoreLib..refresh_messages
                       ,update_table   = work.refresh_messages
                       ,_rc            = rarc
                       );
      %if (&rarc ne 0) %then %do;
         %let &_rc = 200;
         %put %mo_warning Warning messages were not created properly (rarc=&rarc).;
      %end;


      %*------------------------------;
      %* Create scenario directories. ;
      %*------------------------------;
      %local projPath;
      %let projPath = %nrbquote(%mo_pathname(%nrbquote(&moproj)&separator.%nrbquote(&username)&separator.%nrbquote(&project_name)));
      %mo_create_scenario_directories(project_path = %nrbquote(&projPath)
                                     ,scenario_name = %nrbquote(&scenario_name)
                                     ,_rc = csdrc);
                                  
      %if (&csdrc ne 0) %then %do;
         %let &_rc = &csdrc;
         %goto ERROR_EXIT;
      %end;

      %put;
      %mo_no_notes(on)
      %put %mo_note END COMMITMENT AT %sysfunc(putn(%sysfunc(time()), time.));
      %mo_no_notes(off)
   %end;


   %*--------------------------------;
   %* Exit processing                ;
   %*--------------------------------;
   %* Successful completion;
   %mo_no_notes(on)
   %put;
   %if (&action eq COMMIT) %then %do;
      %let &_id = &scenario_id;
      %let &_has_import_msg = 0;
      %put %mo_note Successfully imported scenario for project ID &project_id, scenario name %nrbquote(&scenario_name) from %nrbquote(&fq_import_file_name). New scenario ID is &&&_id.;
   %end;
   %else %do;
      %let &_id =;
      %if (&numRefreshMessages ne 0) %then %do;
         %let &_has_import_msg = 1;
      %end;
      %else %do;
         %let &_has_import_msg = 0;
      %end;
      %put %mo_note Successfully validated scenario definition for project ID &project_id, scenario name %nrbquote(&scenario_name) from %nrbquote(&fq_import_file_name).;
   %end;
   %goto EXIT;


   %ERROR_EXIT:
   %let &_id =;
   %put;
   %if (&action eq COMMIT) %then %do;
      %put %mo_error Error encountered during import of definition for project ID &project_id, scenario name %nrbquote(&scenario_name) from %nrbquote(&fq_import_file_name).;
   %end;
   %else %do;
      %put %mo_error Error encountered during validation of scenario definition for project ID &project_id, scenario name %nrbquote(&scenario_name) from %nrbquote(&fq_import_file_name).;
   %end;

   %EXIT:
   %mo_no_notes(off)
   %if (%symexist(dsLocked) and (&dsLocked eq 1)) %then %do;
      %mo_unlock_dataset(data=mometa.uid)
   %end;
   %if (&action eq COMMIT) %then %do;
      %if (%sysfunc(fileexist(%nrbquote(&fq_import_file_name)))) %then %do;
         %mo_delete_exprt_imprt_file(file_type    = IMPORT
                                    ,fq_file_name = %nrbquote(&fq_import_file_name));
      %end;
   %end;


%mend mo_import_scenario;
