/****************************************************************************************/
/* Copyright (c) 2005-2011, SAS Institute Inc., Cary, NC, USA,                          */
/*                       All Rights Reserved.                                           */
/*                                                                                      */
/* Name:    mo_prepopulate_metadata_tables                                              */
/*                                                                                      */
/* Purpose: Populate metadata tables from input data                                    */
/*                                                                                      */
/* Input:   input_data_id           - Input data ID                                     */
/*          library_metadata_id     - OMR SASLibrary FQID                               */
/*          uses_prioritization     - Flag to indicate priority-based input data        */
/*          uses_contact_history    - Flag to indicate contact history in input data    */
/*          customer_table          - Name of Customer Attributes table                 */
/*          cust_col_metadata_table - Name of Customer Column Metadata table            */
/*          campaign_table          - Name of Campaigns table                           */
/*          communication_table     - Name of Communications table                      */
/*          channel_table           - Name of Channel table            (optional)       */
/*          time_period_table       - Name of Time Period table        (optional)       */
/*          agent_capacity_table    - Name of Agent table              (optional)       */
/*          constraints_table       - Name of Constraints table        (optional)       */
/*          contact_policy_table    - Name of Contact Policy table     (optional)       */
/*          contact_policy_table    - Name of Contact Policies table         (optional) */
/*          contact_history_table   - Name of Contact History table          (optional) */
/*          customer_id_column      - Name of Customer ID column                        */
/*          household_id_column     - Name of Household ID column      (optional)       */
/*          customer_rank_measure   - Name of Customer Rank measure          (optional) */
/*          priority_column         - Name of Priority column                (optional) */
/*          agent_column            - Name of Agent column             (optional)       */
/*                                                                                      */
/* Output:  &_rc = return code                                                          */
/*                                                                                      */
/* Return Codes:                                                                        */
/*                                                                                      */
/*          100   generic error                                                         */
/*          102   invalid or missing input paramater value                              */
/*          103   I/O error updating a metadata table                                   */
/*          104   I/O error creating a metadata table                                   */
/*          106   error obtaining lock                                                  */
/*          107   SQL error                                                             */
/*          108   could not generate unique identifer                                   */
/*          109   libref error                                                          */
/*                                                                                      */
/* HISTORY:                                                                             */
/* 24Mar04 S0237007 Creation (dwh)                                                      */
/* 22Jul04 S0255867 Ignore comm_numeric_attributes with all missing values (dwh)        */
/* 26Jul04 S0262301 Replace hardcoded type values (dwh)                                 */
/* 06Aug04 S0263394 Allow mixed case column names in CUSTOMER and CUSTOMER_CONTROL (dwh)*/
/* 11Aug04 S0255867 Redo this fix using SQL insert (dwh)                                */
/* 12Aug04 S0255867 Ignore customer_attributes with all missing values (dwh)            */
/* 30Aug04 S0266201 Ignore extra numeric columns in CAMPAIGNS table (dwh)               */
/* 03Dec04 S0273454 Scenario specific communication attributes (dwh)                    */
/* 07Dec04 S0268539 Remove leading blanks from group/subgroup (dwh)                     */
/*                                                                                      */
/* 4.3:                                                                                 */
/* 13Jan05 S0273454 Input data defined in OMR (dwh)                                     */
/* 04Mar05 S0273454 Add agent table (dwh)                                               */
/*                                                                                      */
/* 4.4:                                                                                 */
/* 14Dec05 S0310319 Handle customer attributes that contain special characters (dwh)    */
/* 02Mar06 S0347842 Allow non-unique column_name in CUSTOMER_CONTROL table (dwh)        */
/* 22Mar06 S0349405 Make non-required input data columns optional (dwh)                 */
/* 25Apr06 S0355068 Always define comm_unit_cost and outbound_unit_usage in the         */
/*                  datastore whether in the input data or not (dwh)                    */
/* 04May06 S0356919 Remove delimiter conflict with customer attribute values (dwh)      */
/* 09May06 S0357862 Improve customer attribute processing efficiency (dwh)              */
/* 18May06 S0359925 Lowcase list before checking for comm_unit_cost and                 */
/*                  outbound_unit_usage (dwh)                                           */
/* 02Jun06 S0361315 Check for blanks in case reading from Oracle (dwh)                  */
/* 07Jun06 S0361134 Support a SEQUENCE column in the timeperiod table for Oracle (dwh)  */
/*                                                                                      */
/* 5.1:                                                                                 */
/* 13Dec06 S0395314 Changes to input data structure (micopp)                            */
/*                                                                                      */
/* 5.3:                                                                                 */
/* 27May08 S0483854 Introduce target_level as a new field in the                        */
/*         prepopulated_contact_policies and contact_policies tables.                   */
/*         This is a per-contact-policy variant of the household_level                  */
/*         field that was eliminated from the scenario_advanced_options                 */
/*         table. (KEC)                                                                 */
/* 17Jul08 S0483854 MOMA support - Add support for presence of cell_cd column in MOMA   */
/*         5.3 Communications and Customer Control input tables. Treat cell_cd column   */
/*         as a communications character attribute. (KEC)                               */
/* 23Jul08 S0483854 Add support for "report only" constraint type.                      */
/* 11Dec08 S0483854 Add support for a new special value for the time period ID columns  */
/*         in the Prepopulated Contact Policies table to be interpreted as "use the     */
/*         extreme/outermost time period values from the Time Periods table to define   */
/*         the beginning and ending time scope of the optimization" (KEC)               */
/* 11Dec08 S0483854 Add support for the two new columns in the Prepopulated Contact     */
/*         Policies table which are used to 1) define the scope at which a constraint   */
/*         is to be applied (all/aggregate, customer, household) and 2) whether the     */
/*         constraint is to be included in sensitivity analysis (if SA requested for a  */
/*         scenario optimization), respectively. (KEC)                                  */
/* 09Jan09 S0555473 Correct the statement that determines the rolling_period_cnt value  */
/*         using the min function. It incorrectly referenced blocking_period_cnt (KEC)  */
/* 09Jan09 S0555734 Remove use of min function to "correct" rolling_period_cnt and    */
/*         blocking_period_cnt that is larger than the difference between the actual    */
/*         start period and actual end period (KEC)                                     */
/*                                                                                      */
/* 5.4:                                                                                 */
/* 15Feb10 S0639443 Populate communications, comm_numeric_attributes (and related),     */
/*         and input_data_measures tables with mandatory inclusions columns (KEC)       */
/* 20Apr10 Enhancements related to support for scenario import (KEC)                    */
/* 19Aug10 Add support for new prioritization columns, new contact history              */
/*         columns. (KEC)                                                               */
/*                                                                                      */
/* F1MO54, F0MO541                                                                      */
/* 09Jan12 S0823209 Replace use of SQL query of Dictionay Tables with use               */
/*         of SAS Proc Contents.                                                        */
/*                                                                                      */
/****************************************************************************************/

%macro mo_prepopulate_metadata_tables(input_data_id                    =
                                     ,library_metadata_id              =
                                     ,uses_prioritization              =
                                     ,uses_contact_history             =
                                     ,customer_table                   =
                                     ,cust_col_metadata_table          =
                                     ,campaign_table                   =
                                     ,communication_table              =
                                     ,channel_table                    =
                                     ,time_period_table                =
                                     ,agent_capacity_table             =
                                     ,constraints_table                =
                                     ,contact_policy_table             =
                                     ,contact_history_table            =
                                     ,communication_history_table      =
                                     ,customer_id_column               =
                                     ,household_id_column              =
                                     ,customer_rank_measure            =
                                     ,priority_column                  =
                                     ,agent_column                     =
                                     ,_rc                              =
                                     );

   %local macname;
   %let macname=&sysmacroname;
   %let macname=%lowcase(&macname);

   %if %bquote(&_rc) eq %str() %then %do;
      %local rc;
      %let _rc = rc;
   %end;
   %else %global &_rc;
   %let &_rc = 100;      %* Guilty until proven innocent;

   %mo_constants

   %*--------------------------------;
   %* Validate parms                 ;
   %*--------------------------------;
   %if (%bquote(&input_data_id) eq %str()) %then %do;
      %put %mo_error No input_data_id was specified.;
      %let &_rc = 102;
      %goto ERROR_EXIT;
   %end;
   %if (%bquote(&library_metadata_id) eq %str()) %then %do;
      %put %mo_error No library_metadata_id was specified.;
      %let &_rc = 102;
      %goto ERROR_EXIT;
   %end;
   %if (%bquote(&customer_table) eq %str()) %then %do;
      %put %mo_error No customer_table was specified.;
      %let &_rc = 102;
      %goto ERROR_EXIT;
   %end;
   %if (%bquote(&cust_col_metadata_table) eq %str()) %then %do;
      %put %mo_error No cust_col_metadata_table was specified.;
      %let &_rc = 102;
      %goto ERROR_EXIT;
   %end;
   %if (%bquote(&campaign_table) eq %str()) %then %do;
      %put %mo_error No campaign_table was specified.;
      %let &_rc = 102;
      %goto ERROR_EXIT;
   %end;
   %if (%bquote(&communication_table) eq %str()) %then %do;
      %put %mo_error No communication_table was specified.;
      %let &_rc = 102;
      %goto ERROR_EXIT;
   %end;
 
   %mo_no_notes(off)
   libname moinput meta libid=&library_metadata_id;
   %if (&syslibrc ne 0) %then %do;
      %put %mo_error Error assigning MOINPUT libref (syslibrc=&syslibrc);
      %let &_rc = 109;
      %goto ERROR_EXIT;
   %end;


   %*--------------------------------------------------------------------;
   %* Create views of input tables that fill in missing optional columns ;
   %*--------------------------------------------------------------------;
   %let tablelist = campaign cust_col_metadata communication;
   %if (%bquote(&channel_table) ne %str()) %then %let tablelist = &tablelist channel;
   %if (%bquote(&time_period_table) ne %str()) %then %let tablelist = &tablelist time_period;
   %if (%bquote(&agent_capacity_table) ne %str()) %then %let tablelist = &tablelist agent_capacity;
   %if (%bquote(&constraints_table) ne %str()) %then %let tablelist = &tablelist constraints;
   %if (%bquote(&contact_policy_table) ne %str()) %then %let tablelist = &tablelist contact_policy;

   %mo_create_tables(lib=work, table=INPUT, labels=N, _rc=cudrc)
   %if (&cudrc ne 0) %then %do;
      %let &_rc = 100;
      %put %mo_error Error occurred creating INPUT tables in WORK library.;
      %goto ERROR_EXIT;
   %end;

   %do i=1 %to %mo_word_count(&tablelist);
      %let table=%scan(&tablelist,&i);

      %mo_no_notes(off)
      proc datasets library=work nolist nowarn;
         delete &table._in_mo_tbl / memtype=all;
      quit;

      %mo_no_notes(on)
      data work.&table._in_mo_tbl;
         merge moinput.&&&table._table work.&table;
      run;
      %if (&syserr gt 4) %then %do;
         %let &_rc = 100;
         %put %mo_error Error encountered creating work.&table._in_mo_tbl (syserr=&syserr).;
         %goto ERROR_EXIT;
      %end;
   %end;
   %mo_no_notes(off)


   %*-------------------------------------------------------------------------------------;
   %* Reset &communication_table to the table from validation that joined to other tables ;
   %*-------------------------------------------------------------------------------------;
   data work.communication_in_mo_tbl;
      set work.comm_table_joined_others; /* From mo_validate_input_data */
   run;
   %if (&syserr gt 4) %then %do;
      %let &_rc = 100;
      %put %mo_error Error encountered updating work.communication_in_mo_tbl (syserr=&syserr).;
      %goto ERROR_EXIT;
   %end;

   proc contents data=work.communication_in_mo_tbl (drop=_mo_column_sort_order) out=work.comm_out noprint;
   run;
   %if (&syserr gt 4) %then %do;
      %put %mo_note Error encountered getting contents of the communication_in_mo_tbl (syserr=&syserr).;
      %goto ERROR_EXIT;
   %end;

   data work.comm_out;
      set work.comm_out;
      lowcase_name=lowcase(name);
      select (lowcase(name));
         when ('campaign_cd') sort_order=1;
         when ('campaign_desc') sort_order=2;
         when ('communication_cd') sort_order=3;
         when ('communication_desc') sort_order=4;
         when ('cell_cd') sort_order=5;
         when ('channel_cd') sort_order=6;
         when ('time_period') sort_order=7;
         when ('communication_unit_cost') sort_order=8;
         when ('comm_unit_cost') sort_order=9;
         when ('unit_usage') sort_order=10;
         when ('mandatory') sort_order=11;
         when ('exclude_from_constraints') sort_order=12;
         when ('exclude_from_contact_pols') sort_order=13;
         otherwise sort_order=14;
      end;
   run;
   %if (&syserr gt 4) %then %do;
      %put %mo_note Error encountered setting sort order of work.comm_out (syserr=&syserr).;
      %goto ERROR_EXIT;
   %end;

   proc sort data=work.comm_out;
      by sort_order lowcase_name;
   run;
   %if (&syserr gt 4) %then %do;
      %put %mo_note Error encountered sorting work.comm_out (syserr=&syserr).;
      %goto ERROR_EXIT;
   %end;

   %local comm_numeric_fields comm_character_fields;
   proc sql noprint;
     select name into :comm_numeric_fields
        separated by ' '
     from work.comm_out
     where type=1
     order by name;

     select name into :comm_character_fields
        separated by ' '
     from work.comm_out
     where type=2
     order by name;
   quit;
   %if (&sqlrc gt 4) %then %do;
      %put %mo_note SQL error getting variables from communication_in_mo_tbl (sqlrc=&sqlrc).;
      %let &_rc = 107;
      %goto ERROR_EXIT;
   %end;

   /* Determine if mandatory column is present in the communications input    */
   /* data table. The table has passed validation so if the mandatory column  */
   /* is not present, the mandatory, exclude_from_constraints and             */
   /* exclude_from_contact_pols columns in the communication_in_mo_tbl        */
   /* should be excluded from the comm_numeric_attributes list.               */
   %local mandatory_in_comm_table;
   %let mandatory_in_comm_table = 0;
   proc contents data=moinput.&communication_table out=work.isMandInCommTbl(keep=name) noprint;
   run;
   %if (&syserr gt 4) %then %do;
      %put %mo_note Error encountered retrieving contents of the MOINPUT.%upcase(&communication_table) table (syserr=&syserr).;
      %let &_rc = 100;
      %goto ERROR_EXIT;
   %end;

   proc sql noprint;
      select count(*) into :mandatory_in_comm_table
      from work.isMandInCommTbl
      where (upcase(name) = "MANDATORY");
   quit;
   %if (&sqlrc gt 4) %then %do;
      %let &_rc = 107;
      %put %mo_error SQL error encountered obtaining information about moinput.&communication_table (sqlrc=&sqlrc).;
      %goto ERROR_EXIT;
   %end;

   /* "Weed out" communications table columns for which all rows have missing */   /* values, or are optional and are not used.                               */
   %local comm_numeric_fields_missing total;
   %let comm_numeric_fields_missing = ;
   %local wcval;
   %let wcval = %mo_word_count(&comm_numeric_fields);
   %do z=1 %to &wcval;
      %let field=%scan(&comm_numeric_fields, &z);

      proc sql noprint;
         select sum(&field) into :total
         from work.communication_in_mo_tbl;
      quit;
      %if (&sqlrc gt 4) %then %do;
         %put %mo_note Error encountered determining comm_numeric_fields_missing (sqlrc=&sqlrc).;
         %let &_rc = 107;
         %goto ERROR_EXIT;
      %end;

      /* Add column to list of columns to exclude from         */
      /* comm_numeric_attributes if:                           */
      /*    1. All column values are missing (except in the    */
      /*       case of the priority column when prioritization */
      /*       is in use ... keep the priority column anyway)  */
      /*    2. It is a mandatory inclusions-related column     */
      /*       but mandatory inclusions is not in use.         */
      %if (%nrbquote(&total) eq .) %then %do; 
         %if (not ((&uses_prioritization eq 1) and (%upcase(&field) eq %upcase(&priority_column)))) %then %do;
            %let comm_numeric_fields_missing = &comm_numeric_fields_missing &field;
         %end;
      %end;

      %if ((&mandatory_in_comm_table eq 0) and 
           ((%upcase(&field) eq MANDATORY) or 
            (%upcase(&field) eq EXCLUDE_FROM_CONSTRAINTS) or 
            (%upcase(&field) eq EXCLUDE_FROM_CONTACT_POLS))) %then %do;
         %let comm_numeric_fields_missing = &comm_numeric_fields_missing &field;
      %end;
   %end;
   %if (%bquote(&comm_numeric_fields_missing) ne %str()) %then %do;
      %local comm_numeric_fields_new;
      %let comm_numeric_fields_new = ;
      %do z = 1 %to &wcval;
         %let field = %scan(&comm_numeric_fields,&z);
         %if %sysfunc(indexw(&comm_numeric_fields_missing, &field)) eq 0 %then %let comm_numeric_fields_new = &comm_numeric_fields_new &field;
      %end;
      %let comm_numeric_fields = &comm_numeric_fields_new;
   %end;

   %local comm_character_fields_missing total;
   %let comm_character_fields_missing = ;
   %let wcval = %mo_word_count(&comm_character_fields);
   %do z=1 %to &wcval;
      %let field=%scan(&comm_character_fields, &z);

      proc sql noprint;
         select &field into :total separated by ' '
         from work.communication_in_mo_tbl;
      quit;
      %if (&sqlrc gt 4) %then %do;
         %put %mo_note Error encountered determining comm_character_fields_missing (sqlrc=&sqlrc).;
         %let &_rc = 107;
         %goto ERROR_EXIT;
      %end;

      %if (%superq(total) eq %str()) %then %let comm_character_fields_missing = &comm_character_fields_missing &field;
   %end;
   %if (%bquote(&comm_character_fields_missing) ne %str()) %then %do;
      %local comm_character_fields_new;
      %let comm_character_fields_new = ;
      %do z = 1 %to &wcval;
         %let field = %scan(&comm_character_fields,&z);
         %if %sysfunc(indexw(&comm_character_fields_missing, &field)) eq 0 %then %let comm_character_fields_new = &comm_character_fields_new &field;
      %end;
      %let comm_character_fields = &comm_character_fields_new;
   %end;


   %*------------------------------------------------------------------------------------------------;
   %* Create a temporary channel table if no input channel table is specified but channel_cd is used ;
   %*------------------------------------------------------------------------------------------------;

   %if (%bquote(&channel_table) eq %str()) %then %do;

      proc sql noprint;
         create table work.channel_in_mo_tbl as
            select distinct (compbl(trim(left(channel_cd)))) as channel_cd format $32.,
                            '' as uom_cd format $32.
            from work.communication_in_mo_tbl
            where channel_cd ne ''
            %if (%bquote(&agent_column) ne %str()) %then %do;
               and upcase(channel_cd) ne %upcase("&agent_column")
            %end;
         ;
      quit;
      %if &sqlobs > 0 %then %let channel_table = temp_table_channel_cd;
      %if (&sqlrc gt 4) %then %do;
         %put %mo_note SQL error creating work.channel_in_mo_tbl from channel_cd column (sqlrc=&sqlrc).;
         %let &_rc = 107;
         %goto ERROR_EXIT;
      %end;

   %end;


   %*--------------------------------;
   %* Useful macros and macvars      ;
   %*--------------------------------;

   %macro populate_preprocess(metadata_table    =
                             ,metadata_key      =
                             ,_rc               =
                             );

      %local macname;
      %let macname=&sysmacroname;
      %let macname=%lowcase(&macname);

      %if %bquote(&_rc) eq %str() %then %do;
         %local rc;
         %let _rc = rc;
      %end;
      %else %global &_rc;
      %let &_rc = 100;      %* Guilty until proven innocent;

      %mo_no_notes(on)
      %mo_create_tables(lib=work,table=&metadata_table,_rc=cudrc)
      %if (&cudrc ne 0) %then %do;
         %let &_rc = 100;
         %put %mo_error Error occurred creating WORK.%upcase(&metadata_table) table.;
         %goto ERROR_EXIT;
      %end;
      
      %mo_no_notes(off)
      proc sql noprint;
         insert into work.&metadata_table
         select * from mometa.&metadata_table
         where (input_data_id eq &input_data_id);
      quit;
      %if (&sqlrc gt 4) %then %do;
         %put %mo_error SQL error occurred retrieving data from MOMETA.%upcase(&metadata_table) (sqlrc=&sqlrc).;
         %let &_rc = 107;
         %goto ERROR_EXIT;
      %end;

      %if (&sqlobs gt 0) %then %do;
         %mo_record_delete(metadata_table = mometa.&metadata_table
                          ,where          = input_data_id eq &input_data_id
                          ,_rc            = delrc);
         %if (&delrc ne 0) %then %do;
            %let &_rc = 103;
            %put %mo_error I/O error encountered while deleting entries from MOMETA.%upcase(&metadata_table) table with input_data_id = &input_data_id (delrc=&delrc).;
            %goto ERROR_EXIT;
         %end;
      %end;

      %mo_no_notes(off)
      %mo_gen_uid(r_macro=&metadata_key, data=mometa.&metadata_table, id_var=&metadata_key, _rc=uidrc)
      %if (&uidrc ne 0) %then %do;
         %let &_rc = 108;
         %put %mo_error Could not generate a unique &metadata_key.;
         %goto ERROR_EXIT;
      %end;

      %if (&metadata_table eq comm_numeric_attribute_defaults) %then %let meta_table = comm_numeric_defaults;
      %else %if (&metadata_table eq comm_character_distinct_values) %then %let meta_table = comm_distinct_values;
      %else %if (&metadata_table eq comm_character_attribute_values) %then %let meta_table = comm_character_values;
      %else %if (&metadata_table eq prepopulated_contact_policies) %then %let meta_table = prepop_contact_policies;
      %else %let meta_table = &metadata_table;

      %mo_no_notes(off)
      proc contents data=mometa.&metadata_table out=work._&meta_table._ noprint;
      run;
      %if (&syserr gt 4) %then %do;
         %put %mo_note Error encountered retrieving contents of the MOMETA.%upcase(&metadata_table) table (syserr=&syserr).;
         %let &_rc = 100;
         %goto ERROR_EXIT;
      %end;

      %global &meta_table._vars;
      %let &meta_table._vars=;
      proc sql noprint;
        select name into :&meta_table._vars
           separated by ' '
        from work._&meta_table._
      quit;
      %if (&sqlrc gt 4) %then %do;
         %put %mo_note SQL error encountered retrieving contents of the MOMETA.%upcase(&metadata_table) table  (sqlrc=&sqlrc).;
         %let &_rc = 107;
         %goto ERROR_EXIT;
      %end;

      %mo_no_notes(on)
      %mo_create_tables(lib=work,table=&metadata_table,_rc=cudrc)
      %if (&cudrc ne 0) %then %do;
         %let &_rc = 100;
         %put %mo_error Error occurred creating WORK.%upcase(&metadata_table) table.;
         %goto ERROR_EXIT;
      %end;
      %mo_no_notes(off)

      %let &_rc = 0;

      %ERROR_EXIT:

   %mend populate_preprocess;

   %macro populate_postprocess(metadata_table =
                              ,metadata_key   =
                              ,_rc            =
                              );

      %local macname;
      %let macname=&sysmacroname;
      %let macname=%lowcase(&macname);

      %if %bquote(&_rc) eq %str() %then %do;
         %local rc;
         %let _rc = rc;
      %end;
      %else %global &_rc;
      %let &_rc = 100;      %* Guilty until proven innocent;

      %if &mo_validate_append_option = 1 %then %do;
         %mo_record_append(metadata_table = mometa.&metadata_table
                          ,update_table   = work.&metadata_table
                          ,_rc            = updrc);
      %end;
      %else %do;
         %mo_record_insert_multiple(metadata_table = mometa.&metadata_table
                                   ,update_table   = work.&metadata_table
                                   ,by             = input_data_id &metadata_key
                                   ,_rc            = updrc);
      %end;

      %if (&updrc ne 0) %then %do;
         %let &_rc = 103;
         %goto ERROR_EXIT;
      %end;
      %else %do;
         %put %mo_note Successfully populated the MOMETA.%upcase(&metadata_table) table.;
      %end;

      %let &_rc = 0;
      
      %ERROR_EXIT:

   %mend populate_postprocess;


   %*--------------------------------;
   %* CAMPAIGNS                      ;
   %*--------------------------------;
   %populate_preprocess(metadata_table = campaigns
                       ,metadata_key   = campaign_id
                       ,_rc            = pprc);
   %if (&pprc ne 0) %then %do;
      %let &_rc = &pprc;
      %goto ERROR_EXIT;
   %end;

   %mo_no_notes(off)
   data work.campaigns (keep=&campaigns_vars);
      set work.campaigns work.campaign_in_mo_tbl;
      input_data_id = &input_data_id;
      campaign_id = (_N_-1) + &campaign_id;
      campaign_code = compbl(trim(left(campaign_cd)));
      campaign_description = trim(left(campaign_desc));
   run;
   %if (&syserr gt 4) %then %do;
      %put %mo_note Error encountered updating the work.campaigns table (syserr=&syserr).;
      %let &_rc = 100;
      %goto ERROR_EXIT;
   %end;

   %populate_postprocess(metadata_table = campaigns
                        ,metadata_key   = campaign_id
                        ,_rc            = pprc);
   %if (&pprc ne 0) %then %do;
      %let &_rc = &pprc;
      %goto ERROR_EXIT;
   %end;
   %symdel campaign_id;


   %*--------------------------------;
   %* CHANNELS                       ;
   %*--------------------------------;

   %if (%bquote(&channel_table) ne %str()) or (%bquote(&agent_capacity_table) ne %str()) %then %do;

      %populate_preprocess(metadata_table = channels
                          ,metadata_key   = channel_id
                          ,_rc            = pprc);
      %if (&pprc ne 0) %then %do;
         %let &_rc = &pprc;
         %goto ERROR_EXIT;
      %end;

      %mo_no_notes(off)

      %if (%bquote(&agent_capacity_table) ne %str()) %then %do;
         proc sql noprint;
            insert into work.channels
               set channel = "&agent_column",
                   channel_type = &_CHANNEL_TYPE_AGENT_;
         quit;
         %if (&sqlrc gt 4) %then %do;
            %put %mo_note Error encountered updating the work.channels table with agent channel (sqlrc=&sqlrc).;
            %let &_rc = 100;
            %goto ERROR_EXIT;
         %end;
      %end;

      %if (%bquote(&channel_table) ne %str()) %then %do;
         data work.channels (keep = &channels_vars);
            set work.channels work.channel_in_mo_tbl (rename=(channel_cd = channel));
            input_data_id = &input_data_id;
            channel_id = (_N_-1) + &channel_id;
            if channel_type ne &_CHANNEL_TYPE_AGENT_ then channel_type = &_CHANNEL_TYPE_OUTBOUND_;
            channel_units = compbl(trim(left(uom_cd)));
         run;
         %if (&syserr gt 4) %then %do;
            %put %mo_note Error encountered updating the work.channels table (syserr=&syserr).;
            %let &_rc = 100;
            %goto ERROR_EXIT;
         %end;
      %end;
      %else %if (%bquote(&agent_capacity_table) ne %str()) %then %do;
         data work.channels (keep = &channels_vars);
            set work.channels;
            input_data_id = &input_data_id;
            channel_id = (_N_-1) + &channel_id;
            channel_units = '';
         run;
         %if (&syserr gt 4) %then %do;
            %put %mo_note Error encountered updating the work.channels table (syserr=&syserr).;
            %let &_rc = 100;
            %goto ERROR_EXIT;
         %end;         
      %end;

      %populate_postprocess(metadata_table = channels
                           ,metadata_key   = channel_id
                           ,_rc            = pprc);
      %if (&pprc ne 0) %then %do;
         %let &_rc = &pprc;
         %goto ERROR_EXIT;
      %end;
      %symdel channel_id;
   %end;


   %*--------------------------------;
   %* TIME_PERIODS                   ;
   %*--------------------------------;
   %if (%bquote(&time_period_table) ne %str()) %then %do;
      %populate_preprocess(metadata_table = time_periods
                          ,metadata_key   = time_period_id
                          ,_rc            = pprc);
      %if (&pprc ne 0) %then %do;
         %let &_rc = &pprc;
         %goto ERROR_EXIT;
      %end;

      %mo_no_notes(off)
      data work.time_periods (keep=&time_periods_vars);
         set work.time_periods work.time_period_in_mo_tbl (rename=(time_period=period order_no=ordr));
         input_data_id = &input_data_id;
         time_period_id = (_N_-1) + &time_period_id;
         time_period = compbl(trim(left(period)));
         order = ordr;
      run;
      %if (&syserr gt 4) %then %do;
         %put %mo_note Error encountered updating the work.time_period_in_mo_tbl (syserr=&syserr).;
         %let &_rc = 100;
         %goto ERROR_EXIT;
      %end;

      %populate_postprocess(metadata_table = time_periods
                           ,metadata_key   = time_period_id
                           ,_rc            = pprc);
      %if (&pprc ne 0) %then %do;
         %let &_rc = &pprc;
         %goto ERROR_EXIT;
      %end;
      %symdel time_period_id;
   %end;


   %*--------------------------------;
   %* COMMUNICATIONS                 ;
   %*--------------------------------;
   %populate_preprocess(metadata_table = communications
                       ,metadata_key   = communication_id
                       ,_rc            = pprc);
   %if (&pprc ne 0) %then %do;
      %let &_rc = &pprc;
      %goto ERROR_EXIT;
   %end;

   proc datasets nolist;
      modify communication_in_mo_tbl;
      rename channel_cd=channel_mo_comm_table;
   quit;
   %if (&syserr gt 4) %then %do;
      %put %mo_note Error renaming channel_cd on communication_in_mo_tbl table (syserr=&syserr).;
      %let &_rc = 100;
      %goto ERROR_EXIT;
   %end;

   %local working_table;
   %let working_table = work.communication_in_mo_tbl;
   %mo_no_notes(off)
   proc sql noprint;
      create table work.communications_2 as
         select a.*, b.input_data_id, b.campaign_id
         from &working_table as a, mometa.campaigns as b
         where compbl(trim(left(a.campaign_cd))) eq compbl(trim(left(b.campaign_code)))
           and (b.input_data_id eq &input_data_id)
      ;
      %let working_table = work.communications_2;

      %if (%bquote(&channel_table) ne %str()) %then %do;
         create table work.communications_3 as
            select a.*, b.channel_id
            from &working_table as a left join mometa.channels as b
            on compbl(trim(left(a.channel_mo_comm_table))) eq compbl(trim(left(b.channel)))
               and (b.input_data_id eq &input_data_id) and (b.channel_type eq &_CHANNEL_TYPE_OUTBOUND_)
         ;
         %let working_table = work.communications_3;
      %end;

      %if (%bquote(&agent_capacity_table) ne %str()) %then %do;
         create table work.communications_4 as
            select a.*, b.channel_id as agent_channel_id
            from &working_table as a left join mometa.channels as b
            on compbl(trim(left(upcase(a.channel_mo_comm_table)))) eq compbl(trim(left(upcase(b.channel))))
               and (b.input_data_id eq &input_data_id) and (b.channel_type eq &_CHANNEL_TYPE_AGENT_)
         ;
         %let working_table = work.communications_4;
      %end;

      %if (%bquote(&time_period_table) ne %str()) %then %do;
         create table work.communications_5 as
            select a.*, b.time_period_id
            from &working_table as a left join mometa.time_periods as b
            on compbl(trim(left(a.time_period))) eq compbl(trim(left(b.time_period)))
               and (b.input_data_id eq &input_data_id)
         ;
         %let working_table = work.communications_5;
      %end;
   quit;
   %if (&sqlrc gt 4) %then %do;
      %put %mo_note SQL error encountered updating the work.communications tables (sqlrc=&sqlrc).;
      %let &_rc = 107;
      %goto ERROR_EXIT;
   %end;

   proc sort data=&working_table;
      by _mo_column_sort_order;
   run;
   %if (&syserr gt 4) %then %do;
      %put %mo_note Error encountered sorting the &working_table table (syserr=&syserr).;
      %let &_rc = 100;
      %goto ERROR_EXIT;
   %end;

   data &working_table (drop=agent_channel_id _mo_column_sort_order);
      set &working_table;
      if agent_channel_id ne . then channel_id = agent_channel_id;
   run;
   %if (&syserr gt 4) %then %do;
      %put %mo_note Error encountered updating the &working_table table (syserr=&syserr).;
      %let &_rc = 100;
      %goto ERROR_EXIT;
   %end;

   data work.communications (keep = &communications_vars) work.communications_numeric;
      set work.communications &working_table;
      communication_id = (_N_-1) + &communication_id;
      communication_code = trim(left(communication_cd));
      communication_description = trim(left(communication_desc));
   run;
   %if (&syserr gt 4) %then %do;
      %put %mo_note Error encountered updating the work.communications table (syserr=&syserr).;
      %let &_rc = 100;
      %goto ERROR_EXIT;
   %end;

   %populate_postprocess(metadata_table = communications
                        ,metadata_key   = communication_id
                        ,_rc            = pprc);
   %if (&pprc ne 0) %then %do;
      %let &_rc = &pprc;
      %goto ERROR_EXIT;
   %end;
   %symdel communication_id;


   %*-------------------------------------------------------------;
   %* COMM_NUMERIC_ATTRIBUTES                                     ;
   %*-------------------------------------------------------------;
   %let wcval = %mo_word_count(&comm_numeric_fields);
   %if (&wcval gt 0) %then %do;
      %populate_preprocess(metadata_table = comm_numeric_attributes
                          ,metadata_key   = comm_attribute_id
                          ,_rc            = pprc);
      %if (&pprc ne 0) %then %do;
         %let &_rc = &pprc;
         %goto ERROR_EXIT;
      %end;

      %mo_no_notes(off)
      %let wcval = %mo_word_count(&comm_numeric_fields);
      %do s = 1 %to &wcval;
         %let var = %scan(&comm_numeric_fields,&s);
         proc sql noprint;
            insert into work.comm_numeric_attributes
            set
               input_data_id = &input_data_id
              ,comm_attribute_id = (&s-1) + &comm_attribute_id
              ,comm_attribute_name = "&var"
            ;
         quit;
         %if (&sqlrc gt 4) %then %do;
            %put %mo_note Error encountered creating work.comm_numeric_attributes (sqlrc=&sqlrc).;
            %let &_rc = 107;
            %goto ERROR_EXIT;
         %end;
      %end;

      %populate_postprocess(metadata_table = comm_numeric_attributes
                           ,metadata_key   = comm_attribute_id
                           ,_rc            = pprc);
      %if (&pprc ne 0) %then %do;
         %let &_rc = &pprc;
         %goto ERROR_EXIT;
      %end;
      %symdel comm_attribute_id;
   %end;


   %*-------------------------------------------------------------;
   %* COMM_NUMERIC_ATTRIBUTE_DEFAULTS                             ;
   %*-------------------------------------------------------------;
   %let wcval = %mo_word_count(&comm_numeric_fields);
   %if (&wcval gt 0) %then %do;
      %populate_preprocess(metadata_table = comm_numeric_attribute_defaults
                          ,metadata_key   = comm_attribute_value_id
                          ,_rc            = pprc);
      %if (&pprc ne 0) %then %do;
         %let &_rc = &pprc;
         %goto ERROR_EXIT;
      %end;

      %mo_no_notes(off)
      data work.comm_numeric_attribute_defaults2 (keep=&comm_numeric_defaults_vars comm_attribute_name);
         length comm_attribute_name $32.;
         set work.comm_numeric_attribute_defaults work.communications_numeric;
         %let wcval = %mo_word_count(&comm_numeric_fields);
         %do s = 1 %to &wcval;
            %let var = %scan(&comm_numeric_fields,&s);
            comm_attribute_name = "&var";
            comm_attribute_value = &var;
            output;
         %end;
      run;
      %if (&syserr gt 4) %then %do;
         %put %mo_note Error encountered creating work.comm_numeric_attribute_defaults2 (syserr=&syserr).;
         %let &_rc = 100;
         %goto ERROR_EXIT;
      %end;

      data work.comm_numeric_attribute_defaults2;
         set work.comm_numeric_attribute_defaults2;
         comm_attribute_value_id = (_N_-1) + &comm_attribute_value_id;
      run;
      %if (&syserr gt 4) %then %do;
         %put %mo_note Error encountered updating work.comm_numeric_attribute_defaults2 (syserr=&syserr).;
         %let &_rc = 100;
         %goto ERROR_EXIT;
      %end;

      proc sql noprint;
        create table work.comm_numeric_attribute_defaults as
        select a.input_data_id, a.comm_attribute_value, a.comm_attribute_value_id, b.comm_attribute_id,
               a.campaign_id, a.communication_id
        from work.comm_numeric_attribute_defaults2 as a, mometa.comm_numeric_attributes as b
        where a.comm_attribute_name eq b.comm_attribute_name
          and a.input_data_id eq b.input_data_id;
      quit;
      %if (&sqlrc gt 4) %then %do;
         %put %mo_note SQL error encountered updating work.comm_numeric_attribute_defaults (sqlrc=&sqlrc).;
         %let &_rc = 107;
         %goto ERROR_EXIT;
      %end;

      %populate_postprocess(metadata_table = comm_numeric_attribute_defaults
                           ,metadata_key   = comm_attribute_value_id
                           ,_rc            = pprc);
      %if (&pprc ne 0) %then %do;
         %let &_rc = &pprc;
         %goto ERROR_EXIT;
      %end;
      %symdel comm_attribute_value_id;
   %end;


   %*-------------------------------------------------------------;
   %* COMM_CHARACTER_ATTRIBUTES                                   ;
   %*-------------------------------------------------------------;
   %let wcval = %mo_word_count(&comm_character_fields);
   %if (&wcval gt 0) %then %do;
      %populate_preprocess(metadata_table = comm_character_attributes
                          ,metadata_key   = comm_char_attribute_id
                          ,_rc            = pprc);
      %if (&pprc ne 0) %then %do;
         %let &_rc = &pprc;
         %goto ERROR_EXIT;
      %end;

      %mo_no_notes(off)
      %do s = 1 %to &wcval;
         %let var = %scan(&comm_character_fields,&s);
         proc sql noprint;
            insert into work.comm_character_attributes
            set
               input_data_id = &input_data_id
              ,comm_char_attribute_id = (&s-1) + &comm_char_attribute_id
              ,comm_attribute_name = "&var"
            ;
         quit;
         %if (&sqlrc gt 4) %then %do;
            %put %mo_note Error encountered creating work.comm_character_attributes (sqlrc=&sqlrc).;
            %let &_rc = 107;
            %goto ERROR_EXIT;
         %end;
      %end;

      %populate_postprocess(metadata_table = comm_character_attributes
                           ,metadata_key   = comm_char_attribute_id
                           ,_rc            = pprc);
      %if (&pprc ne 0) %then %do;
         %let &_rc = &pprc;
         %goto ERROR_EXIT;
      %end;
   %end;


   %*-------------------------------------------------------------;
   %* COMM_CHARACTER_DISTINCT_VALUES                              ;
   %*-------------------------------------------------------------;
   %if (&wcval gt 0) %then %do;
      %populate_preprocess(metadata_table = comm_character_distinct_values
                          ,metadata_key   = comm_attribute_distinct_value_id
                          ,_rc            = pprc);
      %if (&pprc ne 0) %then %do;
         %let &_rc = &pprc;
         %goto ERROR_EXIT;
      %end;

      %do s = 1 %to &wcval;
         %let var = %scan(&comm_character_fields,&s);
         %if %upcase(&var) = CHANNEL_CD %then %let var = channel_mo_comm_table;

         proc sql noprint;
            create table work.temp_char_distinct_vals as
               select distinct(&var) as comm_attribute_value length=256,
                      &input_data_id as input_data_id,
                      (&s-1) + &comm_char_attribute_id as comm_char_attribute_id,
                      . as comm_attribute_distinct_value_id
               from work.communications_numeric
               where &var ne ''
               order by comm_attribute_value;
         quit;
         %if (&sqlrc gt 4) %then %do;
            %put %mo_note Error encountered creating work.temp_char_distinct_vals (sqlrc=&sqlrc).;
            %let &_rc = 107;
            %goto ERROR_EXIT;
         %end;

         /* If var = time_period, then sort the time periods according to order, not alphabetically */
         %if %upcase(&var) = TIME_PERIOD %then %do;
            %if (%bquote(&time_period_table) ne %str()) %then %do;
               proc sql noprint;
                  create table work.temp_time_period_distinct_vals as
                     select a.*, b.order_no
                     from work.temp_char_distinct_vals as a left join work.time_period_in_mo_tbl as b
                     on a.comm_attribute_value = b.time_period;
               quit;
               %if (&sqlrc gt 4) %then %do;
                  %put %mo_note Error encountered creating work.temp_time_period_distinct_vals (sqlrc=&sqlrc).;
                  %let &_rc = 107;
                  %goto ERROR_EXIT;
               %end;

               proc sort data=work.temp_time_period_distinct_vals 
                         out=work.temp_char_distinct_vals (drop=order_no);
                  by order_no;
               run;
               %if (&syserr gt 4) %then %do;
                  %put %mo_note Error encountered sorting work.temp_time_period_distinct_vals (syserr=&syserr).;
                  %let &_rc = 100;
                  %goto ERROR_EXIT;
               %end;
            %end;
         %end;

         proc append base=work.comm_character_distinct_values data=work.temp_char_distinct_vals;
         run;
         %if (&syserr gt 4) %then %do;
            %put %mo_note Error encountered appending to work.comm_character_distinct_values (syserr=&syserr).;
            %let &_rc = 100;
            %goto ERROR_EXIT;
         %end;
      %end;

      data work.comm_character_distinct_values;
         set work.comm_character_distinct_values;
         comm_attribute_distinct_value_id = (_N_-1) + &comm_attribute_distinct_value_id;
      run;
      %if (&syserr gt 4) %then %do;
         %put %mo_note Error encountered updating work.comm_character_distinct_values (syserr=&syserr).;
         %let &_rc = 100;
         %goto ERROR_EXIT;
      %end;

      %populate_postprocess(metadata_table = comm_character_distinct_values
                           ,metadata_key   = comm_attribute_distinct_value_id
                           ,_rc            = pprc);
      %if (&pprc ne 0) %then %do;
         %let &_rc = &pprc;
         %goto ERROR_EXIT;
      %end;
      %symdel comm_char_attribute_id comm_attribute_distinct_value_id;
   %end;


   %*-------------------------------------------------------------;
   %* COMM_CHARACTER_ATTRIBUTE_VALUES                             ;
   %*-------------------------------------------------------------;
   %let wcval = %mo_word_count(&comm_character_fields);
   %if (&wcval gt 0) %then %do;
      %populate_preprocess(metadata_table = comm_character_attribute_values
                          ,metadata_key   = comm_attribute_value_id
                          ,_rc            = pprc);
      %if (&pprc ne 0) %then %do;
         %let &_rc = &pprc;
         %goto ERROR_EXIT;
      %end;

      %mo_no_notes(off)
      data work.comm_character_attribute_values2 (keep=&comm_character_values_vars comm_attribute_name comm_attribute_value);
         length comm_attribute_name $32. comm_attribute_value $256.;
         set work.comm_character_attribute_values work.communications_numeric;
         %let wcval = %mo_word_count(&comm_character_fields);
         %do s = 1 %to &wcval;
            %let var = %scan(&comm_character_fields,&s);
            %if %upcase(&var) = CHANNEL_CD %then %let var2 = channel_mo_comm_table;
            %else %let var2 = &var;
            comm_attribute_name = "&var";
            comm_attribute_value = compbl(trim(left(&var2)));
            if comm_attribute_value eq '' then comm_attribute_value=' ';
            comm_attribute_value_id = (&s-1) + &comm_attribute_value_id;
            output;
         %end;
      run;
      %if (&syserr gt 4) %then %do;
         %put %mo_note Error encountered creating work.comm_character_attribute_values2 (syserr=&syserr).;
         %let &_rc = 100;
         %goto ERROR_EXIT;
      %end;

      data work.comm_character_attribute_values2;
         set work.comm_character_attribute_values2;
         comm_attribute_value_id = (_N_-1) + &comm_attribute_value_id;
      run;
      %if (&syserr gt 4) %then %do;
         %put %mo_note Error encountered updating work.comm_character_attribute_values2 (syserr=&syserr).;
         %let &_rc = 100;
         %goto ERROR_EXIT;
      %end;

      proc sql noprint;
         create table work.comm_character_attribute_values3 as
            select a.input_data_id, a.campaign_id, a.communication_id, a.comm_attribute_value_id, a.comm_attribute_value,
               b.comm_char_attribute_id
            from work.comm_character_attribute_values2 as a left join work.comm_character_attributes as b
            on a.comm_attribute_name=b.comm_attribute_name;
         create table work.comm_character_attribute_values as
            select a.input_data_id, a.campaign_id, a.communication_id, a.comm_char_attribute_id, a.comm_attribute_value_id,
               b.comm_attribute_distinct_value_id
            from work.comm_character_attribute_values3 as a left join work.comm_character_distinct_values as b
            on a.comm_char_attribute_id=b.comm_char_attribute_id and a.comm_attribute_value=b.comm_attribute_value            
            order by comm_attribute_value_id, campaign_id, communication_id;
      quit;
      %if (&sqlrc gt 4) %then %do;
         %put %mo_note SQL error encountered updating work.comm_character_attribute_values (sqlrc=&sqlrc).;
         %let &_rc = 107;
         %goto ERROR_EXIT;
      %end;

      %populate_postprocess(metadata_table = comm_character_attribute_values
                           ,metadata_key   = comm_attribute_value_id
                           ,_rc            = pprc);
      %if (&pprc ne 0) %then %do;
         %let &_rc = &pprc;
         %goto ERROR_EXIT;
      %end;
      %symdel comm_attribute_value_id;
   %end;


   %*--------------------------------;
   %* CUSTOMER_ATTRIBUTES            ;
   %*--------------------------------;
   %populate_preprocess(metadata_table = customer_attributes
                       ,metadata_key   = customer_attribute_id
                       ,_rc            = pprc);
   %if (&pprc ne 0) %then %do;
      %let &_rc = &pprc;
      %goto ERROR_EXIT;
   %end;

   %mo_no_notes(off)
   /* Use work.extra_columns, which was generated during validate */
   proc sort data=work.extra_columns; /* From mo_validate_input_data */
      by name;
   run;
   %if (&syserr gt 4) %then %do;
      %put %mo_note Error encountered sorting the work.extra_columns table (syserr=&syserr).;
      %let &_rc = 100;
      %goto ERROR_EXIT;
   %end;

   data work.customer_attributes (keep=&customer_attributes_vars);
      set work.customer_attributes work.extra_columns;
      input_data_id = &input_data_id;
      customer_attribute_id = (_N_-1) + &customer_attribute_id;
      customer_attribute = name;
      if (type eq 1) then
         attribute_type = &_CUST_ATTRIBUTE_TYPE_NUM_;
      else attribute_type = &_CUST_ATTRIBUTE_TYPE_CHAR_;
   run;
   %if (&syserr gt 4) %then %do;
      %put %mo_note Error encountered updating the work.customer_attributes table (syserr=&syserr).;
      %let &_rc = 100;
      %goto ERROR_EXIT;
   %end;

   %populate_postprocess(metadata_table = customer_attributes
                        ,metadata_key   = customer_attribute_id
                        ,_rc            = pprc);
   %if (&pprc ne 0) %then %do;
      %let &_rc = &pprc;
      %goto ERROR_EXIT;
   %end;
   %symdel customer_attribute_id;


   %*--------------------------------;
   %* CUSTOMER_ATTRIBUTE_VALUES      ;
   %*--------------------------------;
   %populate_preprocess(metadata_table = customer_attribute_values
                       ,metadata_key   = customer_attribute_value_id
                       ,_rc            = pprc);
   %if (&pprc ne 0) %then %do;
      %let &_rc = &pprc;
      %goto ERROR_EXIT;
   %end;

   %mo_no_notes(off)

   proc sql noprint;
      create table work.mo_cust_attrs_distinct as
         select a.*, b.customer_attribute_id, b.input_data_id
         from work.extra_customer_attribute_values as a left join work.customer_attributes as b
         on a.customer_attribute = b.customer_attribute
         order by customer_attribute_id, customer_attribute_value;
   quit;
   %if (&sqlrc gt 4) %then %do;
      %put %mo_note SQL error encountered creating WORK.MO_CUST_ATTRS_DISTINCT (sqlrc=&sqlrc).;
      %let &_rc = 107;
      %goto ERROR_EXIT;
   %end;

   data work.customer_attribute_values (keep=&customer_attribute_values_vars);
      set work.customer_attribute_values work.mo_cust_attrs_distinct;
      customer_attribute_value_id = (_N_-1) + &customer_attribute_value_id;
   run;
   %if (&syserr gt 4) %then %do;
      %put %mo_note Error encountered updating the work.customer_attribute_values table (syserr=&syserr).;
      %let &_rc = 100;
      %goto ERROR_EXIT;
   %end;

   %populate_postprocess(metadata_table = customer_attribute_values
                        ,metadata_key   = customer_attribute_value_id
                        ,_rc            = pprc);
   %if (&pprc ne 0) %then %do;
      %let &_rc = &pprc;
      %goto ERROR_EXIT;
   %end;
   %symdel customer_attribute_value_id;


   %*--------------------------------;
   %* CUSTOMER_OFFER_ATTRIBUTES      ;
   %*--------------------------------;
   %populate_preprocess(metadata_table = customer_offer_attributes
                       ,metadata_key   = customer_offer_attribute_id
                       ,_rc            = pprc);
   %if (&pprc ne 0) %then %do;
      %let &_rc = &pprc;
      %goto ERROR_EXIT;
   %end;

   %mo_no_notes(off)
   proc contents data=moinput.&customer_table out=work.cust_out noprint;
   run;
   %if (&syserr gt 4) %then %do;
      %put %mo_note Error encountered retrieving contents of moinput.&customer_table (syserr=&syserr).;
      %let &_rc = 100;
      %goto ERROR_EXIT;
   %end;

   proc sql noprint;
      create table work.num_fields as
      select distinct(numeric_measure)
      from work.cust_col_metadata_in_mo_tbl
      order by numeric_measure;
   quit;
   %if (&sqlrc gt 4) %then %do;
      %put %mo_note SQL error encountered creating work.num_fields table (sqlrc=&sqlrc).;
      %let &_rc = 107;
      %goto ERROR_EXIT;
   %end;

   data work.customer_offer_attributes (keep=&customer_offer_attributes_vars);
      set work.customer_offer_attributes work.num_fields;
      input_data_id = &input_data_id;
      customer_offer_attribute_id = (_N_-1) + &customer_offer_attribute_id;
      customer_offer_attribute = numeric_measure;
   run;
   %if (&syserr gt 4) %then %do;
      %put %mo_note Error encountered updating the work.customer_attribute_values table (syserr=&syserr).;
      %let &_rc = 100;
      %goto ERROR_EXIT;
   %end;

   %populate_postprocess(metadata_table = customer_offer_attributes
                        ,metadata_key   = customer_offer_attribute_id
                        ,_rc            = pprc);
   %if (&pprc ne 0) %then %do;
      %let &_rc = &pprc;
      %goto ERROR_EXIT;
   %end;
   %symdel customer_offer_attribute_id;


   %*--------------------------------;
   %* INPUT_DATA_MEASURES            ;
   %*--------------------------------;
   %populate_preprocess(metadata_table = input_data_measures
                       ,metadata_key   = input_data_measure_id
                       ,_rc            = pprc);
   %if (&pprc ne 0) %then %do;
      %let &_rc = &pprc;
      %goto ERROR_EXIT;
   %end;

   data work.input_data_measures_cust (keep=input_data_id attribute_id attribute_type);
      set work.customer_attributes;
      where attribute_type = &_CUST_ATTRIBUTE_TYPE_NUM_;
      rename customer_attribute_id=attribute_id;
      attribute_type = &_ATTRIBUTE_TYPE_CUST_;
   run;
   %if (&syserr gt 4) %then %do;
      %put %mo_note Error encountered creating the work.input_data_measures_cust table (syserr=&syserr).;
      %let &_rc = 100;
      %goto ERROR_EXIT;
   %end;

   %let wcval = %mo_word_count(&comm_numeric_fields);
   %if (&wcval gt 0) %then %do;
      data work.input_data_measures_comm (keep=input_data_id attribute_id attribute_type);
         set work.comm_numeric_attributes;
         rename comm_attribute_id=attribute_id;
         attribute_type = &_ATTRIBUTE_TYPE_COMM_;
      run;
      %if (&syserr gt 4) %then %do;
         %put %mo_note Error encountered creating the work.input_data_measures_comm table (syserr=&syserr).;
         %let &_rc = 100;
         %goto ERROR_EXIT;
      %end;
   %end;

   data work.input_data_measures_custcomm (keep=input_data_id attribute_id attribute_type);
      set work.customer_offer_attributes;
      rename customer_offer_attribute_id=attribute_id;
      attribute_type = &_ATTRIBUTE_TYPE_CUST_COMM_;
   run;
   %if (&syserr gt 4) %then %do;
      %put %mo_note Error encountered creating the work.input_data_measures_custcomm table (syserr=&syserr).;
      %let &_rc = 100;
      %goto ERROR_EXIT;
   %end;

   data work.input_data_measures;
      set work.input_data_measures_cust
          %if (&wcval gt 0) %then %do;        
             work.input_data_measures_comm
          %end;
          work.input_data_measures_custcomm;
      input_data_measure_id = (_N_-1) + &input_data_measure_id;
   run;
   %if (&syserr gt 4) %then %do;
      %put %mo_note Error encountered updating the work.input_data_measures table (syserr=&syserr).;
      %let &_rc = 100;
      %goto ERROR_EXIT;
   %end;

   %populate_postprocess(metadata_table = input_data_measures
                        ,metadata_key   = input_data_measure_id
                        ,_rc            = pprc);
   %if (&pprc ne 0) %then %do;
      %let &_rc = &pprc;
      %goto ERROR_EXIT;
   %end;
   %symdel input_data_measure_id;


   %*--------------------------------;
   %* AGENTS                         ;
   %*--------------------------------;
   %if (%bquote(&agent_capacity_table) ne %str()) %then %do;
      %populate_preprocess(metadata_table = agents
                          ,metadata_key   = agent_id
                          ,_rc            = pprc);
      %if (&pprc ne 0) %then %do;
         %let &_rc = &pprc;
         %goto ERROR_EXIT;
      %end;

      %mo_no_notes(off)
      proc contents data=work.agent_capacity_in_mo_tbl out=work.agent_out noprint;
      run;
      %if (&syserr gt 4) %then %do;
         %put %mo_note Error encountered retrieving contents of work.agent_capacity_in_mo_tbl (syserr=&syserr).;
         %let &_rc = 100;
         %goto ERROR_EXIT;
      %end;

      %local agent_column_type;
      proc sql noprint;
         select type into :agent_column_type
         from work.agent_out
         where upcase(name) eq "%upcase(&agent_column)";
      quit;
      %if (&sqlrc gt 4) %then %do;
         %put %mo_note SQL error encountered retrieving type from work.agent_out table (sqlrc=&sqlrc).;
         %let &_rc = 107;
         %goto ERROR_EXIT;
      %end;

      %mo_no_notes(off)
      proc sql noprint;
         create table work.agent_capacity_table as
         select a.*, b.time_period_id
         from work.agent_capacity_in_mo_tbl as a left join mometa.time_periods as b
         on (compbl(trim(left(a.time_period))) eq compbl(trim(left(b.time_period))))
         and (b.input_data_id eq &input_data_id)
         order by a.&agent_column;
      quit;
      %if (&sqlrc gt 4) %then %do;
         %put %mo_note SQL error encountered creating work.agent_capacity_table table (sqlrc=&sqlrc).;
         %let &_rc = 107;
         %goto ERROR_EXIT;
      %end;

      data work.agents (keep=&agents_vars);
         set work.agents work.agent_capacity_table (rename=(&agent_column=__agent99 agent_id=__agent98 time_period=period capacity=cap));
         input_data_id = &input_data_id;
         agent_id = (_N_-1) + &agent_id;
         %if (&agent_column_type eq 1) %then %do;
            agent = put(__agent99,32.);
         %end;
         %else %do;
            agent = __agent99;
         %end;
         agent = left(agent);
         capacity = cap;
      run;
      %if (&syserr gt 4) %then %do;
         %put %mo_note Error encountered updating the work.agents table (syserr=&syserr).;
         %let &_rc = 100;
         %goto ERROR_EXIT;
      %end;

      %populate_postprocess(metadata_table = agents
                           ,metadata_key   = agent_id
                           ,_rc            = pprc);
      %if (&pprc ne 0) %then %do;
         %let &_rc = &pprc;
         %goto ERROR_EXIT;
      %end;
      %symdel agent_id;
   %end;


   %*--------------------------------;
   %* PREPOPULATED_CONSTRAINTS       ;
   %*--------------------------------;
   %if (%bquote(&constraints_table) ne %str()) %then %do;
      %populate_preprocess(metadata_table = prepopulated_constraints
                          ,metadata_key   = prepopulated_constraint_id
                          ,_rc            = pprc);
      %if (&pprc ne 0) %then %do;
         %let &_rc = &pprc;
         %goto ERROR_EXIT;
      %end;

      %mo_no_notes(off)

      data work.prepopulated_constraints (keep = &prepopulated_constraints_vars);
         set work.prepopulated_constraints 
             work.constraints_in_mo_tbl (rename=(constraint_id=constraint_name
                                                 active_flg=active
                                                 sensitivity_flg=include_in_sensitivity_analysis
                                                 constraint_rhs=rhs));
         input_data_id = &input_data_id;
         prepopulated_constraint_id = (_N_-1) + &prepopulated_constraint_id;

         if (active = .) then do;
            active = 1;
         end;

         if (include_in_sensitivity_analysis = .) then do;
            include_in_sensitivity_analysis = 0;
         end;

         if (upcase(constraint_type_cd) = 'BUDGET') then do;
            constraint_type = &_CONSTRAINT_TYPE_BUDGET_;
         end;
         else if (upcase(constraint_type_cd) = 'CELLSIZE') then do;
            constraint_type = &_CONSTRAINT_TYPE_CELLSIZE_;
         end;
         else if (upcase(constraint_type_cd) = 'CHANNEL') then do;
            constraint_type = &_CONSTRAINT_TYPE_CHANNEL_;
         end;
         else if (upcase(constraint_type_cd) in ('MISCELLANEOUS','MISC','')) then do;
            constraint_type = &_CONSTRAINT_TYPE_MISC_;
         end;
         else if (upcase(constraint_type_cd) = 'ROI') then do;
            constraint_type = &_CONSTRAINT_TYPE_ROI_;
         end;
         else if (upcase(constraint_type_cd) in ('REPORT', 'REPORT ONLY')) then do;
            constraint_type = &_CONSTRAINT_TYPE_REPORT_ONLY_;
         end;
         else do;
            constraint_type = &_CONSTRAINT_TYPE_MISC_;
         end;

         if (upcase(constraint_scope_cd) = 'AGGREGATE') then do;
            constraint_scope = &_CONSTRAINT_SCOPE_AGGREGATE_;
         end;
         else if (upcase(constraint_scope_cd) = 'CUSTOMER') then do;
            constraint_scope = &_CONSTRAINT_SCOPE_CUSTOMER_;
         end;
         else if (upcase(constraint_scope_cd) = 'HOUSEHOLD') then do;
            constraint_scope = &_CONSTRAINT_SCOPE_HOUSEHOLD_;
         end;
         else do;
            constraint_scope = &_CONSTRAINT_SCOPE_AGGREGATE_;
         end;

         if (upcase(constraint_operator) = 'LE') then do;
            operator = &_LE_;
         end;
         else do;
            operator = &_GE_;
         end;

         if (upcase(constraint_computation_cd) = 'NUM_OFFERS') then do;
            computation_type = &_COMPUTATION_TYPE_NUM_OFFERS_;
         end;
         else if (upcase(constraint_computation_cd) = 'SUM') then do;
            computation_type = &_COMPUTATION_TYPE_SUM_;
         end;
         else if (upcase(constraint_computation_cd) = 'AVERAGE') then do;
            computation_type = &_COMPUTATION_TYPE_AVERAGE_;
         end;
         else if (upcase(constraint_computation_cd) = 'RATIO') then do;
            computation_type = &_COMPUTATION_TYPE_RATIO_;
         end;
         else if (constraint_computation_cd = '') then do;
            if (scan(numeric_measure,2) ne '') then do;
               computation_type = &_COMPUTATION_TYPE_RATIO_;
            end;
            else if (scan(numeric_measure,1) ne '') then do;
               computation_type = &_COMPUTATION_TYPE_SUM_;
            end;
            else do;
               computation_type = &_COMPUTATION_TYPE_NUM_OFFERS_;
            end;
         end;

         numerator_attribute_name = scan(numeric_measure,1);
         denominator_attribute_name = scan(numeric_measure,2);
      run;
      %if (&syserr gt 4) %then %do;
         %put %mo_note Error encountered updating the work.prepopulated_constraints table (syserr=&syserr).;
         %let &_rc = 100;
         %goto ERROR_EXIT;
      %end;

      %populate_postprocess(metadata_table = prepopulated_constraints
                           ,metadata_key   = prepopulated_constraint_id
                           ,_rc            = pprc);
      %if (&pprc ne 0) %then %do;
         %let &_rc = &pprc;
         %goto ERROR_EXIT;
      %end;
      %symdel prepopulated_constraint_id;
   %end;


   %*--------------------------------;
   %* PREPOPULATED_CONTACT_POLICIES  ;
   %*--------------------------------;
   %if (%bquote(&contact_policy_table) ne %str()) %then %do;
      %populate_preprocess(metadata_table = prepopulated_contact_policies
                          ,metadata_key   = prepopulated_contact_policy_id
                          ,_rc            = pprc);
      %if (&pprc ne 0) %then %do;
         %let &_rc = &pprc;
         %goto ERROR_EXIT;
      %end;

      %mo_no_notes(off)

      data work.prepopulated_contact_policies (keep = &prepop_contact_policies_vars 
                                                      %if (%bquote(&time_period_table) ne %str()) %then %do;
                                                         start_time_period end_time_period
                                                      %end;
                                                      );
         set work.prepopulated_contact_policies
             work.contact_policy_in_mo_tbl (rename=(contact_policy_id=contact_policy_name
                                                    active_flg=active
                                                    contact_limit_cnt=rhs));
         input_data_id = &input_data_id;
         prepopulated_contact_policy_id = (_N_-1) + &prepopulated_contact_policy_id;

         if (active = .) then do;
            active = 1;
         end;

         if (upcase(target_level_cd) = 'CUSTOMER') then do;
            target_level = &_CONTACT_POLICY_TARGET_CUSTOMER_;
         end;
         else if (upcase(target_level_cd) = 'HOUSEHOLD') then do;
            target_level = &_CONTACT_POLICY_TARGET_HOUSEHLD_;
         end;
         else do;
            target_level = &_CONTACT_POLICY_TARGET_CUSTOMER_;
         end;

         /* We have already validated that the contact policies are the correct form,
            so we know that contact_limit_cnt (rhs) is for maxmin/rolling and missing for blocking. */
         if (rhs ne .) then contact_policy_type = 0; /* MaxMin / Rolling */
         else contact_policy_type = 1; /* Blocking */

         if upcase(contact_policy_operator) = 'LE' then operator = &_LE_;
         else if upcase(contact_policy_operator) = 'GE' then operator = &_GE_;
         else operator = .; /* Blocking */
      run;
      %if (&syserr gt 4) %then %do;
         %put %mo_note Error encountered updating the work.prepopulated_contact_policies table (syserr=&syserr).;
         %let &_rc = 100;
         %goto ERROR_EXIT;
      %end;

      %if (%bquote(&time_period_table) ne %str()) %then %do;
         %let first_time_period_id = 0;
         %let last_time_period_id = 0;
         proc sql noprint;
            create table work.prepop_contact_start as
               select a.*, b.time_period_id as policy_start_id
               from work.prepopulated_contact_policies as a left join work.time_periods as b
               on a.start_time_period = b.time_period;
            create table work.prepopulated_contact_policies as
               select a.*, b.time_period_id as policy_end_id
               from work.prepop_contact_start as a left join work.time_periods as b
               on a.end_time_period = b.time_period
               order by prepopulated_contact_policy_id;
            select min(time_period_id), max(time_period_id)
               into :first_time_period_id, :last_time_period_id
               from work.time_periods;
         quit;
         %if (&sqlrc gt 4) %then %do;
            %put %mo_note SQL error encountered creating work.prepop_contact_start table (sqlrc=&sqlrc).;
            %let &_rc = 107;
            %goto ERROR_EXIT;
         %end;

         data work.prepopulated_contact_policies (drop=policy_start_id policy_end_id start_time_period end_time_period);
            set work.prepopulated_contact_policies;

            if (contact_policy_type = &_CONTACT_POLICY_TYPE_BLOCKING_) then do;
               if (blocking_period_cnt ne .) then do;
                  time_based = 1;

                  if (policy_start_id = .) then do;
                     start_time_period_id = &_CONTACT_POLICY_TP_TABLE_RANGE_;
                  end;
                  else do;
                     start_time_period_id = policy_start_id;
                  end;

                  if (policy_end_id = .) then do;
                     end_time_period_id = &_CONTACT_POLICY_TP_TABLE_RANGE_;
                  end;
                  else do;
                     end_time_period_id = policy_end_id;
                  end;
               end;
               else do;
                  time_based = 0;
                  blocking_period_cnt = .;
                  start_time_period_id = .;
                  end_time_period_id = .;
               end;
            end;
            else do; /* maxmin/rolling */
               if (policy_start_id ne .) or (policy_end_id ne .) or (rolling_period_cnt ne .) then do;
                  time_based = 1;

                  if (policy_start_id = .) then do;
                     start_time_period_id = &_CONTACT_POLICY_TP_TABLE_RANGE_;
                  end;
                  else do;
                     start_time_period_id = policy_start_id;
                  end;

                  if (policy_end_id = .) then do;
                     end_time_period_id = &_CONTACT_POLICY_TP_TABLE_RANGE_;
                  end;
                  else do;
                     end_time_period_id = policy_end_id;
                  end;
               end;
               else do;
                  time_based = 0;
                  rolling_period_cnt = .;
                  start_time_period_id = .;
                  end_time_period_id = .;
               end;
            end;
         run;
         %if (&syserr gt 4) %then %do;
            %put %mo_note Error encountered updating the work.prepopulated_contact_policies table (syserr=&syserr).;
            %let &_rc = 100;
            %goto ERROR_EXIT;
         %end;
      %end;

      %populate_postprocess(metadata_table = prepopulated_contact_policies
                           ,metadata_key   = prepopulated_contact_policy_id
                           ,_rc            = pprc);
      %if (&pprc ne 0) %then %do;
         %let &_rc = &pprc;
         %goto ERROR_EXIT;
      %end;
      %symdel prepopulated_contact_policy_id;
   %end;



   %*--------------------------------;
   %* Exit processing                ;
   %*--------------------------------;
   %* Successful completion;
   %let &_rc = 0;
   %put;
   %put %mo_note Successfully populated metadata tables for input_data_id = &input_data_id..;
   %goto EXIT;


   %ERROR_EXIT:
   %put;
   %put %mo_error Error encountered populating metadata tables for input_data_id = &input_data_id (rc=&&&_rc).;
   %symdel campaign_id channel_id time_period_id communication_id;
   %symdel comm_attribute_id comm_attribute_value_id comm_char_attribute_id comm_attribute_distinct_value_id;
   %symdel customer_attribute_id customer_attribute_value_id customer_offer_attribute_id;
   %symdel input_data_measure_id agent_id prepopulated_constraint_id prepopulated_contact_policy_id;


   %EXIT:
   %mo_no_notes(off)


%mend mo_prepopulate_metadata_tables;
