/****************************************************************************/
/* Copyright (c) 2005-2011, SAS Institute Inc., Cary, NC, USA,              */
/*                       All Rights Reserved.                               */
/*                                                                          */
/* Name:    mo_reassign_project_ownership                                   */
/*                                                                          */
/* Purpose: Given a project ID, a "from" user ID, and a "to" user ID,       */
/*          transfer ownership of the project between the users by:         */
/*          1) copying the the "from" user's project directory to the       */
/*             "to" user's directory project directory                      */
/*          2) updating the job file for scheduled scenario optimization    */
/*             to reflect "to" user/owner                                   */
/*          3) changing the user ID to the "to" user within MO datastore    */
/*             tables                                                       */
/*                                                                          */
/* Input:   project_id = Numeric ID within MO of the project for which      */
/*                       ownership is to be reassigned                      */
/*          from_user_id = Numeric ID within MO of the user that currently  */
/*                         "owns" the specified project and from which the  */
/*                         project is to be reassigned                      */
/*          to_user_id = Numeric ID within MO of the user to which the      */
/*                       specified project is to be reassigned              */
/*                         project is to be reassigned                      */
/* Output:  _rc = Return code indicating success (0) or failure (non-0) of  */
/*                the project reassignment request                          */
/*                                                                          */
/* HISTORY:                                                                 */
/* 5.3:                                                                     */
/* 03Jul08 Creation (KEC)                                                   */
/* 24Sep08 S0483854 Change mechanism for accessing MO Property value when   */
/*         running on 9.2 (Property objects on the MO Application object)   */
/*         (KEC)                                                            */
/* 18Nov08 S0483854 Remove code and references to related to jobs being     */
/*         scheduled through the SAS Batch Datastep Server since this       */
/*         mechanism for scheduling jobs was obsoleted by the Java-based    */
/*         MO Launcher (KEC)                                                */
/* 02Dec08 S0483854 Eliminate home-grown moutil catalog functions CREATE_   */
/*         DIRECTORY and RENAME_DIRECTORY.  Replace with corresponding SAS  */
/*         built-in functions DCREATE and RENAME.                           */
/*                                                                          */
/* F1MO54, F0MO541                                                          */
/* 09Jan12 S0823209 Replace use of SQL query of Dictionay Tables with use   */
/*         of SAS Proc Contents.                                            */
/*                                                                          */
/****************************************************************************/

%macro mo_reassign_project_ownership(project_id       =
                                    ,from_user_id     =
                                    ,to_user_id       =
                                    ,_rc              =
                                    );


   %local macname;
   %let macname=&sysmacroname;
   %let macname=%lowcase(&macname);

   %if %bquote(&_rc) eq %str() %then %do;
      %local rc;
      %let _rc = rc;
   %end;
   %else %do;
      %global &_rc;
   %end;
   %let &_rc = 0;

   %mo_constants;

   %if (%bquote(&project_id) eq %str()) %then %do;
      %put %mo_error No project ID was specified.;
      %let &_rc = 102;
      %goto ERROR_EXIT;
   %end;
   %if (%bquote(from_user_id) eq %str()) %then %do;
      %put %mo_error No FROM user ID was specified.;
      %let &_rc = 102;
      %goto ERROR_EXIT;
   %end;
   %if (%bquote(&to_user_id) eq %str()) %then %do;
      %put %mo_error No TO user ID was specified.;
      %let &_rc = 102;
      %goto ERROR_EXIT;
   %end;
   %if (&from_user_id eq &to_user_id) %then %do;
      %put %mo_warning TO user ID equals FROM user ID - No re-assignment performed.;
      %goto EXIT;
   %end;


   %*---------------------------------;
   %* Useful macros.                  ;
   %*---------------------------------;

   %***** MACRO APPEND_TO_LOG_DATASET *****;
   %macro append_to_log_dataset(timestamp, action, item, _rc);
      %local macname;
      %let macname=&sysmacroname;
      %let macname=%lowcase(&macname);

      %if %bquote(&_rc) eq %str() %then %do;
         %local rc;
         %let _rc = rc;
      %end;
      %else %do;
         %global &_rc;
      %end;
      %let &_rc = 0;

      %if (not %sysfunc(exist(work.activity_log))) %then %do;
         proc sql noprint;
            create table work.activity_log
            (timestamp num format=datetime.,
             action    char(50),
             item      char(256)
			 );
         quit;
      %end;

      data work.temp_rec;
         length action $50;
         length item $256;
         timestamp = &timestamp;
         action = "&action";
         item = "&item";
      run;
      proc append base=work.activity_log data=work.temp_rec;
      run;      
   %mend append_to_log_dataset;


   %***** MACRO CREATE_DIR *****;
   %macro create_dir(parent_dir, new_dir, user_name, _rc);
      %local macname;
      %let macname=&sysmacroname;
      %let macname=%lowcase(&macname);

      %if %bquote(&_rc) eq %str() %then %do;
         %local rc;
         %let _rc = rc;
      %end;
      %else %do;
         %global &_rc;
      %end;
      %let &_rc = 0;

      %let dir = %mo_pathname(&parent_dir.&new_dir.);

      %local cdrc;
      %let cdrc=;
      data _null_;
         newdir = dcreate("&new_dir", "&parent_dir");
         call symput("cdrc", trim(left(newdir)));
      run;
      %if (%bquote(&cdrc) eq %str()) %then %do;
         %let &_rc = 105;
         %put %mo_error I/O error encountered creating directory = %upcase(&dir) for user = %upcase(&user_name).;
      %end;
      %else %do;
         %put %mo_note Successfully created directory = %upcase(&dir) for user = %upcase(&user_name).;
         %append_to_log_dataset(%sysfunc(datetime()), DIRECTORY_CREATED, &dir);
      %end;
   %mend create_dir;


   %***** MACRO COPY_DATASETS *****;
   %macro copy_datasets(from_dir, to_dir, _rc);
      %local macname;
      %let macname=&sysmacroname;
      %let macname=%lowcase(&macname);

      %if %bquote(&_rc) eq %str() %then %do;
         %local rc;
         %let _rc = rc;
      %end;
      %else %do;
         %global &_rc;
      %end;
      %let &_rc = 0;

      %mo_no_notes(off)
      libname _fmlib_ "&from_dir";
      %if (&syslibrc ne 0) %then %do;
         %put %mo_error Error assigning _FMLIB_ libref for directory = %upcase(&from_dir) (syslibrc=&syslibrc);
         %let &_rc = 109;
      %end;
      libname _tolib_ "&to_dir";
      %if (&syslibrc ne 0) %then %do;
         %put %mo_error Error assigning _TOLIB_ libref for directory = %upcase(&to_dir) (syslibrc=&syslibrc);
         %let &_rc = 109;
      %end;
      %mo_no_notes(on)

      proc datasets nolist lib=_fmlib_;
         copy out = _tolib_;
      quit;
      %if (&syserr gt 4) %then %do;
         %let &_rc = 100;
         %put %mo_error I/O error encountered while copying datasets from FROM directory = %upcase(&from_dir) to TO directory = %upcase(&to_dir) (syserr=&syserr).;
      %end;
      %else %do;
         %put %mo_note Successfully copied SAS datasets from FROM directory = %upcase(&from_dir) to TO directory = %upcase(&to_dir).;
         %append_to_log_dataset(%sysfunc(datetime()), DATASETS_COPIED_TO_DIR, &to_dir);
      %end;

      libname _fmlib_ clear;
      libname _tolib_ clear;
   %mend copy_datasets;


   %***** MACRO COPY_FILE *****;
   %macro copy_file(from_dir, to_dir, file_name, _rc);
      %local macname;
      %let macname=&sysmacroname;
      %let macname=%lowcase(&macname);

      %if %bquote(&_rc) eq %str() %then %do;
         %local rc;
         %let _rc = rc;
      %end;
      %else %do;
         %global &_rc;
      %end;
      %let &_rc = 0;

      %let file1 = %mo_pathname(&from_dir.&separator.&file_name);
      %let file2 = %mo_pathname(&to_dir.&separator.&file_name);
      %if (%sysfunc(fileexist(&file1)) eq 1) %then %do;
         data _null_;
            file "&file2" lrecl=256;
            infile "&file1";
            input;
            put _INFILE_;
         run;
         %if (&syserr gt 4) %then %do;
            %let &_rc = 200;
            %put %mo_error Error encountered while copying file = %upcase(&file_name) from directory = %upcase(&from_dir) to directory = %upcase(&to_dir) (syserr=&syserr).;
         %end;
         %else %do;
            %put %mo_note Successfully copied file = %upcase(&file_name) from directory = %upcase(&from_dir) to directory = %upcase(&to_dir).;
            %append_to_log_dataset(%sysfunc(datetime()), FILE_COPIED_TO_DIR, &file2);
         %end;
      %end;
   %mend copy_file;


   %***** MACRO UPDATE_USER_ID *****;
   %macro update_user_id(old_user_id, new_user_id, project_id, metadata_table, _rc);
      %local macname;
      %let macname=&sysmacroname;
      %let macname=%lowcase(&macname);

      %if %bquote(&_rc) eq %str() %then %do;
         %local rc;
         %let _rc = rc;
      %end;
      %else %do;
         %global &_rc;
      %end;
      %let &_rc = 0;

      %mo_create_tables(lib=work, table=&metadata_table, _rc=cudrc)
      %if (&cudrc ne 0) %then %do;
         %let &_rc = 100;
         %put %mo_error Error occurred creating %upcase(WORK.&metadata_table) table.;
      %end;

      %else %do;
         proc sql noprint;
            insert into work.&metadata_table
            select * from mometa.&metadata_table
            where ((user_id = &old_user_id) and 
                   (project_id = &project_id));
         quit;
         %if (&sqlrc gt 4) %then %do;
            %let &_rc = 107;
            %put %mo_error SQL error encountered while updating %upcase(WORK.&metadata_table) table (sqlrc=&sqlrc).;
         %end;

         %else %do;
            data work.&metadata_table;
               set work.&metadata_table;
               user_id = &new_user_id;
            run;
            %if (&syserr gt 4) %then %do;
               %let &_rc = 100;
               %put %mo_error Error encountered while updating %upcase(WORK.&metadata_table) table (syserr=&syserr).;
            %end;

            %else %do;
               proc contents data=mometa.&metadata_table out=work.idColNames_tmp (keep=memname name) noprint nodetails;
               run;

               %local column_name_list;
               %let column_name_list=;
               proc sql noprint;
                  select distinct name into :column_name_list separated by ' '
                  from work.idColNames_tmp
                  where ((index(name, "_id") > 0) and (name ne "user_id"));
               quit;
               %let column_name_list = %lowcase(&column_name_list);

               %if (&metadata_table ne "projects") %then %do;
                  %mo_record_update_multiple(metadata_table = mometa.&metadata_table
                                            ,update_table   = work.&metadata_table
                                            ,by             = &column_name_list
                                            ,_rc            = updrc);
                  %if (&updrc ne 0) %then %do;
                     %let &_rc = 103;
                  %end;      
                  %else %do;
                     %put %mo_note Successfully updated table MOMETA.&metadata_table for project ID = &project_ID FROM user ID = &old_user_id) TO user ID = &new_user_id.;
                     %append_to_log_dataset(%sysfunc(datetime()), USER_ID_UPDATED_IN_TABLE, &metadata_table);
                  %end;
               %end;      
               %else %do;
                  %mo_record_update(metadata_table = mometa.&metadata_table
                                   ,update_table   = work.&metadata_table
                                   ,by             = &column_name_list
                                   ,_rc            = updrc);
                  %if (&updrc ne 0) %then %do;
                     %let &_rc = 103;
                  %end;      
                  %else %do;
                     %put %mo_note Successfully updated table MOMETA.&metadata_table for project ID = &project_ID FROM user ID = &old_user_id) TO user ID = &new_user_id.;
                     %append_to_log_dataset(%sysfunc(datetime()), USER_ID_UPDATED_IN_TABLE, &metadata_table);
                  %end;
               %end;      
            %end;
         %end;
      %end;      
   %mend update_user_id;


   %*---------------------------------;
   %* Obtain lock for metadata update ;
   %*---------------------------------;
   %mo_lock_dataset(data=mometa.uid,_rc=lckrc)
   %if (&lckrc ne 0) %then %do;
      %put %mo_error Error obtaining lock of MOMETA.UID.;
      %let &_rc = 106;
      %goto ERROR_EXIT;
   %end;


   %*--------------------------------;
   %* Retrieve/validate metadata     ;
   %*--------------------------------;
   %mo_no_notes(off)
   %local project_name project_user_id;
   data _null_;
      set mometa.projects;
      if (project_id eq &project_id) then do;
         call symput('project_name', strip(project_name));
         call symput('project_user_id', user_id);
      end;
   run;
   %if (&syserr gt 4) %then %do;
      %let &_rc = 100;
      %put %mo_error Error occurred while reading metadata (syserr=&syserr).;
      %goto ERROR_EXIT;
   %end;
   %if (%bquote(X&project_name) eq %str(X)) %then %do;
      %let &_rc = 101;
      %put %mo_error Invalid metadata detected.;
      %put %mo_error Project name not identified for project ID = &project_id..;
      %goto ERROR_EXIT;
   %end;
   %if (%bquote(&project_user_id) eq %str()) %then %do;
      %let &_rc = 101;
      %put %mo_error Invalid metadata detected.;
      %put %mo_error User ID not identified for project ID = &project_id..;
      %goto ERROR_EXIT;
   %end;
   %if (&project_user_id ne &from_user_id) %then %do;
      %let &_rc = 102;
      %put %mo_error Invalid input parameter.;
      %put %mo_error FROM user ID = &from_user_id is not the owner of project ID = &project_id..;
      %goto ERROR_EXIT;
   %end;

   %local from_user_name to_user_name;
   data _null_;
      set mometa.users;
      if (user_id eq &from_user_id) then do;
         call symput('from_user_name',strip(username));
      end;
      else if (user_id eq &to_user_id) then do;
         call symput('to_user_name',strip(username));
      end;
   run;
   %if (&syserr gt 4) %then %do;
      %let &_rc = 100;
      %put %mo_error Error occurred while reading metadata (syserr=&syserr).;
      %goto ERROR_EXIT;
   %end;
   %if (%bquote(&from_user_name) eq %str()) %then %do;
      %let &_rc = 101;
      %put %mo_error Invalid metadata detected.;
      %put %mo_error FROM username not identified for project ID = &project_id, FROM user ID = &from_user_id..;
      %goto ERROR_EXIT;
   %end;
   %if (%bquote(&to_user_name) eq %str()) %then %do;
      %let &_rc = 101;
      %put %mo_error Invalid metadata detected.;
      %put %mo_error TO username not identified for TO user ID = &to_user_id..;
      %goto ERROR_EXIT;
   %end;
   %mo_no_notes(on)


   %*-----------------------------------;
   %* Determine if any scenarios in the ;
   %* FROM project are being optimized. ;
   %*-----------------------------------;
   proc sql noprint;
      select count(*), scenario_id into :num_submitted, :scenario_id_list separated by ' '
      from mometa.scenarios
      where ((user_id = &from_user_id) and (project_id = &project_id) and 
             (status = &_SUBMITTED_));
   quit;
   %if (&sqlrc gt 4) %then %do;
      %put %mo_error SQL error occurred retrieving data from MOMETA.SCENARIOS (sqlrc=&sqlrc).;
      %let &_rc = 107;
      %goto ERROR_EXIT;
   %end;
   %if (&num_submitted gt 0) %then %do;
      %put %mo_error Cannot re-assign project ID = &project_id because there are &num_submitted optimization jobs currently running for scenario IDs = &scenario_id_list.;
      %let &_rc = 105;
      %goto ERROR_EXIT;
   %end;


   %*---------------------------------;
   %* Validate TO,FROM project dirs.  ;
   %* Create TO project dir if valid. ;
   %*---------------------------------;
   %local from_proj_dir to_proj_dir;
   %let from_proj_dir = %mo_pathname(&moproj.&separator.&from_user_name.&separator.&project_name);
   %let to_proj_dir = %mo_pathname(&moproj.&separator.&to_user_name.&separator.&project_name);

   %if (not %sysfunc(fileexist(&from_proj_dir))) %then %do;
      %let &_rc = 105;
      %put %mo_error Project directory = %upcase(&from_proj_dir) not found for FROM user ID = &from_user_id .;
      %goto ERROR_EXIT;
   %end;

   %local to_dir_already_existed;
   %let to_dir_already_existed = 0;
   %if (%sysfunc(fileexist(&to_proj_dir))) %then %do;
      %let to_dir_already_existed = 1;
      %let &_rc = 105;
      %put %mo_error Project directory = %upcase(&to_proj_dir) already exists for TO user ID = &to_user_id.;
      %goto ERROR_EXIT;
   %end;
   %else %do;
      %create_dir(%mo_pathname(&moproj.&separator.&to_user_name.&separator),
                  %mo_pathname(&project_name), &to_user_name, crtrc);
      %if (&crtrc ne 0) %then %do;
         %let &_rc = &crtrc;
         %goto ERROR_EXIT;
      %end;

      %copy_file(&from_proj_dir, &to_proj_dir, mo_opt_config.sas, cprc);
      %if (&cprc ne 0) %then %do;
         %let &_rc = &cprc;
         %goto ERROR_EXIT;
      %end;
   %end;


   %*--------------------------------;
   %* Create TO scenario dirs, copy  ;
   %* FROM scenario dir contents to  ;
   %* TO scenario dir.               ;
   %*--------------------------------;
   proc sql noprint;
      select scenario_id, scenario_name 
      into :scenario_id_list separated by '*', :scenario_name_list separated by '*'
      from mometa.scenarios
      where ((project_id = &project_id) and (user_id = &from_user_id));
   quit;
   %if (&sqlrc gt 4) %then %do;
      %put %mo_error SQL error occurred retrieving data from MOMETA.SCENARIOS (sqlrc=&sqlrc).;
      %let &_rc = 107;
      %goto ERROR_EXIT;
   %end;

   /* Create a scenario directory with log, saved, solution, and reports     */
   /* sub_directories for each scenario name in the FROM project.            */
   %let num_scenarios = %mo_word_count(%bquote(&scenario_name_list), %str(*));
   %do i = 1 %to &num_scenarios;
      %let scenario_name = %scan(&scenario_name_list, &i, *);
      %let scenario_id = %scan(&scenario_id_list, &i, *);

      %do q = 1 %to 5;
         %if (&q eq 1) %then %do; /* scenario directory */
            %let parent_dir=%mo_pathname(&to_proj_dir.&separator);
            %let new_dir=%mo_pathname(&scenario_name);

            %let to_dir = %mo_pathname(&to_proj_dir.&separator.&scenario_name);
            %let from_dir = %mo_pathname(&from_proj_dir.&separator.&scenario_name);
         %end;
         %else %do;
            %let parent_dir = %mo_pathname(&to_proj_dir.&separator.&scenario_name.&separator);
            %if (&q eq 2) %then %do;
               %let new_dir = solution;
            %end;
            %else %if (&q eq 3) %then %do;
               %let new_dir = saved;
            %end;
            %else %if (&q eq 4) %then %do; /* log subdirectory */
               %let new_dir = log;
            %end;
            %else %if (&q eq 5) %then %do; /* reports subdirectory */
               %let new_dir = reports;
            %end;

            %let to_dir = %mo_pathname(&to_proj_dir.&separator.&scenario_name.&separator.&new_dir);
            %let from_dir = %mo_pathname(&from_proj_dir.&separator.&scenario_name.&separator.&new_dir);
         %end;

         %create_dir(&parent_dir, &new_dir, &to_user_name, crtrc);
         %if (&crtrc ne 0) %then %do;
            %let &_rc = &crtrc;
            %goto ERROR_EXIT;
         %end;

         %copy_datasets(&from_dir, &to_dir, cpdsrc);
         %if (&cpdsrc ne 0) %then %do;
            %let &_rc = &cpdsrc;
            %goto ERROR_EXIT;
         %end;

         /* Copy individual non-dataset files */
         %if (&q eq 1) %then %do; /* scenario directory */
            %copy_file(&from_dir, &to_dir, mo_opt_config.sas, cprc);
            %if (&cprc ne 0) %then %do;
               %let &_rc = &cprc;
               %goto ERROR_EXIT;
            %end;

            %copy_file(&from_dir, &to_dir, mo_optimize.sas, cprc);
            %if (&cprc ne 0) %then %do;
               %let &_rc = &cprc;
               %goto ERROR_EXIT;
            %end;
         %end;
         %else %if (&q eq 4) %then %do; /* log subdirectory */
            %copy_file(&from_dir, &to_dir, optimize.log, cprc);
            %if (&cprc ne 0) %then %do;
               %let &_rc = &cprc;
               %goto ERROR_EXIT;
            %end;
         %end;
      %end;
   %end;


   %*--------------------------------------------------------------------------;
   %* Now that all checks have passed and all directories and files for the    ;
   %* FROM project have been replicated in a new TO project directory,         ;
   %* replace the FROM user_id with the TO user_id in MO datastore tables.     ;
   %*--------------------------------------------------------------------------;
   proc contents data=mometa._all_ out=work.tableNames_tmp (keep=memname name) noprint nodetails;
   run;
   %if (&syserr gt 4) %then %do;
      %let &_rc = 100;
      %put %mo_error Error encountered using proc contents to get the list of tables in MOMETA (syserr=&syserr).;
      %goto ERROR_EXIT;
   %end;

   %local table_list;
   %let table_list = ;
   proc sql noprint;
      select distinct memname into :table_list separated by ' '
      from work.tableNames_tmp
      where ((name = "user_id") and
             (%upcase(memname) not in ("USERS", "USERS_SHARE", "USERS_NOSHARE",
                                       "USERS_PREFERENCES")) and
             (exists (select distinct memname 
                      from work.tableNames_tmp
                      where (name = "project_id"))));
   quit;
   %if (&sqlrc gt 4) %then %do;
      %put %mo_error SQL error occurred retrieving data from WORK.TABLENAMES_TMP (sqlrc=&sqlrc).;
      %let &_rc = 107;
      %goto ERROR_EXIT;
   %end;

   %let table_list = %lowcase(&table_list);
   %let num_tables = %mo_word_count(&table_list);
   %do i = 1 %to &num_tables;
      %let current_table = %scan(&table_list, &i);

      %update_user_id(&from_user_id, &to_user_id, &project_id, &current_table, uuidrc);
      %if (&uuidrc ne 0) %then %do;
         %let &_rc = &uuidrc;
         %goto ERROR_EXIT;
      %end;
   %end;


   %*---------------------------------;
   %* Exit processing                 ;
   %*---------------------------------;
   %* Successful completion;
   %mo_unlock_dataset(data=mometa.uid)
   %put;
   %put %mo_note Successfully reassigned project ID=&project_id from user=&from_user_id to user=&to_user_id.;
   %goto EXIT;


   %ERROR_EXIT:
   %mo_unlock_dataset(data=mometa.uid)
   %put;
   %put %mo_error Error reassigning project ID=&project_id from user=&from_user_id to user=&to_user_id (rc=&&&_rc).;
   %if (&to_dir_already_existed = 0) %then %do;
      %if ((&to_proj_dir ne ) and (%sysfunc(fileexist(&to_proj_dir)))) %then %do;
         %mo_delete_file_item(item = &to_proj_dir, _rc = dirrc)
         %if (&dirrc ne 0) %then %do;
            %put %mo_error I/O error encountered deleting directory %upcase(&to_proj_dir) for TO user = %upcase(&to_user_name) (dirrc=&dirrc).;
         %end;
      %end;
   %end;


   %EXIT:

%mend mo_reassign_project_ownership;
