/************************************************************************/
/* Copyright (c) 2005-2011, SAS Institute Inc., Cary, NC, USA,          */
/*                       All Rights Reserved.                           */
/*                                                                      */
/* Name:          mo_rpt_compare_offer_table                            */
/*                                                                      */
/* Purpose:       create constraint comparison report,                  */
/*                feed the XML to GUI                                   */
/*                                                                      */
/* Input:         numScenarios                                          */
/*                scenarioIDs                                           */
/*                group_name                                            */
/*                                                                      */
/* Output:        &_rc = return code                                    */
/*                                                                      */
/*                report.xml                                            */
/*                                                                      */
/* Return Codes:                                                        */
/*                                                                      */
/*          101   SQL error retrieving data from userRoot.offer_summary */
/*          102   Error executing proc transpose                        */
/*          103   Unknown group_type: &group_type                       */
/*          104   Now rows with group_name "&group_name" were found     */
/*          105   Error generating the html data                        */
/*                                                                      */
/* HISTORY:                                                             */
/* 29Mar04 S0237007 Creation (liao)                                     */
/*                                                                      */
/* 4.4:                                                                 */
/* 10Jan06 S0337584 Right justify numbers in tabular reports (dwh)      */
/* 08Jun06 S0361339 Localize channel type (dwh)                         */
/*                                                                      */
/* 5.3:                                                                 */
/* 27Jun08 S0483854 Add secondary objective to report (KEC)             */
/*                                                                      */
/* 5.4:                                                                 */
/* 15Feb10 S0639443 Include Mandatory column in report when filter      */
/*         (group_name) is COMMUNICATION_CD (KEC)                       */
/* 27Aug10 S0639443 Include Priority column in report when filter       */
/*         (group_name) is COMMUNICATION_CD and scenario is             */
/*         prioritization-based (KEC)                                   */
/************************************************************************/
%macro mo_rpt_compare_offer_table(user_id, numScenarios, scenarioIDs, group_name, _rc);


   %local macname;
   %let macname=&sysmacroname;
   %let macname=%lowcase(&macname);

   %if %bquote(&_rc) eq %str() %then %do;
      %local rc;
      %let _rc = rc;
   %end;
   %else %global &_rc;
   %let &_rc = 0;

   %mo_constants


   %*------------------------------------;
   %* Access solution directory datasets ; 
   %* and the report directory and file  ; 
   %*------------------------------------;

   /* get user name and user root directory */
   proc sql noprint;
      select username into :username
      from mometa.users
      where user_id = &user_id;
   quit;
   %if (&sqlrc gt 4) %then %do;
      %put %mo_error SQL error retrieving metadata.;
      %let &_rc = 101;
      %goto ERROR_EXIT;
   %end;
   %let username=%trim(%nrbquote(%left(%nrbquote(&username))));

   libname userRoot "%mo_pathname(&moproj.&separator.&username.)";
   %if (&syslibrc ne 0) %then %do;
      %let &_rc = 102;
      %put %mo_error Error assigning userRoot libname.;
      %goto ERROR_EXIT;
   %end;

   filename morpt1 "%mo_pathname(&moproj.&separator.&username.&separator.report1.xml)";
   %if (&sysfilrc ne 0 and &sysfilrc ne 1) %then %do;
      %put %mo_error Error assigning MORPT1 filref (sysfilrc=&sysfilrc);
      %let &_rc = 125;
      %goto ERROR_EXIT;
   %end;


   %*----------------------------------------;
   %* The following is for all scenarios ... ;
   %*----------------------------------------;
   %let numScenarios = 0;
   %let inputDataIDs = ;
   %let projectIDs = ;
   %let projectNames = ;
   %let scenarioNames = ;
   %let personNames = ;
   %let scenarioID = %scan(&scenarioIDs , 1);
   %let i = 1;
   %do %while(&scenarioID ne);
      proc sql noprint;
         select b.input_data_id, b.project_id, c.username
         into :input_data_id, :project_id, :scen_username
         from mometa.scenarios as a, mometa.projects as b, mometa.users as c
         where a.scenario_id = &scenarioID
               and a.project_id = b.project_id
               and a.user_id = c.user_id;
      quit;
      %if (&sqlrc gt 4) %then %do;
         %put %mo_error SQL error retrieving metadata.;
         %let &_rc = 101;
         %goto ERROR_EXIT;
      %end;
      %let scen_username=%trim(%nrbquote(%left(%nrbquote(&scen_username))));

      proc sql noprint;
         select project_name, scenario_name, person into
               :project_name, :scenario_name, :person_name
         from userRoot.offer_summary
         where scenario_id = &scenarioId;
      quit;
      %if (&sqlrc gt 4) %then %do;
         %put %mo_error SQL error retrieving data from userRoot.offer_summary.;
         %let &_rc = 101;
         %goto ERROR_EXIT;
      %end;
      %let project_name=%trim(%nrbquote(%left(%nrbquote(&project_name))));
      %let scenario_name=%trim(%nrbquote(%left(%nrbquote(&scenario_name))));

      /* Set up libname per scenario solution directory */
      libname moOut&scenarioID "%mo_pathname(&moproj.&separator.&scen_username.&separator.&project_name.&separator.&scenario_name.&separator.solution)";
      %if (&syslibrc ne 0) %then %do;
         %let &_rc = 102;
         %put %mo_error Error assigning moOut&scenarioID libname.;
         %goto ERROR_EXIT;
      %end;

      %let numScenarios = %eval(&numScenarios + 1);
      %let inputDataIDs = &inputDataIDs.*&input_data_id;
      %let projectIDs = &projectIDs.*&project_id;
      %let projectNames = &projectNames.*&project_name;
      %let scenarioNames = &scenarioNames.*&scenario_name;
      %let personNames = &personNames.*&person_name;
      %let i = %eval(&i+1);
      %let scenarioID = %scan(&scenarioIDs, &i);
   %end;

   data work.temp_determine_labels;
      format project scenario person projscen $256.;
      %do i = 1 %to &numScenarios;
         project="%upcase(%scan(&projectNames, &i, %nrbquote(*)))";
         scenario="%upcase(%scan(&scenarioNames, &i, %nrbquote(*)))";
         person="%upcase(%scan(%bquote(&personNames), &i, %nrbquote(*)))";
         projscen = trim(left(project)) || '/' || trim(left(scenario));
         output;
      %end;
   run;
   %if (&syserr gt 4) %then %do;
      %let &_rc = 102;
      %put %mo_error Error creating work.temp_determine_labels (syserr=&syserr).;
      %goto ERROR_EXIT;
   %end;

   %let num_distinct_scenarios = 1;
   %let num_distinct_projscen = 1;
   proc sql noprint;
      select count(distinct scenario) into :num_distinct_scenarios
         from work.temp_determine_labels;
      select count(distinct projscen) into :num_distinct_projscen
         from work.temp_determine_labels;
   quit;
   %if (&sqlrc gt 4) %then %do;
      %put %mo_error SQL error determining number of distinct scenarios (sqlrc=&sqlrc).;
      %let &_rc = 107;
      %goto ERROR_EXIT;
   %end;

   %let label_level = 0;
   %if &num_distinct_scenarios < &numScenarios %then %let label_level = 1;
   %if &num_distinct_projscen < &numScenarios %then %let label_level = 2;
   %let label_level = %sysfunc(max(&label_level, &mo_compare_report_label_level));


   %*----------------------------------------;
   %* Determine the number of distinct sets  ;
   %* of input data that the scenarios to be ;
   %* compared are based on.                 ;
   %*----------------------------------------;
   data work.temp_inputDataIDs;
      %do i = 1 %to &numScenarios;
         inputDataID = %scan(&inputDataIDs, &i, %nrbquote(*));
         output;
      %end;
   run;
   %if (&syserr gt 4) %then %do;
      %let &_rc = 102;
      %put %mo_error Error creating work.temp_inputDataIDs (syserr=&syserr).;
      %goto ERROR_EXIT;
   %end;

   %let num_distinct_inputDataIDs = 1;
   proc sql noprint;
      create table work.temp_distinct_inputDataIDs as
      select distinct inputDataID
      from work.temp_inputDataIDs;

      select inputDataID into :distinct_inputDataIDs_comma separated by ', '
      from work.temp_distinct_inputDataIDs;

      select count(*) into :num_distinct_inputDataIDs
      from work.temp_distinct_inputDataIDs;
   quit;
   %if (&sqlrc gt 4) %then %do;
      %put %mo_error SQL error determining distinct input data IDs (sqlrc=&sqlrc).;
      %let &_rc = 107;
      %goto ERROR_EXIT;
   %end;


   %*-------------------------------------------------;
   %* Determine which scenarios are based on imported ;
   %* solutions and which are regular scenarios.      ;
   %*-------------------------------------------------;
   %local num_imported_solutions importedScenarioIDs normalScenarioID;
   %let num_imported_scenarios = 0;
   %let importedScenarioIDs = ;
   %let normalScenarioID = ;
   %do i=1 %to &numScenarios;
      %let scenarioID = %scan(&scenarioIDs, &i);
      %local scenario_type_&scenarioID;
      %let scenario_type_&scenarioID = 1;
  
      proc sql noprint;
         select status into :scenario_type_&scenarioID
            from mometa.scenarios
            where scenario_id = &scenarioID;
      quit;
      %if (&sqlrc gt 4) %then %do;
         %put %mo_error SQL error retrieving metadata.;
         %let &_rc = 107;
         %goto ERROR_EXIT;
      %end;

      %if (&&scenario_type_&scenarioID = &_IMPORTED_) %then %do;
         %let num_imported_scenarios = %eval(&num_imported_scenarios + 1);
         %let importedScenarioIDs = &importedScenarioIDs &scenarioID;
      %end;
      %else %do;
         %let normalScenarioID = &scenarioID;
      %end;
   %end;


   %*--------------------------------------------------------------;
   %* Determine if one or more scenarios are prioritization-based  ;
   %* or if all are prioritization-based. This will help determine ;
   %* if only offer count is included in the report.               ;
   %*--------------------------------------------------------------;
   %local anyPrioritizationBased allPrioritizationBased;
   %let anyPrioritizationBased = 0;
   %let allPrioritizationBased = 1;

   %let i = 1;
   %let scenarioID = %scan(&scenarioIDs, &i);
   %let inputDataID = %scan(&inputDataIDs, &i);
   %do %while(&scenarioID ne %str());
      %global prioritizationBased_&scenarioID;
      %let prioritizationBased_&scenarioID = 0;
      %mo_check_if_priority_based(item_type             = SCENARIO
                                 ,item_id               = &scenarioID
                                 ,_prioritization_based = prioritizationBased_&scenarioID
                                 ,_rc                   = cipbrc);
      %if (&cipbrc ne 0) %then %do;
         %let &_rc = &cipbrc;
         %goto ERROR_EXIT;
      %end;

      /* Set anyPrioritizationBased if any one is found to be so. */
      %let anyPrioritizationBased = 
              %eval(&anyPrioritizationBased or &&prioritizationBased_&scenarioID);

      /* Reset allPrioritizationBased if any one is found not to be so. */
      %let allPrioritizationBased = 
              %eval(&allPrioritizationBased and &&prioritizationBased_&scenarioID);

      %local priority_column_&scenarioID;
      %let priority_column_&scenarioID = ;
      %if (&&prioritizationBased_&scenarioID eq 1) %then %do;
         proc sql noprint;
            select priority_column_name into :priority_column_&scenarioID
            from mometa.input_data
            where (input_data_id = &inputDataID);
         quit;
         %if (&sqlrc gt 4) %then %do;
            %put %mo_error SQL error reading PRIORITY_COLUMN_NAME from MOMETA.INPUT_DATA table for input data ID &inputDataID (sqlrc=&sqlrc).;
            %let &_rc = 107;
            %goto ERROR_EXIT;
         %end;
         %let priority_column_&scenarioID = %trim(%left(&&priority_column_&scenarioID));
      %end;

      %let i = %eval(&i+1);
      %let scenarioID = %scan(&scenarioIDs, &i);
      %let inputDataID = %scan(&inputDataIDs, &i);
   %end;


   %*---------------------------------------------------;
   %* Determine if one or more scenarios use input data ;
   %* that has mandatory inclusions.                    ;
   %*---------------------------------------------------;
   %local any_active_mandatory;
   %let any_active_mandatory = 0;

   %let i = 1;
   %let scenarioID = %scan(&scenarioIDs, &i);
   %let inputDataID = %scan(&inputDataIDs, &i);
   %do %while(&scenarioID ne %str());

      %local any_active_mandatory_&scenarioID;
      %let any_active_mandatory_&scenarioID = 0;
      proc sql noprint;
         select count(*) into :any_active_mandatory_&scenarioID
         from mometa.communications
         where (input_data_id = &inputDataID) & (mandatory = 1);
      quit;
      %if (&sqlrc gt 4) %then %do;
         %put %mo_error SQL error reading MANDATORY column in MOMETA.COMMUNICATIONS table for input data ID &inputDataID (sqlrc=&sqlrc).;
         %let &_rc = 107;
         %goto ERROR_EXIT;
      %end;

      %let any_active_mandatory = %eval(&any_active_mandatory + &&any_active_mandatory_&scenarioID); 

      %let i = %eval(&i+1);
      %let scenarioID = %scan(&scenarioIDs, &i);
      %let inputDataID = %scan(&inputDataIDs, &i);
   %end;


   %*-----------------------------------------------------;
   %* For non-prioritization-based scenarios, determine   ;
   %* whether one or more of the scenarios to be compared ;
   %* had ROI calculation activated and whether one or    ;
   %* more of scenarios to be compared had the Secondary  ;
   %* Objective activated.                                ; 
   %*-----------------------------------------------------;
   %local any_active_roi any_active_second_obj;
   %let any_active_second_obj = 0;
   %let any_active_roi = 0;
   %if (&allPrioritizationBased ne 1) %then %do;
      %let i = 1;
      %let scenarioID = %scan(&scenarioIDs, &i);
      %do %while(&scenarioID ne %str());
   
         %if (&&scenario_type_&scenarioID ne &_IMPORTED_) %then %do;
            proc sql noprint;
               select active_roi, active_secondary_obj, 
                      secondary_obj_tolerance
               into :active_roi_&scenarioID, :active_second_obj_&scenarioID, 
                    :second_obj_tolerance_&scenarioID
               from mometa.scenario_advanced_options
               where scenario_id = &scenarioID;
            quit;
            %if (&sqlrc gt 4) %then %do;
               %put %mo_error SQL error obtaining metadata (sqlrc=&sqlrc).;
               %let &_rc = 107;
               %goto ERROR_EXIT;
            %end;
         %end;
         %else %do; 
            /* global variables activeROI, activeSecondaryObjective, */
            /* secondObjTolerance were set by mo_initialize, which   */
            /* was called by mo_rpt_compare_with_imported, which was */
            /* called by mo_rpt_compare, which called this macro.    */
            %let active_roi_&scenarioID = &activeROI; 
            %let active_second_obj_&scenarioID = &activeSecondaryObjective;
            %let second_obj_tolerance_&scenarioID = .;
         %end;

         %let any_active_roi = %eval(&any_active_roi + &&active_roi_&scenarioID);
         %let any_active_second_obj = %eval(&any_active_second_obj + &&active_second_obj_&scenarioID);
   
         %let i = %eval(&i + 1);
         %let scenarioID = %scan(&scenarioIDs, &i);
      %end;
   %end;


   %*----------------------------------------;
   %* Prepare the working dataset from which ;
   %* the report will be generated.          ;
   %*----------------------------------------;
   %local oneOrMoreBlankComponents;
   %let oneOrMoreBlankComponents = 0;
   data work.offer_summary;
      set userRoot.offer_summary;
      where upcase(group_name) = "%upcase(&group_name)";

      if (component = "") then do;
         component = "[BLANK]";
         call symput("oneOrMoreBlankComponents", put(1,1.));
      end; 

      group_name = upcase(group_name);

      /* If the offer_summary datasets were created prior to introduction of */
      /* the Dual Objectives feature, there will be no secondary_obj_value   */
      /* column. In that case, add a secondary_obj_value column set to       */
      /* missing values.                                                     */
      if (secondary_obj_value = .) then secondary_obj_value = .;

      /* If the scenario is prioritization-based, make sure */
      /* non-offer-related columns are set to missing.      */
      %do sn = 1 %to &numScenarios;
         %let scenarioID = %scan(&scenarioIDs, &sn);
         if (scenario_id eq &scenarioID) then do; 
            %if (&&prioritizationBased_&scenarioID eq 1) %then %do;
               obj_value = .;
               secondary_obj_value = .;
               roi_profit = .;
               roi_cost = .;
               roi_value = .;
            %end;
         end;
      %end;
   run;
   %if (&syserr gt 4) %then %do;
      %let &_rc = 104;
      %put %mo_error No rows with group_name "&group_name" were found.;
      %goto ERROR_EXIT;
   %end;


   %*-----------------------------------------------;
   %* Screen for all missing values for objective,  ;
   %* secondary objective, and ROI because report   ;
   %* columns for these should be suppressed if     ;
   %* there are no values.                          ;
   %*-----------------------------------------------;
   %let any_roi_vals = 0;
   %let any_objective_vals = 0;
   %let any_second_obj_vals = 0;
   %do i = 1 %to &numScenarios;
      %let scenarioID = %scan(&scenarioIDs, &i);

      %let has_roi = 0;
      %let has_objective = 0;
      %let has_second_obj = 0;
      proc sql noprint;
         select count(roi_value), count(obj_value), count(secondary_obj_value)
            into :has_roi, :has_objective, :has_second_obj 
            from work.offer_summary
            where scenario_id = &scenarioID;
      quit;
      %if (&sqlrc gt 4) %then %do;
         %put %mo_error SQL error retrieving metadata.;
         %let &_rc = 107;
         %goto ERROR_EXIT;
      %end;
      
      %if &has_roi > 0 %then %let any_roi_vals = 1;
      %if &has_objective > 0 %then %let any_objective_vals = 1;
      %if &has_second_obj > 0 %then %let any_second_obj_vals = 1;
   %end;


   %*-------------------------------------;
   %* Continue with working dataset prep. ;
   %*-------------------------------------;
   proc sort data=work.offer_summary;
      by group_name component scenario_id;
   run;
   %if (&syserr gt 4) %then %do;
      %let &_rc = 102;
      %put %mo_error Error executing proc sort.;
      %goto ERROR_EXIT;
   %end;

   proc transpose data=work.offer_summary out=work.temp_1;
      by group_name component scenario_id project_name scenario_name;
      var num_eligible offers obj_value secondary_obj_value roi_profit roi_cost roi_value;
   run;
   %if (&syserr gt 4) %then %do;
      %let &_rc = 102;
      %put %mo_error Error executing proc transpose.;
      %goto ERROR_EXIT;
   %end;

   data work.temp;
      length column_name $64.;
      set work.temp_1;
      column_name = trim(_name_)||"_"||strip(put(scenario_id, 16.));
      drop _name_;
   run;
   %if (&syserr gt 4) %then %do;
      %let &_rc = 102;
      %put %mo_error Error executing data step.;
      %goto ERROR_EXIT;
   %end;

   proc sort data=work.temp;
      by group_name component;
   run;
   %if (&syserr gt 4) %then %do;
      %let &_rc = 102;
      %put %mo_error Error executing proc sort.;
      %goto ERROR_EXIT;
   %end;

   proc transpose data=work.temp out=work.offer_compare_0(drop=_name_);
      by group_name component;
      var col1;
      id column_name;
   run;
   %if (&syserr gt 4) %then %do;
      %let &_rc = 102;
      %put %mo_error Error executing proc transpose.;
      %goto ERROR_EXIT;
   %end;


   %*--------------------------------------------;
   %* Determine if selected group_name values    ;
   %* are unique over the set of communications  ;
   %* for each scenario being compared.          ;
   %*--------------------------------------------;
   %let anyScenHasDistinctValuesPerComm = 0;
   %let allScenHaveDistinctValuesPerComm = 1;

   %local numCommunications numDistinctGroupNameValues;
   %let i = 1;
   %let scenarioID = %scan(&scenarioIDs, &i);
   %let inputDataID = %scan(&inputDataIDs, &i);
   %do %while(&scenarioID ne %str());
      %let numCommunications = 0;
      %let numDistinctGroupNameValues = 0;
      proc sql noprint;
         select count(*) into :numCommunications
         from moOut&scenarioID..offer_summary_nway;

         select count(*) into :numDistinctGroupNameValues
         from userRoot.offer_summary
         where ((upcase(group_name) = "%upcase(&group_name)") and
                (scenario_id = &scenarioID));
      quit;
      %if (&sqlrc gt 4) %then %do;
         %put %mo_error SQL error counting rows in MOOUT&scenarioID..OFFER_SUMMARY_NWAY and WORK.OFFER_COMPARE_0 for scenario ID &scenarioID (sqlrc=&sqlrc).;
         %let &_rc = 107;
         %goto ERROR_EXIT;
      %end;

      %local distinctValuesPerComm_&scenarioID;
      %let distinctValuesPerComm_&scenarioID = 0;
      %if (&numDistinctGroupNameValues eq &numCommunications) %then %do;
         %let distinctValuesPerComm_&scenarioID = 1;
      %end;

      /* Reset anyScenHasDistinctValuesPerComm if one found not to be so. */
      %let anyScenHasDistinctValuesPerComm = 
              %eval(&anyScenHasDistinctValuesPerComm or &&distinctValuesPerComm_&scenarioID);

      /* Reset allScenHaveDistinctValuesPerComm if one found not to be so. */
      %let allScenHaveDistinctValuesPerComm = 
              %eval(&allScenHaveDistinctValuesPerComm and &&distinctValuesPerComm_&scenarioID);

      %let i = %eval(&i+1);
      %let scenarioID = %scan(&scenarioIDs, &i);
   %end;


   %*-------------------------------------------------------------------;
   %* If there are any mandatory communications in the input data, then ;
   %* determine whether it is possible to display the MANDATORY column  ;
   %* in the report. It will be displayed only if the values for the    ;
   %* selected group_name are unique over the set of communications.    ;
   %* Analogous to mandatory inclusions, determine if the scenario is   ;
   %* prioritization-based.  If so, then determine whether is is        ;
   %* possible to display the PRIORITY column in the report. It will    ;
   %* be displayed only if the values for the selected group_name       ;
   %* are unique over the set of communications.                        ;
   %*-------------------------------------------------------------------;
   %local any_priority_vals;
   %let any_priority_vals = 0;

   %if (((&any_active_mandatory gt 0) or (&anyPrioritizationBased eq 1)) and
        (&anyScenHasDistinctValuesPerComm eq 1)) %then %do;
      %local cnt prior_cnt scen_id id_id;
      %do cnt=1 %to &numScenarios;
         %let prior_cnt = %eval(&cnt - 1);
         %let id_id = %scan(&inputDataIDs, &cnt);
         %let proj_id = %scan(&projectIDs, &cnt);
         %let scen_id = %scan(&scenarioIDs, &cnt);

         %if (&&distinctValuesPerComm_&scen_id eq 0) %then %do;
            data work.offer_compare_&cnt;
               set work.offer_compare_&prior_cnt;
               %if (&any_active_mandatory gt 0) %then %do;
                  offer_is_mandatory_&scen_id = " ";
               %end;
               %if (&anyPrioritizationBased eq 1) %then %do;
                  priority_&scen_id = .;
               %end;
            run;
         %end;
         %else %do;

            /* Reconstruct the Communications table as it pertains to this  */
            /* scenario from the piece-parts that were persisted in the MO  */
            /* Datastore. Then join against this working table to determine */
            /* the Mandatory and Priority column settings.                  */

            /* First build the intermediate working tables from which it    */
            /* will be built.                                               */
            proc sql noprint;
               create table work.comm_character_attributes as
                  select * from mometa.comm_character_attributes 
                  where (input_data_id = &id_id);

               create table comm_character_attribute_values as
                  select * from mometa.comm_character_attribute_values 
                  where (input_data_id = &id_id);
   
               create table comm_character_distinct_values as
                  select * from mometa.comm_character_distinct_values
                  where (input_data_id = &id_id);
   
               create table comm_numeric_attributes as
                  select * from mometa.comm_numeric_attributes
                  where (input_data_id = &id_id);

               %if &&scenario_type_&scen_id ne &_IMPORTED_ %then %do;
                  create table comm_numeric_attribute_values as
                     select * from mometa.comm_numeric_attribute_values
                     where (scenario_id = &scen_id);
               %end;
               %else %do;
                  create table comm_numeric_attribute_values as
                     select &user_id as user_id, &proj_id as project_id,
                            &scen_id as scenario_id,
                            * from mometa.comm_numeric_attribute_defaults
                     where (input_data_id = &id_id);
               %end;
            quit;
            %if (&sqlrc gt 4) %then %do;
               %put %mo_error SQL error creating temp tables from which to build WORK.COMMUNICATIONS via call to mo_gui_communication (sqlrc=&sqlrc).;
               %let &_rc = 107;
               %goto ERROR_EXIT;
            %end;

            /* Build work.communications by calling mo_gui_communication.   */
            /* mo_gui_communication assumes the above tables exist in WORK. */
            %let modebug_optengine_log = 0;
            %mo_gui_communication;

            proc sql noprint;
               select communication_code into :comm_codes_&id_id separated by '", "'
               from  mometa.communications
               where (input_data_id = &id_id);

               create table work.offer_compare_&cnt as
               select distinct a.* 
                  %if (&any_active_mandatory gt 0) %then %do;
                     , case
                           when ((b.communication_cd in ("&&comm_codes_&id_id")) &
                                 (b.mandatory = 1)) then upcase("&mo_yes_value")
                           when ((b.communication_cd in ("&&comm_codes_&id_id")) &
                                 (b.mandatory = 0)) then upcase("&mo_no_value")
                           else " "
                      end as offer_is_mandatory_&scen_id
                   %end;
                   %if (&anyPrioritizationBased eq 1) %then %do;
                      ,
                      %if (&&prioritizationBased_&scen_id eq 1) %then %do;
                         b.&&priority_column_&scen_id as priority_&scen_id
                      %end;
                      %else %do;
                         . as priority_&scen_id
                      %end;
                   %end;
               from work.offer_compare_&prior_cnt as a left join work.communications as b
               on (a.component = b.&group_name)
               order by b.&group_name;
            quit;
            %if (&sqlrc gt 4) %then %do;
               %put %mo_error SQL error creating offer compare &cnt by communication codes (input_data_id=&id_id, sqlrc=&sqlrc).;
               %let &_rc = 104;
               %goto ERROR_EXIT;
            %end;

            %if (&&prioritizationBased_&scen_id eq 1) %then %do;
               %let any_priority_vals_&cnt = 0;
               proc sql noprint;
                  select count(&&priority_column_&scen_id) into :any_priority_vals_&cnt
                  from work.communications;
               quit;
               %if (&sqlrc gt 4) %then %do;
                  %put %mo_error SQL error retrieving metadata.;
                  %let &_rc = 107;
                  %goto ERROR_EXIT;
               %end;
               %let any_priority_vals = %eval(&any_priority_vals + &&any_priority_vals_&cnt); 
            %end;
         %end;
      %end;

      /* Finalize working dataset. */
      proc sql noprint;
         create table work.offer_compare_0 as
         select * from work.offer_compare_&numScenarios;
      quit;
      %if (&sqlrc gt 4) %then %do;
         %put %mo_error SQL error creating offer compare (sqlrc=&sqlrc).;
         %let &_rc = 104;
         %goto ERROR_EXIT;
      %end;

      /* If all scenarios being compared are based on the same input data, */
      /* then collapse the offer_is_mandatory and/or priority columns,     */
      /* when present, down to a single column each since the per-scenario */
      /* values are identical.                                             */
      %if (&num_distinct_inputDataIDs eq 1) %then %do;
         %local drop_offer_is_mandatory_col_list drop_priority_col_list;
         %let dropOfferIsMandatoryColList = ;
         %let dropPriorityColList =;

         %do cnt=2 %to &numScenarios;
            %let scen_id = %scan(&scenarioIDs, &cnt);
            %if (&any_active_mandatory gt 0) %then %do;
               %let dropOfferIsMandatoryColList = &dropOfferIsMandatoryColList offer_is_mandatory_&scen_id;
            %end;
            %if (&anyPrioritizationBased eq 1) %then %do;
               %let dropPriorityColList = &dropPriorityColList priority_&scen_id;
            %end;
         %end;

         %let scen_id = %scan(&scenarioIDs, 1);
         data work.offer_compare_0;
            set work.offer_compare_0(rename=(
                                          %if (&any_active_mandatory gt 0) %then %do;
                                             offer_is_mandatory_&scen_id=offer_is_mandatory
                                          %end;
                                          %if (&anyPrioritizationBased eq 1) %then %do;
                                             priority_&scen_id=priority
                                          %end;
                                            )

                                     drop=
                                          %if (&any_active_mandatory gt 0) %then %do;
                                             &dropOfferIsMandatoryColList
                                          %end;
                                          %if (&anyPrioritizationBased eq 1) %then %do;
                                             &dropPriorityColList
                                          %end;
                                            );
         run;
      %end;
   %end;


   %*---------------------------------------------------;
   %* Else, if the selected group_name is time_periods, ;
   %* order output per specified order in datastore.    ;
   %*---------------------------------------------------;
   %if ("%upcase(&group_name)" = "TIME_PERIOD") %then %do;
      proc sql noprint;
         create table work.offer_compare_1 as
         select a.*
         from work.offer_compare_0 as a left join mometa.time_periods as b
         on ((b.input_data_id in (&distinct_inputDataIDs_comma)) &
             (upcase(a.component) = upcase(b.time_period)))
         order by b.input_data_id, b.order;

         select count(*), count(distinct component) into :num_components, :num_distinct_components
         from work.offer_compare_1;

         select distinct component into :distinct_components separated by '" "'
         from work.offer_compare_1;
      quit;
      %if (&sqlrc gt 4) %then %do;
         %put %mo_error SQL error creating offer compare by time periods (sqlrc=&sqlrc).;
         %let &_rc = 107;
         %goto ERROR_EXIT;
      %end;

      %if (&num_distinct_components lt &num_components) %then %do;
         data work.offer_compare (drop=i);
            array distinct_components{&num_distinct_components} $ _temporary_ ("&distinct_components");
            array found_components{&num_distinct_components} _temporary_;

            set work.offer_compare_1;

            do i = 1 to &num_distinct_components;
               put i= component= distinct_components{i}= found_components{i}=;
               if ((strip(component) = strip(distinct_components{i})) & 
                   (found_components{i} = .)) then do;
                  found_components{i} = 1;
                  output;
                  leave;
               end;
            end;
         run;
      %end; 
      %else %do;
         data work.offer_compare;
            set work.offer_compare_1;
         run;
      %end;
   %end;


   %*-------------------------------------------------;
   %* Otherwise, for selected group_names other than  ;
   %* COMMUNICATION_CD or TIME_PERIOD, no additional  ;
   %* massaging of the working dataset is needed.     ;
   %*-------------------------------------------------;
   %else %do;
      proc sql noprint;
         create table work.offer_compare as
         select * from work.offer_compare_0;
      quit;
      %if (&sqlrc gt 4) %then %do;
         %put %mo_error SQL error creating offer compare (sqlrc=&sqlrc).;
         %let &_rc = 104;
         %goto ERROR_EXIT;
      %end;
   %end;


   %*--------------------------------------------------;
   %* Determine if the set of scenarios to be compared ;
   %* includes one or more imported solutions and set  ;
   %* up formats for the objective values accordingly. ;
   %*--------------------------------------------------;

   /* In the presence of a mixture of imported scenarios and regular   */
   /* optimized scenarios, adjust objective type, operator, and format */
   /* to be that of a regular optimized scenario (the last one found). */
   %do i=1 %to &numScenarios;
      %let scenarioID = %scan(&scenarioIDs, &i);
      %local objFormat_&scenarioID secObjFormat_&scenarioID;
      %let objFormat_&scenarioID = ;
      %let secObjFormat_&scenarioID = ;
   %end;

   %if ((&any_objective_vals = 1) or (&any_second_obj_vals = 1)) %then %do;
      %do i=1 %to &numScenarios;
         %let scenarioID = %scan(&scenarioIDs, &i);

         %if &&scenario_type_&scenarioID ne &_IMPORTED_ %then %do;
            %if (&any_objective_vals = 1) %then %do;
               %let attribute_role = ;
               %mo_rpt_get_attribute_role(mometa.scenarios
                                         ,&scenarioID
                                         ,objective_attribute_name
                                         ,attribute_role);
               %let objFormat_&scenarioID =;
               %if ((&attribute_role = &_ATTRIBUTE_ROLE_COST_) or 
                    (&attribute_role = &_ATTRIBUTE_ROLE_REVENUE_) or 
                    (&attribute_role = &_ATTRIBUTE_ROLE_PROFIT_)) %then %do;
                  %let objFormat_&scenarioID = currency=True;
               %end;
            %end;

            %if (&any_second_obj_vals = 1) %then %do;
               %let second_obj_attribute_role = ;
               %mo_rpt_get_attribute_role(mometa.scenario_advanced_options
                                         ,&scenarioID
                                         ,secondary_obj_attribute_name
                                         ,second_obj_attribute_role);
               %let secObjFormat_&scenarioID =;
               %if ((&second_obj_attribute_role = &_ATTRIBUTE_ROLE_COST_) or 
                    (&second_obj_attribute_role = &_ATTRIBUTE_ROLE_REVENUE_) or 
                    (&second_obj_attribute_role = &_ATTRIBUTE_ROLE_PROFIT_)) %then %do;
                  %let secObjFormat_&scenarioID = currency=True;
               %end;
            %end;
         %end;
      %end;

      %do k = 1 %to &num_imported_scenarios;
         %let importedScenario = %scan(&importedScenarioIDs, &k);
         %if (&any_objective_vals = 1) %then %do;
            %let objFormat_&importedScenario = &&objFormat_&normalScenarioID;
         %end;
         %if (&any_second_obj_vals = 1) %then %do;
            %let secObjFormat_&importedScenario = &&secObjFormat_&normalScenarioID;
         %end;
      %end;
   %end;
   
   /* Set the num eligible to missing for the imported scenarios */
   %if &num_imported_scenarios > 0 %then %do;
      data work.offer_compare;
         set work.offer_compare;
         %do k = 1 %to &num_imported_scenarios;
            %let importedScenario = %scan(&importedScenarioIDs, &k);
            num_eligible_&importedScenario = .;
         %end;
      run;
   %end;


   %*--------------------------------------------;
   %* Create XML file from which report will be  ;
   %* surfaced using working dataset from above. ;
   %*--------------------------------------------;
   
   /* Insert %nrstr(&nbsp;) into column names in place of spaces */
   %let offer_summary_eligibles  = %mo_nbsp_string(&offer_summary_eligibles);
   %let offer_summary_offers     = %mo_nbsp_string(&offer_summary_offers);
   %let offer_summary_objective  = %mo_nbsp_string(&offer_summary_objective);
   %let offer_summary_second_obj = %mo_nbsp_string(&offer_summary_second_obj);
   %let offer_summary_roi        = %mo_nbsp_string(&offer_summary_roi);
   %let offer_summary_mandatory  = %mo_nbsp_string(&offer_summary_mandatory);
   %let offer_summary_priority   = %mo_nbsp_string(&offer_summary_priority);

   data _null_;
      length 
         %do i = 1 %to &numScenarios;
            %let scenarioID = %scan(&scenarioIDs,&i);
            typeDef_&scenarioID $32.
         %end;
      ;
      length name $256;

      /* Suppress uninitialized variable warnings when no primary objective. */
      %do i = 1 %to &numScenarios;
         %let scenarioID = %scan(&scenarioIDs,&i);
         typeDef_&scenarioID = "";
      %end;
      name = "";

      file morpt1 encoding='utf-8';
      set work.offer_compare end=eof;
      if _n_ = 1 then do;
         put '<table title="' "&offer_comparison_title" '">';
         put '<tr>';
         put '   <th type="String">' "%upcase(&group_name)" '</th>';

         %do i=1 %to &numScenarios;
            %let projectName = %scan(&projectNames, &i, %nrbquote(*));
            %let scenarioName = %scan(&scenarioNames, &i, %nrbquote(*));
            %let personName = %scan(%bquote(&personNames), &i, %nrbquote(*));
            %if &label_level = 0 %then %let projScen = (&scenarioName);
            %else %if &label_level = 1 %then %let projScen = (&projectName / &scenarioName);
            %else %let projScen = (&personName / &projectName / &scenarioName);
            %let projScen = %qsysfunc(tranwrd(%bquote(&projScen),%str( ),%nrstr(&nbsp;)));

            put '   <th type="Integer">' "%upcase(&offer_summary_eligibles)%nrstr(&nbsp;)%upcase(&projScen)" '</th>';
         %end;
         %do i=1 %to &numScenarios;
            %let projectName = %scan(&projectNames, &i, %nrbquote(*));
            %let scenarioName = %scan(&scenarioNames, &i, %nrbquote(*));
            %let personName = %scan(%bquote(&personNames), &i, %nrbquote(*));
            %if &label_level = 0 %then %let projScen = (&scenarioName);
            %else %if &label_level = 1 %then %let projScen = (&projectName / &scenarioName);
            %else %let projScen = (&personName / &projectName / &scenarioName);
            %let projScen = %qsysfunc(tranwrd(%bquote(&projScen),%str( ),%nrstr(&nbsp;)));

            put '   <th type="Integer">' "%upcase(&offer_summary_offers)%nrstr(&nbsp;)%upcase(&projScen)" '</th>';
         %end;
         %if &any_objective_vals %then %do;
            %do i=1 %to &numScenarios;
               %let scenarioID = %scan(&scenarioIDs, &i);
               %let projectName = %scan(&projectNames, &i, %nrbquote(*));
               %let scenarioName = %scan(&scenarioNames, &i, %nrbquote(*));
               %let personName = %scan(%bquote(&personNames), &i, %nrbquote(*));
               %if &label_level = 0 %then %let projScen = (&scenarioName);
               %else %if &label_level = 1 %then %let projScen = (&projectName / &scenarioName);
               %else %let projScen = (&personName / &projectName / &scenarioName);
               %let projScen = %qsysfunc(tranwrd(%bquote(&projScen),%str( ),%nrstr(&nbsp;)));

               typeDef_&scenarioID = 'type="Double"';
               typeDef_&scenarioID = strip(typeDef_&scenarioID)||" &&objFormat_&scenarioID";

               %if (&any_active_second_obj gt 0) %then %do;
                  name = "%upcase(&offer_summary_prim_obj)%nrstr(&nbsp;)%upcase(&projScen)";
               %end;
               %else %do;
                  name = "%upcase(&offer_summary_objective)%nrstr(&nbsp;)%upcase(&projScen)";
               %end;

               put '   <th ' typeDef_&scenarioID '>' name '</th>';
            %end;
         %end;
         %if ((&any_active_second_obj gt 0) and (&any_second_obj_vals)) %then %do;
            %do i=1 %to &numScenarios;
               %let scenarioID = %scan(&scenarioIDs, &i);
               %let projectName = %scan(&projectNames, &i, %nrbquote(*));
               %let scenarioName = %scan(&scenarioNames, &i, %nrbquote(*));
               %let personName = %scan(%bquote(&personNames), &i, %nrbquote(*));
               %if &label_level = 0 %then %let projScen = (&scenarioName);
               %else %if &label_level = 1 %then %let projScen = (&projectName / &scenarioName);
               %else %let projScen = (&personName / &projectName / &scenarioName);
               %let projScen = %qsysfunc(tranwrd(%bquote(&projScen),%str( ),%nrstr(&nbsp;)));
               %let tolVal = (%trim(%left(&&second_obj_tolerance_&scenarioID))%nrbquote(%) tolerance);

               typeDef_&scenarioID = 'type="Double"';
               typeDef_&scenarioID = strip(typeDef_&scenarioID)||" &&secObjFormat_&scenarioID";

               name = "%upcase(&offer_summary_second_obj)%nrstr(&nbsp;)%upcase(&projScen)";
               put '   <th ' typeDef_&scenarioID '>' name '</th>';
            %end;
         %end;
         %if ((&any_active_roi gt 0) and (&any_roi_vals)) %then %do;
            %do i=1 %to &numScenarios;
               %let projectName = %scan(&projectNames, &i, %nrbquote(*));
               %let scenarioName = %scan(&scenarioNames, &i, %nrbquote(*));
               %let personName = %scan(%bquote(&personNames), &i, %nrbquote(*));
               %if &label_level = 0 %then %let projScen = (&scenarioName);
               %else %if &label_level = 1 %then %let projScen = (&projectName / &scenarioName);
               %else %let projScen = (&personName / &projectName / &scenarioName);
               %let projScen = %qsysfunc(tranwrd(%bquote(&projScen),%str( ),%nrstr(&nbsp;)));

               put '   <th type="Double" percent="True">' "%upcase(&offer_summary_roi)%nrstr(&nbsp;)%upcase(&projScen)" '</th>';
            %end;
         %end;
         %if ((&anyScenHasDistinctValuesPerComm eq 1) and
              (&any_active_mandatory ne 0)) %then %do;
            %if (&num_distinct_inputDataIDs gt 1) %then %do;
               %do i=1 %to &numScenarios;
                  %let projectName = %scan(&projectNames, &i, %nrbquote(*));
                  %let scenarioName = %scan(&scenarioNames, &i, %nrbquote(*));
                  %let personName = %scan(%bquote(&personNames), &i, %nrbquote(*));
                  %if &label_level = 0 %then %let projScen = (&scenarioName);
                  %else %if &label_level = 1 %then %let projScen = (&projectName / &scenarioName);
                  %else %let projScen = (&personName / &projectName / &scenarioName);
                  %let projScen = %qsysfunc(tranwrd(%bquote(&projScen),%str( ),%nrstr(&nbsp;)));

                  put '   <th type="String" align="right">' "%upcase(&offer_summary_mandatory)%nrstr(&nbsp;)%upcase(&projScen)" '</th>';
               %end;
            %end;
            %else %do;
               put '   <th type="String" align="right">' "%upcase(&offer_summary_mandatory)" '</th>';
            %end;
         %end;
         %if ((&anyPrioritizationBased eq 1) and 
              (&any_priority_vals gt 0)) %then %do;
            %if (&num_distinct_inputDataIDs gt 1) %then %do;
               %do i=1 %to &numScenarios;
                  %let projectName = %scan(&projectNames, &i, %nrbquote(*));
                  %let scenarioName = %scan(&scenarioNames, &i, %nrbquote(*));
                  %let personName = %scan(%bquote(&personNames), &i, %nrbquote(*));
                  %if &label_level = 0 %then %let projScen = (&scenarioName);
                  %else %if &label_level = 1 %then %let projScen = (&projectName / &scenarioName);
                  %else %let projScen = (&personName / &projectName / &scenarioName);
                  %let projScen = %qsysfunc(tranwrd(%bquote(&projScen),%str( ),%nrstr(&nbsp;)));

                  put '   <th type="Integer" align="right">' "%upcase(&offer_summary_priority)%nrstr(&nbsp;)%upcase(&projScen)" '</th>';
               %end;
            %end;
            %else %do;
               put '   <th type="Integer" align="right">' "%upcase(&offer_summary_priority)" '</th>';
            %end;
         %end;
         put '</tr>';
      end;
      put '<tr>';
      put '   <td>' component '</td>';
      %do i=1 %to &numScenarios;
         %let scenarioID = %scan(&scenarioIDs, &i);
         %let num_eligible = num_eligible_&scenarioID;
         put '   <td>' &num_eligible '</td>';
      %end;
      %do i=1 %to &numScenarios;
         %let scenarioID = %scan(&scenarioIDs, &i);
         %let offers = offers_&scenarioID;
         put '   <td>' &offers '</td>';
      %end;
      %if &any_objective_vals %then %do;
         %do i=1 %to &numScenarios;
            %let scenarioID = %scan(&scenarioIDs, &i);
            %let obj_value = obj_value_&scenarioID;
            put '   <td>' &obj_value '</td>';
         %end;
      %end;
      %if ((&any_active_second_obj gt 0) and (&any_second_obj_vals)) %then %do;
         %do i=1 %to &numScenarios;
            %let scenarioID = %scan(&scenarioIDs, &i);
            %let secondary_obj_value = secondary_obj_value_&scenarioID;
            put '   <td>' &secondary_obj_value '</td>';
         %end;
      %end;
      %if ((&any_active_roi gt 0) and (&any_roi_vals)) %then %do;
         %do i=1 %to &numScenarios;
            %let scenarioID = %scan(&scenarioIDs, &i);
            %let roi_value = roi_value_&scenarioID;
            put '   <td>' &roi_value '</td>';
         %end;
      %end;
      %if ((&anyScenHasDistinctValuesPerComm eq 1) and
           (&any_active_mandatory ne 0)) %then %do;
         %if (&num_distinct_inputDataIDs gt 1) %then %do;
            %do i=1 %to &numScenarios;
               %let scenarioID = %scan(&scenarioIDs, &i);
               %let offer_is_mandatory = offer_is_mandatory_&scenarioID;
               put '   <td>' &offer_is_mandatory '</td>';
            %end;
         %end;
         %else %do;
            put '   <td>' offer_is_mandatory '</td>';
         %end;
      %end;
      %if ((&anyPrioritizationBased eq 1) and 
           (&any_priority_vals gt 0)) %then %do;
         %if (&num_distinct_inputDataIDs gt 1) %then %do;
            %do i=1 %to &numScenarios;
               %let scenarioID = %scan(&scenarioIDs, &i);
               %let priority = priority_&scenarioID;
               put '   <td>' &priority '</td>';
            %end;
         %end;
         %else %do;
            put '   <td>' priority '</td>';
         %end;
      %end;
      put '</tr>';
      if eof then do;
         put '</table>';
      end;
   run;
   %if (&syserr gt 4) %then %do;
      %let &_rc = 105;
      %put %mo_error Error generating the html data.;
      %goto ERROR_EXIT;
   %end;

   %let morpt1 = %sysfunc(pathname(morpt1));

   %ERROR_EXIT:
%mend;
