/************************************************************************/
/* Copyright (c) 2005-2011, SAS Institute Inc., Cary, NC, USA,          */
/*                       All Rights Reserved.                           */
/*                                                                      */
/* Name:          mo_rpt_compare_with_imported                          */
/*                                                                      */
/* Purpose:       create reports for an imported solution               */
/*                                                                      */
/* Input:         MOScenarioID                                          */
/*                importedScenarioID                                    */
/*                reportID                                              */
/*                                                                      */
/* Output:        &_rc = return code                                    */
/*                                                                      */
/* Return Codes:                                                        */
/*                                                                      */
/*          100   generic error                                         */
/*          101   invalid metadata detected                             */
/*          102   missing data table                                    */
/*          107   SQL error                                             */
/*          181   optengine error                                       */
/*                                                                      */
/* HISTORY:                                                             */
/* 15Mar04 S0237007 Creation (liao)                                     */
/* 12Jun08 S0522541 Added reportID                                      */
/*                                                                      */
/************************************************************************/
%macro mo_rpt_compare_with_imported(MOScenarioID, importedScenarioID, reportID, _rc);

   %local macname;
   %let macname=&sysmacroname;
   %let macname=%lowcase(&macname);

   %if %bquote(&_rc) eq %str() %then %do;
      %local rc;
      %let _rc = rc;
   %end;
   %else %global &_rc;
   %let &_rc = 0;

   /* retrieve the information of these two scenarios */
   %global username project_name scenario_name;
   proc sql noprint;
      select library_metadata_id, customer_table_name, control_table_name, customer_id_column_name,
             username, project_name, scenario_name
      into :metadata_id, :customer_table, :customer_control_table, :cust_id_col_name,
           :username, :project_name, :scenario_name
      from mometa.scenarios as a, mometa.projects as b, mometa.input_data as c, mometa.users as d
      where a.scenario_id = &MOScenarioID
            and a.project_id = b.project_id
            and b.input_data_id = c.input_data_id
            and a.user_id = d.user_id;
      
      select username, project_name, scenario_name
      into :im_username, :im_project_name, :im_scenario_name
      from mometa.scenarios as a, mometa.projects as b, mometa.input_data as c, mometa.users as d
      where a.scenario_id = &importedScenarioID
            and   a.project_id  = b.project_id
            and   b.input_data_id = c.input_data_id
            and   a.user_id = d.user_id;
   quit;
   %if (&sqlrc gt 4) %then %do;
      %put %mo_error SQL error retrieving metadata.;
      %let &_rc = 107;
      %goto ERROR_EXIT;
   %end;

   %let username=%trim(%nrbquote(%left(%nrbquote(&username))));
   %let project_name=%trim(%nrbquote(%left(%nrbquote(&project_name))));
   %let scenario_name=%trim(%nrbquote(%left(%nrbquote(&scenario_name))));

   %let im_username=%trim(%nrbquote(%left(%nrbquote(&im_username))));
   %let im_project_name=%trim(%nrbquote(%left(%nrbquote(&im_project_name))));
   %let im_scenario_name=%trim(%nrbquote(%left(%nrbquote(&im_scenario_name))));


   /* store the path name of the original moMeta */
   %let metaPath = %sysfunc(pathname(moMeta));

   /* now prepare the parameters for the optEngine run */
   %mo_no_notes(on)
   libname rootMeta "&metaPath";
   %if (&syslibrc ne 0) %then %do;
      %let &_rc = 101;
      %put %mo_error Error assigning ROOTMETA libname.;
      %goto ERROR_EXIT;
   %end;

   libname moinput meta libid=&metadata_id;
   %if (&syslibrc ne 0) %then %do;
      %let &_rc = 101;
      %put %mo_error Error assigning MOINPUT libname.;
      %goto ERROR_EXIT;
   %end;

   libname moOutput "%mo_pathname(&moproj.&separator.&username.&separator.&project_name.&separator.&scenario_name.&separator.solution)";
   %if (&syslibrc ne 0) %then %do;
      %let &_rc = 101;
      %put %mo_error Error assigning MOOUTPUT libname.;
      %goto ERROR_EXIT;
   %end;

   libname moImport "%mo_pathname(&moproj.&separator.&im_username.&separator.&im_project_name.&separator.&im_scenario_name.&separator.solution)";
   %if (&syslibrc ne 0) %then %do;
      %let &_rc = 101;
      %put %mo_error Error assigning MOOUTPUT libname.;
      %goto ERROR_EXIT;
   %end;
   
   proc datasets nolist lib=moImport; 
      delete constraint_summary objective_summary offer_summary offer_summary_nway;
   quit;

   libname moSaved "%mo_pathname(&moproj.&separator.&im_username.&separator.&im_project_name.&separator.&im_scenario_name.&separator.saved)";
   %if (&syslibrc ne 0) %then %do;
      %let &_rc = 101;
      %put %mo_error Error assigning MOSAVED libname.;
      %goto ERROR_EXIT;
   %end;

   libname moLog "%mo_pathname(&moproj.&separator.&im_username.&separator.&im_project_name.&separator.&im_scenario_name.&separator.log)";
   %if (&syslibrc ne 0) %then %do;
      %let &_rc = 101;
      %put %mo_error Error assigning MOSAVED libname.;
      %goto ERROR_EXIT;
   %end;

   %if ~%sysfunc(exist(moinput.&customer_table)) and 
       ~%sysfunc(exist(moinput.&customer_table , VIEW)) %then %do;
      %let &_rc = 102;
      %put %mo_error The customer table &customer_table does not exist.;
      %goto ERROR_EXIT;
   %end;

   %if ~%sysfunc(exist(moinput.&customer_control_table)) and
       ~%sysfunc(exist(moinput.&customer_control_table , VIEW)) %then %do;
      %let &_rc = 102;
      %put %mo_error The customer control table &customer_control_table does not exist.;
      %goto ERROR_EXIT;
   %end;

   %global customerdata customercontrol scenario_id;
   %let customerdata    = moinput.&customer_table;
   %let customercontrol = moinput.&customer_control_table;
   %let scenario_id     = &MOScenarioID;
   
   %global compare_imported_solution compare_imported_library_name compare_imported_table_name;
   %let compare_imported_solution = 1;
   %let compare_imported_library_name = moImport;
   %let compare_imported_table_name = mo_solution;

   /* If the input data uses a multi-part key, need to add mo_single_key_cust_id to the imported solution */
   /* This is ignored for HPMO because multi-part key is only valid with MOMA input data.
   %let cust_table_multi_part_key = %index(&cust_id_col_name, *);
   %if &cust_table_multi_part_key %then %do;
      %let num_multi_part_keys = %mo_word_count(&cust_id_col_name, %str(*));
      %do i = 1 %to &num_multi_part_keys;
         %let multi_part_key_&i = %scan(&cust_id_col_name, &i, *);
      %end;
      
      proc sql noprint;
         create table work.multi_part_key_solution as
            select a.campaign_cd, a.communication_cd, b.mo_single_key_cust_id
            from moOutput.mo_solution as a left join &customerData as b
            on a.&multi_part_key_1 = b.&multi_part_key_1
               %do i = 2 %to &num_multi_part_keys;
                  and a.&&multi_part_key_&i = b.&&multi_part_key_&i
               %end;
            order by mo_single_key_cust_id;
      quit;
   %end;
   */
   
   /* Ignore this check for HPMO:
   %let dsid = %sysfunc(open(&customerData, i));
   %let varnum = %sysfunc(varnum(&dsid, &custID));
   %let vartype_raw = %sysfunc(vartype(&dsid, &varnum));
   %let rc=%sysfunc(close(&dsid));
   
   %let dsid = %sysfunc(open(&external_solution_name , i));
   %let varnum = %sysfunc(varnum(&dsid, &custID));
   %if &varnum > 0 %then %do;
      %let vartype_sol = %sysfunc(vartype(&dsid, &varnum));
      %let rc=%sysfunc(close(&dsid));

      %if &vartype_raw ne &vartype_sol %then %do;
         %let &_rc = 140;
         %put %mo_error the customer id column of the imported solution is not compatible with the MO solution.;
         %goto ERROR_EXIT;
      %end;
   %end;
   */
   
   %global compare_imported_objective compare_imported_constraints compare_imported_offers;
   %let compare_imported_objective = 0;
   %let compare_imported_constraints = 0;
   %let compare_imported_offers = 0;
   
   %if &reportID = &_REPORT_TYPE_OBJECTIVE_ %then %let compare_imported_objective = 1;
   %else %if &reportID = &_REPORT_TYPE_CONSTRAINT_ %then %let compare_imported_constraints = 1;
   %else %if &reportID = &_REPORT_TYPE_OFFER_TABLE_ or &reportID = &_REPORT_TYPE_OFFER_CHART_ %then %let compare_imported_offers = 1;

   %mo_main;

   %if &reportID = &_REPORT_TYPE_OBJECTIVE_ %then %do;
      data moImport.objective_summary;
         set temp_objective_summary;
      run;
      %mo_error_tracking;
   %end;   
   %else %if &reportID = &_REPORT_TYPE_CONSTRAINT_ %then %do;
      data moImport.constraint_summary;
         set temp_constraint_summary;
      run;
      %mo_error_tracking;
   %end;
   %else %if &reportID = &_REPORT_TYPE_OFFER_TABLE_ or &reportID = &_REPORT_TYPE_OFFER_CHART_ %then %do;

      /* Get the list of character columns from the communications table */
      proc contents data=communications out=comm_contents_char_columns noprint;
      run;
      %mo_error_tracking;
    
      proc sql noprint;
         select "a." || trim(left(name)) into :char_comm_columns_list separated by ','
         from comm_contents_char_columns
         where type = 2;
      quit;
      %mo_error_tracking;
   
      proc sql noprint;
         create table moImport.offer_summary_nway as
            select &char_comm_columns_list, b.*, . as num_eligible
            from communications as a, temp_offer_summary_nway as b
            where a.offer_id=b.offer_id;
         alter table moImport.offer_summary_nway
            drop offer_id;
      quit;
      %mo_error_tracking;
      
      libname moOutput "%sysfunc(pathname(moImport))";
      %let username = &im_username;
      %let project_name = &im_project_name;
      %let scenario_name = &im_scenario_name;
      %mo_rpt_offer_summary;
   %end;

   %if (&moerror ne 0) %then %do;
      %let &_rc = 181;
      %put %mo_error Error creating reports for the imported solution.;
      %goto ERROR_EXIT;
   %end;

   %mo_no_notes(on)
   libname rootmeta clear;
   libname moinput clear;
   libname mooutput clear;
   libname mosaved clear;
   libname moimport clear;
   libname molog clear;

%ERROR_EXIT:
%mend;
