/************************************************************************/
/* Copyright (c) 2005-2011, SAS Institute Inc., Cary, NC, USA,          */
/*                       All Rights Reserved.                           */
/*                                                                      */
/* Name:          mo_rpt_offer_summary                                  */
/*                                                                      */
/* Purpose:       create offer summary by groups: campaign,             */
/*                communication, product, outbound_channel/team,        */
/*                inbound_channel/team, offer group,                    */
/*                offer_group/subgroup, and time period                 */
/* Input:                                                               */
/*                mo_solution: the temporary solution data set stored   */
/*                             in the work library                      */
/*                commDict: the communication dictionary                */
/*                                                                      */
/* Output:        &_rc = return code                                    */
/*                                                                      */
/*                moOutput.offer_summary                                */
/*                - group_name: the group name                          */
/*                - component1: the components within a group           */
/*                - component2: the components within a group           */
/*                - offers: the total number of offers for this         */
/*                          component                                   */
/*                - obj_value: the objective value for this component   */
/*                - secondary_obj_value: the secondary objective value  */
/*                  for this component, if secondary objective was      */
/*                  active; otherwise, set to missing (.)               */
/*                - roi_profit: the roi profit value for this component */
/*                - roi_cost: the roi cost value for this component     */
/*                - roi_value: the roi value (roi_profit/roi_cost) for  */
/*                             this component                           */
/*                                                                      */
/* Return Codes:                                                        */
/*                                                                      */
/*          101   SQL error                                             */
/*          102   Error encountered while executing                     */
/*                %mo_rpt_offer_summary_by_group                        */
/*                                                                      */
/* HISTORY:                                                             */
/* 15Mar04 S0237007 Creation (liao)                                     */
/*                                                                      */
/*                                                                      */
/* 5.3:                                                                 */
/* 23Jun08 S0483854 Add secondary objective to offer_summary dataset    */
/*         (KEC)                                                        */
/* 17Jul08 S0483854 MOMA support - Include cell_cd column in offer      */
/*         summary reports if present in the MO solution. (KEC)         */
/*                                                                      */
/* 5.4:                                                                 */
/* 16Nov10 S0712522 Remove exclusion of attributes (group_names) for    */
/*         which all values are blank in the solution.                  */
/************************************************************************/
%macro mo_rpt_offer_summary(_rc);
   %local macname;
   %let macname=&sysmacroname;
   %let macname=%lowcase(&macname);

   %if %bquote(&_rc) eq %str() %then %do;
      %local rc;
      %let _rc = rc;
   %end;
   %else %global &_rc;
   %let &_rc = 0;

   /* create an empty offer summary table in the moOutput libraray */
   proc sql noprint;
      create table moOutput.offer_summary
         (group_name          char(32),
          component           char(256),
          num_eligible        num,
          offers              num,
          obj_value           num,
          secondary_obj_value num,
          roi_profit          num,
          roi_cost            num,
          roi_value           num
         );
   quit;
   %if (&sqlrc gt 4) %then %do;
      %let &_rc = 101;
      %put %mo_error SQL error encountered while creating moOutput.offer_summary.;
      %goto ERROR_EXIT;
   %end;

   /* Do proc contents to get all the character variables from the communications table that are
      also in the solution table or offer_summary_nway_table. If you have any attributes called 
      Num_Eligible, Offers, Obj_Value, Secondary_Obj_Value, ROI_Profit, ROI_Cost, or ROI_Value, 
      then they won't be shown in the filter, because these are reserved names for the reports. */
   proc contents data=communications out=communications_contents (keep=name type) noprint;
   run;
   %if (&syserr gt 4) %then %do;
      %let &_rc = 100;
      %put %mo_error Error encountered obtaining contents of communications table (syserr=&syserr).;
      %goto ERROR_EXIT;
   %end;

   %let comm_char_vars = ;
   proc sql noprint;
      delete from communications_contents
         where type=1 
            or upcase(name) in ('NUM_ELIGIBLE','OFFERS','OBJ_VALUE','SECONDARY_OBJ_VALUE','ROI_PROFIT','ROI_COST','ROI_VALUE');
      select 'b.'||trim(left(name)) into :comm_char_vars separated by ', '
         from communications_contents;
   quit;
   %if (&sqlrc gt 4) %then %do;
      %let &_rc = 101;
      %put %mo_error SQL error encountered while deleting rows from communications_contents.;
      %goto ERROR_EXIT;
   %end;

   %if not(%sysfunc(exist(moOutput.offer_summary_nway))) %then %do;

      proc contents data=&hpmo_solution_library..mo_solution out=mo_solution_contents (keep=name type) noprint;
      run;
      %if (&syserr gt 4) %then %do;
         %let &_rc = 100;
         %put %mo_error Error encountered obtaining contents of &hpmo_solution_library..mo_solution table (syserr=&syserr).;
         %goto ERROR_EXIT;
      %end;

      %let objInSolution = 0;
      %let secondaryObjInSolution = 0;
      %let ROInumInSolution = 0;
      %let ROIdenomInSolution = 0;
      %let cellCdColInSolution = 0;
      proc sql noprint;
         select count(*) into :objInSolution from mo_solution_contents where upcase(name)="%upcase(&objField)";
         select count(*) into :secondaryObjInSolution from mo_solution_contents where upcase(name)="%upcase(&secondObjField)";
         select count(*) into :ROInumInSolution from mo_solution_contents where upcase(name)="%upcase(&roiNumerator)";
         select count(*) into :ROIdenomInSolution from mo_solution_contents where upcase(name)="%upcase(&roiDenominator)";
         select count(*) into :cellCdColInSolution from mo_solution_contents where lowcase(name)="cell_cd";
      quit;
      %if (&sqlrc gt 4) %then %do;
         %let &_rc = 101;
         %put %mo_error SQL error encountered while retrieving objective information.;
         %goto ERROR_EXIT;
      %end;

      proc sql noprint;
         /* Create an offer summary by communication, if cell not present, or by */
         /* cell, if present. The idea is to summarize by the lowest level of    */
         /* any group, because "offer summary by group" can be derived from it.  */
        create table offer_summary_by_comm as
            select campaign_cd, communication_cd
                %if &cellCdColInSolution %then %do;
                   ,cell_cd
                %end;
                   ,
                   count(*) as offers,
                   %if &objInSolution %then %do;
                      sum(&objField) as obj_value,
                   %end;
                   %else %do;
                      . as obj_value,
                   %end;
                   %if &secondaryObjInSolution %then %do;
                      sum(&secondObjField) as secondary_obj_value,
                   %end;
                   %else %do;
                      . as secondary_obj_value,
                   %end;
                   %if &ROInumInSolution=1 and &ROIdenomInSolution=1 %then %do;
                      sum(&roiNumerator) as roi_profit,
                      sum(&roiDenominator) as roi_cost,
                      sum(missing(&roiNumerator)) as num_missing_roi_profit,
                      sum(missing(&roiDenominator)) as num_missing_roi_cost
                   %end;
                   %else %do;
                      . as roi_profit,
                      . as roi_cost,
                      0 as num_missing_roi_profit,
                      0 as num_missing_roi_cost
                   %end;
            from &hpmo_solution_lib..mo_solution
            group by campaign_cd, communication_cd
                  %if &cellCdColInSolution %then %do;
                     ,cell_cd
                  %end;
            ;

         /* Merge the "offer summary by communication (cell)" with other         */
         /* communication attributes.                                            */
         create table offer_summary2 as
            select &comm_char_vars,
               case
                  when a.offers ne . then a.offers
                  else 0
               end as offers,
               case
                  when a.obj_value ne . then a.obj_value
                  else 0
               end as obj_value,
               case
                  when a.secondary_obj_value ne . then a.secondary_obj_value
                  else 0
               end as secondary_obj_value,
               case
                  when a.num_missing_roi_profit = 0 then a.roi_profit
                  else .
               end as roi_profit,
               case
                  when a.num_missing_roi_cost = 0 then a.roi_cost
                  else .
               end as roi_cost
            from offer_summary_by_comm as a right join communications as b
            on ((a.campaign_cd=b.campaign_cd) and 
                (a.communication_cd=b.communication_cd)
             %if (&cellCdColInSolution eq 1) %then %do;
                and (a.cell_cd=b.cell_cd)
             %end;
               );
      quit;
      %if (&sqlrc gt 4) %then %do;
         %let &_rc = 101;
         %put %mo_error SQL error encountered while creating offer_summary2.;
         %goto ERROR_EXIT;
      %end;

      %if &objInSolution = 0 %then %do;
         data offer_summary2;
            set offer_summary2;
            obj_value = .;
         run;
         %if (&syserr gt 4) %then %do;
            %let &_rc = 100;
            %put %mo_error Error encountered updating obj_value of offer_summary2.;
            %goto ERROR_EXIT;
         %end;
      %end;

      %if &secondaryObjInSolution = 0 %then %do;
         data offer_summary2;
            set offer_summary2;
            secondary_obj_value = .;
         run;
         %if (&syserr gt 4) %then %do;
            %let &_rc = 100;
            %put %mo_error Error encountered updating secondary_obj_value of offer_summary2.;
            %goto ERROR_EXIT;
         %end;
      %end;

      %if %sysfunc(exist(moOutput.num_of_eligible_offers)) %then %do;
         proc contents data=moOutput.num_of_eligible_offers out=mo_eligible_contents (keep=name type) noprint;
         run;
         %if (&syserr gt 4) %then %do;
            %let &_rc = 100;
            %put %mo_error Error encountered obtaining contents of %trim(moOutput.num_of_eligible_offers) table (syserr=&syserr).;
            %goto ERROR_EXIT;
         %end;
         
         data work.num_of_eligible_offers;
            set mooutput.num_of_eligible_offers;
            if num_eligible = . then num_eligible = eligibles;
         run;
         %mo_error_tracking;
      
         %local cellCdInNumEligible;
         proc sql noprint;
            select count(*) into :cellCdInNumEligible from work.mo_eligible_contents where lowcase(name)="cell_cd";

            create table mooutput.offer_summary_nway as
               select a.*, a.roi_profit/a.roi_cost as roi_value, b.num_eligible
               from offer_summary2 as a, work.num_of_eligible_offers as b
               where ((a.campaign_cd=b.campaign_cd) and 
                      (a.communication_cd=b.communication_cd)
                   %if ((&cellCdColInSolution eq 1) and
                        (&cellCdInNumEligible eq 1)) %then %do;
                      and (a.cell_cd=b.cell_cd)
                   %end;
                     );
         quit;
         %if (&sqlrc gt 4) %then %do;
            %let &_rc = 101;
            %put %mo_error SQL error encountered while updating number of eligible customers.;
            %goto ERROR_EXIT;
         %end;
      %end;
      %else %do;
         proc sql noprint;
            create table mooutput.offer_summary_nway as
               select *, roi_profit/roi_cost as roi_value, . as num_eligible
               from offer_summary2;
         quit;
         %if (&sqlrc gt 4) %then %do;
            %let &_rc = 101;
            %put %mo_error SQL error encountered while creating offer_summary_nway.;
            %goto ERROR_EXIT;
         %end;
      %end;
   %end;

   /* now create different by group reports */
   %let comm_char_vars = ;
   %let num_comm_char_vars = 0;
   proc sql noprint;
      select trim(left(name)) into :comm_char_vars separated by ' '
         from communications_contents;
      select count(*) into :num_comm_char_vars
         from communications_contents;
   quit;
   %if (&sqlrc gt 4) %then %do;
      %let &_rc = 101;
      %put %mo_error SQL error encountered while retrieving character variables from communications_contents.;
      %goto ERROR_EXIT;
   %end;

   %do i = 1 %to &num_comm_char_vars;
      %let char_var = %scan(&comm_char_vars,&i);
      
      %mo_rpt_offer_summary_by_group(&char_var, rc1);
      %if (&rc1 ne 0) %then %do;
         %let &_rc = 102;
         %put %mo_error Error encountered while creating offer summary per &char_var..;
         %goto ERROR_EXIT;
      %end;
   %end;

   %put;
   %put %mo_note Successfully generated data for offer summary reports.;
   %put;
   %goto EXIT;


   %ERROR_EXIT:
   %put;
   %put %mo_error Offer summary report job encountered an error (rc=&&&_rc).;

   %EXIT:

%mend;
