/************************************************************************/
/* Copyright (c) 2005-2011, SAS Institute Inc., Cary, NC, USA,          */
/*                       All Rights Reserved.                           */
/*                                                                      */
/* Name:          mo_rpt_offer_summary_by_group                         */
/*                                                                      */
/* Purpose:       generate offer summary data set for a particular      */
/*                communication group                                   */
/*                                                                      */
/* Input:         group_name                                            */
/*                                                                      */
/* Output:        &_rc = return code                                    */
/*                                                                      */
/*                moOutput.offer_summary                                */
/*                - group_name: the group name                          */
/*                - component:  the components within a group           */
/*                - offers: the total number of offers for this         */
/*                          component                                   */
/*                - obj_value: the objective value for this component   */
/*                - secondary_obj_value: the secondary objective value  */
/*                  for this component, if secondary objective was      */
/*                  active; otherwise, set to missing (.)               */
/*                - roi_profit: the roi profit value for this component */
/*                - roi_cost: the roi cost value for this component     */
/*                - roi_value: the roi value (roi_profit/roi_cost) for  */
/*                             this component                           */
/*                                                                      */
/* Return Codes:                                                        */
/*                                                                      */
/*          101   SQL error                                             */
/*                                                                      */
/* HISTORY:                                                             */
/* 15Mar04 S0237007 Creation (liao)                                     */
/*                                                                      */
/* 5.3                                                                  */
/* 23Jun08 S0483854 Add secondary objective to offer_summary dataset    */
/*         (KEC)                                                        */
/************************************************************************/
%macro mo_rpt_offer_summary_by_group(group_name, _rc);
   %local macname;
   %let macname=&sysmacroname;
   %let macname=%lowcase(&macname);

   %if %bquote(&_rc) eq %str() %then %do;
      %local rc;
      %let _rc = rc;
   %end;
   %else %global &_rc;
   %let &_rc = 0;

   proc sql noprint;
      /* create the by group report */
      create table temp as
         select &group_name,
            sum(num_eligible) as num_eligible,
            sum(offers) as offers,
            sum(obj_value) as obj_value,
            sum(secondary_obj_value) as secondary_obj_value,
            sum(roi_profit) as roi_profit,
            sum(roi_cost) as roi_cost,
            sum(missing(roi_profit)*offers) as num_missing_roi_profit,
            sum(missing(roi_cost)*offers) as num_missing_roi_cost
      from mooutput.offer_summary_nway
      group by &group_name;

      /* insert the by group report to moOuput.offer_summary */
      insert into moOutput.offer_summary
         select "&group_name" as group_name,
            &group_name as component,
            num_eligible as num_eligible,
            offers as offers,
            obj_value as obj_value,
            secondary_obj_value as secondary_obj_value,
            case
               when num_missing_roi_profit = 0 then roi_profit
               else .
            end as roi_profit,
            case
               when num_missing_roi_cost = 0 then roi_cost
               else .
            end as roi_cost,
            case
               when num_missing_roi_profit = 0 and num_missing_roi_cost = 0 then 100*roi_profit/roi_cost
               else .
            end as roi_value
         from temp;
   quit;

   %if (&sqlrc gt 4) %then %do;
      %let &_rc = 101;
      %put %mo_error SQL error encountered while creating offer summary by group.;
      %goto ERROR_EXIT;
   %end;

   %ERROR_EXIT:

%mend;
