/************************************************************************/
/* Copyright (c) 2005-2011, SAS Institute Inc., Cary, NC, USA,          */
/*                       All Rights Reserved.                           */
/*                                                                      */
/* Name:          mo_rpt_offer_summary_table                            */
/*                                                                      */
/* Purpose:       create offer summary table report, feed the XML to GUI*/
/*                                                                      */
/* Input:         scenario_id: the scenario id                          */
/*                group_name: name of customer attribute to filter by   */
/*                                                                      */
/* Output:        &_rc = return code                                    */
/*                                                                      */
/*                report.xml                                            */
/*                                                                      */
/* Return Codes:                                                        */
/*                                                                      */
/*          101   SQL error retrieving metadata                         */
/*          102   Error assigning moOutput libname                      */
/*          103   Unknown group_type: &group_type                       */
/*          105   Now rows with group_name "&group_name" were found     */
/*          106   Error generating HTML data                            */
/*                                                                      */
/* HISTORY:                                                             */
/* 29Mar04 S0237007 Creation (liao)                                     */
/* 03Dec04 S0273454 Scenario specific communication attributes (dwh)    */
/*                                                                      */
/* 4.4:                                                                 */
/* 10Jan06 S0337584 Right justify numbers in tabular reports (dwh)      */
/* 08Jun06 S0361339 Localize channel type (dwh)                         */
/*                                                                      */
/* 5.3:                                                                 */
/* 23Jun08 S0483854 Add secondary objective to report (KEC)             */
/*                                                                      */
/* 5.4:                                                                 */
/* 15Feb10 S0639443 Include Mandatory column in report when scenario    */
/*         is based on input data for which Mandatory Inclusions are    */
/*         defined and the selected filter (group_name) values are      */
/*         unique over the set of communications e.g., COMMUNICATION_CD */
/*         (KEC)                                                        */
/* 27Aug10 Include Priority column in report when scenario is           */
/*         Prioritization-based and the selected filter (group_name)    */
/*         values are unique over the set of communications e.g.,       */
/*         COMMUNICATION_CD (KEC)                                       */
/************************************************************************/
%macro mo_rpt_offer_summary_table(scenario_id, group_name, locale, _rc);
   %local macname;
   %let macname=&sysmacroname;
   %let macname=%lowcase(&macname);

   %if %bquote(&_rc) eq %str() %then %do;
      %local rc;
      %let _rc = rc;
   %end;
   %else %global &_rc;
   %let &_rc = 0;


   %*----------------------------------------;
   %* Gather user, project, and scenario     ;
   %* names for all scenarios.               ;
   %*----------------------------------------;
   %local username project_name scenario_name input_data_id;
   %let username = ;
   %let project_name = ;
   %let scenario_name = ;
   %let input_data_id = .;

   %local active_sample sample_ratio;
   %let active_sample = 0;
   %let sample_ratio = 1;

   proc sql noprint;
      select username, project_name, scenario_name, b.input_data_id  
         into :username, :project_name, :scenario_name, :input_data_id
         from mometa.scenarios as a, mometa.projects as b, mometa.users as c
         where a.scenario_id = &scenario_id
               and a.project_id = b.project_id
               and a.user_id = c.user_id;

      select active_sample, round(sample_ratio)
         into :active_sample, :sample_ratio 
         from mometa.scenario_advanced_options
         where scenario_id = &scenario_id;
   quit;
   %if (&sqlrc gt 4) %then %do;
      %put %mo_error SQL error retrieving metadata.;
      %let &_rc = 101;
      %goto ERROR_EXIT;
   %end;

   %let username=%trim(%left(&username));
   %let project_name=%trim(%left(&project_name));
   %let scenario_name=%trim(%left(&scenario_name));

   %let samples =;
   %if &active_sample = 1 %then %do;
      %let sample_ratio = &sample_ratio;
      %let samples = (&sample_ratio.% &sample_label);
   %end;


   %*------------------------------------;
   %* Access solution directory datasets ; 
   %* and the report directory and file. ; 
   %*------------------------------------;
   libname moOutput "%mo_pathname(&moproj.&separator.&username.&separator.&project_name.&separator.&scenario_name.&separator.solution)";
   %if (&syslibrc ne 0) %then %do;
      %let &_rc = 102;
      %put %mo_error Error assigning moOutput libname.;
      %goto ERROR_EXIT;
   %end;

   filename morpt1 "%mo_pathname(&moproj.&separator.&username.&separator.&project_name.&separator.&scenario_name.&separator.reports&separator.report1.xml)";
   %if (&sysfilrc ne 0 and &sysfilrc ne 1) %then %do;
      %put %mo_error Error assigning MORPT1 filref (sysfilrc=&sysfilrc);
      %let &_rc = 125;
      %goto ERROR_EXIT;
   %end;


   %*----------------------------------------------------;
   %* Determine if the scenario is prioritization-based. ;
   %*----------------------------------------------------;
   %global prioritizationBased;
   %let prioritizationBased = 0;
   %mo_check_if_priority_based(item_type             = SCENARIO
                              ,item_id               = &scenario_id
                              ,_prioritization_based = prioritizationBased
                              ,_rc                   = cipbrc);
   %if (&cipbrc ne 0) %then %do;
      %let &_rc = &cipbrc;
      %goto ERROR_EXIT;
   %end;

   %local priority_column;
   %let priority_column = ;
   %if (&prioritizationBased eq 1) %then %do;
      proc sql noprint;
         select priority_column_name into :priority_column
         from mometa.input_data
         where (input_data_id = &input_data_id);
      quit;
      %if (&sqlrc gt 4) %then %do;
         %put %mo_error SQL error reading PRIORITY_COLUMN_NAME from MOMETA.INPUT_DATA table (sqlrc=&sqlrc).;
         %let &_rc = 107;
         %goto ERROR_EXIT;
      %end;
      %let priority_column = %trim(%left(&priority_column));
   %end;


   %*-----------------------------------------------------;
   %* Determine if the scenario has mandatory inclusions. ;
   %*-----------------------------------------------------;
   %local any_active_mandatory;
   %let any_active_mandatory = 0;
   proc sql noprint;
      select count(*) into :any_active_mandatory
      from mometa.communications
      where (input_data_id = &input_data_id) & (mandatory = 1);
   quit;
   %if (&sqlrc gt 4) %then %do;
      %put %mo_error SQL error reading MANDATORY column in MOMETA.COMMUNICATIONS table (sqlrc=&sqlrc).;
      %let &_rc = 107;
      %goto ERROR_EXIT;
   %end;


   %*------------------------------------------;
   %* For a non-prioritization-based scenario, ;
   %* obtain the objective and ROI values.     ;
   %*------------------------------------------;
   %local objective_attribute_name;
   %local active_roi active_secondary_obj;
   %local secondary_obj_attribute_name;
   %local attribute_role secondary_obj_attribute_role;
   %let objective_attribute_name = ;
   %let active_roi = 0;
   %let active_secondary_obj = 0;
   %let secondary_obj_attribute_name = ;
   %let attribute_role = ;
   %let secondary_obj_attribute_role = ;

   %if (&prioritizationBased ne 1) %then %do;
      proc sql noprint;
         select objective_attribute_name
            into :objective_attribute_name
            from mometa.scenarios
            where scenario_id = &scenario_id;
      
         select active_roi, active_secondary_obj, secondary_obj_attribute_name
            into :active_roi, :active_secondary_obj, :secondary_obj_attribute_name
            from mometa.scenario_advanced_options
            where scenario_id = &scenario_id;
         quit;
      %if (&sqlrc gt 4) %then %do;
         %put %mo_error SQL error retrieving metadata.;
         %let &_rc = 101;
         %goto ERROR_EXIT;
      %end;
      %let objective_attribute_name=%trim(%nrbquote(%left(%nrbquote(&objective_attribute_name))));
      %let secondary_obj_attribute_name=%trim(%nrbquote(%left(%nrbquote(&secondary_obj_attribute_name))));

      %mo_rpt_get_attribute_role(mometa.scenarios
                                ,&scenario_id
                                ,objective_attribute_name
                                ,attribute_role);

      %mo_rpt_get_attribute_role(mometa.scenario_advanced_options
                                ,&scenario_id
                                ,secondary_obj_attribute_name
                                ,secondary_obj_attribute_role);
   %end;


   %*------------------------------------;
   %* Prepare working dataset from which ;
   %* report will be generated.          ;
   %*------------------------------------;
   %local oneOrMoreBlankComponents;
   %let oneOrMoreBlankComponents = 0;
   data work.offer_summary_0;
      set moOutput.offer_summary;
      where upcase(group_name) = "%upcase(&group_name)";

      if (component = "") then do;
         component = "[BLANK]";
         call symput("oneOrMoreBlankComponents", put(1,1.));
      end; 

      /* If the offer_summary dataset was created prior to introduction of   */
      /* the Dual Objectives feature, there will be no secondary_obj_value   */
      /* column. In that case, add a secondary_obj_value column set to       */
      /* missing values.                                                     */
      if (secondary_obj_value = .) then secondary_obj_value = .;

      %if (&prioritizationBased eq 1) %then %do;
         obj_value = .;
         secondary_obj_value = .;
         roi_profit = .;
         roi_cost = .;
         roi_value = .;
      %end;
   run;
   %if (&syserr gt 4) %then %do;
      %let &_rc = 104;
      %put %mo_error No rows with group_name "&group_name" were found.;
      %goto ERROR_EXIT;
   %end;


   %*--------------------------------------------;
   %* Determine if selected group_name values    ;
   %* are unique over the set of communications. ;
   %*--------------------------------------------;
   %local numCommunications numDistinctGroupNameValues;
   %let numCommunications = 0;
   %let numDistinctGroupNameValues = 0;
   proc sql noprint;
      select count(*) into :numCommunications
      from moOutput.offer_summary_nway;

      select count(*) into :numDistinctGroupNameValues
      from moOutput.offer_summary
      where (upcase(group_name) = "%upcase(&group_name)");
   quit;
   %if (&sqlrc gt 4) %then %do;
      %put %mo_error SQL error counting rows in MOOUTPUT.OFFER_SUMMARY_NWAY and MOOUTPUT.OFFER_SUMMARY (sqlrc=&sqlrc).;
      %let &_rc = 107;
      %goto ERROR_EXIT;
   %end;

   %local distinctValuesPerComm;
   %let distinctValuesPerComm = 0;
   %if (&numDistinctGroupNameValues eq &numCommunications) %then %do;
      %let distinctValuesPerComm = 1;
   %end;
   

   %*-------------------------------------------------------------------;
   %* If there are any mandatory communications in the input data, then ;
   %* determine whether it is possible to display the MANDATORY column  ;
   %* in the report. It will be displayed only if the values for the    ;
   %* selected group_name are unique over the set of communications.    ;
   %* Analogous to mandatory inclusions, determine if the scenario is   ;
   %* prioritization-based.  If so, then determine whether is is        ;
   %* possible to display the PRIORITY column in the report. It will    ;
   %* be displayed only if the values for the selected group_name       ;
   %* are unique over the set of communications.                        ;
   %*-------------------------------------------------------------------;
   %if (((&any_active_mandatory gt 0) or (&prioritizationBased eq 1)) and
         (&distinctValuesPerComm eq 1)) %then %do;
   
      /* Reconstruct the Communications table as it pertains to this  */
      /* scenario from the piece-parts that were persisted in the MO  */
      /* Datastore. Then join against this working table to determine */
      /* the Mandatory and Priority column settings.                  */

      /* First build the intermediate working tables from which it    */
      /* will be built.                                               */
      proc sql noprint;
         create table work.comm_character_attributes as
            select * from mometa.comm_character_attributes 
            where (input_data_id = &input_data_id);

         create table comm_character_attribute_values as
            select * from mometa.comm_character_attribute_values 
            where (input_data_id = &input_data_id);

         create table comm_character_distinct_values as
            select * from mometa.comm_character_distinct_values
            where (input_data_id = &input_data_id);

         create table comm_numeric_attributes as
            select * from mometa.comm_numeric_attributes
            where (input_data_id = &input_data_id);

         create table comm_numeric_attribute_values as
            select * from mometa.comm_numeric_attribute_values 
            where (scenario_id = &scenario_id);
      quit;
      %if (&sqlrc gt 4) %then %do;
         %put %mo_error SQL error creating temp tables from which to build WORK.COMMUNICATIONS via call to mo_gui_communication (sqlrc=&sqlrc).;
         %let &_rc = 107;
         %goto ERROR_EXIT;
      %end;

      /* Build work.communications by calling mo_gui_communication.   */
      /* mo_gui_communication assumes the above tables exist in WORK. */
      %let modebug_optengine_log = 0;
      %mo_gui_communication;

      proc sql noprint;
         create table work.offer_summary_1 as
         select a.*
             %if (&any_active_mandatory gt 0) %then %do;
                , case
                     when (b.mandatory = 1) then upcase("&mo_yes_value")
                     else upcase("&mo_no_value")
                  end as offer_is_mandatory
             %end;
             %if (&prioritizationBased eq 1) %then %do;
                , b.&priority_column as priority
             %end; 
         from work.offer_summary_0 as a left join work.communications as b
         on (upcase(a.component) = upcase(b.&group_name));
      quit;
      %if (&sqlrc gt 4) %then %do;
         %put %mo_error SQL error creating offer summary by communication codes (sqlrc=&sqlrc).;
         %let &_rc = 104;
         %goto ERROR_EXIT;
      %end;
   %end;
   %else %do;
      proc sql noprint;
         create table work.offer_summary_1 as
         select * from work.offer_summary_0;
      quit;
      %if (&sqlrc gt 4) %then %do;
         %put %mo_error SQL error creating offer summary (sqlrc=&sqlrc).;
         %let &_rc = 104;
         %goto ERROR_EXIT;
      %end;
   %end;


   %*------------------------------------------------;
   %* When the selected group_name is time_periods,  ;
   %* order output per specified order in datastore. ;
   %*------------------------------------------------;
   %if "%upcase(&group_name)" = "TIME_PERIOD" %then %do;
      proc sql noprint;
         create table work.offer_summary as
            select a.*
            from work.offer_summary_1 as a,
                 mometa.time_periods as b
            where b.input_data_id = &input_data_id
                  and a.component = b.time_period
            order by b.order;
      quit;
      %if (&sqlrc gt 4) %then %do;
         %put %mo_error SQL error creating offer summary by time periods (sqlrc=&sqlrc).;
         %let &_rc = 104;
         %goto ERROR_EXIT;
      %end;
   %end;
   %else %do;
      proc sql noprint;
         create table work.offer_summary as
         select * from work.offer_summary_1;
      quit;
      %if (&sqlrc gt 4) %then %do;
         %put %mo_error SQL error creating offer summary (sqlrc=&sqlrc).;
         %let &_rc = 104;
         %goto ERROR_EXIT;
      %end;
   %end;


   %*------------------------------------------------;
   %* Screen for all missing values for objective,   ;
   %* secondary objective, ROI, and priority because ;
   %* report columns for these should be suppressed  ;
   %* if there are no values.                        ;
   %*------------------------------------------------;
   %let has_objective_vals = 0;
   %let has_secondary_obj_vals = 0;
   %let has_roi_vals = 0;
   %let has_priority_vals = 0;
   proc sql noprint;
      select count(obj_value) into :has_objective_vals from work.offer_summary;

      %if (&active_secondary_obj eq 1) %then %do;
         select count(secondary_obj_value) into :has_secondary_obj_vals from work.offer_summary;
      %end;

      %if (&active_roi eq 1) %then %do;
         select count(roi_value) into :has_roi_vals from work.offer_summary;
      %end;

      %if ((&prioritizationBased eq 1) and
           (&distinctValuesPerComm eq 1)) %then %do;
         select count(priority) into :has_priority_vals from work.offer_summary;
      %end;
   quit;
   %if (&sqlrc gt 4) %then %do;
      %put %mo_error SQL error retrieving objective information from offer summary (sqlrc=&sqlrc).;
      %let &_rc = 104;
      %goto ERROR_EXIT;
   %end;


   %*--------------------------------------------;
   %* Create XML file from which report will be  ;
   %* surfaced using working dataset from above. ;
   %*--------------------------------------------;
   options locale=&locale;

   /* Insert %nrstr(&nbsp;) into column names in place of spaces */
   %let offer_summary_eligibles = %mo_nbsp_string(&offer_summary_eligibles);
   %let offer_summary_offers    = %mo_nbsp_string(&offer_summary_offers);
   %let offer_summary_roi       = %mo_nbsp_string(&offer_summary_roi);
   %let offer_summary_mandatory = %mo_nbsp_string(&offer_summary_mandatory);
   %let offer_summary_priority  = %mo_nbsp_string(&offer_summary_priority);

   data _null_;
      file morpt1 encoding='utf-8';
      set work.offer_summary end=eof;
      if _n_ = 1 then do;
         put '<table title="' "&offer_summary_title &left_parentheses.&project_name/&scenario_name.&right_parentheses"
         %if "&samples" ne "" %then %do; "&samples" %end; '">';
         put '<tr>';
         put '   <th type="String">' "%upcase(&group_name)" '</th>';
         put '   <th type="Integer">' "%upcase(&offer_summary_eligibles)" '</th>';
         put '   <th type="Integer">' "%upcase(&offer_summary_offers)" '</th>';

         %if &has_objective_vals %then %do;
            %if ((&attribute_role = &_ATTRIBUTE_ROLE_COST_) or 
                 (&attribute_role = &_ATTRIBUTE_ROLE_REVENUE_) or 
                 (&attribute_role = &_ATTRIBUTE_ROLE_PROFIT_)) %then %do;
               put '   <th type="Double" currency="true">' "%upcase(&objective_attribute_name)" '</th>';
            %end;
            %else %do;
               put '   <th type="Double">' "%upcase(&objective_attribute_name)" '</th>';
            %end;
         %end;

         %if ((&active_secondary_obj eq 1) and (&has_secondary_obj_vals)) %then %do;
            %if ((&secondary_obj_attribute_role = &_ATTRIBUTE_ROLE_COST_) or 
                 (&secondary_obj_attribute_role = &_ATTRIBUTE_ROLE_REVENUE_) or 
                 (&secondary_obj_attribute_role = &_ATTRIBUTE_ROLE_PROFIT_)) %then %do;
               put '   <th type="Double" currency="true">' "%upcase(&secondary_obj_attribute_name)" '</th>';
            %end;
            %else %do;
               put '   <th type="Double">' "%upcase(&secondary_obj_attribute_name)" '</th>';
            %end;
         %end;

         %if ((&active_roi eq 1) and (&has_roi_vals)) %then %do;
            put '   <th type="Double" percent="true">' "%upcase(&offer_summary_roi)" ' </th>';
         %end;
         %if ((&any_active_mandatory gt 0) and
              (&distinctValuesPerComm eq 1)) %then %do;
            put '   <th type="String" align="right">' "%upcase(&offer_summary_mandatory)" '</th>';
         %end;
         %if ((&prioritizationBased eq 1) and 
              (&distinctValuesPerComm eq 1) and 
              (&has_priority_vals)) %then %do;
            put '   <th type="Integer" align="right">' "%upcase(&offer_summary_priority)" '</th>';
         %end;
         put '</tr>';
      end;
      put '<tr>';
      put '   <td>' component '</td>';
      put '   <td>' num_eligible '</td>';
      put '   <td>' offers '</td>';
      %if &has_objective_vals %then %do;
         put '   <td>' obj_value '</td>';
      %end;
      %if ((&active_secondary_obj eq 1) and (&has_secondary_obj_vals)) %then %do;
         put '   <td>' secondary_obj_value '</td>';
      %end;
      %if ((&active_roi eq 1) and (&has_roi_vals)) %then %do;
         put '   <td>' roi_value '</td>';
      %end;
      %if ((&any_active_mandatory gt 0) and
           (&distinctValuesPerComm eq 1)) %then %do;
         put '   <td align="right">' offer_is_mandatory '</td>';
      %end;
      %if ((&prioritizationBased eq 1) and 
           (&distinctValuesPerComm eq 1) and
           (&has_priority_vals)) %then %do;
         put '   <td align="right">' priority '</td>';
      %end;
      put '</tr>';
      if eof then do;
         put '</table>';
      end;
   run;
   %if (&syserr gt 4) %then %do;
      %let &_rc = 106;
      %put %mo_error Error generating HTML data.;
      %goto ERROR_EXIT;
   %end;

   %let morpt1 = %sysfunc(pathname(morpt1));

   libname mooutput clear;

   %ERROR_EXIT:
%mend;
